@@ -79,6 +79,9 @@ void GC_push_all_stacks() {
 #elif defined(I386)
   i386_thread_state_t state;
   mach_msg_type_number_t thread_state_count = i386_THREAD_STATE_COUNT;
+#elif defined(ARM)
+  arm_thread_state_t state;
+  mach_msg_type_number_t thread_state_count = ARM_THREAD_STATE_COUNT;
 #else
 # error FIXME for non-x86 || ppc architectures
   mach_msg_type_number_t thread_state_count = MACHINE_THREAD_STATE_COUNT;
@@ -102,6 +105,17 @@ void GC_push_all_stacks() {
 	if(r != KERN_SUCCESS) ABORT("thread_get_state failed");
 	
 #if defined(I386)
+#if defined(_STRUCT_X86_EXCEPTION_STATE32)
+	lo = state.__esp;
+
+	GC_push_one(state.__eax); 
+	GC_push_one(state.__ebx); 
+	GC_push_one(state.__ecx); 
+	GC_push_one(state.__edx); 
+	GC_push_one(state.__edi); 
+	GC_push_one(state.__esi); 
+	GC_push_one(state.__ebp); 
+#else
 	lo = state.esp;
 
 	GC_push_one(state.eax); 
@@ -111,7 +125,43 @@ void GC_push_all_stacks() {
 	GC_push_one(state.edi); 
 	GC_push_one(state.esi); 
 	GC_push_one(state.ebp); 
+#endif
 #elif defined(POWERPC)
+#if defined(_STRUCT_PPC_EXCEPTION_STATE)
+	lo = (void*)(state.__r1 - PPC_RED_ZONE_SIZE);
+        
+	GC_push_one(state.__r0); 
+	GC_push_one(state.__r2); 
+	GC_push_one(state.__r3); 
+	GC_push_one(state.__r4); 
+	GC_push_one(state.__r5); 
+	GC_push_one(state.__r6); 
+	GC_push_one(state.__r7); 
+	GC_push_one(state.__r8); 
+	GC_push_one(state.__r9); 
+	GC_push_one(state.__r10); 
+	GC_push_one(state.__r11); 
+	GC_push_one(state.__r12); 
+	GC_push_one(state.__r13); 
+	GC_push_one(state.__r14); 
+	GC_push_one(state.__r15); 
+	GC_push_one(state.__r16); 
+	GC_push_one(state.__r17); 
+	GC_push_one(state.__r18); 
+	GC_push_one(state.__r19); 
+	GC_push_one(state.__r20); 
+	GC_push_one(state.__r21); 
+	GC_push_one(state.__r22); 
+	GC_push_one(state.__r23); 
+	GC_push_one(state.__r24); 
+	GC_push_one(state.__r25); 
+	GC_push_one(state.__r26); 
+	GC_push_one(state.__r27); 
+	GC_push_one(state.__r28); 
+	GC_push_one(state.__r29); 
+	GC_push_one(state.__r30); 
+	GC_push_one(state.__r31);
+#else
 	lo = (void*)(state.r1 - PPC_RED_ZONE_SIZE);
         
 	GC_push_one(state.r0); 
@@ -145,6 +195,27 @@ void GC_push_all_stacks() {
 	GC_push_one(state.r29); 
 	GC_push_one(state.r30); 
 	GC_push_one(state.r31);
+#endif
+#elif defined(ARM)
+        lo = (void*)state.__sp;
+
+        GC_push_one(state.__r[0]);
+        GC_push_one(state.__r[1]);
+        GC_push_one(state.__r[2]);
+        GC_push_one(state.__r[3]);
+        GC_push_one(state.__r[4]);
+        GC_push_one(state.__r[5]);
+        GC_push_one(state.__r[6]);
+        GC_push_one(state.__r[7]);
+        GC_push_one(state.__r[8]);
+        GC_push_one(state.__r[9]);
+        GC_push_one(state.__r[10]);
+        GC_push_one(state.__r[11]);
+        GC_push_one(state.__r[12]);
+        /* GC_push_one(state.__sp);  */
+        GC_push_one(state.__lr);
+        GC_push_one(state.__pc);
+        GC_push_one(state.__cpsr);
 #else
 # error FIXME for non-x86 || ppc architectures
 #endif
@@ -199,6 +270,42 @@ void GC_push_all_stacks() {
 			     (natural_t *)&info, &outCount);
 	if(r != KERN_SUCCESS) continue;
 
+#if defined(_STRUCT_PPC_EXCEPTION_STATE)
+	lo = (void*)(info.__r1 - PPC_RED_ZONE_SIZE);
+	hi = (ptr_t)FindTopOfStack(info.__r1);
+
+	GC_push_one(info.__r0); 
+	GC_push_one(info.__r2); 
+	GC_push_one(info.__r3); 
+	GC_push_one(info.__r4); 
+	GC_push_one(info.__r5); 
+	GC_push_one(info.__r6); 
+	GC_push_one(info.__r7); 
+	GC_push_one(info.__r8); 
+	GC_push_one(info.__r9); 
+	GC_push_one(info.__r10); 
+	GC_push_one(info.__r11); 
+	GC_push_one(info.__r12); 
+	GC_push_one(info.__r13); 
+	GC_push_one(info.__r14); 
+	GC_push_one(info.__r15); 
+	GC_push_one(info.__r16); 
+	GC_push_one(info.__r17); 
+	GC_push_one(info.__r18); 
+	GC_push_one(info.__r19); 
+	GC_push_one(info.__r20); 
+	GC_push_one(info.__r21); 
+	GC_push_one(info.__r22); 
+	GC_push_one(info.__r23); 
+	GC_push_one(info.__r24); 
+	GC_push_one(info.__r25); 
+	GC_push_one(info.__r26); 
+	GC_push_one(info.__r27); 
+	GC_push_one(info.__r28); 
+	GC_push_one(info.__r29); 
+	GC_push_one(info.__r30); 
+	GC_push_one(info.__r31);
+#else
 	lo = (void*)(info.r1 - PPC_RED_ZONE_SIZE);
 	hi = (ptr_t)FindTopOfStack(info.r1);
 
@@ -233,7 +340,8 @@ void GC_push_all_stacks() {
 	GC_push_one(info.r29); 
 	GC_push_one(info.r30); 
 	GC_push_one(info.r31);
-#      else
+#endif
+#      elif defined(I386) /* !POWERPC */
 	/* FIXME: Remove after testing:	*/
 	WARN("This is completely untested and likely will not work\n", 0);
 	i386_thread_state_t info;
@@ -242,6 +350,26 @@ void GC_push_all_stacks() {
 			     (natural_t *)&info, &outCount);
 	if(r != KERN_SUCCESS) continue;
 
+#if defined(_STRUCT_X86_EXCEPTION_STATE32)
+	lo = (void*)info.__esp;
+	hi = (ptr_t)FindTopOfStack(info.__esp);
+
+	GC_push_one(info.__eax); 
+	GC_push_one(info.__ebx); 
+	GC_push_one(info.__ecx); 
+	GC_push_one(info.__edx); 
+	GC_push_one(info.__edi); 
+	GC_push_one(info.__esi); 
+	GC_push_one(info.__ebp);
+	/* GC_push_one(info.__esp);  */
+	GC_push_one(info.__ss); 
+	GC_push_one(info.__eip); 
+	GC_push_one(info.__cs); 
+	GC_push_one(info.__ds); 
+	GC_push_one(info.__es); 
+	GC_push_one(info.__fs); 
+	GC_push_one(info.__gs); 
+#else
 	lo = (void*)info.esp;
 	hi = (ptr_t)FindTopOfStack(info.esp);
 
@@ -251,7 +379,7 @@ void GC_push_all_stacks() {
 	GC_push_one(info.edx); 
 	GC_push_one(info.edi); 
 	GC_push_one(info.esi); 
-	/* GC_push_one(info.ebp);  */
+	GC_push_one(info.ebp);
 	/* GC_push_one(info.esp);  */
 	GC_push_one(info.ss); 
 	GC_push_one(info.eip); 
@@ -260,7 +388,35 @@ void GC_push_all_stacks() {
 	GC_push_one(info.es); 
 	GC_push_one(info.fs); 
 	GC_push_one(info.gs); 
-#      endif /* !POWERPC */
+#endif
+#      elif defined(ARM) /* !I386 */
+	arm_thread_state_t info;
+	mach_msg_type_number_t outCount = THREAD_STATE_MAX;
+	r = thread_get_state(thread, GC_MACH_THREAD_STATE_FLAVOR,
+			     (natural_t *)&info, &outCount);
+	if(r != KERN_SUCCESS) continue;
+
+	lo = (void*)info.__sp;
+	hi = (ptr_t)FindTopOfStack(info.__sp);
+
+	GC_push_one(info.__r[0]); 
+	GC_push_one(info.__r[1]); 
+	GC_push_one(info.__r[2]); 
+	GC_push_one(info.__r[3]); 
+	GC_push_one(info.__r[4]); 
+	GC_push_one(info.__r[5]); 
+	GC_push_one(info.__r[6]); 
+	GC_push_one(info.__r[7]); 
+	GC_push_one(info.__r[8]); 
+	GC_push_one(info.__r[9]); 
+	GC_push_one(info.__r[10]); 
+	GC_push_one(info.__r[11]); 
+	GC_push_one(info.__r[12]); 
+	/* GC_push_one(info.__sp);  */
+	GC_push_one(info.__lr); 
+	GC_push_one(info.__pc); 
+	GC_push_one(info.__cpsr); 
+#      endif /* !ARM */
       }
 #     if DEBUG_THREADS
        GC_printf3("Darwin: Stack for thread 0x%lx = [%lx,%lx)\n",
@@ -280,13 +436,14 @@ void GC_push_all_stacks() {
 static mach_port_t GC_mach_handler_thread;
 static int GC_use_mach_handler_thread = 0;
 
-static struct GC_mach_thread GC_mach_threads[THREAD_TABLE_SZ];
+#define SUSPEND_THREADS_SIZE 2048
+static struct GC_mach_thread GC_mach_threads[SUSPEND_THREADS_SIZE];
 static int GC_mach_threads_count;
 
 void GC_stop_init() {
   int i;
 
-  for (i = 0; i < THREAD_TABLE_SZ; i++) {
+  for (i = 0; i < SUSPEND_THREADS_SIZE; i++) {
     GC_mach_threads[i].thread = 0;
     GC_mach_threads[i].already_suspended = 0;
   }
