@@ -26,6 +26,8 @@ using System.ComponentModel;
 using System.Collections;
 using System.Collections.Generic;
 using System.Reflection;
+using System.Text;
+using System.Text.RegularExpressions;
 
 namespace System.Windows.Forms {
 
@@ -39,10 +41,15 @@ namespace System.Windows.Forms {
 		IList, ISupportInitializeNotification, ICollection,
 		IComponent, ICurrencyManagerProvider, IEnumerable 
 	{
-		ISite site;
+		bool is_initialized = true;
 
 		IList list;
-		bool list_defaulted;
+		CurrencyManager currency_manager;
+		Dictionary<string,CurrencyManager> related_currency_managers = new Dictionary<string,CurrencyManager> ();
+		//bool list_defaulted;
+		internal Type item_type;
+		bool item_has_default_ctor;
+		bool list_is_ibinding;
 
 		object datasource;
 		string datamember;
@@ -51,9 +58,12 @@ namespace System.Windows.Forms {
 
 		bool allow_new_set;
 		bool allow_new;
-		bool suspended;
 
-		int position = -1;
+		bool add_pending;
+		int pending_add_index;
+
+		string filter;
+		string sort;
 
 		public BindingSource (IContainer container) : this ()
 		{
@@ -68,16 +78,11 @@ namespace System.Windows.Forms {
 			raise_list_changed_events = true;
 
 			ResetList ();
+			ConnectCurrencyManager ();
 		}
 
-		public BindingSource ()
+		public BindingSource () : this (null, String.Empty)
 		{
-			datasource = null;
-			datamember = "";
-
-			raise_list_changed_events = true;
-
-			ResetList ();
 		}
 
 		IList GetListFromEnumerable (IEnumerable enumerable)
@@ -116,31 +121,83 @@ namespace System.Windows.Forms {
 			return l;
 		}
 
+		void ConnectCurrencyManager ()
+		{
+			currency_manager = new CurrencyManager (this);
+
+			currency_manager.PositionChanged += delegate (object o, EventArgs args) { OnPositionChanged (args); };
+			currency_manager.CurrentChanged += delegate (object o, EventArgs args) { OnCurrentChanged (args); };
+			currency_manager.BindingComplete += delegate (object o, BindingCompleteEventArgs args) { OnBindingComplete (args); };
+			currency_manager.DataError += delegate (object o, BindingManagerDataErrorEventArgs args) { OnDataError (args); };
+			currency_manager.CurrentChanged += delegate (object o, EventArgs args) { OnCurrentChanged (args); };
+			currency_manager.CurrentItemChanged += delegate (object o, EventArgs args) { OnCurrentItemChanged (args); };
+		}
+
 		void ResetList ()
 		{
+			if (!is_initialized)
+				return;
+
 			IList l;
+			object source = ListBindingHelper.GetList (datasource, datamember);
+
+			// 
+			// If original source is null, then create a new object list
+			// Otherwise, try to infer the list item type
+			//
 
 			if (datasource == null) {
 				l = new BindingList<object>();
-				list_defaulted = true;
-			}
-			else if (datasource is IList) {
-				l = (IList)datasource;
-			}
-			else if (datasource is IEnumerable) {
-				IList new_list = GetListFromEnumerable ((IEnumerable)datasource);
-				l = new_list == null ? list : new_list;
-			}
-			else {
-				Type t = typeof (BindingList<>).MakeGenericType (new Type[] { datasource.GetType() });
+				//list_defaulted = true;
+			} else if (source == null) {
+				//Infer type based on datasource and datamember,
+				// where datasource is an empty IEnumerable
+				// and need to find out the datamember type
+
+				Type property_type = ListBindingHelper.GetListItemProperties (datasource) [datamember].PropertyType;
+				Type t = typeof (BindingList<>).MakeGenericType (new Type [] { property_type } );
 				l = (IList)Activator.CreateInstance (t);
+			} else if (source is IList) {
+				l = (IList)source;
+			} else if (source is IEnumerable) {
+				IList new_list = GetListFromEnumerable ((IEnumerable)source);
+				l = new_list == null ? list : new_list;
+			} else if (source is Type) {
+				Type t = typeof (BindingList<>).MakeGenericType (new Type [] { (Type)source });
+				l = (IList)Activator.CreateInstance (t);
+			} else {
+				Type t = typeof (BindingList<>).MakeGenericType (new Type[] { source.GetType() });
+				l = (IList)Activator.CreateInstance (t);
+				l.Add (source);
 			}
 
-			// XXX
+			SetList (l);
+		}
+
+		void SetList (IList l)
+		{
+			if (list is IBindingList)
+				((IBindingList) list).ListChanged -= IBindingListChangedHandler;
+
 			list = l;
+			item_type = ListBindingHelper.GetListItemType (list);
+			item_has_default_ctor = item_type.GetConstructor (Type.EmptyTypes) != null;
 
-			if (datamember != "") {
+			list_is_ibinding = list is IBindingList;
+			if (list_is_ibinding) {
+				((IBindingList) list).ListChanged += IBindingListChangedHandler;
+
+				if (list is IBindingListView)
+					((IBindingListView)list).Filter = filter;
 			}
+			
+			ResetBindings (true);
+		}
+
+		void IBindingListChangedHandler (object o, ListChangedEventArgs args)
+		{
+			if (raise_list_changed_events)
+				OnListChanged (args);
 		}
 
 		[Browsable (false)]
@@ -164,25 +221,32 @@ namespace System.Windows.Forms {
 				if (allow_new_set)
 					return allow_new;
 
-				if (list.IsFixedSize || list.IsReadOnly)
-					return false;
-
 				if (list is IBindingList)
 					return ((IBindingList)list).AllowNew;
 
-				// XXX we need to check the element
-				// type to see if it has a default
-				// constructor
+				if (list.IsFixedSize || list.IsReadOnly || !item_has_default_ctor)
+					return false;
+
 				return true;
 			}
 			set {
-				if (allow_new != value) {
-					allow_new_set = true;
-					allow_new = value;
+				if (value == allow_new && allow_new_set)
+					return;
 
-					if (raise_list_changed_events)
-						OnListChanged (new ListChangedEventArgs (ListChangedType.Reset, -1));
-				}
+				if (value && (list.IsReadOnly || list.IsFixedSize))
+					throw new InvalidOperationException ();
+
+				allow_new_set = true;
+				allow_new = value;
+
+				if (raise_list_changed_events)
+					OnListChanged (new ListChangedEventArgs (ListChangedType.Reset, -1));
+			}
+		}
+
+		bool IsAddingNewHandled {
+			get {
+				return Events [AddingNewEvent] != null;
 			}
 		}
 
@@ -205,20 +269,24 @@ namespace System.Windows.Forms {
 		[Browsable (false)]
 		public virtual int Count {
 			get {
-				if (list == null)
-					return -1;
 				return list.Count;
 			}
 		}
 
 		[Browsable (false)]
 		public virtual CurrencyManager CurrencyManager {
-			get { throw new NotImplementedException (); }
+			get {
+				return currency_manager;
+			}
 		}
 
 		[Browsable (false)]
 		public object Current {
-			get { throw new NotImplementedException (); }
+			get {
+				if (currency_manager.Count > 0)
+					return currency_manager.Current;
+				return null;
+			}
 		}
 
 		[DefaultValue ("")]
@@ -229,7 +297,7 @@ namespace System.Windows.Forms {
 			set {
 				/* we don't allow null DataMembers */
 				if (value == null)
-					value = "";
+					value = String.Empty;
 
 				if (datamember != value) {
 					this.datamember = value;
@@ -248,13 +316,12 @@ namespace System.Windows.Forms {
 			get { return datasource; }
 			set {
 				if (datasource != value) {
-					this.datasource = value;
-					datamember = "";
+					if (datasource == null)
+						datamember = String.Empty;
 
+					datasource = value;
 					ResetList ();
 
-					position = 0;
-
 					OnDataSourceChanged (EventArgs.Empty);
 				}
 			}
@@ -262,13 +329,20 @@ namespace System.Windows.Forms {
 
 		[DefaultValue (null)]
 		public virtual string Filter {
-			get { throw new NotImplementedException (); }
-			set { throw new NotImplementedException (); }
+			get {
+				return filter;
+			}
+			set {
+				if (SupportsFiltering)
+					((IBindingListView)list).Filter = value;
+
+				filter = value;
+			}
 		}
 
 		[Browsable (false)]
 		public bool IsBindingSuspended {
-			get { return suspended; }
+			get { return currency_manager.IsBindingSuspended; }
 		}
 
 		[Browsable (false)]
@@ -288,13 +362,18 @@ namespace System.Windows.Forms {
 
 		[Browsable (false)]
 		public virtual bool IsSynchronized {
-			get { throw new NotImplementedException (); }
+			get {
+				return list.IsSynchronized;
+			}
 		}
 
 		[Browsable (false)]
 		public virtual object this [int index] {
-			get { throw new NotImplementedException (); }
-			set { throw new NotImplementedException (); }
+			get { return list[index]; }
+			set
+			{
+				list[index] = value;
+			}
 		}
 
 		[Browsable (false)]
@@ -305,15 +384,14 @@ namespace System.Windows.Forms {
 		[DefaultValue (-1)]
 		[Browsable (false)]
 		public int Position {
-			get { return position; }
+			get {
+				return currency_manager.Position;
+			}
 			set {
 				if (value >= Count) value = Count - 1;
 				if (value < 0) value = 0;
 
-				if (position != value) {
-					position = value;
-					OnPositionChanged (EventArgs.Empty);
-				}
+				currency_manager.Position = value;
 			}
 		}
 
@@ -326,8 +404,72 @@ namespace System.Windows.Forms {
 
 		[DefaultValue (null)]
 		public string Sort {
-			get { throw new NotImplementedException (); }
-			set { throw new NotImplementedException (); }
+			get {
+				return sort;
+			}
+			set {
+				if (value == null || value.Length == 0) {
+					if (list_is_ibinding && SupportsSorting)
+						RemoveSort ();
+				
+					sort = value;
+					return;
+				}
+
+				if (!list_is_ibinding || !SupportsSorting)
+					throw new ArgumentException ("value");
+
+				ProcessSortString (value);
+				sort = value;
+			}
+		}
+
+		// NOTE: Probably the parsing can be improved
+		void ProcessSortString (string sort)
+		{
+			// Only keep simple whitespaces in the middle
+			sort = Regex.Replace (sort, "( )+", " ");
+
+			string [] properties = sort.Split (',');
+			PropertyDescriptorCollection prop_descs = GetItemProperties (null);
+			if (properties.Length == 1) {
+				ListSortDescription sort_desc = GetListSortDescription (prop_descs, properties [0]);
+				ApplySort (sort_desc.PropertyDescriptor, sort_desc.SortDirection);
+			} else {
+				if (!SupportsAdvancedSorting)
+					throw new ArgumentException ("value");
+
+				ListSortDescription [] sort_descs = new ListSortDescription [properties.Length];
+				for (int i = 0; i < properties.Length; i++)
+					sort_descs [i] = GetListSortDescription (prop_descs, properties [i]);
+
+				ApplySort (new ListSortDescriptionCollection (sort_descs));
+			}
+
+		}
+
+		ListSortDescription GetListSortDescription (PropertyDescriptorCollection prop_descs, string property)
+		{
+			property = property.Trim ();
+			string [] p = property.Split (new char [] {' '}, 2);
+
+			string prop_name = p [0];
+			PropertyDescriptor prop_desc = prop_descs [prop_name];
+			if (prop_desc == null)
+				throw new ArgumentException ("value");
+
+			ListSortDirection sort_direction = ListSortDirection.Ascending;
+			if (p.Length > 1) {
+				string dir_s = p [1];
+				if (String.Compare (dir_s, "ASC", true) == 0)
+					sort_direction = ListSortDirection.Ascending;
+				else if (String.Compare (dir_s, "DESC", true) == 0)
+					sort_direction = ListSortDirection.Descending;
+				else
+					throw new ArgumentException ("value");
+			}
+
+			return new ListSortDescription (prop_desc, sort_direction);
 		}
 
 		[Browsable (false)]
@@ -392,7 +534,9 @@ namespace System.Windows.Forms {
 
 		[Browsable (false)]
 		public virtual object SyncRoot {
-			get { throw new NotImplementedException (); }
+			get { 
+				return list.SyncRoot;
+			}
 		}
 
 		static object AddingNewEvent = new object ();
@@ -452,114 +596,191 @@ namespace System.Windows.Forms {
 
 		public virtual int Add (object value)
 		{
-			throw new NotImplementedException ();
+			// 
+			// First (re)create the BindingList<T> based on value
+			// if datasource is null and the current list is empty
+			//
+			if (datasource == null && list.Count == 0 && value != null) {
+				Type t = typeof (BindingList<>).MakeGenericType (new Type [] { value.GetType () } );
+				IList l = (IList) Activator.CreateInstance (t);
+				SetList (l);
+			}
+
+			if (value != null && !item_type.IsAssignableFrom (value.GetType ()))
+				throw new InvalidOperationException ("Objects added to the list must all be of the same type.");
+			if (list.IsReadOnly)
+				throw new NotSupportedException ("Collection is read-only.");
+			if (list.IsFixedSize)
+				throw new NotSupportedException ("Collection has a fixed size.");
+
+			int idx = list.Add (value);
+			if (raise_list_changed_events && !list_is_ibinding)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.ItemAdded, idx));
+
+			return idx;
 		}
 
 		public virtual object AddNew ()
 		{
-			if (list.IsFixedSize ||
-			    list.IsReadOnly ||
-			    (list is IBindingList && !((IBindingList)list).AllowNew))
-				throw new InvalidOperationException ("Item cannot be added to a read-onlyor fixed-size list.");
+			if (!AllowEdit)
+				throw new InvalidOperationException ("Item cannot be added to a read-only or fixed-size list.");
+			if (!AllowNew) 
+				throw new InvalidOperationException ("AddNew is set to false.");
 
 			EndEdit ();
 
 			AddingNewEventArgs args = new AddingNewEventArgs ();
 			OnAddingNew (args);
 
-			if (args.NewObject != null) {
-				// XXX verify the type here and make sure it matches our internal list
-				list.Add (args.NewObject);
-				return args.NewObject;
-			}
+			object new_object = args.NewObject;
+			if (new_object != null) {
+				if (!item_type.IsAssignableFrom (new_object.GetType ()))
+					throw new InvalidOperationException ("Objects added to the list must all be of the same type.");
+			} else if (list is IBindingList) {
+				object newObj = ((IBindingList)list).AddNew ();
+				add_pending = true;
+				pending_add_index = list.IndexOf (newObj);
+				return newObj;
+			} else if (!item_has_default_ctor)
+				throw new InvalidOperationException ("AddNew cannot be called on '" + item_type.Name +
+						", since it does not have a public default ctor. Set AllowNew to true " +
+						", handling AddingNew and creating the appropriate object.");
+			else // fallback to default .ctor
+				new_object = Activator.CreateInstance (item_type);
 
-			// if the list is a IBindingList, try to use IBindingList.AddNew
-			if (list is IBindingList)
-				return ((IBindingList)list).AddNew ();
+			int idx = list.Add (new_object);
+			if (raise_list_changed_events && !list_is_ibinding)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.ItemAdded, idx));
 
-			throw new NotImplementedException ();
+			add_pending = true;
+			pending_add_index = idx;
+
+			return new_object;
 		}
 
 		[EditorBrowsable (EditorBrowsableState.Never)]
 		public virtual void ApplySort (PropertyDescriptor property, ListSortDirection sort)
 		{
-			throw new NotImplementedException ();
+			if (!list_is_ibinding)
+				throw new NotSupportedException ("This operation requires an IBindingList.");
+
+			IBindingList iblist = (IBindingList)list;
+			iblist.ApplySort (property, sort);
 		}
 
 		[EditorBrowsable (EditorBrowsableState.Never)]
 		public virtual void ApplySort (ListSortDescriptionCollection sorts)
 		{
-			throw new NotImplementedException ();
+			if (!(list is IBindingListView))
+				throw new NotSupportedException ("This operation requires an IBindingListView.");
+
+			IBindingListView iblist_view = (IBindingListView)list;
+			iblist_view.ApplySort (sorts);
 		}
 
 		public void CancelEdit ()
 		{
-			throw new NotImplementedException ();
+			currency_manager.CancelCurrentEdit ();
 		}
 
 		public virtual void Clear ()
 		{
-			throw new NotImplementedException ();
+			if (list.IsReadOnly)
+				throw new NotSupportedException ("Collection is read-only.");
+
+			list.Clear ();
+			if (raise_list_changed_events && !list_is_ibinding)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.Reset, -1));
 		}
 
 		public virtual bool Contains (object value)
 		{
-			throw new NotImplementedException ();
+			return list.Contains (value);
 		}
 
 		public virtual void CopyTo (Array arr, int index)
 		{
-			throw new NotImplementedException ();
+			list.CopyTo (arr, index);
 		}
 
 		protected override void Dispose (bool disposing)
 		{
-			throw new NotImplementedException ();
+			base.Dispose (disposing);
 		}
 
 		public void EndEdit ()
 		{
-			throw new NotImplementedException ();
+			currency_manager.EndCurrentEdit ();
 		}
 
 		public int Find (string propertyName, object key)
 		{
-			throw new NotImplementedException ();
+			PropertyDescriptor property = GetItemProperties (null).Find (propertyName, true);
+			if (property == null)
+				throw new ArgumentException ("propertyName");
+
+			return Find (property, key);
 		}
 
 		public virtual int Find (PropertyDescriptor prop, object key)
 		{
-			throw new NotImplementedException ();
+			if (!list_is_ibinding)
+				throw new NotSupportedException ();
+
+			return ((IBindingList)list).Find (prop, key);
 		}
 
 		public virtual IEnumerator GetEnumerator ()
 		{
-			throw new NotImplementedException ();
+			return this.List.GetEnumerator ();
 		}
 
 		public virtual PropertyDescriptorCollection GetItemProperties (PropertyDescriptor[] listAccessors)
 		{
-			throw new NotImplementedException ();
+			return ListBindingHelper.GetListItemProperties (list, listAccessors);
 		}
 
 		public virtual string GetListName (PropertyDescriptor[] listAccessors)
 		{
-			throw new NotImplementedException ();
+			return ListBindingHelper.GetListName (list, listAccessors);
 		}
 
-		public virtual CurrencyManager GetRelatedCurrencyManager (string dataMamber)
+		public virtual CurrencyManager GetRelatedCurrencyManager (string dataMember)
 		{
-			throw new NotImplementedException ();
+			if (dataMember == null || dataMember.Length == 0)
+				return currency_manager;
+
+			if (related_currency_managers.ContainsKey (dataMember))
+				return related_currency_managers [dataMember];
+
+			// FIXME - Why passing invalid dataMembers containing a . return
+			// a null value?
+			if (dataMember.IndexOf ('.') != -1)
+				return null;
+
+			BindingSource source = new BindingSource (this, dataMember);
+			related_currency_managers [dataMember] = source.CurrencyManager;
+
+			return source.CurrencyManager;
 		}
 
 		public virtual int IndexOf (object value)
 		{
-			throw new NotImplementedException ();
+			return list.IndexOf (value);
 		}
 
 		public virtual void Insert (int index, object value)
 		{
-			throw new NotImplementedException ();
+			if (index < 0 || index > list.Count)
+				throw new ArgumentOutOfRangeException ("index");
+			if (list.IsReadOnly || list.IsFixedSize)
+				throw new NotSupportedException ();
+			if (!item_type.IsAssignableFrom (value.GetType ()))
+				throw new ArgumentException ("value");
+
+			list.Insert (index, value);
+			if (raise_list_changed_events && !list_is_ibinding)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.ItemAdded, index));
 		}
 
 		public void MoveFirst ()
@@ -647,27 +868,52 @@ namespace System.Windows.Forms {
 
 		public virtual void Remove (object value)
 		{
-			throw new NotImplementedException ();
+			if (list.IsReadOnly)
+				throw new NotSupportedException ("Collection is read-only.");
+			if (list.IsFixedSize)
+				throw new NotSupportedException ("Collection has a fixed size.");
+
+			int idx = list_is_ibinding ? - 1 : list.IndexOf (value);
+			list.Remove (value);
+
+			if (idx != -1 && raise_list_changed_events)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.ItemDeleted, idx));
 		}
 
 		public virtual void RemoveAt (int index)
 		{
-			throw new NotImplementedException ();
+			if (index < 0 || index > list.Count)
+				throw new ArgumentOutOfRangeException ("index");
+			if (list.IsReadOnly || list.IsFixedSize)
+				throw new InvalidOperationException ();
+
+			list.RemoveAt (index);
+			if (raise_list_changed_events && !list_is_ibinding)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.ItemDeleted, index));
 		}
 
 		public void RemoveCurrent ()
 		{
-			throw new NotImplementedException ();
+			if (Position < 0)
+				throw new InvalidOperationException ("Cannot remove item because there is no current item.");
+			if (!AllowRemove)
+				throw new InvalidOperationException ("Cannot remove item because list does not allow removal of items.");
+
+			RemoveAt (Position);
 		}
 
 		public virtual void RemoveFilter ()
 		{
-			throw new NotImplementedException ();
+			Filter = null;
 		}
 
 		public virtual void RemoveSort ()
 		{
-			throw new NotImplementedException ();
+			if (!list_is_ibinding)
+				return;
+
+			sort = null;
+			((IBindingList)list).RemoveSort ();
 		}
 
 		[EditorBrowsable (EditorBrowsableState.Advanced)]
@@ -696,34 +942,71 @@ namespace System.Windows.Forms {
 
 		public void ResumeBinding ()
 		{
-			suspended = false;
-			throw new NotImplementedException ();
+			currency_manager.ResumeBinding ();
 		}
 
 		public void SuspendBinding ()
 		{
-			suspended = true;
-			throw new NotImplementedException ();
+			currency_manager.SuspendBinding ();
 		}
 
 		/* explicit interface implementations */
 
-		void ICancelAddNew.CancelNew (int itemIndex)
+		void ICancelAddNew.CancelNew (int position)
 		{
-			throw new NotImplementedException ();
+			if (!add_pending)
+				return;
+
+			if (position != pending_add_index)
+				return;
+
+			add_pending = false;
+			list.RemoveAt (position);
+
+			if (raise_list_changed_events && !list_is_ibinding)
+				OnListChanged (new ListChangedEventArgs (ListChangedType.ItemDeleted, position));
 		}
 
-		void ICancelAddNew.EndNew (int itemIndex)
+		void ICancelAddNew.EndNew (int position)
 		{
-			throw new NotImplementedException ();
+			if (!add_pending)
+				return;
+
+			if (position != pending_add_index)
+				return;
+
+			add_pending = false;
 		}
 		
 		void ISupportInitialize.BeginInit ()
 		{
+			is_initialized = false;
+		}
+
+		void DataSourceEndInitHandler (object o, EventArgs args)
+		{
+			((ISupportInitializeNotification)datasource).Initialized -= DataSourceEndInitHandler;
+
+			ISupportInitializeNotification inotif = (ISupportInitializeNotification)this;
+			inotif.EndInit ();
 		}
 
 		void ISupportInitialize.EndInit ()
 		{
+			if (datasource != null && datasource is ISupportInitializeNotification) {
+				ISupportInitializeNotification inotif = (ISupportInitializeNotification)datasource;
+				if (!inotif.IsInitialized) {
+					inotif.Initialized += DataSourceEndInitHandler;
+					return;
+				}
+			}
+
+			is_initialized = true;
+			ResetList ();
+
+			EventHandler eh = (EventHandler) Events [InitializedEvent];
+			if (eh != null)
+				eh (this, EventArgs.Empty);
 		}
 
 		void IBindingList.AddIndex (PropertyDescriptor property)
@@ -734,18 +1017,18 @@ namespace System.Windows.Forms {
 			((IBindingList)list).AddIndex (property);
 		}
 
-		void IBindingList.RemoveIndex (PropertyDescriptor property)
+		void IBindingList.RemoveIndex (PropertyDescriptor prop)
 		{
 			if (!(list is IBindingList))
 				throw new NotSupportedException();
 
-			((IBindingList)list).RemoveIndex (property);
+			((IBindingList)list).RemoveIndex (prop);
 		}
 
 		bool ISupportInitializeNotification.IsInitialized {
-			// XXX this is likely wrong, but i can't
-			// figure out how to make it return false
-			get { return true; }
+			get { 
+				return is_initialized;
+			}
 		}
 
 		static object InitializedEvent = new object ();
@@ -754,23 +1037,7 @@ namespace System.Windows.Forms {
 			add { Events.AddHandler (InitializedEvent, value); }
 			remove { Events.RemoveHandler (InitializedEvent, value); }
 		}
-
-		ISite IComponent.Site {
-			get { return site; }
-			set { site = value; }
-		}
-
-		event EventHandler IComponent.Disposed {
-			add { throw new NotImplementedException (); }
-			remove { throw new NotImplementedException (); }
-		}
-
-		void IDisposable.Dispose ()
-		{
-			throw new NotImplementedException ();
-		}
 	}
-
 }
 
 #endif
