@@ -98,6 +98,7 @@ namespace System.Windows.Forms {
 		private Bitmap		mask;
 		private Bitmap		cursor;
 		internal string		name;
+		private StdCursor	std_cursor = (StdCursor) (-1);
 
 #if NET_2_0
 		private object tag;
@@ -119,7 +120,11 @@ namespace System.Windows.Forms {
 				this.cursor = ToBitmap(true, true);
 			}
 		}
-
+	
+		internal Cursor (StdCursor cursor) : this (XplatUI.DefineStdCursor (cursor)) {
+			std_cursor = cursor;
+		}
+		
 		private Cursor(SerializationInfo info, StreamingContext context) {
 		}
 
@@ -139,7 +144,11 @@ namespace System.Windows.Forms {
 			CreateCursor(stream);
 		}
 
-		public Cursor(string fileName) : this (new FileStream (fileName, FileMode.Open)) {
+		public Cursor (string fileName)
+		{
+			using (FileStream fs = File.OpenRead (fileName)) {
+				CreateCursor (fs);
+			}
 		}
 
 		public Cursor(Type type, string resource) {
@@ -188,7 +197,6 @@ namespace System.Windows.Forms {
 			}
 		}
 
-		[MonoTODO("Implement setting a null cursor, and add XplatUI method to get current cursor")]
 		public static Cursor Current {
 			get {
 				if (current != null) {
@@ -200,7 +208,7 @@ namespace System.Windows.Forms {
 			set {
 				if (current != value) {
 					current = value;
-					if (current == null || current == Cursors.Default){
+					if (current == null){
 						// FIXME - define and set empty cursor
 						XplatUI.OverrideCursor(IntPtr.Zero);
 					} else
@@ -231,6 +239,17 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[MonoTODO ("Implemented for Win32, X11 always returns 0,0")]
+		public Point HotSpot {
+			get {
+				int cursor_w, cursor_h, hot_x, hot_y;
+				XplatUI.GetCursorInfo (Handle, out cursor_w, out cursor_h, out hot_x, out hot_y);
+
+				return new Point (hot_x, hot_y);
+			}
+		}
+#endif
 		public Size Size {
 			get {
 				return size;
@@ -256,7 +275,7 @@ namespace System.Windows.Forms {
 		}
 
 		public static void Show() {
-			XplatUI.ShowCursor(false);
+			XplatUI.ShowCursor(true);
 		}
 
 		public static bool operator !=(Cursor left, Cursor right) {
@@ -316,12 +335,20 @@ namespace System.Windows.Forms {
 		}
 
 		public void Draw(Graphics g, Rectangle targetRect) {
+			if (this.cursor == null && std_cursor != (StdCursor) (-1)) {
+				this.cursor = XplatUI.DefineStdCursorBitmap (std_cursor);
+			}
 			if (this.cursor != null) {
-				g.DrawImage(this.cursor, targetRect);
+				// Size of the targetRect is not considered at all
+				g.DrawImage (this.cursor, targetRect.X, targetRect.Y);
 			}
 		}
 
-		public void DrawStretched(Graphics g, Rectangle targetRect) {
+		public void DrawStretched (Graphics g, Rectangle targetRect)
+		{
+			if (this.cursor == null && std_cursor != (StdCursor)(-1)) {
+				this.cursor = XplatUI.DefineStdCursorBitmap (std_cursor);
+			}
 			if (this.cursor != null) {
 				g.DrawImage(this.cursor, targetRect, new Rectangle(0, 0, this.cursor.Width, this.cursor.Height), GraphicsUnit.Pixel);
 			}
