@@ -31,6 +31,7 @@ using System;
 using System.Drawing;
 using System.Collections;
 using System.ComponentModel;
+using System.Globalization;
 using System.Runtime.InteropServices;
 using System.Windows.Forms;
 
@@ -96,7 +97,11 @@ namespace System.Windows.Forms {
 
 		// variables for determining how to format the string
 		internal PartData[]					part_data;
+		internal int						editing_part_index = -1;
+		internal int 						editing_number = -1;
+		internal string						editing_text;
 
+		bool drop_down_button_entered;
 		#endregion	// Local variables
 		
 		#region DateTimePickerAccessibleObject Subclass
@@ -190,9 +195,13 @@ namespace System.Windows.Forms {
 			updown_timer.Tick += new EventHandler (UpDownTimerTick);
 			KeyPress += new KeyPressEventHandler (KeyPressHandler);
 			KeyDown += new KeyEventHandler (KeyDownHandler);
+			GotFocus += new EventHandler (GotFocusHandler);
 			LostFocus += new EventHandler (LostFocusHandler);
 			MouseDown += new MouseEventHandler (MouseDownHandler);			
 			MouseUp += new MouseEventHandler (MouseUpHandler);
+			MouseEnter += new EventHandler (OnMouseEnter);
+			MouseLeave += new EventHandler (OnMouseLeave);
+			MouseMove += new MouseEventHandler (OnMouseMove);
 			Paint += new PaintEventHandler (PaintHandler);
 			Resize += new EventHandler (ResizeHandler);
 			SetStyle (ControlStyles.UserPaint | ControlStyles.StandardClick, false);
@@ -307,8 +316,10 @@ namespace System.Windows.Forms {
 					// invalidate the value inside this control
 					if (ShowCheckBox) {
 						for (int i = 0; i < part_data.Length; i++)
-							part_data [i].is_selected = false;
+							part_data [i].Selected = false;
 						Invalidate (date_area_rect);
+						OnUIAChecked ();
+						OnUIASelectionChanged ();
 					}
 				}
 			}
@@ -395,20 +406,34 @@ namespace System.Windows.Forms {
 		public DateTime MaxDate {
 			set {
 				if (value < min_date) {
-					throw new ArgumentException ();
+					string msg = string.Format (CultureInfo.CurrentCulture,
+						"'{0}' is not a valid value for 'MaxDate'. 'MaxDate' "
+						+ "must be greater than or equal to MinDate.",
+						value.ToString ("G"));
+#if NET_2_0
+					throw new ArgumentOutOfRangeException ("MaxDate", msg);
+#else
+					throw new ArgumentException (msg);
+#endif
 				}
 				if (value > MaxDateTime) {
-					throw new SystemException ();
+					string msg = string.Format (CultureInfo.CurrentCulture,
+						"DateTimePicker does not support dates after {0}.",
+						MaxDateTime.ToString ("G", CultureInfo.CurrentCulture));
+#if NET_2_0
+					throw new ArgumentOutOfRangeException ("MaxDate", msg);
+#else
+					throw new ArgumentException (msg, "value");
+#endif
 				}
 				if (max_date != value) {
 					max_date = value;
-					
-					// TODO: verify this is correct behaviour when value > max date
 					if (Value > max_date) {
 						Value = max_date;
 						// invalidate the value inside this control
 						this.Invalidate (date_area_rect);
 					}
+					OnUIAMaximumChanged ();
 				}
 			}
 			get {
@@ -426,21 +451,45 @@ namespace System.Windows.Forms {
 		
 		public DateTime MinDate {
 			set {
-				if (value < min_date) {
-					throw new ArgumentException ();
+				// If the user tries to set DateTime.MinValue, fix it to
+				// DateTimePicker's minimum.
+				if (value == DateTime.MinValue)
+					value = MinDateTime;
+					
+#if NET_2_0
+				if (value > MaxDate) {
+#else
+				if (value >= MaxDate) {
+#endif
+					string msg = string.Format (CultureInfo.CurrentCulture,
+						"'{0}' is not a valid value for 'MinDate'. 'MinDate' "
+						+ "must be less than MaxDate.",
+						value.ToString ("G"));
+#if NET_2_0
+					throw new ArgumentOutOfRangeException ("MinDate", msg);
+#else
+					throw new ArgumentException (msg);
+#endif
 				}
 				if (value < MinDateTime) {
-					throw new SystemException ();
+					string msg = string.Format (CultureInfo.CurrentCulture,
+						"DateTimePicker does not support dates before {0}.",
+						MinDateTime.ToString ("G", CultureInfo.CurrentCulture));
+#if NET_2_0
+					throw new ArgumentOutOfRangeException ("MinDate", msg);
+#else
+					throw new ArgumentException (msg, "value");
+#endif
+
 				}
 				if (min_date != value) {
 					min_date = value;
-					
-					// TODO: verify this is correct behaviour when value > max date
 					if (Value < min_date) {
 						Value = min_date;
 						// invalidate the value inside this control
 						this.Invalidate (date_area_rect);
 					}
+					OnUIAMinimumChanged ();
 				}
 			}
 			get {
@@ -498,6 +547,7 @@ namespace System.Windows.Forms {
 					show_check_box = value;
 					// invalidate the value inside this control
 					this.Invalidate (date_area_rect);
+					OnUIAShowCheckBoxChanged ();
 				}
 			}
 			get {
@@ -513,6 +563,7 @@ namespace System.Windows.Forms {
 					show_up_down = value;
 					// need to invalidate the whole control
 					this.Invalidate ();
+					OnUIAShowUpDownChanged ();
 				}
 			}
 			get {
@@ -569,6 +620,13 @@ namespace System.Windows.Forms {
 		public DateTime Value {
 			set {
 				if (date_value != value) {
+					if (value < MinDate || value > MaxDate)
+#if NET_2_0
+						throw new ArgumentOutOfRangeException ("value", "value must be between MinDate and MaxDate");
+#else
+						throw new ArgumentException ("value", "value must be between MinDate and MaxDate");
+#endif
+
 					date_value = value;
 					this.OnValueChanged (EventArgs.Empty);
 					this.Invalidate (date_area_rect);
@@ -878,24 +936,20 @@ namespace System.Windows.Forms {
 			if (eh != null)
 				eh (this, eventargs);
 		}
-		
+
+#if !NET_2_0
 		// overridden to set the bounds of this control properly
 		protected override void SetBoundsCore(int x, int y, int width, int height, BoundsSpecified specified) {
-			// TODO: ensure I implemented the bounds core setting properly.
-			if ((specified & BoundsSpecified.Height) == BoundsSpecified.Height ||
-				(specified & BoundsSpecified.Size) == BoundsSpecified.Size)  {
-				base.SetBoundsCore (x, y, width, DefaultSize.Height, specified);
-			} else {
-				base.SetBoundsCore (x, y, width, height, specified);
-			}
-			
-			// need to set the rectangles for all the support internal rects
-			// this is done here as a optimisation since this is an array of rects
-			if ((specified & BoundsSpecified.X) == BoundsSpecified.X ||
-				(specified & BoundsSpecified.Y) == BoundsSpecified.Y) {
-				// TODO set up all the datepart rects
-			}
+			base.SetBoundsCore (x, y, width, DefaultSize.Height, specified);
 		}
+#else
+		// SetBoundsCore was removed from the 2.0 public API, so
+		// I had to do this hack instead.  :/
+		internal override int OverrideHeight (int height)
+		{
+			return DefaultSize.Height;
+		}
+#endif
 
 		// not sure why we're overriding this
 		protected override void WndProc (ref Message m) {
@@ -909,26 +963,7 @@ namespace System.Windows.Forms {
 		// this is the region that the date and the check box is drawn on
 		internal Rectangle date_area_rect {
 			get {
-				Rectangle rect = this.ClientRectangle;
-				if (ShowUpDown) {
-					// set the space to the left of the up/down button
-					if (rect.Width > (up_down_width + 4)) {
-						rect.Width -= (up_down_width + 4);
-					} else {
-						rect.Width = 0;
-					}
-				} else {
-					// set the space to the left of the up/down button
-					// TODO make this use up down button
-					if (rect.Width > (SystemInformation.VerticalScrollBarWidth + 4)) {
-						rect.Width -= SystemInformation.VerticalScrollBarWidth;
-					} else {
-						rect.Width = 0;
-					}
-				}
-				
-				rect.Inflate (-2, -2);
-				return rect;
+				return ThemeEngine.Current.DateTimePickerGetDateArea (this);
 			}
 		}
 
@@ -943,16 +978,7 @@ namespace System.Windows.Forms {
 		// the rectangle for the drop down arrow
 		internal Rectangle drop_down_arrow_rect {
 			get {
-				Rectangle rect = this.ClientRectangle;
-				rect.X = rect.Right - SystemInformation.VerticalScrollBarWidth - 2;
-				if (rect.Width > (SystemInformation.VerticalScrollBarWidth + 2)) {
-					rect.Width = SystemInformation.VerticalScrollBarWidth;
-				} else {
-					rect.Width = Math.Max (rect.Width - 2, 0);
-				}
-				
-				rect.Inflate (0, -2);
-				return rect;
+				return ThemeEngine.Current.DateTimePickerGetDropDownButtonArea (this);
 			}
 		}
 		
@@ -963,6 +989,10 @@ namespace System.Windows.Forms {
 				return Rectangle.Empty;
 			}
 		}	
+
+		internal bool DropDownButtonEntered {
+			get { return drop_down_button_entered; }
+		}
 			
 		#endregion
 		
@@ -1125,7 +1155,7 @@ namespace System.Windows.Forms {
 					case 'g': // Spec says nothing about g, but it seems to be treated like spaces.
 						if (!(lastch == ch || lastch == 0) && literal.Length != 0)
 						{
-							formats.Add (new PartData(literal.ToString (), false));
+							formats.Add (new PartData(literal.ToString (), false, this));
 							literal.Length = 0;
 						}
 						literal.Append (ch);
@@ -1140,24 +1170,24 @@ namespace System.Windows.Forms {
 							is_literal = !is_literal;
 							break;
 						}
-						formats.Add (new PartData (literal.ToString (), is_literal));
+						formats.Add (new PartData (literal.ToString (), is_literal, this));
 						literal.Length = 0;
 						is_literal = !is_literal;
 						break;
 					default:
 						if (literal.Length != 0)
 						{
-							formats.Add (new PartData(literal.ToString (), false));
+							formats.Add (new PartData(literal.ToString (), false, this));
 							literal.Length = 0;
 						}
-						formats.Add (new PartData (ch.ToString(), true));
+						formats.Add (new PartData (ch.ToString(), true, this));
 						break;
 
 				}
 				lastch = ch;
 			}
 			if (literal.Length >= 0)
-				formats.Add (new PartData (literal.ToString (), is_literal));
+				formats.Add (new PartData (literal.ToString (), is_literal, this));
 
 			part_data = new PartData [formats.Count];
 			formats.CopyTo (part_data);
@@ -1184,7 +1214,9 @@ namespace System.Windows.Forms {
 			}
 
 			// since the parent of the month calendar is the form, adjust accordingly.
-			screen_location = month_calendar.Parent.PointToClient(screen_location);
+			if (month_calendar.Parent != null) {
+				screen_location = month_calendar.Parent.PointToClient(screen_location);
+			}
 
 			return screen_location;
 		}
@@ -1198,6 +1230,7 @@ namespace System.Windows.Forms {
 		// drop the calendar down
 		internal void DropDownMonthCalendar ()
 		{
+			EndDateEdit (true);
 			// ensure the right date is set for the month_calendar
 			month_calendar.SetDate (this.date_value);
 			// get a rectangle that has the dimensions of the text area,
@@ -1206,7 +1239,6 @@ namespace System.Windows.Forms {
 			align_area.Y = this.ClientRectangle.Y;
 			align_area.Height = this.ClientRectangle.Height;
 
-			month_calendar.Parent = this.FindForm ();
 			// establish the month calendar's location
 			month_calendar.Location = CalculateDropDownLocation (
 				align_area,
@@ -1238,74 +1270,78 @@ namespace System.Windows.Forms {
 		{
 			for (int i = 0; i < part_data.Length; i++)
 			{
-				if (part_data[i].is_selected && !part_data[i].is_literal)
+				if (part_data[i].Selected && !part_data[i].is_literal)
 					return i;
 			}
 			return -1;
 		}
 
-		private void IncrementSelectedPart(int delta)
+		internal void IncrementSelectedPart(int delta)
 		{
 			int selected_index = GetSelectedPartIndex();
 
 			if (selected_index == -1) {
 				return;
 			}
-			
-			switch (part_data[selected_index].value)
+
+			EndDateEdit (false);
+
+			DateTimePart dt_part = part_data [selected_index].date_time_part;
+			switch (dt_part)
 			{
-				case "d":
-				case "dd": // number day formats
+				case DateTimePart.Day:
 					if (delta < 0) {
 						if (Value.Day == 1)
-							SetPart(DateTime.DaysInMonth(Value.Year, Value.Month), 'd');
+							SetPart(DateTime.DaysInMonth(Value.Year, Value.Month), dt_part);
 						else
-							SetPart(Value.Day + delta, 'd');
+							SetPart(Value.Day + delta, dt_part);
 					} else {
 						if (Value.Day == DateTime.DaysInMonth(Value.Year, Value.Month))
-							SetPart(1, 'd');
+							SetPart(1, dt_part);
 						else
-							SetPart(Value.Day + delta, 'd') ;
+							SetPart(Value.Day + delta, dt_part) ;
 					}
 					break;
-				case "ddd":
-				case "dddd": // text day formats
+				case DateTimePart.DayName:
 					Value = Value.AddDays(delta);
 					break;
-				case "h":
-				case "hh":
-				case "H":
-				case "HH": // hour formats
-					SetPart(Value.Hour + delta, 'h');
+				case DateTimePart.AMPMHour:
+				case DateTimePart.Hour:
+					SetPart(Value.Hour + delta, dt_part);
 					break;
-				case "m":
-				case "mm": // minute formats
-					SetPart(Value.Minute + delta, 'm');
+				case DateTimePart.Minutes:
+					SetPart(Value.Minute + delta, dt_part);
 					break;
-				case "M":
-				case "MM":
-				case "MMM":
-				case "MMMM": // month formats
-					SetPart(Value.Month + delta, 'M');
+				case DateTimePart.Month:
+					SetPart (Value.Month + delta, dt_part, true);
 					break;
-				case "s":
-				case "ss": // second format
-					SetPart(Value.Second + delta, 's');
+				case DateTimePart.Seconds:
+					SetPart(Value.Second + delta, dt_part);
 					break;
-				case "t":
-				case "tt": // AM / PM specifier
-					SetPart(Value.Hour + delta * 12, 'h');
+				case DateTimePart.AMPMSpecifier:
+					int hour = Value.Hour;
+					hour = hour >= 0 && hour <= 11 ? hour + 12 : hour - 12;
+					SetPart (hour, DateTimePart.Hour);
 					break;
-				case "y":
-				case "yy":
-				case "yyy":
-				case "yyyy":
-					SetPart(Value.Year + delta, 'y');
+				case DateTimePart.Year:
+					SetPart(Value.Year + delta, dt_part);
 					break;
 			}
 		}
+		
+		internal void SelectPart (int index)
+		{
+			is_checkbox_selected = false;
+			for (int i = 0; i < part_data.Length; i++)
+			{
+				part_data[i].Selected = (i == index);
+			}
 
-		private void SelectNextPart()
+			Invalidate ();
+			OnUIASelectionChanged ();
+		}
+
+		internal void SelectNextPart()
 		{
 			int selected_index;
 			if (is_checkbox_selected) {
@@ -1314,7 +1350,7 @@ namespace System.Windows.Forms {
 					if (!part_data[i].is_literal)
 					{
 						is_checkbox_selected = false;
-						part_data[i].is_selected = true;
+						part_data[i].Selected = true;
 						Invalidate();
 						break;
 					}
@@ -1322,12 +1358,13 @@ namespace System.Windows.Forms {
 			} else {
 				selected_index = GetSelectedPartIndex();
 				if (selected_index >= 0)
-					part_data[selected_index].is_selected = false;
+					part_data [selected_index].Selected = false;
+
 				for (int i = selected_index + 1; i < part_data.Length; i++)
 				{
 					if (!part_data[i].is_literal)
 					{
-						part_data[i].is_selected = true;
+						part_data [i].Selected = true;
 						Invalidate();
 						break;
 					}
@@ -1345,7 +1382,7 @@ namespace System.Windows.Forms {
 						{
 							if (!part_data[i].is_literal)
 							{
-								part_data[i].is_selected = true;
+								part_data[i].Selected = true;
 								Invalidate();
 								break;
 							}
@@ -1354,9 +1391,10 @@ namespace System.Windows.Forms {
 				}
 			}
 
+			OnUIASelectionChanged ();
 		}
 
-		private void SelectPreviousPart()
+		internal void SelectPreviousPart()
 		{
 			if (is_checkbox_selected)
 			{
@@ -1365,7 +1403,7 @@ namespace System.Windows.Forms {
 					if (!part_data[i].is_literal)
 					{
 						is_checkbox_selected = false;
-						part_data[i].is_selected = true;
+						part_data[i].Selected = true;
 						Invalidate();
 						break;
 					}
@@ -1376,13 +1414,13 @@ namespace System.Windows.Forms {
 				int selected_index = GetSelectedPartIndex();
 
 				if (selected_index >= 0)
-					part_data[selected_index].is_selected = false;
+					part_data[selected_index].Selected = false;
 
 				for (int i = selected_index - 1; i >= 0; i--)
 				{
 					if (!part_data[i].is_literal)
 					{
-						part_data[i].is_selected = true;
+						part_data[i].Selected = true;
 						Invalidate();
 						break;
 					}
@@ -1400,7 +1438,7 @@ namespace System.Windows.Forms {
 						{
 							if (!part_data[i].is_literal)
 							{
-								part_data[i].is_selected = true;
+								part_data[i].Selected = true;
 								Invalidate();
 								break;
 							}
@@ -1408,6 +1446,8 @@ namespace System.Windows.Forms {
 					}
 				}
 			}
+
+			OnUIASelectionChanged ();
 		}
 
 		// raised by key down events.
@@ -1450,8 +1490,11 @@ namespace System.Windows.Forms {
 						break;
 					}
 				case Keys.F4:
-					if (!is_drop_down_visible)
-						DropDownMonthCalendar();
+					if (!e.Alt && !is_drop_down_visible) {
+						DropDownMonthCalendar ();
+						e.Handled = true;
+					}
+
 					break;
 			}
 		}
@@ -1461,10 +1504,8 @@ namespace System.Windows.Forms {
 		{
 			switch (e.KeyChar) {
 				case ' ':
-					if (is_checkbox_selected)
-					{
+					if (show_check_box && is_checkbox_selected)
 						Checked = !Checked;
-					}
 					break;
 				case '0':
 				case '1':
@@ -1482,55 +1523,36 @@ namespace System.Windows.Forms {
 						break;
 					if (!part_data[selected_index].is_numeric_format)
 						break;
-					switch (part_data[selected_index].value)
-					{
-						case "d":
-						case "dd":
-							int newDay = Value.Day * 10 + number;
-							if (DateTime.DaysInMonth(Value.Year, Value.Month) < newDay)
-								newDay = number;
-							SetPart(newDay, 'd');
-							break;
-						case "M":
-						case "MM":
-							int newMonth = Value.Month * 10 + number;
-							if (newMonth > 12)
-								newMonth = number;
-							SetPart(newMonth, 'M');
-							break;
-						case "y":
-						case "yy":
-						case "yyyy":
-							int newYear = Value.Year * 10 + number;
-							if (newYear > 9999)
-								newYear = number;
-							SetPart(newYear, 'y');
-							break;
-						case "h":
-						case "hh":
-						case "H":
-						case "HH":
-							int newHour = Value.Hour * 10 + number;
-							if (newHour >= 24)
-								newHour = number;
-							SetPart(newHour, 'h');
-							break;
-						case "m":
-						case "mm":
-							int newMinute = Value.Minute* 10 + number;
-							if (newMinute >= 60)
-								newMinute = number;
-							SetPart(newMinute, 'm');
-							break;
-						case "s":
-						case "ss":
-							int newSecond = Value.Second * 10 + number;
-							if (newSecond >= 60)
-								newSecond = number;
-							SetPart(newSecond, 's');
-							break;
 
+					DateTimePart dt_part = part_data [selected_index].date_time_part;
+					if (editing_part_index < 0) {
+						editing_part_index = selected_index;
+						editing_number = 0;
+						editing_text = String.Empty;
 					}
+
+					editing_text += number.ToString ();
+					int date_part_max_length = 0;
+
+					switch (dt_part) {
+						case DateTimePart.Day:
+						case DateTimePart.Month:
+						case DateTimePart.Seconds:
+						case DateTimePart.Minutes:
+						case DateTimePart.AMPMHour:
+						case DateTimePart.Hour:
+							date_part_max_length = 2;
+							break;
+						case DateTimePart.Year:
+							date_part_max_length = 4;
+							break;
+					}
+
+					editing_number = editing_number * 10 + number;
+					if (editing_text.Length >= date_part_max_length)
+						EndDateEdit (false);
+
+					Invalidate (date_area_rect);
 					break;
 				default:
 					break;
@@ -1538,65 +1560,128 @@ namespace System.Windows.Forms {
 			e.Handled = true;
 		}
 
-		// set the specified part of the date to the specified value
-		private void SetPart(int value, char part)
+		private void EndDateEdit (bool invalidate)
 		{
-			switch (part)
+			if (editing_part_index == -1)
+				return;
+
+			PartData part = part_data [editing_part_index];
+			if (part.date_time_part == DateTimePart.Year) { // Special case
+				// Infer, like .Net does
+				if (editing_number > 0 && editing_number < 30)
+					editing_number += 2000;
+				else if (editing_number >= 30 && editing_number < 100)
+					editing_number += 1900;
+			}
+
+			SetPart (editing_number, part.date_time_part);
+			editing_part_index = editing_number = -1;
+			editing_text = null;
+
+			if (invalidate)
+				Invalidate (date_area_rect);
+		}
+
+		internal void SetPart (int value, DateTimePart dt_part)
+		{
+			SetPart (value, dt_part, false);
+		}
+
+		// set the specified part of the date to the specified value
+		internal void SetPart (int value, DateTimePart dt_part, bool adjust)
+		{
+			switch (dt_part)
 			{
-				case 's': // seconds
-					value %= 60;
+				case DateTimePart.Seconds:
 					if (value == -1)
 						value = 59;
 					if (value >= 0 && value <= 59)
 						Value = new DateTime(Value.Year, Value.Month, Value.Day, Value.Hour, Value.Minute, value, Value.Millisecond);
 					break;
-				case 'm': // minutes
-					value %= 60;
+				case DateTimePart.Minutes:
 					if (value == -1)
 						value = 59;
 					if (value >= 0 && value <= 59)
 						Value = new DateTime(Value.Year, Value.Month, Value.Day, Value.Hour, value, Value.Second, Value.Millisecond);
 					break;
-				case 'h':
-				case 'H': // hours
-					value %= 24;
+				case DateTimePart.AMPMHour:
+					if (value == -1)
+						value = 23;
+					if (value >= 0 && value <= 23) {
+						int prev_hour = Value.Hour;
+						if ((prev_hour >= 12 && prev_hour <= 23) && value < 12) // Adjust to p.m.
+							value += 12;
+						Value = new DateTime (Value.Year, Value.Month, Value.Day, value, Value.Minute, Value.Second, Value.Millisecond);
+					}
+					break;
+				case DateTimePart.Hour:
 					if (value == -1)
 						value = 23;
 					if (value >= 0 && value <= 23)
 						Value = new DateTime(Value.Year, Value.Month, Value.Day, value, Value.Minute, Value.Second, Value.Millisecond);
 					break;
-				case 'd': // days
+				case DateTimePart.Day:
 					int max_days = DateTime.DaysInMonth(Value.Year, Value.Month);
-					if (value > max_days)
-						Value = new DateTime(Value.Year, Value.Month, max_days, Value.Hour, Value.Minute, Value.Second, Value.Millisecond);
-					if (value >= 1 && value <= 31)
+					if (value >= 1 && value <= 31 && value <= max_days)
 						Value = new DateTime(Value.Year, Value.Month, value, Value.Hour, Value.Minute, Value.Second, Value.Millisecond);
 					break;
-				case 'M': // months
-					value %= 12;
-					if (value == 0)
-						value = 12;
-					if (value >= 1 && value <= 12)
-						Value = new DateTime(Value.Year, value, Value.Day, Value.Hour, Value.Minute, Value.Second, Value.Millisecond);
+				case DateTimePart.Month:
+					DateTime date = Value;
+
+					if (adjust) {
+						if (value == 0) {
+							date = date.AddYears (-1);
+							value = 12;
+						} else if (value == 13) {
+							date = date.AddYears (1);
+							value = 1;
+						}
+					}
+
+					if (value >= 1 && value <= 12) {
+						// if we move from say december to november with days on 31, we must
+						// remap to the maximum number of days
+						int days_in_new_month = DateTime.DaysInMonth (date.Year, value);
+						
+						if (date.Day > days_in_new_month)
+							Value = new DateTime (date.Year, value, days_in_new_month, date.Hour, date.Minute, date.Second, date.Millisecond);
+						else
+							Value = new DateTime (date.Year, value, date.Day, date.Hour, date.Minute, date.Second, date.Millisecond);
+					}
 					break;
-				case 'y': // years
-					value %= 10000;
-					if (value > 0 && value <= 9999)
-						Value = new DateTime(value, Value.Month, Value.Day, Value.Hour, Value.Minute, Value.Second, Value.Millisecond);
+				case DateTimePart.Year:
+					if (value >= min_date.Year && value <= max_date.Year) {
+						// if we move to a leap year, the days in month could throw an exception
+						int days_in_new_month = DateTime.DaysInMonth (value, Value.Month);
+						
+						if (Value.Day > days_in_new_month)
+							Value = new DateTime (value, Value.Month, days_in_new_month, Value.Hour, Value.Minute, Value.Second, Value.Millisecond);
+						else
+							Value = new DateTime (value, Value.Month, Value.Day, Value.Hour, Value.Minute, Value.Second, Value.Millisecond);
+					}
 					break;
 			}
 		}
 
+		private void GotFocusHandler (object sender, EventArgs e)
+		{
+			if (ShowCheckBox) {
+				is_checkbox_selected = true;
+				Invalidate (CheckBoxRect);
+			}
+		}
+
 		// if we loose focus deselect any selected parts.
 		private void LostFocusHandler (object sender, EventArgs e) 
 		{
 			int selected_index = GetSelectedPartIndex ();
 			if (selected_index != -1)
 			{
-				part_data [selected_index].is_selected = false;
+				part_data [selected_index].Selected = false;
 				Rectangle invalidate_rect = Rectangle.Ceiling (part_data [selected_index].drawing_rectangle);
 				invalidate_rect.Inflate (2, 2);
 				Invalidate (invalidate_rect);
+				OnUIASelectionChanged ();
 			}
 			else if (is_checkbox_selected)
 			{
@@ -1652,8 +1737,6 @@ namespace System.Windows.Forms {
 			if (e.Button != MouseButtons.Left)
 				return;
 
-			is_checkbox_selected = false;
-
 			if (ShowCheckBox && CheckBoxRect.Contains(e.X, e.Y))
 			{
 				is_checkbox_selected = true;
@@ -1661,6 +1744,10 @@ namespace System.Windows.Forms {
 				return;
 			}
 
+			// Deselect the checkbox only if the pointer is not on it
+			// *and* the other parts are enabled (Checked as true)
+			if (Checked)
+				is_checkbox_selected = false;
 
 			if (ShowUpDown && drop_down_arrow_rect.Contains (e.X, e.Y))
 			{
@@ -1680,11 +1767,7 @@ namespace System.Windows.Forms {
 					updown_timer.Enabled = true;
 				}
 			} else if (is_drop_down_visible == false && drop_down_arrow_rect.Contains (e.X, e.Y)) {
-				is_drop_down_visible = true;
-				if (!Checked)
-					Checked = true;
-				Invalidate (drop_down_arrow_rect);
-				DropDownMonthCalendar ();
+				DropDownButtonClicked ();
     			} else {
     				// mouse down on this control anywhere else collapses it
     				if (is_drop_down_visible) {    				
@@ -1696,26 +1779,41 @@ namespace System.Windows.Forms {
 					// go through the parts to see if the click is in any of them
 					bool invalidate_afterwards = false;
 					for (int i = 0; i < part_data.Length; i++) {
-						bool old = part_data [i].is_selected;
+						bool old = part_data [i].Selected;
 
 						if (part_data [i].is_literal)
 							continue;
 
 						if (part_data [i].drawing_rectangle.Contains (e.X, e.Y)) {
-							part_data [i].is_selected = true;
-						} else {
-							part_data [i].is_selected = false;
-						}
-						if (old != part_data [i].is_selected) 
+							part_data [i].Selected = true;
+						} else
+							part_data [i].Selected = false;
+
+						if (old != part_data [i].Selected) 
 							invalidate_afterwards = true;
 					}
-					if (invalidate_afterwards)
+					if (invalidate_afterwards) {
 						Invalidate ();
+						OnUIASelectionChanged ();
+					}
 				}
 				
 			}
 		}
 		
+		internal void DropDownButtonClicked ()
+		{
+			if (!is_drop_down_visible) {
+				is_drop_down_visible = true;
+				if (!Checked)
+					Checked = true;
+				Invalidate (drop_down_arrow_rect);
+				DropDownMonthCalendar ();
+			} else {
+				HideMonthCalendar ();
+				this.Focus ();
+			}
+		}
 		
 		// paint this control now
 		private void PaintHandler (object sender, PaintEventArgs pe) {
@@ -1725,15 +1823,52 @@ namespace System.Windows.Forms {
 			Draw (pe.ClipRectangle, pe.Graphics);
 		}
 		
+		void OnMouseEnter (object sender, EventArgs e)
+		{
+			if (ThemeEngine.Current.DateTimePickerBorderHasHotElementStyle)
+				Invalidate ();
+		}
+
+		void OnMouseLeave (object sender, EventArgs e)
+		{
+			drop_down_button_entered = false;
+			if (ThemeEngine.Current.DateTimePickerBorderHasHotElementStyle)
+				Invalidate ();
+		}
+
+		void OnMouseMove (object sender, MouseEventArgs e)
+		{
+			if (!is_drop_down_visible &&
+				ThemeEngine.Current.DateTimePickerDropDownButtonHasHotElementStyle &&
+				drop_down_button_entered != drop_down_arrow_rect.Contains (e.Location)) {
+				drop_down_button_entered = !drop_down_button_entered;
+				Invalidate (drop_down_arrow_rect);
+			}
+		}
 		#endregion		
 
 		#region internal classes
+		internal enum DateTimePart {
+			Seconds,
+			Minutes,
+			AMPMHour,
+			Hour,
+			Day,
+			DayName,
+			Month,
+			Year,
+			AMPMSpecifier,
+			Literal
+		}
+
 		internal class PartData
 		{
 			internal string value;
 			internal bool is_literal;
-			internal bool is_selected;
+			bool is_selected;
 			internal RectangleF drawing_rectangle;
+			internal DateTimePart date_time_part;
+			DateTimePicker owner;
 
 			internal bool is_numeric_format
 			{
@@ -1767,10 +1902,25 @@ namespace System.Windows.Forms {
 				}
 			}
 
-			internal PartData(string value, bool is_literal)
+			internal PartData(string value, bool is_literal, DateTimePicker owner)
 			{
 				this.value = value;
 				this.is_literal = is_literal;
+				this.owner = owner;
+				date_time_part = GetDateTimePart (value);
+			}
+
+			internal bool Selected {
+				get {
+					return is_selected;
+				}
+				set {
+					if (value == is_selected)
+						return;
+
+					owner.EndDateEdit (false);
+					is_selected = value;
+				}
 			}
 
 			// calculate the string to show for this data
@@ -1783,17 +1933,143 @@ namespace System.Windows.Forms {
 				}
 			}
 
+			static DateTimePart GetDateTimePart (string value)
+			{
+				switch (value) {
+					case "s":
+					case "ss":
+						return DateTimePart.Seconds;
+					case "m":
+					case "mm":
+						return DateTimePart.Minutes;
+					case "h":
+					case "hh":
+						return DateTimePart.AMPMHour;
+					case "H":
+					case "HH":
+						return DateTimePart.Hour;
+					case "d":
+					case "dd":
+						return DateTimePart.Day;
+					case "ddd":
+					case "dddd":
+						return DateTimePart.DayName;
+					case "M":
+					case "MM":
+					case "MMMM":
+						return DateTimePart.Month;
+					case "y":
+					case "yy":
+					case "yyy":
+					case "yyyy":
+						return DateTimePart.Year;
+					case "t":
+					case "tt":
+						return DateTimePart.AMPMSpecifier;
+				}
+					
+				return DateTimePart.Literal;
+			}
+
 			static internal string GetText(DateTime date, string format)
 			{
 				if (format.StartsWith ("g")) 
 					return " ";
 				else if (format.Length == 1)
 					return date.ToString ("%" + format);
-				else
+				else if (format == "yyyyy" || format == "yyyyyy" || format == "yyyyyyy" || format == "yyyyyyyy")
+					return date.ToString ("yyyy");
+				else if (format.Length > 1)
 					return date.ToString (format);
+				else
+					return string.Empty;
 			}
 		}
 		
 		#endregion		
+
+	
+		#region UIA Framework: Methods, Properties and Events
+
+		static object UIAMinimumChangedEvent = new object ();
+		static object UIAMaximumChangedEvent = new object ();
+		static object UIASelectionChangedEvent = new object ();
+		static object UIACheckedEvent = new object ();
+		static object UIAShowCheckBoxChangedEvent = new object ();
+		static object UIAShowUpDownChangedEvent = new object ();
+
+		internal event EventHandler UIAMinimumChanged {
+			add { Events.AddHandler (UIAMinimumChangedEvent, value); }
+			remove { Events.RemoveHandler (UIAMinimumChangedEvent, value); }
+		}
+
+		internal event EventHandler UIAMaximumChanged {
+			add { Events.AddHandler (UIAMinimumChangedEvent, value); }
+			remove { Events.RemoveHandler (UIAMinimumChangedEvent, value); }
+		}
+
+		internal event EventHandler UIASelectionChanged {
+			add { Events.AddHandler (UIASelectionChangedEvent, value); }
+			remove { Events.RemoveHandler (UIASelectionChangedEvent, value); }
+		}
+
+		internal event EventHandler UIAChecked {
+			add { Events.AddHandler (UIACheckedEvent, value); }
+			remove { Events.RemoveHandler (UIACheckedEvent, value); }
+		}
+
+		internal event EventHandler UIAShowCheckBoxChanged {
+			add { Events.AddHandler (UIAShowCheckBoxChangedEvent, value); }
+			remove { Events.RemoveHandler (UIAShowCheckBoxChangedEvent, value); }
+		}
+
+		internal event EventHandler UIAShowUpDownChanged {
+			add { Events.AddHandler (UIAShowUpDownChangedEvent, value); }
+			remove { Events.RemoveHandler (UIAShowUpDownChangedEvent, value); }
+		}
+
+		internal void OnUIAMinimumChanged ()
+		{
+			EventHandler eh = (EventHandler)(Events [UIAMinimumChangedEvent]);
+			if (eh != null)
+				eh (this, EventArgs.Empty);
+		}
+
+		internal void OnUIAMaximumChanged ()
+		{
+			EventHandler eh = (EventHandler)(Events [UIAMaximumChangedEvent]);
+			if (eh != null)
+				eh (this, EventArgs.Empty);
+		}
+
+		internal void OnUIASelectionChanged ()
+		{
+			EventHandler eh = (EventHandler)(Events [UIASelectionChangedEvent]);
+			if (eh != null)
+				eh (this, EventArgs.Empty);
+		}
+
+		internal void OnUIAChecked ()
+		{
+			EventHandler eh = (EventHandler)(Events [UIACheckedEvent]);
+			if (eh != null)
+				eh (this, EventArgs.Empty);
+		}
+
+		internal void OnUIAShowCheckBoxChanged ()
+		{
+			EventHandler eh = (EventHandler)(Events [UIAShowCheckBoxChangedEvent]);
+			if (eh != null)
+				eh (this, EventArgs.Empty);
+		}
+
+		internal void OnUIAShowUpDownChanged ()
+		{
+			EventHandler eh = (EventHandler)(Events [UIAShowUpDownChangedEvent]);
+			if (eh != null)
+				eh (this, EventArgs.Empty);
+		}
+		
+		#endregion
 	}
 }
