@@ -21,6 +21,7 @@
 //
 // Authors:
 //	Alexander Olk (alex.olk@googlemail.com)
+//	Gert Driesen (drieseng@users.sourceforge.net)
 //
 //
 
@@ -38,9 +39,8 @@ namespace System.Windows.Forms {
 	public sealed class FolderBrowserDialog : CommonDialog
 	{
 		#region Local Variables
-		private string description = "";
 		private Environment.SpecialFolder rootFolder = Environment.SpecialFolder.Desktop;
-		private string selectedPath = "";
+		private string selectedPath = string.Empty;
 		private bool showNewFolderButton = true;
 		
 		private Label descriptionLabel;
@@ -51,7 +51,7 @@ namespace System.Windows.Forms {
 		private ContextMenu folderBrowserTreeViewContextMenu;
 		private MenuItem newFolderMenuItem;
 		
-		private string old_selectedPath = "";
+		private string old_selectedPath = string.Empty;
 		
 		private readonly string folderbrowserdialog_string = "FolderBrowserDialog";
 		private readonly string width_string = "Width";
@@ -63,6 +63,7 @@ namespace System.Windows.Forms {
 		#region Public Constructors
 		public FolderBrowserDialog ()
 		{
+			form = new DialogForm (this);
 			Size formConfigSize = Size.Empty;
 			Point formConfigLocation = Point.Empty;
 			
@@ -92,7 +93,7 @@ namespace System.Windows.Forms {
 			form.SuspendLayout ();
 			form.ClientSize = new Size (322, 324);
 			form.MinimumSize = new Size (310, 254);
-			form.Text = "Search Folder";
+			form.Text = "Browse For Folder";
 			form.SizeGripStyle = SizeGripStyle.Show;
 
 			newFolderMenuItem = new MenuItem("New Folder", new EventHandler (OnClickNewFolderButton));
@@ -104,7 +105,7 @@ namespace System.Windows.Forms {
 			descriptionLabel.Location = new Point (15, 14);
 			descriptionLabel.Size = new Size (292, 40);
 			descriptionLabel.TabIndex = 0;
-			descriptionLabel.Text = "";
+			descriptionLabel.Text = string.Empty;
 			
 			// folderBrowserTreeView
 			folderBrowserTreeView.Anchor = ((AnchorStyles)((((AnchorStyles.Top | AnchorStyles.Bottom)
@@ -114,7 +115,7 @@ namespace System.Windows.Forms {
 			folderBrowserTreeView.Location = new Point (15, 60);
 			folderBrowserTreeView.SelectedImageIndex = -1;
 			folderBrowserTreeView.Size = new Size (292, 212);
-			folderBrowserTreeView.TabIndex = 1;
+			folderBrowserTreeView.TabIndex = 3;
 			folderBrowserTreeView.ShowLines = false;
 			folderBrowserTreeView.ShowPlusMinus = true;
 			folderBrowserTreeView.HotTracking = true;
@@ -127,16 +128,16 @@ namespace System.Windows.Forms {
 			newFolderButton.FlatStyle = FlatStyle.System;
 			newFolderButton.Location = new Point (15, 285);
 			newFolderButton.Size = new Size (105, 23);
-			newFolderButton.TabIndex = 2;
-			newFolderButton.Text = "New Folder";
+			newFolderButton.TabIndex = 4;
+			newFolderButton.Text = "Make New Folder";
 			newFolderButton.Enabled = true;
 			
 			// okButton
 			okButton.Anchor = ((AnchorStyles)((AnchorStyles.Bottom | AnchorStyles.Right)));
 			okButton.FlatStyle = FlatStyle.System;
-			okButton.Location = new Point (142, 285);
+			okButton.Location = new Point (135, 285);
 			okButton.Size = new Size (80, 23);
-			okButton.TabIndex = 3;
+			okButton.TabIndex = 1;
 			okButton.Text = "OK";
 			
 			// cancelButton
@@ -145,7 +146,7 @@ namespace System.Windows.Forms {
 			cancelButton.FlatStyle = FlatStyle.System;
 			cancelButton.Location = new Point (227, 285);
 			cancelButton.Size = new Size (80, 23);
-			cancelButton.TabIndex = 4;
+			cancelButton.TabIndex = 2;
 			cancelButton.Text = "Cancel";
 			
 			form.Controls.Add (cancelButton);
@@ -167,6 +168,8 @@ namespace System.Windows.Forms {
 			okButton.Click += new EventHandler (OnClickOKButton);
 			cancelButton.Click += new EventHandler (OnClickCancelButton);
 			newFolderButton.Click += new EventHandler (OnClickNewFolderButton);
+
+			form.VisibleChanged += new EventHandler (OnFormVisibleChanged);
 			
 			RootFolder = rootFolder;
 		}
@@ -179,31 +182,31 @@ namespace System.Windows.Forms {
 		[Localizable(true)]
 		public string Description {
 			set {
-				description = value;
-				descriptionLabel.Text = description;
+				descriptionLabel.Text = value;
 			}
 			
 			get {
-				return description;
+				return descriptionLabel.Text;
 			}
 		}
 		
 		[Browsable(true)]
 		[DefaultValue(Environment.SpecialFolder.Desktop)]
 		[Localizable(false)]
+#if NET_2_0
+		[TypeConverter (typeof (SpecialFolderEnumConverter))]
+#endif
 		public Environment.SpecialFolder RootFolder {
 			set {
 				int v = (int)value;
+
+				Type enumType = typeof (Environment.SpecialFolder);
+				if (!Enum.IsDefined (enumType, v))
+					throw new InvalidEnumArgumentException (
+						"value", v, enumType);
 				
-				if (!Enum.IsDefined(typeof(Environment.SpecialFolder), v))
-					throw new InvalidEnumArgumentException ();
-				
-				if (rootFolder != value)
-					rootFolder = value;
-				
-				folderBrowserTreeView.RootFolder = rootFolder;
+				rootFolder = value;
 			}
-			
 			get {
 				return rootFolder;
 			}
@@ -215,14 +218,11 @@ namespace System.Windows.Forms {
 		[Localizable(true)]
 		public string SelectedPath {
 			set {
-				if (!Path.IsPathRooted(value))
-					return;
-				
+				if (value == null)
+					value = string.Empty;
 				selectedPath = value;
 				old_selectedPath = value;
-				folderBrowserTreeView.SelectedPath = selectedPath;
 			}
-			
 			get {
 				return selectedPath;
 			}
@@ -234,11 +234,8 @@ namespace System.Windows.Forms {
 		public bool ShowNewFolderButton {
 			set {
 				if (value != showNewFolderButton) {
+					newFolderButton.Visible = value;
 					showNewFolderButton = value;
-					if (showNewFolderButton)
-						newFolderButton.Show ();
-					else
-						newFolderButton.Hide ();
 				}
 			}
 			
@@ -251,14 +248,17 @@ namespace System.Windows.Forms {
 		#region Public Instance Methods
 		public override void Reset ()
 		{
-			Description = "";
+			Description = string.Empty;
 			RootFolder = Environment.SpecialFolder.Desktop;
-			selectedPath = "";
+			selectedPath = string.Empty;
 			ShowNewFolderButton = true;
 		}
 		
-		protected override bool RunDialog (IntPtr hwndOwner)
+		protected override bool RunDialog (IntPtr hWndOwner)
 		{
+			folderBrowserTreeView.RootFolder = RootFolder;
+			folderBrowserTreeView.SelectedPath = SelectedPath;
+
 			form.Refresh ();
 			
 			return true;
@@ -285,6 +285,12 @@ namespace System.Windows.Forms {
 		{
 			folderBrowserTreeView.CreateNewFolder ();
 		}
+
+		void OnFormVisibleChanged (object sender, EventArgs e)
+		{
+			if (form.Visible && okButton.Enabled)
+				okButton.Select ();
+		}
 		
 		private void WriteConfigValues ()
 		{
@@ -312,12 +318,12 @@ namespace System.Windows.Forms {
 			private ImageList imageList = new ImageList ();
 			private Environment.SpecialFolder rootFolder;
 			private bool dont_enable = false;
-			
-			private int platform = (int) Environment.OSVersion.Platform;
+			private TreeNode node_under_mouse;
 			
 			public FolderBrowserTreeView (FolderBrowserDialog parent_dialog)
 			{
 				parentDialog = parent_dialog;
+				HideSelection = false;
 				ImageList = imageList;
 				SetupImageList ();
 			}
@@ -326,15 +332,19 @@ namespace System.Windows.Forms {
 				set {
 					rootFolder = value;
 					
-					string root_path = "";
+					string root_path = string.Empty;
 					
 					switch (rootFolder) {
-						default:
 						case Environment.SpecialFolder.Desktop:
 							root_node = new FBTreeNode ("Desktop");
 							root_node.RealPath = ThemeEngine.Current.Places (UIIcon.PlacesDesktop);
 							root_path = MWFVFS.DesktopPrefix;
 							break;
+						case Environment.SpecialFolder.Recent:
+							root_node = new FBTreeNode ("My Recent Documents");
+							root_node.RealPath = ThemeEngine.Current.Places (UIIcon.PlacesRecentDocuments);
+							root_path = MWFVFS.RecentlyUsedPrefix;
+							break;
 						case Environment.SpecialFolder.MyComputer:
 							root_node = new FBTreeNode ("My Computer");
 							root_path = MWFVFS.MyComputerPrefix;
@@ -344,11 +354,20 @@ namespace System.Windows.Forms {
 							root_path = MWFVFS.PersonalPrefix;
 							root_node.RealPath = ThemeEngine.Current.Places (UIIcon.PlacesPersonal);
 							break;
+						default:
+							root_node = new FBTreeNode (rootFolder.ToString ());
+							root_node.RealPath = Environment.GetFolderPath (rootFolder);
+							root_path = root_node.RealPath;
+							break;
 					}
 					
 					root_node.Tag = root_path;
 					root_node.ImageIndex = NodeImageIndex (root_path);
-					
+
+					BeginUpdate ();
+					Nodes.Clear ();
+					EndUpdate ();
+
 					FillNode (root_node);
 					
 					root_node.Expand ();
@@ -359,31 +378,40 @@ namespace System.Windows.Forms {
 			
 			public string SelectedPath {
 				set {
-					if (Check_if_path_is_child_of_RootFolder (value)) {
-						SetSelectedPath (Path.GetFullPath (value));
+					if (value.Length == 0)
+						return;
+
+					if (!Path.IsPathRooted (value))
+						return;
+
+					try {
+						if (Check_if_path_is_child_of_RootFolder (value))
+							SetSelectedPath (Path.GetFullPath (value));
+					} catch (Exception) {
+						// If we can't set the user's requested path, the
+						// best we can do is not crash and reset to the default
+						EndUpdate ();
+						RootFolder = rootFolder;
 					}
 				}
 			}
 			
 			private string parent_real_path;
-			private bool dont_do_onbeforeexpand = false;
+			private bool dont_do_onbeforeexpand;
 			
 			public void CreateNewFolder ()
 			{
-				FBTreeNode fbnode = SelectedNode as FBTreeNode;
+				FBTreeNode fbnode = node_under_mouse == null ? SelectedNode as FBTreeNode : node_under_mouse as FBTreeNode;
 				
-				if (fbnode == null || fbnode.Parent == null)
-					return;
-				
-				if (fbnode.RealPath == null)
+				if (fbnode == null || fbnode.RealPath == null)
 					return;
 				
 				string tmp_filename = "New Folder";
 				
 				if (Directory.Exists (Path.Combine (fbnode.RealPath, tmp_filename))) {
 					int i = 1;
-					
-					if ((platform == 4) || (platform == 128)) {
+
+					if (XplatUI.RunningOnUnix) {
 						tmp_filename = tmp_filename + "-" + i;
 					} else {
 						tmp_filename = tmp_filename + " (" + i + ")";
@@ -391,29 +419,34 @@ namespace System.Windows.Forms {
 					
 					while (Directory.Exists (Path.Combine (fbnode.RealPath, tmp_filename))) {
 						i++;
-						if ((platform == 4) || (platform == 128)) {
+						if (XplatUI.RunningOnUnix) {
 							tmp_filename = "New Folder" + "-" + i;
 						} else {
 							tmp_filename = "New Folder" + " (" + i + ")";
 						}
 					}
 				}
-				
+
 				parent_real_path = fbnode.RealPath;
 				
-				FBTreeNode new_node = new FBTreeNode (tmp_filename);
-				new_node.ImageIndex = NodeImageIndex(tmp_filename);
-				
 				FillNode (fbnode);
 				dont_do_onbeforeexpand = true;
 				fbnode.Expand ();
 				dont_do_onbeforeexpand = false;
-				
+
+				// to match MS, immediately create the new folder
+				// and rename it once the label edit completes
+				string fullPath = Path.Combine (fbnode.RealPath, tmp_filename);
+				if (!vfs.CreateFolder (fullPath))
+					return;
+
+				FBTreeNode new_node = new FBTreeNode (tmp_filename);
+				new_node.ImageIndex = NodeImageIndex (tmp_filename);
+				new_node.Tag = new_node.RealPath = fullPath;
 				fbnode.Nodes.Add (new_node);
-				
+
 				LabelEdit = true;
-				if (!new_node.IsEditing)
-					new_node.BeginEdit();
+				new_node.BeginEdit();
 			}
 			
 			protected override void OnAfterLabelEdit (NodeLabelEditEventArgs e)
@@ -421,24 +454,33 @@ namespace System.Windows.Forms {
 				if (e.Label != null) {
 					if (e.Label.Length > 0) {
 						FBTreeNode fbnode = e.Node as FBTreeNode;
-						
-						fbnode.RealPath = Path.Combine(parent_real_path, e.Label);
-						
-						if (vfs.CreateFolder (fbnode.RealPath)) {
-							SelectedNode = e.Node;
+
+						string originalPath = fbnode.RealPath;
+						string newPath = Path.Combine (parent_real_path, e.Label);
+
+						if (vfs.MoveFolder (originalPath, newPath)) {
+							fbnode.Tag = fbnode.RealPath = newPath;
 						} else {
-							SelectedNode = e.Node.Parent;
-							e.Node.Parent.Nodes.Remove(e.Node);
+							e.CancelEdit = true;
+							e.Node.BeginEdit ();
+							return;
 						}
-						
-						e.Node.EndEdit (false);
 					} else {
 						e.CancelEdit = true;
-						e.Node.BeginEdit();
+						e.Node.BeginEdit ();
+						return;
 					}
-					
-					LabelEdit = false;
 				}
+
+				// select new folder only if both the curren node under
+				// mouse pointer and SelectedNode are the same (match .Net)
+				if (node_under_mouse == SelectedNode)
+					SelectedNode = e.Node;
+
+				// disable LabelEdit when either edit has finished
+				// or has been cancelled, to prevent the user from
+				// editing label of existing folders
+				LabelEdit = false;
 			}
 			
 			private void SetSelectedPath (string path)
@@ -450,53 +492,60 @@ namespace System.Windows.Forms {
 				if (node == null) {
 					Stack stack = new Stack ();
 					
-					string path_cut = path.Substring (0, path.LastIndexOf (Path.AltDirectorySeparatorChar));
+					string path_cut = path.Substring (0, path.LastIndexOf (Path.DirectorySeparatorChar));
+					if (!XplatUI.RunningOnUnix && path_cut.Length == 2)
+						path_cut += Path.DirectorySeparatorChar;
 					
 					while (node == null && path_cut.Length > 0) {
 						node = FindPathInNodes (path_cut, Nodes);
 						
 						if (node == null) {
-							string path_cut_new = path_cut.Substring (0, path_cut.LastIndexOf (Path.AltDirectorySeparatorChar));
-							string leftover = path_cut.Replace (path_cut_new, "");
+							string path_cut_new = path_cut.Substring (0, path_cut.LastIndexOf (Path.DirectorySeparatorChar));
+							string leftover = path_cut.Replace (path_cut_new, string.Empty);
 							
 							stack.Push (leftover);
 							
 							path_cut = path_cut_new;
 						}
 					}
+
+					// If we didn't find anything, just display the full, unselected tree
+					if (node == null) {
+						EndUpdate ();
+						RootFolder = rootFolder;
+						return;
+					}
 					
-					if (stack.Count > 0) {
-						FillNode (node);
-						node.Expand ();
-						
-						// walk through the subdirs and fill the nodes
-						while (stack.Count > 0) {
-							string part_name = stack.Pop () as string;
-							
-							foreach (TreeNode treeNode in node.Nodes) {
-								FBTreeNode fbnode = treeNode as FBTreeNode;
-								
-								if (path_cut + part_name == fbnode.RealPath) {
-									node = fbnode;
-									path_cut += part_name;
-									
-									FillNode (node);
-									node.Expand ();
-									break;
-								}
-							}
-						}
-						
-						// finally find the node for the complete path
+					FillNode (node);
+					node.Expand ();
+
+					// walk through the subdirs and fill the nodes
+					while (stack.Count > 0) {
+						string part_name = stack.Pop () as string;
+
 						foreach (TreeNode treeNode in node.Nodes) {
 							FBTreeNode fbnode = treeNode as FBTreeNode;
-							
-							if (path == fbnode.RealPath) {
+
+							if (path_cut + part_name == fbnode.RealPath) {
 								node = fbnode;
+								path_cut += part_name;
+
+								FillNode (node);
+								node.Expand ();
 								break;
 							}
 						}
 					}
+
+					// finally find the node for the complete path
+					foreach (TreeNode treeNode in node.Nodes) {
+						FBTreeNode fbnode = treeNode as FBTreeNode;
+
+						if (path == fbnode.RealPath) {
+							node = fbnode;
+							break;
+						}
+					}
 				}
 				
 				if (node != null) {
@@ -509,6 +558,13 @@ namespace System.Windows.Forms {
 			
 			private FBTreeNode FindPathInNodes (string path, TreeNodeCollection nodes)
 			{
+				// On Windows the devices can be passed as C: yet match
+				// the C:\ form
+				//
+				// Hackish, but works
+				if (!XplatUI.RunningOnUnix && path.Length == 2)
+					path += Path.DirectorySeparatorChar;
+
 				foreach (TreeNode node in nodes) {
 					FBTreeNode fbnode = node as FBTreeNode;
 					
@@ -517,7 +573,9 @@ namespace System.Windows.Forms {
 							return fbnode;
 					}
 					
-					return FindPathInNodes (path, node.Nodes);
+					FBTreeNode n = FindPathInNodes (path, node.Nodes);
+					if (n != null)
+						return n;
 				}
 				
 				return null;
@@ -525,9 +583,9 @@ namespace System.Windows.Forms {
 			
 			private bool Check_if_path_is_child_of_RootFolder (string path)
 			{
-				string root_path = (string)root_node.RealPath;
+				string root_path = (string) root_node.RealPath;
 				
-				if (root_path != null) {
+				if (root_path != null || rootFolder == Environment.SpecialFolder.MyComputer) {
 					try {
 						if (!Directory.Exists (path))
 							return false;
@@ -602,17 +660,13 @@ namespace System.Windows.Forms {
 				
 				if (path == MWFVFS.DesktopPrefix)
 					index = 1;
-				else 
-				if (path == MWFVFS.RecentlyUsedPrefix)
+				else if (path == MWFVFS.RecentlyUsedPrefix)
 					index = 0;
-				else
-				if (path == MWFVFS.PersonalPrefix)
+				else if (path == MWFVFS.PersonalPrefix)
 					index = 2;
-				else
-				if (path == MWFVFS.MyComputerPrefix)
+				else if (path == MWFVFS.MyComputerPrefix)
 					index = 3;
-				else
-				if (path == MWFVFS.MyNetworkPrefix)
+				else if (path == MWFVFS.MyNetworkPrefix)
 					index = 4;
 				
 				return index;
@@ -651,6 +705,12 @@ namespace System.Windows.Forms {
 				
 				base.OnBeforeExpand (e);
 			}
+
+			protected override void OnMouseDown (MouseEventArgs e)
+			{
+				node_under_mouse = GetNodeAt (e.X, e.Y);
+				base.OnMouseDown (e);
+			}
 			
 			protected override void OnMouseUp (MouseEventArgs e)
 			{
@@ -664,6 +724,8 @@ namespace System.Windows.Forms {
 					parentDialog.newFolderButton.Enabled = true;
 					parentDialog.newFolderMenuItem.Enabled = true;
 				}
+
+				node_under_mouse = null;
 				
 				base.OnMouseUp (e);
 			}
@@ -689,4 +751,25 @@ namespace System.Windows.Forms {
 			}
 		}
 	}
+	
+#if NET_2_0
+	internal class SpecialFolderEnumConverter : TypeConverter
+	{
+		public override object ConvertFrom (ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
+		{
+			if ((value == null) || !(value is String))
+				return base.ConvertFrom (context, culture, value);
+
+			return Enum.Parse (typeof (Environment.SpecialFolder), (string)value, true);
+		}
+
+		public override object ConvertTo (ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
+		{
+			if ((value == null) || !(value is Environment.SpecialFolder) || (destinationType != typeof (string)))
+				return base.ConvertTo (context, culture, value, destinationType);
+				
+			return ((Environment.SpecialFolder)value).ToString ();
+		}
+	}
+#endif
 }
