@@ -28,9 +28,15 @@ using System.Collections;
 using System.ComponentModel;
 using System.ComponentModel.Design;
 using System.Drawing;
-
+using System.Runtime.InteropServices;
+using System.Windows.Forms.Theming;
+using System.Windows.Forms.VisualStyles;
 
 namespace System.Windows.Forms {
+#if NET_2_0
+	[ComVisibleAttribute (true)]
+	[ClassInterfaceAttribute (ClassInterfaceType.AutoDispatch)]
+#endif
 	[DefaultEvent("SelectedIndexChanged")]
 	[DefaultProperty("TabPages")]
 	[Designer("System.Windows.Forms.Design.TabControlDesigner, " + Consts.AssemblySystem_Design, "System.ComponentModel.Design.IDesigner")]
@@ -50,9 +56,14 @@ namespace System.Windows.Forms {
 		private bool show_tool_tips;
 		private TabSizeMode size_mode;
 		private bool show_slider = false;
-		private ButtonState right_slider_state;
-		private ButtonState left_slider_state;
+		private PushButtonState right_slider_state = PushButtonState.Normal;
+		private PushButtonState left_slider_state = PushButtonState.Normal;
 		private int slider_pos = 0;
+		TabPage entered_tab_page;
+		bool mouse_down_on_a_tab_page;
+#if NET_2_0
+		private bool rightToLeftLayout;
+#endif		
 		#endregion	// Fields
 
 		#region Public Constructors
@@ -64,6 +75,8 @@ namespace System.Windows.Forms {
 			item_size = ThemeEngine.Current.TabControlDefaultItemSize;
 
 			MouseDown += new MouseEventHandler (MouseDownHandler);
+			MouseLeave += new EventHandler (OnMouseLeave);
+			MouseMove += new MouseEventHandler (OnMouseMove);
 			MouseUp += new MouseEventHandler (MouseUpHandler);
 			SizeChanged += new EventHandler (SizeChangedHandler);
 		}
@@ -112,12 +125,29 @@ namespace System.Windows.Forms {
 			set { base.BackgroundImage = value; }
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public override ImageLayout BackgroundImageLayout {
+			get { return base.BackgroundImageLayout; }
+			set { base.BackgroundImageLayout = value; }
+		}
+#endif
+		
 		public override Rectangle DisplayRectangle {
 			get {
 				return ThemeEngine.Current.TabControlGetDisplayRectangle (this);
 			}
 		}
 
+#if NET_2_0
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		protected override bool DoubleBuffered {
+			get { return base.DoubleBuffered; }
+			set { base.DoubleBuffered = value; }
+		}
+#endif
+
 		[DefaultValue(TabDrawMode.Normal)]
 		public TabDrawMode DrawMode {
 			get { return draw_mode; }
@@ -147,6 +177,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[RefreshProperties (RefreshProperties.Repaint)]
+#endif
 		[DefaultValue(null)]
 		public ImageList ImageList {
 			get { return image_list; }
@@ -180,7 +213,11 @@ namespace System.Windows.Forms {
 		}
 
 		[Localizable(true)]
-		public new Point Padding {
+		public
+#if NET_2_0
+		new
+#endif
+		Point Padding {
 			get { return padding; }
 			set {
 				if (value.X < 0 || value.Y < 0)
@@ -193,6 +230,21 @@ namespace System.Windows.Forms {
 
 		}
 
+#if NET_2_0
+		[MonoTODO ("Saves the value and raises event, but needs actual implementation call")]
+		[Localizable (true)]
+		[DefaultValue (false)]
+		public virtual bool RightToLeftLayout {
+			get { return this.rightToLeftLayout; }
+			set {
+				if (value != this.rightToLeftLayout) {
+					this.rightToLeftLayout = value;
+					this.OnRightToLeftLayoutChanged (EventArgs.Empty);
+				}
+			}
+		}
+#endif
+
 		[Browsable(false)]
 		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
 		public int RowCount {
@@ -269,7 +321,7 @@ namespace System.Windows.Forms {
 
 				if (new_index != -1) {
 					int le = TabPages[new_index].TabBounds.Right;
-					int re = ThemeEngine.Current.TabControlGetLeftScrollRect (this).Left;
+					int re = LeftScrollButtonArea.Left;
 					if (show_slider && le > re) {
 						int i = 0;
 						for (i = 0; i < new_index - 1; i++) {
@@ -316,9 +368,11 @@ namespace System.Windows.Forms {
 				} else {
 					SizeTabs ();
 					// The lines are drawn on the edges of the tabs so the invalid area should
-					// needs to include the extra pixels of line width.
+					// needs to include the extra pixels of line width (but should not
+					// overflow the control bounds).
 					if (appearance == TabAppearance.Normal) {
 						invalid.Inflate (6, 4);
+						invalid.Intersect (ClientRectangle);
 					}
 					Invalidate (invalid);
 				}
@@ -373,7 +427,11 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Editor ("System.Windows.Forms.Design.TabPageCollectionEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+#else
 		[DefaultValue(null)]
+#endif
 		[DesignerSerializationVisibility(DesignerSerializationVisibility.Hidden)]
 		[MergableProperty(false)]
 		public TabPageCollection TabPages {
@@ -387,7 +445,6 @@ namespace System.Windows.Forms {
 			get { return base.Text; }
 			set { base.Text = value; }
 		}
-
 		#endregion	// Public Instance Properties
 
 		#region Internal Properties
@@ -400,14 +457,57 @@ namespace System.Windows.Forms {
 			get { return slider_pos; }
 		}
 
-		internal ButtonState RightSliderState {
+		internal PushButtonState RightSliderState {
 			get { return right_slider_state; }
+			private set {
+				if (right_slider_state == value)
+					return;
+				PushButtonState old_value = right_slider_state;
+				right_slider_state = value;
+				if (NeedsToInvalidateScrollButton (old_value, value))
+					Invalidate (RightScrollButtonArea);
+			}
 		}
 
-		internal ButtonState LeftSliderState {
+		internal PushButtonState LeftSliderState {
 			get { return left_slider_state; }
+			set {
+				if (left_slider_state == value)
+					return;
+				PushButtonState old_value = left_slider_state;
+				left_slider_state = value;
+				if (NeedsToInvalidateScrollButton (old_value, value))
+					Invalidate (LeftScrollButtonArea);
+			}
 		}
 
+		bool NeedsToInvalidateScrollButton (PushButtonState oldState, PushButtonState newState)
+		{
+			if ((oldState == PushButtonState.Hot && newState == PushButtonState.Normal) ||
+				(oldState == PushButtonState.Normal && newState == PushButtonState.Hot))
+				return HasHotElementStyles;
+			return true;
+		}
+
+		internal TabPage EnteredTabPage {
+			get { return entered_tab_page; }
+			private set {
+				if (entered_tab_page == value)
+					return;
+				if (HasHotElementStyles) {
+					Region area_to_invalidate = new Region ();
+					area_to_invalidate.MakeEmpty ();
+					if (entered_tab_page != null)
+						area_to_invalidate.Union (entered_tab_page.TabBounds);
+					entered_tab_page = value;
+					if (entered_tab_page != null)
+						area_to_invalidate.Union (entered_tab_page.TabBounds);
+					Invalidate (area_to_invalidate);
+					area_to_invalidate.Dispose ();
+				} else
+					entered_tab_page = value;
+			}
+		}
 		#endregion	// Internal Properties
 
 		#region Protected Instance Properties
@@ -437,6 +537,22 @@ namespace System.Windows.Forms {
 		}
 
 #if NET_2_0
+		public void SelectTab (TabPage tabPage)
+		{
+			if (tabPage == null)
+				throw new ArgumentNullException ("tabPage");
+
+			SelectTab (this.tab_pages [tabPage]);
+		}
+
+		public void SelectTab (string tabPageName)
+		{
+			if (tabPageName == null)
+				throw new ArgumentNullException ("tabPageName");
+
+			SelectTab (this.tab_pages [tabPageName]);
+		}
+
 		public void SelectTab (int index)
 		{
 			if (index < 0 || index > this.tab_pages.Count - 1)
@@ -444,6 +560,33 @@ namespace System.Windows.Forms {
 				
 			SelectedIndex = index;
 		}
+
+		public void DeselectTab (TabPage tabPage)
+		{
+			if (tabPage == null)
+				throw new ArgumentNullException ("tabPage");
+
+			DeselectTab (this.tab_pages [tabPage]);
+		}
+
+		public void DeselectTab (string tabPageName)
+		{
+			if (tabPageName == null)
+				throw new ArgumentNullException ("tabPageName");
+
+			DeselectTab (this.tab_pages [tabPageName]);
+		}
+
+		public void DeselectTab (int index)
+		{
+			if (index == SelectedIndex) {
+				if (index >= 0 && index < this.tab_pages.Count - 1)
+					SelectedIndex = ++index;
+				else
+					SelectedIndex = 0;
+			}
+		}
+
 #endif
 
 		public override string ToString ()
@@ -460,16 +603,13 @@ namespace System.Windows.Forms {
 		#endregion	// Public Instance Methods
 
 		#region Protected Instance Methods
+
+		#region Handles
 		protected override Control.ControlCollection CreateControlsInstance ()
 		{
 			return new TabControl.ControlCollection (this);
 		}
 
-		protected void UpdateTabSelection (bool uiselected)
-		{
-			ResizeTabPages ();
-		}
-
 		protected override void CreateHandle ()
 		{
 			base.CreateHandle ();
@@ -494,6 +634,14 @@ namespace System.Windows.Forms {
 			base.OnHandleDestroyed (e);
 		}
 
+		protected override void Dispose (bool disposing)
+		{
+			base.Dispose (disposing);
+		}
+
+		#endregion
+
+		#region Events
 		protected virtual void OnDrawItem (DrawItemEventArgs e)
 		{
 			if (DrawMode != TabDrawMode.OwnerDrawFixed)
@@ -525,53 +673,170 @@ namespace System.Windows.Forms {
 			base.OnStyleChanged (e);
 		}
 
+		protected virtual void OnSelectedIndexChanged (EventArgs e)
+		{
+			EventHandler eh = (EventHandler) (Events[SelectedIndexChangedEvent]);
+			if (eh != null)
+				eh (this, e);
+		}
+
+		internal override void OnPaintInternal (PaintEventArgs pe)
+		{
+			if (GetStyle (ControlStyles.UserPaint))
+				return;
+
+			Draw (pe.Graphics, pe.ClipRectangle);
+			pe.Handled = true;
+		}
+
+#if NET_2_0
+		protected override void OnEnter (EventArgs e)
+		{
+			base.OnEnter (e);
+			if (SelectedTab != null)
+				SelectedTab.FireEnter ();
+		}
+
+		protected override void OnLeave (EventArgs e)
+		{
+			if (SelectedTab != null)
+				SelectedTab.FireLeave ();
+			base.OnLeave (e);
+		}
+
+		[EditorBrowsable (EditorBrowsableState.Advanced)]
+		protected virtual void OnRightToLeftLayoutChanged (EventArgs e)
+		{
+			EventHandler eh = (EventHandler) (Events[RightToLeftLayoutChangedEvent]);
+			if (eh != null)
+				eh (this, e);
+		}
+
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		protected override void ScaleCore (float dx, float dy)
+		{
+			base.ScaleCore (dx, dy);
+		}
+
+		protected virtual void OnDeselecting (TabControlCancelEventArgs e)
+		{
+			TabControlCancelEventHandler eh = (TabControlCancelEventHandler) (Events[DeselectingEvent]);
+			if (eh != null)
+				eh (this, e);
+
+			if (!e.Cancel)
+				OnDeselected (new TabControlEventArgs (SelectedTab, selected_index, TabControlAction.Deselected));
+		}
+
+		protected virtual void OnDeselected (TabControlEventArgs e)
+		{
+			TabControlEventHandler eh = (TabControlEventHandler) (Events[DeselectedEvent]);
+			if (eh != null)
+				eh (this, e);
+
+			if (this.SelectedTab != null)
+				this.SelectedTab.FireLeave ();
+		}
+
+		protected virtual void OnSelecting (TabControlCancelEventArgs e)
+		{
+			TabControlCancelEventHandler eh = (TabControlCancelEventHandler) (Events[SelectingEvent]);
+			if (eh != null)
+				eh (this, e);
+
+			if (!e.Cancel)
+				OnSelected (new TabControlEventArgs (SelectedTab, selected_index, TabControlAction.Selected));
+		}
+
+		protected virtual void OnSelected (TabControlEventArgs e)
+		{
+			TabControlEventHandler eh = (TabControlEventHandler) (Events[SelectedEvent]);
+			if (eh != null)
+				eh (this, e);
+
+			if (this.SelectedTab != null)
+				this.SelectedTab.FireEnter ();
+		}
+#endif
+
+		#endregion
+
+		#region Keys
 		protected override bool ProcessKeyPreview (ref Message m)
 		{
 			return base.ProcessKeyPreview (ref m);
 		}
 
-		protected override void OnKeyDown (KeyEventArgs e)
+		protected override void OnKeyDown (KeyEventArgs ke)
 		{
-			if (e.KeyCode == Keys.Tab && (e.KeyData & Keys.Control) != 0) {
-				if ((e.KeyData & Keys.Shift) == 0)
+			if (ke.KeyCode == Keys.Tab && (ke.KeyData & Keys.Control) != 0) {
+				if ((ke.KeyData & Keys.Shift) == 0)
 					SelectedIndex = (SelectedIndex + 1) % TabCount;
 				else
-					SelectedIndex = (SelectedIndex - 1) % TabCount;
-				e.Handled = true;
-			} else if (e.KeyCode == Keys.Home) {
+					SelectedIndex = (SelectedIndex + TabCount - 1) % TabCount;
+				ke.Handled = true;
+			} else if (ke.KeyCode == Keys.Home) {
 				SelectedIndex = 0;
-				e.Handled = true;
-			} else if (e.KeyCode == Keys.End) {
+				ke.Handled = true;
+			} else if (ke.KeyCode == Keys.End) {
 				SelectedIndex = TabCount - 1;
-				e.Handled = true;
-			} else if (e.KeyCode == Keys.Left && SelectedIndex > 0) {
-				SelectedIndex--;
-				e.Handled = true;
-			} else if (e.KeyCode == Keys.Right && SelectedIndex < TabCount - 1) {
-				SelectedIndex++;
-				e.Handled = true;
-			}
+				ke.Handled = true;
+			} else if (NavigateTabs (ke.KeyCode))
+				ke.Handled = true;
 
-			base.OnKeyDown (e);
+			base.OnKeyDown (ke);
 		}
 
-		protected override bool IsInputKey (Keys key)
+		protected override bool IsInputKey (Keys keyData)
 		{
-			switch (key & Keys.KeyCode) {
+			switch (keyData & Keys.KeyCode) {
 			case Keys.Home:
 			case Keys.End:
 			case Keys.Left:
 			case Keys.Right:
+			case Keys.Up:
+			case Keys.Down:
 				return true;
 			}
-			return base.IsInputKey (key);
+			return base.IsInputKey (keyData);
 		}
-
-		protected override void Dispose (bool disposing)
+		
+		private bool NavigateTabs (Keys keycode)
 		{
-			base.Dispose (disposing);
+			bool move_left = false;
+			bool move_right = false;
+			
+			if (alignment == TabAlignment.Bottom || alignment == TabAlignment.Top) {
+				if (keycode == Keys.Left)
+					move_left = true;
+				else if (keycode == Keys.Right)
+					move_right = true;
+			} else {
+				if (keycode == Keys.Up)
+					move_left = true;
+				else if (keycode == Keys.Down)
+					move_right = true;
+			}
+				
+			if (move_left) {
+				if (SelectedIndex > 0) {
+					SelectedIndex--;
+					return true;
+				}
+			}
+			
+			if (move_right) {
+				if (SelectedIndex < TabCount - 1) {
+					SelectedIndex++;
+					return true;
+				}
+			}
+			
+			return false;
 		}
+		#endregion
 
+		#region Pages Collection
 		protected void RemoveAll ()
 		{
 			Controls.Clear ();
@@ -584,12 +849,22 @@ namespace System.Windows.Forms {
 			return pages;
 		}
 
-		protected virtual object [] GetItems (Type type)
+		protected virtual object [] GetItems (Type baseType)
 		{
-			object [] pages = (object []) Array.CreateInstance (type, Controls.Count);
+			object[] pages = (object[])Array.CreateInstance (baseType, Controls.Count);
 			Controls.CopyTo (pages, 0);
 			return pages;
 		}
+		#endregion
+
+#if NET_2_0
+		protected void UpdateTabSelection (bool updateFocus)
+#else
+		protected void UpdateTabSelection (bool uiselected)
+#endif
+		{
+			ResizeTabPages ();
+		}
 
 		protected string GetToolTipText (object item)
 		{
@@ -618,60 +893,6 @@ namespace System.Windows.Forms {
 			}
 		}
 
-#if NET_2_0
-		protected virtual void OnDeselecting (TabControlCancelEventArgs e)
-		{
-			TabControlCancelEventHandler eh = (TabControlCancelEventHandler) (Events[DeselectingEvent]);
-			if (eh != null)
-				eh (this, e);
-
-			if (!e.Cancel)
-				OnDeselected (new TabControlEventArgs (SelectedTab, selected_index, TabControlAction.Deselected));
-		}
-
-		protected virtual void OnDeselected (TabControlEventArgs e)
-		{
-			TabControlEventHandler eh = (TabControlEventHandler) (Events[DeselectedEvent]);
-			if (eh != null)
-				eh (this, e);
-
-			if (this.SelectedTab != null)
-				this.SelectedTab.FireLeave ();
-		}
-
-		protected virtual void OnSelecting (TabControlCancelEventArgs e)
-		{
-			TabControlCancelEventHandler eh = (TabControlCancelEventHandler) (Events[SelectingEvent]);
-			if (eh != null)
-				eh (this, e);
-			
-			if (!e.Cancel)
-				OnSelected (new TabControlEventArgs (SelectedTab, selected_index, TabControlAction.Selected));
-		}
-
-		protected virtual void OnSelected (TabControlEventArgs e)
-		{
-			TabControlEventHandler eh = (TabControlEventHandler)(Events[SelectedEvent]);
-			if (eh != null)
-				eh (this, e);
-
-			if (this.SelectedTab != null)
-				this.SelectedTab.FireEnter ();
-		}
-#endif
-
-		protected virtual void OnSelectedIndexChanged (EventArgs e)
-		{
-			EventHandler eh = (EventHandler)(Events [SelectedIndexChangedEvent]);
-			if (eh != null)
-				eh (this, e);
-		}
-
-		internal override void OnPaintInternal (PaintEventArgs pe)
-		{
-			Draw (pe.Graphics, pe.ClipRectangle);
-			pe.Handled = true;
-		}
 		#endregion	// Protected Instance Methods
 
 		#region Internal & Private Methods
@@ -687,11 +908,14 @@ namespace System.Windows.Forms {
 
 		private void MouseDownHandler (object sender, MouseEventArgs e)
 		{
+			if ((e.Button & MouseButtons.Left) == 0)
+				return;
+
 			if (ShowSlider) {
-				Rectangle right = ThemeEngine.Current.TabControlGetRightScrollRect (this);
-				Rectangle left = ThemeEngine.Current.TabControlGetLeftScrollRect (this);
+				Rectangle right = RightScrollButtonArea;
+				Rectangle left = LeftScrollButtonArea;
 				if (right.Contains (e.X, e.Y)) {
-					right_slider_state = ButtonState.Pushed;
+					right_slider_state = PushButtonState.Pressed;
 					if (CanScrollRight) {
 						slider_pos++;
 						SizeTabs ();
@@ -716,7 +940,7 @@ namespace System.Windows.Forms {
 					}
 					return;
 				} else if (left.Contains (e.X, e.Y)) {
-					left_slider_state = ButtonState.Pushed;
+					left_slider_state = PushButtonState.Pressed;
 					if (CanScrollLeft) {
 						slider_pos--;
 						SizeTabs ();
@@ -747,35 +971,55 @@ namespace System.Windows.Forms {
 				if (!GetTabRect (i).Contains (e.X, e.Y))
 					continue;
 				SelectedIndex = i;
+				mouse_down_on_a_tab_page = true;
 				break;
 			}
 		}
 
 		private void MouseUpHandler (object sender, MouseEventArgs e)
 		{
-			if (ShowSlider && (left_slider_state == ButtonState.Pushed || right_slider_state == ButtonState.Pushed)) {
+			mouse_down_on_a_tab_page = false;
+			if (ShowSlider && (left_slider_state == PushButtonState.Pressed || right_slider_state == PushButtonState.Pressed)) {
 				Rectangle invalid;
-				if (left_slider_state == ButtonState.Pushed)
-					invalid = ThemeEngine.Current.TabControlGetLeftScrollRect (this);
-				else
-					invalid = ThemeEngine.Current.TabControlGetRightScrollRect (this);
-				left_slider_state = ButtonState.Normal;
-				right_slider_state = ButtonState.Normal;
-
+				if (left_slider_state == PushButtonState.Pressed) {
+					invalid = LeftScrollButtonArea;
+					left_slider_state = GetScrollButtonState (invalid, e.Location);
+				} else {
+					invalid = RightScrollButtonArea;
+					right_slider_state = GetScrollButtonState (invalid, e.Location);
+				}
 				Invalidate (invalid);
 			}
 		}
 
+		bool HasHotElementStyles {
+			get {
+				return ThemeElements.CurrentTheme.TabControlPainter.HasHotElementStyles (this);
+			}
+		}
+
+		Rectangle LeftScrollButtonArea {
+			get {
+				return ThemeElements.CurrentTheme.TabControlPainter.GetLeftScrollRect (this);
+			}
+		}
+
+		Rectangle RightScrollButtonArea {
+			get {
+				return ThemeElements.CurrentTheme.TabControlPainter.GetRightScrollRect (this);
+			}
+		}
+
+		static PushButtonState GetScrollButtonState (Rectangle scrollButtonArea, Point cursorLocation)
+		{
+			return scrollButtonArea.Contains (cursorLocation) ? PushButtonState.Hot : PushButtonState.Normal;
+		}
+
 		private void SizeChangedHandler (object sender, EventArgs e)
 		{
 			Redraw ();
 		}
 
-		internal void UpdateTabpage (TabPage page)
-		{
-
-		}
-
 		internal int IndexForTabPage (TabPage page)
 		{
 			for (int i = 0; i < tab_pages.Count; i++) {
@@ -830,15 +1074,13 @@ namespace System.Windows.Forms {
 				row_count = 1;
 			show_slider = false;
 			
-			int scrollerWidth = ThemeEngine.Current.TabControlScrollerWidth * 2;
-
 			for (int i = 0; i < TabPages.Count; i++) {
 				TabPage page = TabPages [i];
 				int aux = 0;
-				SizeTab (page, i, row_width - scrollerWidth, ref xpos, ref ypos, spacing, 0, ref aux, true);
+				SizeTab (page, i, row_width, ref xpos, ref ypos, spacing, 0, ref aux, true);
 			}
 
-			if (SelectedIndex != -1 && TabPages [SelectedIndex].Row != BottomRow)
+			if (SelectedIndex != -1 && TabPages.Count > SelectedIndex && TabPages[SelectedIndex].Row != BottomRow)
 				DropRow (TabPages [SelectedIndex].Row);
 		}
 
@@ -849,13 +1091,7 @@ namespace System.Windows.Forms {
 		private int Direction
 		{
 			get {
-				switch (Alignment) {
-				case TabAlignment.Right:
-				case TabAlignment.Bottom:
-					return -1;
-				default:
-					return 1;
-				}
+				return 1;
 			}
 		}
 
@@ -922,7 +1158,6 @@ namespace System.Windows.Forms {
 			Size spacing = TabSpacing;
 			int begin_prev = 0;
 
-			row_width -= ThemeEngine.Current.TabControlScrollerWidth * 2;
 			if (TabPages.Count == 0)
 				return;
 
@@ -955,7 +1190,7 @@ namespace System.Windows.Forms {
 			}
 
 			if (SelectedIndex != -1) {
-				ExpandSelected (TabPages [SelectedIndex], 2, row_width - 1);
+				ExpandSelected (TabPages [SelectedIndex], 0, row_width - 1);
 			}
 		}
 		
@@ -968,11 +1203,14 @@ namespace System.Windows.Forms {
 				width = item_size.Width;
 			} else {			
 				width = MeasureStringWidth (DeviceContext, page.Text, Font);
-				width += (Padding.X * 2) - 2; // don't know why, but tabpages on ms.net have the right padding smaller :p
+				width += (Padding.X * 2) + 2;
 
 				if (ImageList != null && page.ImageIndex >= 0 && page.ImageIndex < ImageList.Images.Count) {
 					width += ImageList.ImageSize.Width + ThemeEngine.Current.TabControlImagePadding.X;
-					item_size.Height = ImageList.ImageSize.Height + ThemeEngine.Current.TabControlImagePadding.Y;
+
+					int image_size = ImageList.ImageSize.Height + ThemeEngine.Current.TabControlImagePadding.Y;
+					if (item_size.Height < image_size)
+						item_size.Height = image_size;
 				}
 			}
 
@@ -1061,7 +1299,10 @@ namespace System.Windows.Forms {
 			
 				if (page.Row != prev_row) {
 					if (SizeMode == TabSizeMode.FillToRight && !ShowSlider) {
-						FillRow (begin_prev, i - 1, ((row_width - TabPages [i - 1].TabBounds.Right) / (i - begin_prev)), spacing);
+						bool vertical = alignment == TabAlignment.Right || alignment == TabAlignment.Left;
+						int offset = vertical ? TabPages [i - 1].TabBounds.Bottom : TabPages [i - 1].TabBounds.Right;
+						FillRow (begin_prev, i - 1, ((row_width - offset) / (i - begin_prev)), spacing,
+							vertical);
 					}
 					begin_prev = i;
 				}	
@@ -1124,7 +1365,7 @@ namespace System.Windows.Forms {
 					break;
 				case TabAlignment.Right:
 					r.Y -= ThemeEngine.Current.TabControlSelectedDelta.Y;
-					r.X += ThemeEngine.Current.TabControlSelectedDelta.X;
+					r.X -= ThemeEngine.Current.TabControlSelectedDelta.X;
 					break;
 			}
 
@@ -1132,7 +1373,10 @@ namespace System.Windows.Forms {
 			r.Height += ThemeEngine.Current.TabControlSelectedDelta.Height;
 			if (r.Left < left_edge)
 				r.X = left_edge;
-			if (r.Right > right_edge && SizeMode != TabSizeMode.Normal)
+			// Adjustment can't be used for right alignment, since it is
+			// the only one that has a different X origin than 0
+			if (r.Right > right_edge && SizeMode != TabSizeMode.Normal &&
+					alignment != TabAlignment.Right)
 				r.Width = right_edge - r.X;
 			page.TabBounds = r;
 		}
@@ -1149,10 +1393,23 @@ namespace System.Windows.Forms {
 
 		private void SetTab (int index, TabPage value)
 		{
-			((IList) Controls).Insert (index, value);
+			if (!tab_pages.Contains (value)) {
+				this.Controls.Add (value);
+			}
+			this.Controls.RemoveAt (index);
+			this.Controls.SetChildIndex (value, index);
 			Redraw ();
 		}
-
+#if NET_2_0
+		private void InsertTab (int index, TabPage value)
+		{
+			if (!tab_pages.Contains (value)) {
+				this.Controls.Add (value);
+			}
+			this.Controls.SetChildIndex (value, index);
+			Redraw ();
+		}
+#endif
 		internal void Redraw ()
 		{
 			if (!IsHandleCreated)
@@ -1180,6 +1437,47 @@ namespace System.Windows.Forms {
 			return (int)(rect.Width);
 		}
 
+		void OnMouseMove (object sender, MouseEventArgs e)
+		{
+			if (!mouse_down_on_a_tab_page && ShowSlider) {
+				if (LeftSliderState == PushButtonState.Pressed ||
+					RightSliderState == PushButtonState.Pressed)
+					return;
+				if (LeftScrollButtonArea.Contains (e.Location)) {
+					LeftSliderState = PushButtonState.Hot;
+					RightSliderState = PushButtonState.Normal;
+					EnteredTabPage = null;
+					return;
+				}
+				if (RightScrollButtonArea.Contains (e.Location)) {
+					RightSliderState = PushButtonState.Hot;
+					LeftSliderState = PushButtonState.Normal;
+					EnteredTabPage = null;
+					return;
+				}
+				LeftSliderState = PushButtonState.Normal;
+				RightSliderState = PushButtonState.Normal;
+			}
+			if (EnteredTabPage != null && EnteredTabPage.TabBounds.Contains (e.Location))
+				return;
+			for (int index = 0; index < TabCount; index++) {
+				TabPage tab_page = TabPages[index];
+				if (tab_page.TabBounds.Contains (e.Location)) {
+					EnteredTabPage = tab_page;
+					return;
+				}
+			}
+			EnteredTabPage = null;
+		}
+
+		void OnMouseLeave (object sender, EventArgs e)
+		{
+			if (ShowSlider) {
+				LeftSliderState = PushButtonState.Normal;
+				RightSliderState = PushButtonState.Normal;
+			}
+			EnteredTabPage = null;
+		}
 		#endregion	// Internal & Private Methods
 
 		#region Events
@@ -1197,6 +1495,16 @@ namespace System.Windows.Forms {
 			remove { base.BackgroundImageChanged -= value; }
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+		[EditorBrowsable (EditorBrowsableState.Never)]
+		public new event EventHandler BackgroundImageLayoutChanged
+		{
+			add { base.BackgroundImageLayoutChanged += value; }
+			remove { base.BackgroundImageLayoutChanged -= value; }
+		}
+#endif
+
 		[Browsable(false)]
 		[EditorBrowsable(EditorBrowsableState.Never)]
 		public new event EventHandler ForeColorChanged {
@@ -1262,11 +1570,21 @@ namespace System.Windows.Forms {
 			add { Events.AddHandler (DeselectingEvent, value); }
 			remove { Events.RemoveHandler (DeselectingEvent, value); }
 		}
+
+		static object RightToLeftLayoutChangedEvent = new object ();
+		public event EventHandler RightToLeftLayoutChanged
+		{
+			add { Events.AddHandler (RightToLeftLayoutChangedEvent, value); }
+			remove { Events.RemoveHandler (RightToLeftLayoutChangedEvent, value); }
+		}
 #endif
 		#endregion	// Events
 
 
 		#region Class TaControl.ControlCollection
+#if NET_2_0
+		[ComVisible (false)]
+#endif
 		public new class ControlCollection : System.Windows.Forms.Control.ControlCollection {
 
 			private TabControl owner;
@@ -1293,13 +1611,27 @@ namespace System.Windows.Forms {
 
 			public override void Remove (Control value)
 			{
+				bool change_index = false;
+				
 				TabPage page = value as TabPage;
-				if (page != null) {
+				if (page != null && owner.Controls.Contains (page)) {
 					int index = owner.IndexForTabPage (page);
 					if (index < owner.SelectedIndex || owner.SelectedIndex == Count - 1)
-						owner.SelectedIndex--;
+						change_index = true;
 				}
+				
 				base.Remove (value);
+				
+				// We don't want to raise SelectedIndexChanged until after we
+				// have removed from the collection, so TabCount will be
+				// correct for the user.
+				if (change_index && Count > 0)
+					owner.SelectedIndex--;
+				else if (change_index) {
+					owner.selected_index = -1;
+					owner.OnSelectedIndexChanged (EventArgs.Empty);
+				} else
+					owner.Redraw ();
 			}
 		}
 		#endregion	// Class TabControl.ControlCollection
@@ -1331,7 +1663,34 @@ namespace System.Windows.Forms {
 				}
 				set {
 					owner.SetTab (index, value);
-				}    
+				}
+			}
+#if NET_2_0
+			public virtual TabPage this [string key] {
+				get {
+					if (string.IsNullOrEmpty (key))
+						return null;
+
+					int index = this.IndexOfKey (key);
+					if (index < 0 || index >= this.Count)
+						return null;
+					
+					return this[index];
+				}
+			}
+#endif
+
+			internal int this[TabPage tabPage] {
+				get {
+					if (tabPage == null)
+						return -1;
+
+					for (int i = 0; i < this.Count; i++)
+						if (this[i].Equals (tabPage))
+							return i;
+
+					return -1;
+				}
 			}
 
 			bool ICollection.IsSynchronized {
@@ -1355,13 +1714,45 @@ namespace System.Windows.Forms {
 				}
 			}
 
-			public void Add (TabPage page)
+			public void Add (TabPage value)
 			{
-				if (page == null)
+				if (value == null)
 					throw new ArgumentNullException ("Value cannot be null.");
-				owner.Controls.Add (page);
+				owner.Controls.Add (value);
 			}
 
+#if NET_2_0
+			public void Add (string text)
+			{
+				TabPage page = new TabPage (text);
+				this.Add (page);
+			}
+
+			public void Add (string key, string text)
+			{
+				TabPage page = new TabPage (text);
+				page.Name = key;
+				this.Add (page);
+			}
+
+			public void Add (string key, string text, int imageIndex)
+			{
+				TabPage page = new TabPage (text);
+				page.Name = key;
+				page.ImageIndex = imageIndex;
+				this.Add (page);
+			}
+
+			// .Net sets the ImageKey, but does not show the image when this is used
+			public void Add (string key, string text, string imageKey)
+			{
+				TabPage page = new TabPage (text);
+				page.Name = key;
+				page.ImageKey = imageKey;
+				this.Add (page);
+			}
+#endif
+
 			public void AddRange (TabPage [] pages)
 			{
 				if (pages == null)
@@ -1372,6 +1763,7 @@ namespace System.Windows.Forms {
 			public virtual void Clear ()
 			{
 				owner.Controls.Clear ();
+				owner.Invalidate ();
 			}
 
 			public bool Contains (TabPage page)
@@ -1381,6 +1773,14 @@ namespace System.Windows.Forms {
 				return owner.Controls.Contains (page);
 			}
 
+#if NET_2_0
+			public virtual bool ContainsKey (string key)
+			{
+				int index = this.IndexOfKey (key);
+				return (index >= 0 && index < this.Count);
+			}
+#endif
+
 			public IEnumerator GetEnumerator ()
 			{
 				return owner.Controls.GetEnumerator ();
@@ -1391,16 +1791,44 @@ namespace System.Windows.Forms {
 				return owner.Controls.IndexOf (page);
 			}
 
-			public void Remove (TabPage page)
+#if NET_2_0
+			public virtual int IndexOfKey(string key)
 			{
-				owner.Controls.Remove (page);
+				if (string.IsNullOrEmpty (key))
+					return -1;
+
+				for (int i = 0; i < this.Count; i++) {
+					if (string.Compare (this[i].Name, key, true, 
+						System.Globalization.CultureInfo.InvariantCulture) == 0) {
+						return i;
+					}
+				}
+
+				return -1;
+			}
+#endif
+
+			public void Remove (TabPage value)
+			{
+				owner.Controls.Remove (value);
+				owner.Invalidate ();
 			}
 
 			public void RemoveAt (int index)
 			{
 				owner.Controls.RemoveAt (index);
+				owner.Invalidate ();
 			}
 
+#if NET_2_0
+			public virtual void RemoveByKey (string key)
+			{
+				int index = this.IndexOfKey (key);
+				if (index >= 0 && index < this.Count)
+					this.RemoveAt (index);
+			}
+#endif
+
 			void ICollection.CopyTo (Array dest, int index)
 			{
 				owner.Controls.CopyTo (dest, index);
@@ -1412,30 +1840,68 @@ namespace System.Windows.Forms {
 				if (value == null)
 					throw new ArgumentException ("value");
 				owner.Controls.Add (page);
-                                return owner.Controls.IndexOf (page);
+				return owner.Controls.IndexOf (page);
 			}
 
-			bool IList.Contains (object value)
+			bool IList.Contains (object page)
 			{
-				TabPage page = value as TabPage;
-				if (page == null)
+				TabPage tabPage = page as TabPage;
+				if (tabPage == null)
 					return false;
-				return Contains (page);
+				return Contains (tabPage);
 			}
 
-			int IList.IndexOf (object value)
+			int IList.IndexOf (object page)
 			{
-				TabPage page = value as TabPage;
-				if (page == null)
+				TabPage tabPage = page as TabPage;
+				if (tabPage == null)
 					return -1;
-				return IndexOf ((TabPage) page);
+				return IndexOf (tabPage);
 			}
 
+#if NET_2_0
+			void IList.Insert (int index, object tabPage)
+#else
 			void IList.Insert (int index, object value)
+#endif
 			{
 				throw new NotSupportedException ();
 			}
 
+#if NET_2_0
+			public void Insert (int index, string text)
+			{
+				owner.InsertTab (index, new TabPage (text));
+			}
+			
+			public void Insert (int index, TabPage tabPage)
+			{
+				owner.InsertTab (index, tabPage);
+			}
+
+			public void Insert (int index, string key, string text)
+			{
+				TabPage page = new TabPage(text);
+				page.Name = key;
+				owner.InsertTab (index, page);
+			}
+
+			public void Insert (int index, string key, string text, int imageIndex) 
+			{
+				TabPage page = new TabPage(text);
+				page.Name = key;
+				owner.InsertTab (index, page);
+				page.ImageIndex = imageIndex;
+			}
+
+			public void Insert (int index, string key, string text, string imageKey) 
+			{
+				TabPage page = new TabPage(text);
+				page.Name = key;
+				owner.InsertTab (index, page);
+				page.ImageKey = imageKey;
+			}
+#endif
 			void IList.Remove (object value)
 			{
 				TabPage page = value as TabPage;
