@@ -29,10 +29,15 @@ using System.Drawing;
 using System.Runtime.Serialization;
 using System.Text;
 
-namespace System.Windows.Forms {
+namespace System.Windows.Forms
+{
+#if NET_2_0
+	[DefaultProperty ("Text")]
+#endif
 	[TypeConverter(typeof(TreeNodeConverter))]
 	[Serializable]
-	public class TreeNode : MarshalByRefObject, ICloneable, ISerializable {
+	public class TreeNode : MarshalByRefObject, ICloneable, ISerializable
+	{
 		#region Fields
 		private TreeView tree_view;
 		internal TreeNode parent;
@@ -40,9 +45,16 @@ namespace System.Windows.Forms {
 		private string text;
 		private int image_index = -1;
 		private int selected_image_index = -1;
-        private string image_key = String.Empty;
-        private string selected_image_key = String.Empty;
-        internal TreeNodeCollection nodes;
+#if NET_2_0
+		private ContextMenu context_menu;
+		private ContextMenuStrip context_menu_strip;
+		private string image_key = String.Empty;
+		private string selected_image_key = String.Empty;
+		private int state_image_index = -1;
+		private string state_image_key = String.Empty;
+		private string tool_tip_text = String.Empty;
+#endif
+		internal TreeNodeCollection nodes;
 		internal TreeViewAction check_reason = TreeViewAction.Unknown;
 
 		internal int visible_order = 0;
@@ -61,22 +73,27 @@ namespace System.Windows.Forms {
 #endif
 		#endregion	// Fields
 
-		#region Internal Constructors		
+		#region Internal Constructors
 		internal TreeNode (TreeView tree_view) : this ()
 		{
 			this.tree_view = tree_view;
 			is_expanded = true;
 		}
 
-		private TreeNode (SerializationInfo info, StreamingContext context) : this ()
+#if NET_2_0
+		protected
+#else
+		private
+#endif
+		TreeNode (SerializationInfo serializationInfo, StreamingContext context) : this ()
 		{
 			SerializationInfoEnumerator	en;
 			SerializationEntry		e;
 			int				children;
 
-			en = info.GetEnumerator();
+			en = serializationInfo.GetEnumerator();
 			children = 0;
-			while (en.MoveNext()) { 
+			while (en.MoveNext()) {
 				e = en.Current;
 				switch(e.Name) {
 					case "Text": Text = (string)e.Value; break;
@@ -90,7 +107,7 @@ namespace System.Windows.Forms {
 			}
 			if (children > 0) {
 				for (int i = 0; i < children; i++) {
-					TreeNode node = (TreeNode) info.GetValue ("children" + i, typeof (TreeNode));
+					TreeNode node = (TreeNode) serializationInfo.GetValue ("children" + i, typeof (TreeNode));
 					Nodes.Add (node);
 				}
 			}
@@ -113,14 +130,15 @@ namespace System.Windows.Forms {
 			Nodes.AddRange (children);
 		}
 
-		public TreeNode (string text, int image_index, int selected_image_index) : this (text)
+		public TreeNode (string text, int imageIndex, int selectedImageIndex) : this (text)
 		{
-			this.image_index = image_index;
-			this.selected_image_index = selected_image_index;
+			this.image_index = imageIndex;
+			this.selected_image_index = selectedImageIndex;
 		}
 
-		public TreeNode (string text, int image_index, int selected_image_index,
-				TreeNode [] children) : this (text, image_index, selected_image_index)
+		public TreeNode (string text, int imageIndex, int selectedImageIndex,
+				TreeNode[] children)
+			: this (text, imageIndex, selectedImageIndex)
 		{
 			Nodes.AddRange (children);
 		}
@@ -145,31 +163,62 @@ namespace System.Windows.Forms {
 		#endregion	// ICloneable Members
 
 		#region ISerializable Members
-		void ISerializable.GetObjectData (SerializationInfo info, StreamingContext context)
+		void ISerializable.GetObjectData (SerializationInfo si, StreamingContext context)
 		{
-			info.AddValue ("Text", Text);
-			info.AddValue ("prop_bag", prop_bag, typeof (OwnerDrawPropertyBag));
-			info.AddValue ("ImageIndex", ImageIndex);
-			info.AddValue ("SelectedImageIndex", SelectedImageIndex);
-			info.AddValue ("Tag", Tag);
-			info.AddValue ("Checked", Checked);
-			
-			info.AddValue ("NumberOfChildren", Nodes.Count);
+			si.AddValue ("Text", Text);
+			si.AddValue ("prop_bag", prop_bag, typeof (OwnerDrawPropertyBag));
+			si.AddValue ("ImageIndex", ImageIndex);
+			si.AddValue ("SelectedImageIndex", SelectedImageIndex);
+			si.AddValue ("Tag", Tag);
+			si.AddValue ("Checked", Checked);
+
+			si.AddValue ("NumberOfChildren", Nodes.Count);
 			for (int i = 0; i < Nodes.Count; i++)
-				info.AddValue ("Child-" + i, Nodes [i], typeof (TreeNode));
+				si.AddValue ("Child-" + i, Nodes [i], typeof (TreeNode));
 		}
+
+#if NET_2_0
+		protected virtual void Deserialize (SerializationInfo serializationInfo, StreamingContext context)
+		{
+			Text = serializationInfo.GetString ("Text");
+			prop_bag = (OwnerDrawPropertyBag)serializationInfo.GetValue ("prop_bag", typeof (OwnerDrawPropertyBag));
+			ImageIndex = serializationInfo.GetInt32 ("ImageIndex");
+			SelectedImageIndex = serializationInfo.GetInt32 ("SelectedImageIndex");
+			Tag = serializationInfo.GetValue ("Tag", typeof (Object));
+			Checked = serializationInfo.GetBoolean ("Checked");
+			
+			int count = serializationInfo.GetInt32 ("NumberOfChildren");
+			
+			for (int i = 0; i < count; i++)
+				Nodes.Add ((TreeNode)serializationInfo.GetValue ("Child-" + i, typeof (TreeNode)));
+		}
+		
+		protected virtual void Serialize (SerializationInfo si,  StreamingContext context)
+		{
+			si.AddValue ("Text", Text);
+			si.AddValue ("prop_bag", prop_bag, typeof (OwnerDrawPropertyBag));
+			si.AddValue ("ImageIndex", ImageIndex);
+			si.AddValue ("SelectedImageIndex", SelectedImageIndex);
+			si.AddValue ("Tag", Tag);
+			si.AddValue ("Checked", Checked);
+
+			si.AddValue ("NumberOfChildren", Nodes.Count);
+			for (int i = 0; i < Nodes.Count; i++)
+				si.AddValue ("Child-" + i, Nodes[i], typeof (TreeNode));
+		}
+#endif
 		#endregion	// ISerializable Members
 
 		#region Public Instance Properties
 		public Color BackColor {
-			get { 
+			get {
 				if (prop_bag != null)
 					return prop_bag.BackColor;
 				if (TreeView != null)
 					return TreeView.BackColor;
 				return Color.Empty;
 			}
-			set { 
+			set {
 				if (prop_bag == null)
 					prop_bag = new OwnerDrawPropertyBag ();
 				prop_bag.BackColor = value;
@@ -180,6 +229,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public Rectangle Bounds {
 			get {
 				if (TreeView == null)
@@ -210,6 +262,10 @@ namespace System.Windows.Forms {
 			int indent_level = IndentLevel;
 			int roots = (TreeView.ShowRootLines ? 1 : 0);
 			int cb = (TreeView.CheckBoxes ? 19 : 0);
+#if NET_2_0
+			if (!TreeView.CheckBoxes && StateImage != null)
+				cb = 19;
+#endif
 			int imgs = (TreeView.ImageList != null ?  TreeView.ImageList.ImageSize.Width + 3 : 0);
 			return ((indent_level + roots) * TreeView.Indent) + cb + imgs - TreeView.hbar_offset;
 		}
@@ -222,7 +278,11 @@ namespace System.Windows.Forms {
 
 		internal int GetImageX ()
 		{
+#if NET_2_0
+			return GetLinesX () + (TreeView.CheckBoxes || StateImage != null ? 19 : 0);
+#else
 			return GetLinesX () + (TreeView.CheckBoxes ? 19 : 0);
+#endif
 		}
 
 		// In theory we should be able to track this instead of computing
@@ -241,12 +301,15 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[DefaultValue (false)]
+#endif
 		public bool Checked {
 			get { return check; }
 			set {
 				if (check == value)
 					return;
-			        TreeViewCancelEventArgs args = new TreeViewCancelEventArgs (this, false, check_reason);
+				TreeViewCancelEventArgs args = new TreeViewCancelEventArgs (this, false, check_reason);
 				if (TreeView != null)
 					TreeView.OnBeforeCheck (args);
 				if (!args.Cancel) {
@@ -260,6 +323,21 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[DefaultValue (null)]
+		public virtual ContextMenu ContextMenu {
+			get { return context_menu; }
+			set { context_menu = value; }
+		}
+		
+		[DefaultValue (null)]
+		public virtual ContextMenuStrip ContextMenuStrip {
+			get { return context_menu_strip; }
+			set { context_menu_strip = value; }
+		}
+		
+		[Browsable (false)]
+#endif
 		public TreeNode FirstNode {
 			get {
 				if (nodes.Count > 0)
@@ -287,6 +365,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public string FullPath {
 			get {
 				if (TreeView == null)
@@ -302,6 +383,13 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[DefaultValue (-1)]
+		[RelatedImageList ("TreeView.ImageList")]
+		[TypeConverter (typeof (TreeViewImageIndexConverter))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+#endif
 		[Localizable(true)]
 		public int ImageIndex {
 			get { return image_index; }
@@ -309,6 +397,9 @@ namespace System.Windows.Forms {
 				if (image_index == value)
 					return;
 				image_index = value;
+#if NET_2_0
+				image_key = string.Empty;
+#endif
 				TreeView tree = TreeView;
 				if (tree != null)
 					tree.UpdateNode (this);
@@ -316,23 +407,31 @@ namespace System.Windows.Forms {
 		}
 
 #if NET_2_0
-        [Localizable(true)]
-        public string ImageKey
-        {
-            get { return image_key; }
-            set
-            {
-                if (image_key == value)
-                    return;
-                image_key = value;
-                TreeView tree = TreeView;
-                if (tree != null)
-                    tree.UpdateNode(this);
-            }
-        }
+		[Localizable(true)]
+		[DefaultValue ("")]
+		[RelatedImageList ("TreeView.ImageList")]
+		[TypeConverter (typeof (TreeViewImageKeyConverter))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+		public string ImageKey {
+			get { return image_key; }
+			set {
+				if (image_key == value)
+					return;
+				image_key = value;
+				image_index = -1;
+
+				TreeView tree = TreeView;
+				if (tree != null)
+				tree.UpdateNode(this);
+			}
+		}
 #endif
 
-        public bool IsEditing {
+#if NET_2_0
+		[Browsable (false)]
+#endif
+		public bool IsEditing {
 			get {
 				TreeView tv = TreeView;
 				if (tv == null)
@@ -341,6 +440,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public bool IsExpanded {
 			get {
 				TreeView tv = TreeView;
@@ -356,11 +458,14 @@ namespace System.Windows.Forms {
 					if (!found)
 						return false;
 				}
-					
+
 				return is_expanded;
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public bool IsSelected {
 			get {
 				if (TreeView == null || !TreeView.IsHandleCreated)
@@ -369,24 +474,24 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public bool IsVisible {
 			get {
 				if (TreeView == null || !TreeView.IsHandleCreated || !TreeView.Visible)
 					return false;
 
-				if (visible_order < TreeView.skipped_nodes || visible_order - TreeView.skipped_nodes > TreeView.VisibleCount)
+				if (visible_order <= TreeView.skipped_nodes || visible_order - TreeView.skipped_nodes > TreeView.VisibleCount)
 					return false;
 
-				TreeNode parent = Parent;
-				while (parent != null) {
-					if (!parent.is_expanded)
-						return false;
-					parent = parent.Parent;
-				}
-				return true;
+				return ArePreviousNodesExpanded;
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public TreeNode LastNode {
 			get {
 				return (nodes == null || nodes.Count == 0) ? null : nodes [nodes.Count - 1];
@@ -394,6 +499,11 @@ namespace System.Windows.Forms {
 		}
 
 #if NET_2_0
+		[Browsable (false)]
+		public int Level {
+			get { return IndentLevel; }
+		}
+		
 		public string Name
 		{
 			get { return this.name; }
@@ -404,6 +514,9 @@ namespace System.Windows.Forms {
 		}
 #endif
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public TreeNode NextNode {
 			get {
 				if (parent == null)
@@ -414,10 +527,15 @@ namespace System.Windows.Forms {
 				return null;
 			}
 		}
-		
+
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public TreeNode NextVisibleNode {
 			get {
 				OpenTreeNodeEnumerator o = new OpenTreeNodeEnumerator (this);
+				o.MoveNext (); // move to the node itself
+
 				if (!o.MoveNext ())
 					return null;
 				TreeNode c = o.CurrentNode;
@@ -427,7 +545,10 @@ namespace System.Windows.Forms {
 			}
 		}
 
-		[Localizable(true)]
+#if NET_2_0
+		[DefaultValue (null)]
+#endif
+		[Localizable (true)]
 		public Font NodeFont {
 			get {
 				if (prop_bag != null)
@@ -444,7 +565,10 @@ namespace System.Windows.Forms {
 			}
 		}
 
-		[ListBindable(false)]
+#if NET_2_0
+		[Browsable (false)]
+#endif
+		[ListBindable (false)]
 		public TreeNodeCollection Nodes {
 			get {
 				if (nodes == null)
@@ -453,6 +577,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public TreeNode Parent {
 			get {
 				TreeView tree_view = TreeView;
@@ -462,6 +589,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public TreeNode PrevNode {
 			get {
 				if (parent == null)
@@ -473,9 +603,14 @@ namespace System.Windows.Forms {
 			}
 		}
 
+#if NET_2_0
+		[Browsable (false)]
+#endif
 		public TreeNode PrevVisibleNode {
 			get {
 				OpenTreeNodeEnumerator o = new OpenTreeNodeEnumerator (this);
+				o.MovePrevious (); // move to the node itself
+
 				if (!o.MovePrevious ())
 					return null;
 				TreeNode c = o.CurrentNode;
@@ -485,22 +620,67 @@ namespace System.Windows.Forms {
 			}
 		}
 
-		[Localizable(true)]
+#if NET_2_0
+		[DefaultValue (-1)]
+		[RelatedImageList ("TreeView.ImageList")]
+		[TypeConverter (typeof (TreeViewImageIndexConverter))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+#endif
+		[Localizable (true)]
 		public int SelectedImageIndex {
 			get { return selected_image_index; }
 			set { selected_image_index = value; }
 		}
 
 #if NET_2_0
-        [Localizable(true)]
-        public string SelectedImageKey
-        {
-            get { return selected_image_key; }
-            set { selected_image_key = value; }
-        }
+		[Localizable (true)]
+		[DefaultValue ("")]
+		[RelatedImageList ("TreeView.ImageList")]
+		[TypeConverter (typeof (TreeViewImageKeyConverter))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+		public string SelectedImageKey {
+			get { return selected_image_key; }
+			set { selected_image_key = value; }
+		}
+
+		[Localizable (true)]
+		[DefaultValue (-1)]
+		[RelatedImageList ("TreeView.StateImageList")]
+		[TypeConverter (typeof (NoneExcludedImageIndexConverter))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+		public int StateImageIndex {
+			get { return state_image_index; }
+			set {
+				if (state_image_index != value) {
+					state_image_index = value;
+					state_image_key = string.Empty;
+					InvalidateWidth ();
+				}
+			}
+		}
+
+		[Localizable (true)]
+		[DefaultValue ("")]
+		[RelatedImageList ("TreeView.StateImageList")]
+		[TypeConverter (typeof (ImageKeyConverter))]
+		[RefreshProperties (RefreshProperties.Repaint)]
+		[Editor ("System.Windows.Forms.Design.ImageIndexEditor, " + Consts.AssemblySystem_Design, typeof (System.Drawing.Design.UITypeEditor))]
+		public string StateImageKey {
+			get { return state_image_key; }
+			set {
+				if (state_image_key != value) {
+					state_image_key = value;
+					state_image_index = -1;
+					InvalidateWidth ();
+				}
+			}
+		}
 #endif
 
-        [Bindable(true)]
+		[Bindable(true)]
 		[Localizable(false)]
 		[TypeConverter(typeof(System.ComponentModel.StringConverter))]
 		[DefaultValue(null)]
@@ -521,9 +701,25 @@ namespace System.Windows.Forms {
 					return;
 				text = value;
 				InvalidateWidth ();
+#if NET_2_0
+				// UIA Framework Event: Text Changed
+				TreeView view = TreeView;
+				if (view != null)
+					view.OnUIANodeTextChanged (new TreeViewEventArgs (this));
+#endif
 			}
 		}
 
+#if NET_2_0
+		[DefaultValue ("")]
+		[Localizable (false)]
+		public string ToolTipText {
+			get { return tool_tip_text; }
+			set { tool_tip_text = value; }
+		}
+		
+		[Browsable (false)]
+#endif
 		public TreeView TreeView {
 			get {
 				if (tree_view != null)
@@ -540,7 +736,10 @@ namespace System.Windows.Forms {
 			}
 		}
 
-                public IntPtr Handle {
+#if NET_2_0
+		[Browsable (false)]
+#endif
+		public IntPtr Handle {
 			get {
 				// MS throws a NullReferenceException if the TreeView isn't set...
 				if (handle == IntPtr.Zero && TreeView != null)
@@ -561,36 +760,43 @@ namespace System.Windows.Forms {
 		}
 
 		#region Public Instance Methods
-		public void BeginEdit () {
+		public void BeginEdit ()
+		{
 			TreeView tv = TreeView;
 			if (tv != null)
 				tv.BeginEdit (this);
 		}
 
-		public void Collapse () {
+		public void Collapse ()
+		{
 			CollapseInternal (false);
 		}
 
 #if NET_2_0
-		public void Collapse (bool ignore_children)
+		public void Collapse (bool ignoreChildren)
 		{
-			if (ignore_children)
+			if (ignoreChildren)
 				Collapse ();
 			else
 				CollapseRecursive (this);
 		}
 #endif
-		public void EndEdit (bool cancel) {
+		public void EndEdit (bool cancel)
+		{
 			TreeView tv = TreeView;
 			if (!cancel && tv != null)
 				tv.EndEdit (this);
+			else if (cancel && tv != null)
+				tv.CancelEdit (this);
 		}
 
-		public void Expand () {
-			Expand(false);
+		public void Expand ()
+		{
+			Expand (false);
 		}
 
-		public void ExpandAll () {
+		public void ExpandAll ()
+		{
 			ExpandRecursive (this);
 			if(TreeView != null)
 				TreeView.UpdateNode (TreeView.root_node);
@@ -612,8 +818,9 @@ namespace System.Windows.Forms {
 			}
 		}
 
-		public int GetNodeCount (bool include_subtrees) {
-			if (!include_subtrees)
+		public int GetNodeCount (bool includeSubTrees)
+		{
+			if (!includeSubTrees)
 				return Nodes.Count;
 
 			int count = 0;
@@ -622,21 +829,24 @@ namespace System.Windows.Forms {
 			return count;
 		}
 
-		public void Remove () {
+		public void Remove ()
+		{
 			if (parent == null)
 				return;
 			int index = Index;
 			parent.Nodes.RemoveAt (index);
 		}
 
-		public void Toggle () {
+		public void Toggle ()
+		{
 			if (is_expanded)
 				Collapse ();
 			else
 				Expand ();
 		}
 
-		public override String ToString () {
+		public override String ToString ()
+		{
 			return String.Concat ("TreeNode: ", Text);
 		}
 
@@ -644,6 +854,19 @@ namespace System.Windows.Forms {
 
 		#region Internal & Private Methods and Properties
 
+		internal bool ArePreviousNodesExpanded {
+			get {
+				TreeNode parent = Parent;
+				while (parent != null) {
+					if (!parent.is_expanded)
+						return false;
+					parent = parent.Parent;
+				}
+
+				return true;
+			}
+		}
+
 		internal bool IsRoot {
 			get {
 				TreeView tree_view = TreeView;
@@ -698,14 +921,13 @@ namespace System.Windows.Forms {
 					tree_view.OnAfterExpand (new TreeViewEventArgs (this));
 
 					tree_view.RecalculateVisibleOrder (this);
-					tree_view.UpdateScrollBars ();
+					tree_view.UpdateScrollBars (false);
 
-					if (IsVisible)
+					// ExpandBelow if we affect the visible area
+					if (visible_order < tree_view.skipped_nodes + tree_view.VisibleCount + 1 && ArePreviousNodesExpanded)
 						tree_view.ExpandBelow (this, count_to_next);
 				}
 			}
-
-			
 		}
 
 		private void CollapseInternal (bool byInternal)
@@ -732,13 +954,22 @@ namespace System.Windows.Forms {
 				if (tree_view != null) {
 					tree_view.OnAfterCollapse (new TreeViewEventArgs (this));
 
-					tree_view.RecalculateVisibleOrder (this);
-					tree_view.UpdateScrollBars ();
+					bool hbar_visible = tree_view.hbar.Visible;
+					bool vbar_visible = tree_view.vbar.Visible;
 
-					if (IsVisible)
+					tree_view.RecalculateVisibleOrder (this);
+					tree_view.UpdateScrollBars (false);
+
+					// CollapseBelow if we affect the visible area
+					if (visible_order < tree_view.skipped_nodes + tree_view.VisibleCount + 1 && ArePreviousNodesExpanded)
 						tree_view.CollapseBelow (this, count_to_next);
 					if(!byInternal && HasFocusInChildren ())
 						tree_view.SelectedNode = this;
+
+					// If one or both of our scrollbars disappeared,
+					// invalidate everything
+					if ((hbar_visible & !tree_view.hbar.Visible) || (vbar_visible & !tree_view.vbar.Visible))
+						tree_view.Invalidate ();
 				}
 			}
 		}
@@ -766,7 +997,8 @@ namespace System.Windows.Forms {
 
 		private bool HasFocusInChildren()
 		{
-			if(TreeView == null) return false;
+			if (TreeView == null)
+				return false;
 			foreach (TreeNode node in nodes) {
 				if(node == TreeView.SelectedNode)
 					return true;
@@ -779,9 +1011,8 @@ namespace System.Windows.Forms {
 		private void ExpandRecursive (TreeNode node)
 		{
 			node.Expand (true);
-			foreach (TreeNode child in node.Nodes) {
+			foreach (TreeNode child in node.Nodes)
 				ExpandRecursive (child);
-			}
 		}
 
 		private void ExpandParentRecursive (TreeNode node)
@@ -804,18 +1035,16 @@ namespace System.Windows.Forms {
 		private void CollapseRecursive (TreeNode node)
 		{
 			node.Collapse ();
-			foreach (TreeNode child in node.Nodes) {
+			foreach (TreeNode child in node.Nodes)
 				CollapseRecursive (child);
-			}
 		}
 
 		private void CollapseUncheckRecursive (TreeNode node)
 		{
 			node.Collapse ();
 			node.Checked = false;
-			foreach (TreeNode child in node.Nodes) {
+			foreach (TreeNode child in node.Nodes)
 				CollapseUncheckRecursive (child);
-			}
 		}
 
 		internal void SetNodes (TreeNodeCollection nodes)
@@ -826,9 +1055,8 @@ namespace System.Windows.Forms {
 		private void GetNodeCountRecursive (TreeNode node, ref int count)
 		{
 			count += node.Nodes.Count;
-			foreach (TreeNode child in node.Nodes) {
+			foreach (TreeNode child in node.Nodes)
 				GetNodeCountRecursive (child, ref count);
-			}
 		}
 
 		internal bool NeedsWidth {
@@ -851,8 +1079,7 @@ namespace System.Windows.Forms {
 			this.parent = parent;
 		}
 
-		private bool IsInClippingRect
-		{
+		private bool IsInClippingRect {
 			get {
 				if (TreeView == null)
 					return false;
@@ -862,9 +1089,63 @@ namespace System.Windows.Forms {
 				return true;
 			}
 		}
-		#endregion	// Internal & Private Methods and Properties
 
+#if NET_2_0
+		internal Image StateImage {
+			get {
+				if (TreeView != null) {
+					if (TreeView.StateImageList == null)
+						return null;
+					if (state_image_index >= 0)
+						return TreeView.StateImageList.Images[state_image_index];
+					if (state_image_key != string.Empty)
+						return TreeView.StateImageList.Images[state_image_key];
+				}
+
+				return null;
+			}
+		}
+#endif
+
+		// Order of operation:
+		// 1) Node.Image[Key|Index]
+		// 2) TreeView.Image[Key|Index]
+		// 3) First image in TreeView.ImageList
+		internal int Image {
+			get {
+				if (TreeView == null || TreeView.ImageList == null)
+					return -1;
+					
+				if (IsSelected) {
+					if (selected_image_index >= 0)
+						return selected_image_index;
+#if NET_2_0
+					if (!string.IsNullOrEmpty (selected_image_key))
+						return TreeView.ImageList.Images.IndexOfKey (selected_image_key);
+					if (!string.IsNullOrEmpty (TreeView.SelectedImageKey))
+						return TreeView.ImageList.Images.IndexOfKey (TreeView.SelectedImageKey);
+#endif
+					if (TreeView.SelectedImageIndex >= 0)
+						return TreeView.SelectedImageIndex;
+				} else {
+					if (image_index >= 0)
+						return image_index;
+#if NET_2_0
+					if (!string.IsNullOrEmpty (image_key))
+						return TreeView.ImageList.Images.IndexOfKey (image_key);
+					if (!string.IsNullOrEmpty (TreeView.ImageKey))
+						return TreeView.ImageList.Images.IndexOfKey (TreeView.ImageKey);
+#endif
+					if (TreeView.ImageIndex >= 0)
+						return TreeView.ImageIndex;
+				}
+
+				if (TreeView.ImageList.Images.Count > 0)
+					return 0;
+					
+				return -1;
+			}
+		}
+		#endregion	// Internal & Private Methods and Properties
 	}
 }
-
-
