@@ -38,13 +38,13 @@ using System.Collections;
 using System.Diagnostics;
 using System.Runtime.InteropServices;
 using System.Threading;
-using System.Windows.Forms.X11Internal;
 
 namespace System.Windows.Forms {
 	internal class XplatUI {
 		#region Local Variables
 		static XplatUIDriver		driver;
 		static String			default_class_name;
+		internal static ArrayList key_filters = new ArrayList ();
 		#endregion	// Local Variables
 
 		#region Private Methods
@@ -83,16 +83,27 @@ namespace System.Windows.Forms {
 			// and try to register same class name we fail.
 			default_class_name = "SWFClass" + System.Threading.Thread.GetDomainID().ToString();
 
-			// check for Unix platforms - see FAQ for more details
-			// http://www.mono-project.com/FAQ:_Technical#How_to_detect_the_execution_platform_.3F
-			int platform = (int) Environment.OSVersion.Platform;
-			if ((platform == 4) || (platform == 128)) {
-				if (Environment.GetEnvironmentVariable ("not_supported_MONO_MWF_USE_QUARTZ_BACKEND") != null)
-					driver=XplatUIOSX.GetInstance ();
-				else if (Environment.GetEnvironmentVariable ("not_supported_MONO_MWF_USE_NEW_X11_BACKEND") != null)
-					driver=XplatUIX11_new.GetInstance ();
-				else
-					driver=XplatUIX11.GetInstance ();
+			if (RunningOnUnix) {
+				//if (Environment.GetEnvironmentVariable ("not_supported_MONO_MWF_USE_NEW_X11_BACKEND") != null) {
+				//        driver=XplatUIX11_new.GetInstance ();
+				//} else 
+				if (Environment.GetEnvironmentVariable ("MONO_MWF_MAC_FORCE_X11") != null) {
+					driver = XplatUIX11.GetInstance ();
+				} else {
+					IntPtr buf = Marshal.AllocHGlobal (8192);
+					// This is a hacktastic way of getting sysname from uname ()
+					if (uname (buf) != 0) {
+						// WTF: We cannot run uname
+						driver=XplatUIX11.GetInstance ();
+					} else {
+						string os = Marshal.PtrToStringAnsi (buf);
+						if (os == "Darwin")
+							driver=XplatUICarbon.GetInstance ();
+						else
+							driver=XplatUIX11.GetInstance ();
+					}
+					Marshal.FreeHGlobal (buf);
+				}
 			} else {
 				driver=XplatUIWin32.GetInstance();
 			}
@@ -101,10 +112,30 @@ namespace System.Windows.Forms {
 
 			// Initialize things that need to be done after the driver is ready
 			DataFormats.GetFormat(0);
+
+#if NET_2_0
+			// Signal that the Application loop can be run.
+			// This allows UIA to initialize a11y support for MWF
+			// before the main loop begins.
+			Application.FirePreRun ();
+#endif
 		}
 		#endregion	// Constructor & Destructor
 
 		#region Public Static Properties
+
+		public static bool RunningOnUnix {
+			get {
+				int p = (int) Environment.OSVersion.Platform;
+				
+				return (p == 4 || p == 6 || p == 128);
+			}
+		}
+
+		public static int ActiveWindowTrackingDelay {
+			get { return driver.ActiveWindowTrackingDelay; }
+		}
+
 		internal static string DefaultClassName {
 			get {
 				return default_class_name;
@@ -115,18 +146,51 @@ namespace System.Windows.Forms {
 			}
 		}
 
-		static public int CaptionHeight {
+		static public Size Border3DSize {
 			get {
-				return driver.Caption;
+				return driver.Border3DSize;
 			}
 		}
 
+		static public Size BorderSize {
+			get {
+				return driver.BorderSize;
+			}
+		}
+
+		static public Size CaptionButtonSize {
+			get {
+				return driver.CaptionButtonSize;
+			}
+		}
+
+		static public int CaptionHeight {
+			get {
+				return driver.CaptionHeight;
+			}
+		}
+
+		public static int CaretBlinkTime { get { return driver.CaretBlinkTime; } }
+		public static int CaretWidth { get { return driver.CaretWidth; } }
+		
 		static public Size CursorSize {
 			get {
 				return driver.CursorSize;
 			}
 		}
 
+		static public Size DoubleClickSize {
+			get {
+				return driver.DoubleClickSize;
+			}
+		}
+
+		static public int DoubleClickTime {
+			get {
+				return driver.DoubleClickTime;
+			}
+		}
+
 		static public bool DragFullWindows {
 			get {
 				return driver.DragFullWindows;
@@ -139,12 +203,23 @@ namespace System.Windows.Forms {
 			}
 		}
 
+		static public Size FixedFrameBorderSize {
+			get {
+				return driver.FixedFrameBorderSize;
+			}
+		}
+
+		public static int FontSmoothingContrast { get { return driver.FontSmoothingContrast; } }
+		public static int FontSmoothingType { get { return driver.FontSmoothingType; } }
+
 		public static Size FrameBorderSize {
 			get {
 				return driver.FrameBorderSize;
 			}
 		}
 
+		public static int HorizontalResizeBorderThickness { get { return driver.HorizontalResizeBorderThickness; } }
+		
 		static public int HorizontalScrollBarHeight {
 			get {
 				return driver.HorizontalScrollBarHeight;
@@ -157,6 +232,22 @@ namespace System.Windows.Forms {
 			}
 		}
 
+		public static bool IsActiveWindowTrackingEnabled { get { return driver.IsActiveWindowTrackingEnabled; } }
+		public static bool IsComboBoxAnimationEnabled { get { return driver.IsComboBoxAnimationEnabled; } }
+		public static bool IsDropShadowEnabled { get { return driver.IsDropShadowEnabled; } }
+		public static bool IsFontSmoothingEnabled { get { return driver.IsFontSmoothingEnabled; } }
+		public static bool IsHotTrackingEnabled { get { return driver.IsHotTrackingEnabled; } }
+		public static bool IsIconTitleWrappingEnabled { get { return driver.IsIconTitleWrappingEnabled; } }
+		public static bool IsKeyboardPreferred { get { return driver.IsKeyboardPreferred; } }
+		public static bool IsListBoxSmoothScrollingEnabled { get { return driver.IsListBoxSmoothScrollingEnabled; } }
+		public static bool IsMenuAnimationEnabled { get { return driver.IsMenuAnimationEnabled; } }
+		public static bool IsMenuFadeEnabled { get { return driver.IsMenuFadeEnabled; } }
+		public static bool IsMinimizeRestoreAnimationEnabled { get { return driver.IsMinimizeRestoreAnimationEnabled; } }
+		public static bool IsSelectionFadeEnabled { get	{ return driver.IsSelectionFadeEnabled; } }
+		public static bool IsSnapToDefaultEnabled { get { return driver.IsSnapToDefaultEnabled; } }
+		public static bool IsTitleBarGradientEnabled { get { return driver.IsTitleBarGradientEnabled; } }
+		public static bool IsToolTipAnimationEnabled { get { return driver.IsToolTipAnimationEnabled; } }
+		
 		static public int KeyboardSpeed {
 			get {
 				return driver.KeyboardSpeed;
@@ -175,6 +266,22 @@ namespace System.Windows.Forms {
 			}
 		}
 
+		static public bool MenuAccessKeysUnderlined {
+			get {
+				return driver.MenuAccessKeysUnderlined;
+			}
+		}
+
+		static public Size MenuBarButtonSize { get { return driver.MenuBarButtonSize; } }
+
+		public static Size MenuButtonSize {
+			get {
+				return driver.MenuButtonSize;
+			}
+		}
+
+		static public int MenuShowDelay { get { return driver.MenuShowDelay; } }
+		
 		static public Size MinimizedWindowSize {
 			get {
 				return driver.MinimizedWindowSize;
@@ -193,12 +300,28 @@ namespace System.Windows.Forms {
 			}
 		}
 
+		static public Size MinimumFixedToolWindowSize {
+			get { return driver.MinimumFixedToolWindowSize; }
+		}
+
+		static public Size MinimumSizeableToolWindowSize {
+			get { return driver.MinimumSizeableToolWindowSize; }
+		}
+
+		static public Size MinimumNoBorderWindowSize {
+			get { return driver.MinimumNoBorderWindowSize; }
+		}
+
 		static public Size MinWindowTrackSize {
 			get {
 				return driver.MinWindowTrackSize;
 			}
 		}
 
+		public static int MouseSpeed {
+			get { return driver.MouseSpeed; }
+		}
+		
 		static public Size SmallIconSize {
 			get {
 				return driver.SmallIconSize;
@@ -247,12 +370,42 @@ namespace System.Windows.Forms {
 			}
 		}
 
+		public static LeftRightAlignment PopupMenuAlignment {
+			get { return driver.PopupMenuAlignment; }
+		}
+		
+#if NET_2_0
+		public static PowerStatus PowerStatus {
+			get { return driver.PowerStatus; }
+		}
+#endif
+
+		public static bool RequiresPositiveClientAreaSize {
+			get {
+				return driver.RequiresPositiveClientAreaSize;
+			}
+		}
+
+		public static int SizingBorderWidth {
+			get { return driver.SizingBorderWidth; }
+		}
+		
+		public static Size SmallCaptionButtonSize {
+			get { return driver.SmallCaptionButtonSize; }
+		}
+		
+		public static bool UIEffectsEnabled {
+			get { return driver.UIEffectsEnabled; }
+		}
+		
 		static public bool UserClipWontExposeParent {
 			get {
 				return driver.UserClipWontExposeParent;
 			}
 		}
 
+		public static int VerticalResizeBorderThickness { get { return driver.VerticalResizeBorderThickness; } }
+
 		static public int VerticalScrollBarWidth {
 			get {
 				return driver.VerticalScrollBarWidth;
@@ -276,20 +429,31 @@ namespace System.Windows.Forms {
 				return XplatUI.driver.ThemesEnabled;
 			}
 		}
- 
 
+		public static int ToolWindowCaptionHeight {
+			get {
+				return driver.ToolWindowCaptionHeight;
+			}
+		}
+
+		public static Size ToolWindowCaptionButtonSize {
+			get {
+				return driver.ToolWindowCaptionButtonSize;
+			}
+		}
 		#endregion	// Public Static Properties
 
 		#region Events
-                internal static event EventHandler Idle {
-                        add {
-                                driver.Idle += value;
-                        }
-                        remove {
-                                driver.Idle -= value;
-                        }
-                }
-                
+
+		internal static event EventHandler Idle {
+			add {
+				driver.Idle += value;
+			}
+			remove {
+				driver.Idle -= value;
+			}
+		}
+
 		#endregion	// Events
 
 		#region Public Static Methods
@@ -407,6 +571,10 @@ namespace System.Windows.Forms {
 		internal static IntPtr DefineStdCursor(StdCursor id) {
 			return driver.DefineStdCursor(id);
 		}
+		
+		internal static Bitmap DefineStdCursorBitmap(StdCursor id) {
+			return driver.DefineStdCursorBitmap(id);
+		}
 
 		internal static IntPtr DefWndProc(ref Message msg) {
 			return driver.DefWndProc(ref msg);
@@ -564,6 +732,10 @@ namespace System.Windows.Forms {
 			return driver.GetParent(handle);
 		}
 
+		internal static IntPtr GetPreviousWindow(IntPtr handle) {
+			return driver.GetPreviousWindow(handle);
+		}
+
 		internal static bool GetText(IntPtr handle, out string text) {
 			#if DriverDebug
 				Console.WriteLine("GetText({0}): Called", Window(handle));
@@ -632,6 +804,14 @@ namespace System.Windows.Forms {
 			return driver.IsEnabled(handle);
 		}
 
+		internal static bool IsKeyLocked (VirtualKeys key)
+		{
+			#if DriverDebug || DriverDebugState
+				Console.WriteLine("IsKeyLocked({0}): Called, Result={1}", key, driver.IsKeyLocked(key));
+			#endif
+			return driver.IsKeyLocked (key);
+		}
+
 		internal static bool IsVisible(IntPtr handle) {
 			#if DriverDebug || DriverDebugState
 				Console.WriteLine("IsVisible({0}): Called, Result={1}", Window(handle), driver.IsVisible(handle));
@@ -661,18 +841,18 @@ namespace System.Windows.Forms {
 			driver.OverrideCursor(cursor);
 		}
 
-		internal static void PaintEventEnd(IntPtr handle, bool client) {
+		internal static void PaintEventEnd (ref Message msg, IntPtr handle, bool client) {
 			#if DriverDebug || DriverDebugPaint
-				Console.WriteLine("PaintEventEnd({0}, {1}): Called from thread {2}", Window(handle), client, Thread.CurrentThread.GetHashCode());
+				Console.WriteLine("PaintEventEnd({0}, {1}, {2}): Called from thread {3}", msg, Window(handle), client, Thread.CurrentThread.GetHashCode());
 			#endif
-			driver.PaintEventEnd(handle, client);
+			driver.PaintEventEnd (ref msg, handle, client);
 		}
 
-		internal static PaintEventArgs PaintEventStart(IntPtr handle, bool client) {
+		internal static PaintEventArgs PaintEventStart (ref Message msg, IntPtr handle, bool client) {
 			#if DriverDebug || DriverDebugPaint
-				Console.WriteLine("PaintEventStart({0}, {1}): Called from thread {2}", Window(handle), client, Thread.CurrentThread.GetHashCode());
+				Console.WriteLine("PaintEventStart({0}, {1}, {2}): Called from thread {3}", msg, Window(handle), client, Thread.CurrentThread.GetHashCode());
 			#endif
-			return driver.PaintEventStart(handle, client);
+			return driver.PaintEventStart (ref msg, handle, client);
 		}
 
 		internal static bool PeekMessage(Object queue_id, ref MSG msg, IntPtr hWnd, int wFilterMin, int wFilterMax, uint flags) {
@@ -700,6 +880,15 @@ namespace System.Windows.Forms {
 			driver.PostQuitMessage(exitCode);
 		}
 
+		internal static void RaiseIdle (EventArgs e)
+		{
+			#if DriverDebug
+				Console.WriteLine("RaiseIdle({0}): Called", e.ToString ());
+			#endif
+			
+			driver.RaiseIdle (e);
+		}
+		
 		internal static void RequestAdditionalWM_NCMessages(IntPtr handle, bool hover, bool leave) {
 			#if DriverDebug
 				Console.WriteLine("RequestAdditionalWM_NCMessages({0}, {1}, {2}): Called", Window(handle), hover, leave);
@@ -744,7 +933,7 @@ namespace System.Windows.Forms {
 
 		internal static void ScrollWindow(IntPtr handle, int XAmount, int YAmount, bool with_children) {
 			#if DriverDebug
-				Console.WriteLine("ScrollWindow({0}, {2}, {3}, {4}): Called", Window(handle), XAmount, YAmount, with_children);
+				Console.WriteLine("ScrollWindow({0}, {1}, {2}, {3}): Called", Window(handle), XAmount, YAmount, with_children);
 			#endif
 			driver.ScrollWindow(handle, XAmount, YAmount, with_children);
 		}
@@ -827,6 +1016,13 @@ namespace System.Windows.Forms {
 			driver.SetFocus(handle);
 		}
 
+		internal static void SetForegroundWindow(IntPtr handle) {
+			#if DriverDebug
+				Console.WriteLine("SetForegroundWindow({0}): Called", Window(handle));
+			#endif
+			driver.SetForegroundWindow(handle);
+		}
+
 		internal static void SetIcon(IntPtr handle, Icon icon) {
 			#if DriverDebug
 				Console.WriteLine("SetIcon({0}, {1}): Called", Window(handle), icon);
@@ -1064,6 +1260,25 @@ namespace System.Windows.Forms {
 		internal static void Version() {
 			Console.WriteLine("Xplat version $Revision: $");
 		}
+
+		internal static void AddKeyFilter (IKeyFilter value)
+		{
+			lock (key_filters) {
+				key_filters.Add (value);
+			}
+		}
+
+		internal static bool FilterKey (KeyFilterData key)
+		{
+			lock (key_filters) {
+				for (int i = 0; i < key_filters.Count; i++) {
+					IKeyFilter filter = (IKeyFilter) key_filters[i];
+					if (filter.PreFilterKey (key))
+						return true;
+				}
+			}
+			return false;
+		}
 		#endregion	// Public Static Methods
 
 		#region	Delegates
@@ -1071,5 +1286,7 @@ namespace System.Windows.Forms {
 		public delegate bool ObjectToClipboard(ref int type, object obj, out byte[] data);
 		#endregion	// Delegates
 
+		[DllImport ("libc")]
+		static extern int uname (IntPtr buf);
 	}
 }
