@@ -16,7 +16,7 @@ using NUnit.Framework;
 namespace MonoTests.System.Windows.Forms
 {
 	[TestFixture]
-	public class TabControlTest
+	public class TabControlTest : TestHelper
 	{
 		private int _selected_index_changed = 0;
 
@@ -34,9 +34,9 @@ namespace MonoTests.System.Windows.Forms
 		}
 
 		[SetUp]
-		public void SetUp ()
-		{
+		protected override void SetUp () {
 			_selected_index_changed = 0;
+			base.SetUp ();
 		}
 
 		[Test]
@@ -611,6 +611,17 @@ namespace MonoTests.System.Windows.Forms
 			form.Dispose ();
 		}
 
+		[Test] // bug #81802. should not throw an exception
+		public void NoTabPages2 ()
+		{
+			Form form = new Form ();
+			form.Font = new Font ("Tahoma", 8.25F, FontStyle.Regular, GraphicsUnit.Point, ((byte) (0)));
+			TabControl tab = new TabControl ();
+			form.Controls.Add (tab);
+			tab.Dispose ();
+			form.Dispose ();
+		}
+
 		private void SelectedIndexChanged (object sender, EventArgs e)
 		{
 			_selected_index_changed++;
@@ -621,12 +632,41 @@ namespace MonoTests.System.Windows.Forms
 		public void MethodSelectTab ()
 		{
 			TabControl tc = new TabControl ();
-			tc.TabPages.Add (new TabPage ("One"));
-			tc.TabPages.Add (new TabPage ("Two"));
+			tc.TabPages.Add ("One", "One");
+			tc.TabPages.Add ("Two", "Two");
+			tc.TabPages.Add ("Three", "Three");
 			
 			tc.SelectTab (1);
 			Assert.AreEqual (1, tc.SelectedIndex, "A1");
-		}
+
+			tc.SelectTab (tc.TabPages[2]);
+			Assert.AreEqual (2, tc.SelectedIndex, "A2");
+
+			tc.SelectTab ("One");
+			Assert.AreEqual (0, tc.SelectedIndex, "A3");
+
+			tc.SelectTab (1);
+			tc.DeselectTab (1);
+			Assert.AreEqual (2, tc.SelectedIndex, "A4");
+
+			tc.DeselectTab (2);
+			Assert.AreEqual (0, tc.SelectedIndex, "A5");
+
+			tc.DeselectTab (tc.TabPages[0]);
+			Assert.AreEqual (1, tc.SelectedIndex, "A6");		
+
+			tc.DeselectTab (tc.TabPages[0]);
+			Assert.AreEqual (1, tc.SelectedIndex, "A7");		
+
+			tc.DeselectTab (tc.TabPages[1]);
+			Assert.AreEqual (2, tc.SelectedIndex, "A7");		
+
+			tc.DeselectTab ("Two");
+			Assert.AreEqual (2, tc.SelectedIndex, "A8");		
+
+			tc.DeselectTab ("Three");
+			Assert.AreEqual (0, tc.SelectedIndex, "A8");		
+}
 
 		[Test]
 		[ExpectedException (typeof (ArgumentOutOfRangeException))]
@@ -685,6 +725,69 @@ namespace MonoTests.System.Windows.Forms
 			Assert.AreEqual (1, tcea2.TabPageIndex, "A4");
 			f.Close ();
 		}
+
+		[Test]
+		[Category ("NotWorking")]
+		public void EnterLeaveEvents ()
+		{
+			Form f = new Form ();
+			TabControl tc = new TabControl ();
+			TabPage p1 = new TabPage ();
+			string events = string.Empty;
+			tc.Enter += new EventHandler (delegate (Object obj, EventArgs e) { events += ("tc_OnEnter;"); });
+			tc.Leave += new EventHandler (delegate (Object obj, EventArgs e) { events += ("tc_OnLeave;"); });
+			p1.Enter += new EventHandler (delegate (Object obj, EventArgs e) { events += ("p1_OnEnter;");});
+			p1.Leave += new EventHandler (delegate (Object obj, EventArgs e) { events += ("p1_OnLeave;"); });
+			tc.TabPages.Add (p1);
+
+			Button b1 = new Button ();
+
+			tc.TabIndex = 0;
+			b1.TabIndex = 1;
+
+			f.Controls.Add (b1);
+			f.Controls.Add (tc);
+
+			f.Show ();
+			Assert.AreEqual ("tc_OnEnter;p1_OnEnter;", events, "A1");
+			b1.Focus ();
+			Assert.AreEqual ("tc_OnEnter;p1_OnEnter;p1_OnLeave;tc_OnLeave;", events, "A2");
+		}
+
+		
+
 #endif
 	}
+
+	[TestFixture]
+	public class TabPageCollectionTest : TestHelper
+	{
+		[Test]
+		public void Indexer ()
+		{
+			TabControl tab = new TabControl ();
+			TabControl.TabPageCollection tabPages = new TabControl.TabPageCollection (tab);
+			TabPage tabPageA = new TabPage ();
+			TabPage tabPageB = new TabPage ();
+			TabPage tabPageC = new TabPage ();
+			TabPage tabPageD = new TabPage ();
+			tabPages.Add (tabPageA);
+			Assert.AreSame (tabPageA, tabPages [0], "#1");
+			tabPages [0] = tabPageB;
+			Assert.AreSame (tabPageB, tabPages [0], "#2");
+			tabPages.Add (tabPageC);
+			Assert.AreSame (tabPageB, tabPages [0], "#3");
+			Assert.AreSame (tabPageC, tabPages [1], "#4");
+			tabPages.Remove (tabPageB);
+			Assert.AreSame (tabPageC, tabPages [0], "#5");
+			tabPages [0] = tabPageD;
+			Assert.AreSame (tabPageD, tabPages [0], "#6");
+
+			Form form = new Form ();
+			form.ShowInTaskbar = false;
+			form.Controls.Add (tab);
+			form.Show ();
+			form.Dispose ();
+		}
+	}
 }
