@@ -172,7 +172,7 @@ namespace System.Data.OracleClient.Oci
 
 			[DllImport ("oci")]
 			internal static extern int OCIDefineByPos (IntPtr stmtp,
-				ref IntPtr defnpp,
+				out IntPtr defnpp,
 				IntPtr errhp,
 				[MarshalAs (UnmanagedType.U4)] int position,
 				IntPtr valuep,
@@ -660,7 +660,7 @@ namespace System.Data.OracleClient.Oci
 			out uint fsec);
 
 		internal static int OCIDefineByPos (IntPtr stmtp,
-			ref IntPtr defnpp,
+			out IntPtr defnpp,
 			IntPtr errhp,
 			int position,
 			IntPtr valuep,
@@ -674,7 +674,7 @@ namespace System.Data.OracleClient.Oci
 			#if TRACE
 			Trace.WriteLineIf(traceOci, "OCIDefineByPos", "OCI");
 			#endif
-			return OciNativeCalls.OCIDefineByPos (stmtp, ref defnpp, errhp, position, valuep,
+			return OciNativeCalls.OCIDefineByPos (stmtp, out defnpp, errhp, position, valuep,
 				value_sz, dty, ref indp, ref rlenp, rcodep, mode);
 		}
 
@@ -1151,7 +1151,7 @@ namespace System.Data.OracleClient.Oci
 		#region AllocateClear
 
 		private static bool IsUnix =
-		(int) Environment.OSVersion.Platform == 4 || (int) Environment.OSVersion.Platform == 128;
+			(int) Environment.OSVersion.Platform == 4 || (int) Environment.OSVersion.Platform == 128 || (int) Environment.OSVersion.Platform == 6;
 
 		[DllImport("libc")]
 		private static extern IntPtr calloc (int nmemb, int size);
