@@ -7,6 +7,7 @@
 //
 
 using System;
+using System.IO;
 using System.Net;
 using System.Text;
 using System.Threading;
@@ -43,13 +44,19 @@ namespace MonoTests.Remoting
 			ShutdownServer ();
 		}
 
+		public static AppDomain CreateDomain (string friendlyName)
+		{
+			// return AppDomain.CreateDomain (friendlyName);
+			return AppDomain.CreateDomain (friendlyName, null, Directory.GetCurrentDirectory (), ".", false);
+		}
+
 		protected virtual int CreateServer ()
 		{
 			ChannelManager cm = CreateChannelManager ();
 			chs = cm.CreateClientChannel ();
 			ChannelServices.RegisterChannel (chs);
 
-			AppDomain domain = AppDomain.CreateDomain ("testdomain");
+			AppDomain domain = BaseCallTest.CreateDomain ("testdomain");
 			server = (CallsDomainServer) domain.CreateInstanceAndUnwrap(GetType().Assembly.FullName,"MonoTests.Remoting.CallsDomainServer");
 			remoteUris = server.Start (cm);
 			return server.GetDomId ();
@@ -507,16 +514,33 @@ namespace MonoTests.Remoting
 	
 	public class RemoteObject : AbstractRemoteObject, IRemoteObject
 	{
+		int inc = 0;
+		
 		public override int Simple ()
 		{
 			return 130772 + Thread.GetDomainID();
 		}
-
+		
+		public int ReturnOne ()
+		{
+			return 1;
+		}
+		
+		public int Increment ()
+		{
+			return inc++;
+		}
+		
 		public override string PrimitiveParams (int a, uint b, char c, string d)
 		{
 			return "" + a + "-" + b + "-" + c + "-" + d + "@" + Thread.GetDomainID();
 		}
 
+		// declare an overload for bug #77191
+		public void PrimitiveParams ()
+		{
+		}
+
 		public override string PrimitiveParamsInOut (ref int a1, out int a2, ref float b1, out float b2, int filler, ref char c1, out char c2, ref string d1, out string d2)
 		{
 			string res = "" + a1 + "-" + b1.ToString(CultureInfo.InvariantCulture) + "-" + c1 + "-" + d1 + "@" + Thread.GetDomainID();
