@@ -88,18 +88,18 @@ namespace System.Web.UI.WebControls
 		DataKey key;
 		DataKey oldEditValues;
 		
-		private static readonly object PageIndexChangedEvent = new object();
-		private static readonly object PageIndexChangingEvent = new object();
-		private static readonly object ItemCommandEvent = new object();
-		private static readonly object ItemCreatedEvent = new object();
-		private static readonly object ItemDeletedEvent = new object();
-		private static readonly object ItemDeletingEvent = new object();
-		private static readonly object ItemInsertedEvent = new object();
-		private static readonly object ItemInsertingEvent = new object();
-		private static readonly object ModeChangingEvent = new object();
-		private static readonly object ModeChangedEvent = new object();
-		private static readonly object ItemUpdatedEvent = new object();
-		private static readonly object ItemUpdatingEvent = new object();
+		static readonly object PageIndexChangedEvent = new object();
+		static readonly object PageIndexChangingEvent = new object();
+		static readonly object ItemCommandEvent = new object();
+		static readonly object ItemCreatedEvent = new object();
+		static readonly object ItemDeletedEvent = new object();
+		static readonly object ItemDeletingEvent = new object();
+		static readonly object ItemInsertedEvent = new object();
+		static readonly object ItemInsertingEvent = new object();
+		static readonly object ModeChangingEvent = new object();
+		static readonly object ModeChangedEvent = new object();
+		static readonly object ItemUpdatedEvent = new object();
+		static readonly object ItemUpdatingEvent = new object();
 		
 		// Control state
 		int pageIndex;
@@ -464,7 +464,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate EditItemTemplate {
 			get { return editItemTemplate; }
-			set { editItemTemplate = value; RequireBinding (); }
+			set { editItemTemplate = value; }
 		}
 
 		[WebCategoryAttribute ("Styles")]
@@ -505,7 +505,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate EmptyDataTemplate {
 			get { return emptyDataTemplate; }
-			set { emptyDataTemplate = value; RequireBinding (); }
+			set { emptyDataTemplate = value; }
 		}
 		
 		[LocalizableAttribute (true)]
@@ -538,7 +538,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate FooterTemplate {
 			get { return footerTemplate; }
-			set { footerTemplate = value; RequireBinding (); }
+			set { footerTemplate = value; }
 		}
 
 		[LocalizableAttribute (true)]
@@ -616,7 +616,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate HeaderTemplate {
 			get { return headerTemplate; }
-			set { headerTemplate = value; RequireBinding (); }
+			set { headerTemplate = value; }
 		}
 
 		[LocalizableAttribute (true)]
@@ -653,7 +653,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate InsertItemTemplate {
 			get { return insertItemTemplate; }
-			set { insertItemTemplate = value; RequireBinding (); }
+			set { insertItemTemplate = value; }
 		}
 
 		[WebCategoryAttribute ("Styles")]
@@ -678,7 +678,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate ItemTemplate {
 			get { return itemTemplate; }
-			set { itemTemplate = value; RequireBinding (); }
+			set { itemTemplate = value; }
 		}
 		
 		[BrowsableAttribute (false)]
@@ -746,7 +746,7 @@ namespace System.Web.UI.WebControls
 		[Browsable (false)]
 		public virtual ITemplate PagerTemplate {
 			get { return pagerTemplate; }
-			set { pagerTemplate = value; RequireBinding (); }
+			set { pagerTemplate = value; }
 		}
 		
 		[DesignerSerializationVisibilityAttribute (DesignerSerializationVisibility.Hidden)]
@@ -865,7 +865,7 @@ namespace System.Web.UI.WebControls
 				if (RequiresDataBinding) {
 					OnDataBinding (EventArgs.Empty);
 					RequiresDataBinding = false;
-					PerformDataBinding (null);
+					InternalPerformDataBinding (null);
 					MarkAsDataBound ();
 					OnDataBound (EventArgs.Empty);
 				}
@@ -987,10 +987,10 @@ namespace System.Web.UI.WebControls
 				table.Rows.Add (bottomPagerRow);
 			}
 
+			OnItemCreated (EventArgs.Empty);
+			
 			if (dataBinding)
 				DataBind (false);
-			
-			OnItemCreated (EventArgs.Empty);
 
 			return PageCount;
 		}
@@ -1111,6 +1111,7 @@ namespace System.Web.UI.WebControls
 			
 			if (bt != null) {
 				IOrderedDictionary values = bt.ExtractValues (Row.Cells [0]);
+				if (values != null)
 				foreach (DictionaryEntry e in values) {
 					if (includeKeys || Array.IndexOf (DataKeyNames, e.Key) == -1)
 						fieldValues [e.Key] = e.Value;
