@@ -47,6 +47,7 @@ namespace System.Web.UI.WebControls
 	[DefaultEvent ("MenuItemClick")]
 	[ControlValueProperty ("SelectedValue")]
 	[Designer ("System.Web.UI.Design.WebControls.MenuDesigner, " + Consts.AssemblySystem_Design, "System.ComponentModel.Design.IDesigner")]
+	[SupportsEventValidation]
 	public class Menu : HierarchicalDataBoundControl, IPostBackEventHandler, INamingContainer
 	{
 		MenuItemStyle dynamicMenuItemStyle;
@@ -85,8 +86,8 @@ namespace System.Web.UI.WebControls
 		Style dynamicHoverLinkStyle;
 		Style staticHoverLinkStyle;
 
-		private static readonly object MenuItemClickEvent = new object();
-		private static readonly object MenuItemDataBoundEvent = new object();
+		static readonly object MenuItemClickEvent = new object();
+		static readonly object MenuItemDataBoundEvent = new object();
 		
 		public static readonly string MenuItemClickCommandName = "Click";
 		
@@ -368,7 +369,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style PopOutBoxStyle {
+		Style PopOutBoxStyle {
 			get {
 				if (popOutBoxStyle == null) {
 					popOutBoxStyle = new Style ();
@@ -378,7 +379,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style ControlLinkStyle {
+		Style ControlLinkStyle {
 			get {
 				if (controlLinkStyle == null) {
 					controlLinkStyle = new Style ();
@@ -388,7 +389,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style DynamicMenuItemLinkStyle {
+		Style DynamicMenuItemLinkStyle {
 			get {
 				if (dynamicMenuItemLinkStyle == null) {
 					dynamicMenuItemLinkStyle = new Style ();
@@ -397,7 +398,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style StaticMenuItemLinkStyle {
+		Style StaticMenuItemLinkStyle {
 			get {
 				if (staticMenuItemLinkStyle == null) {
 					staticMenuItemLinkStyle = new Style ();
@@ -406,7 +407,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style DynamicSelectedLinkStyle {
+		Style DynamicSelectedLinkStyle {
 			get {
 				if (dynamicSelectedLinkStyle == null) {
 					dynamicSelectedLinkStyle = new Style ();
@@ -415,7 +416,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style StaticSelectedLinkStyle {
+		Style StaticSelectedLinkStyle {
 			get {
 				if (staticSelectedLinkStyle == null) {
 					staticSelectedLinkStyle = new Style ();
@@ -424,7 +425,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style DynamicHoverLinkStyle {
+		Style DynamicHoverLinkStyle {
 			get {
 				if (dynamicHoverLinkStyle == null) {
 					dynamicHoverLinkStyle = new Style ();
@@ -433,7 +434,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private Style StaticHoverLinkStyle {
+		Style StaticHoverLinkStyle {
 			get {
 				if (staticHoverLinkStyle == null) {
 					staticHoverLinkStyle = new Style ();
@@ -856,7 +857,7 @@ namespace System.Web.UI.WebControls
 			EnsureChildControlsDataBound ();
 		}
 
-		private void FillBoundChildrenRecursive (IHierarchicalEnumerable hEnumerable, MenuItemCollection itemCollection)
+		void FillBoundChildrenRecursive (IHierarchicalEnumerable hEnumerable, MenuItemCollection itemCollection)
 		{
 			if (hEnumerable == null)
 				return;
@@ -901,6 +902,7 @@ namespace System.Web.UI.WebControls
 		
 		protected internal virtual void RaisePostBackEvent (string eventArgument)
 		{
+			ValidateEvent (UniqueID, eventArgument);
 			if (!Enabled)
 				return;
 
@@ -1070,7 +1072,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private void CreateChildControlsForItems () {
+		void CreateChildControlsForItems () {
 			Controls.Clear ();
 			// Check for HasChildViewState to avoid unnecessary calls to ClearChildViewState.
 			if (HasChildViewState)
@@ -1080,7 +1082,7 @@ namespace System.Web.UI.WebControls
 			_requiresChildControlsDataBinding = true;
 		}
 
-		private void CreateChildControlsForItems (MenuItemCollection items ) {
+		void CreateChildControlsForItems (MenuItemCollection items ) {
 			foreach (MenuItem item in items) {
 				bool isDynamicItem = IsDynamicItem (item);
 				if (isDynamicItem && dynamicItemTemplate != null) {
@@ -1107,7 +1109,7 @@ namespace System.Web.UI.WebControls
 			EnsureChildControlsDataBound ();
 		}
 
-		private void EnsureChildControlsDataBound () {
+		void EnsureChildControlsDataBound () {
 			if (!_requiresChildControlsDataBinding)
 				return;
 			DataBindChildren ();
@@ -1516,7 +1518,7 @@ namespace System.Web.UI.WebControls
 			}
 		}
 
-		private void FillMenuStyle (bool dynamic, int menuLevel, SubMenuStyle style) {
+		void FillMenuStyle (bool dynamic, int menuLevel, SubMenuStyle style) {
 			if (Page.Header != null) {
 				// styles are registered
 				if (!dynamic && staticMenuStyle != null) {
@@ -1597,11 +1599,11 @@ namespace System.Web.UI.WebControls
 			}
 		}
 		
-		private bool IsDynamicItem (MenuItem item) {
+		bool IsDynamicItem (MenuItem item) {
 			return item.Depth + 1 > StaticDisplayLevels;
 		}
 
-		private bool DisplayChildren (MenuItem item) {
+		bool DisplayChildren (MenuItem item) {
 			return (item.Depth + 1 < StaticDisplayLevels + MaximumDynamicDisplayLevels) && item.ChildItems.Count > 0;
 		}
 		
@@ -1990,10 +1992,10 @@ namespace System.Web.UI.WebControls
 		{
 			return ClientID + "_" + item.Path + sufix;
 		}
-							
+		
 		string GetClientEvent (MenuItem item)
 		{
-			return Page.ClientScript.GetPostBackClientHyperlink (this, item.Path);
+			return Page.ClientScript.GetPostBackClientHyperlink (this, item.Path, true);
 		}
 
 		class MenuTemplateWriter : TextWriter
@@ -2031,7 +2033,7 @@ namespace System.Web.UI.WebControls
 					_buffer [_ptr++] = value [i];
 			}
 
-			private void EnsureCapacity ()
+			void EnsureCapacity ()
 			{
 				char [] tmpBuffer = new char [_buffer.Length * 2];
 				Array.Copy (_buffer, tmpBuffer, _buffer.Length);
