@@ -66,6 +66,7 @@ using System.IO;
 using System.Collections;
 using System.ComponentModel;
 using System.Configuration;
+using System.Diagnostics;
 using System.Globalization;
 using System.Reflection;
 using System.Security.Permissions;
@@ -92,6 +93,11 @@ namespace System.Web {
 		static readonly object disposedEvent = new object ();
 		static readonly object errorEvent = new object ();
 		
+		// we do this static rather than per HttpApplication because
+		// mono's perfcounters use the counter instance parameter for
+		// the process to access shared memory.
+		internal static PerformanceCounter requests_total_counter = new PerformanceCounter ("ASP.NET", "Requests Total");
+		
 		internal static readonly string [] BinDirs = {"Bin", "bin"};
 		object this_lock = new object();
 		
@@ -197,7 +203,7 @@ namespace System.Web {
 				
 #if NET_2_0
 				HttpModulesSection modules;
-				modules = (HttpModulesSection) WebConfigurationManager.GetSection ("system.web/httpModules", HttpRuntime.AppDomainAppVirtualPath);
+				modules = (HttpModulesSection) WebConfigurationManager.GetWebApplicationSection ("system.web/httpModules");
 #else
 				ModulesConfiguration modules;
 
@@ -328,7 +334,12 @@ namespace System.Web {
 
 				if (context == null)
 					throw new HttpException (Locale.GetText ("No context is available."));
-				return context.Session;
+
+				HttpSessionState ret = context.Session;
+				if (ret == null)
+					throw new HttpException (Locale.GetText ("Session state is not available in the context."));
+				
+				return ret;
 			}
 		}
 
@@ -967,10 +978,11 @@ namespace System.Web {
 
 		void async_handler_complete_cb (IAsyncResult ar)
 		{
-			IHttpAsyncHandler async_handler = ((IHttpAsyncHandler) ar.AsyncState);
+			IHttpAsyncHandler async_handler = ar != null ? ar.AsyncState as IHttpAsyncHandler : null;
 
 			try {
-				async_handler.EndProcessRequest (ar);
+				if (async_handler != null)
+					async_handler.EndProcessRequest (ar);
 			} catch (Exception e){
 				ProcessError (e);
 			}
@@ -1085,6 +1097,9 @@ namespace System.Web {
 
 			try {
 				OutputPage ();
+			} catch (ThreadAbortException taex) {
+				ProcessError (taex);
+				Thread.ResetAbort ();
 			} catch (Exception e) {
 				Console.WriteLine ("Internal error: OutputPage threw an exception " + e);
 			} finally {
@@ -1107,8 +1122,51 @@ namespace System.Web {
 				begin_iar.Complete ();
 			else
 				done.Set ();
+
+			requests_total_counter.Increment ();
 		}
-		
+
+		class Tim {
+			string name;
+			DateTime start;
+
+			public Tim () {
+			}
+
+			public Tim (string name) {
+				this.name = name;
+			}
+
+			public string Name {
+				get { return name; }
+				set { name = value; }
+			}
+
+			public void Start () {
+				start = DateTime.UtcNow;
+			}
+
+			public void Stop () {
+				Console.WriteLine ("{0}: {1}ms", name, (DateTime.UtcNow - start).TotalMilliseconds);
+			}
+		}
+
+		Tim tim;
+		[Conditional ("PIPELINE_TIMER")]
+		void StartTimer (string name)
+		{
+			if (tim == null)
+				tim = new Tim ();
+			tim.Name = name;
+			tim.Start ();
+		}
+
+		[Conditional ("PIPELINE_TIMER")]
+		void StopTimer ()
+		{
+			tim.Stop ();
+		}
+
 		//
 		// Events fired as described in `Http Runtime Support, HttpModules,
 		// Handling Public Events'
@@ -1119,56 +1177,75 @@ namespace System.Web {
 			if (stop_processing)
 				yield return true;
 
+			StartTimer ("BeginRequest");
 			eventHandler = Events [BeginRequestEvent];
 			if (eventHandler != null) {
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
 			}
+			StopTimer ();
 			
+			StartTimer ("AuthenticateRequest");
 			eventHandler = Events [AuthenticateRequestEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 
+			StartTimer ("DefaultAuthentication");
 			if (DefaultAuthentication != null)
 				foreach (bool stop in RunHooks (DefaultAuthentication))
 					yield return stop;
+			StopTimer ();
 
 #if NET_2_0
+			StartTimer ("PostAuthenticateRequest");
 			eventHandler = Events [PostAuthenticateRequestEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #endif
+			StartTimer ("AuthorizeRequest");
 			eventHandler = Events [AuthorizeRequestEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #if NET_2_0
+			StartTimer ("PostAuthorizeRequest");
 			eventHandler = Events [PostAuthorizeRequestEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #endif
 
+			StartTimer ("ResolveRequestCache");
 			eventHandler = Events [ResolveRequestCacheEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 
 #if NET_2_0
+			StartTimer ("PostResolveRequestCache");
 			eventHandler = Events [PostResolveRequestCacheEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 
+			StartTimer ("MapRequestHandler");
 			// As per http://msdn2.microsoft.com/en-us/library/bb470252(VS.90).aspx
 			eventHandler = Events [MapRequestHandlerEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #endif
 			
+			StartTimer ("GetHandler");
 			// Obtain the handler for the request.
 			IHttpHandler handler = null;
 			try {
@@ -1194,28 +1271,35 @@ namespace System.Web {
 				ProcessError (e);
 			}
 
+			StopTimer ();
 			if (stop_processing)
 				yield return true;
 
 #if NET_2_0
+			StartTimer ("PostMapRequestHandler");
 			eventHandler = Events [PostMapRequestHandlerEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #endif
 
+			StartTimer ("AcquireRequestState");
 			eventHandler = Events [AcquireRequestStateEvent];
 			if (eventHandler != null){
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
 			}
+			StopTimer ();
 			
 #if NET_2_0
+			StartTimer ("PostAcquireRequestState");
 			eventHandler = Events [PostAcquireRequestStateEvent];
 			if (eventHandler != null){
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
 			}
+			StopTimer ();
 #endif
 			
 			//
@@ -1223,11 +1307,13 @@ namespace System.Web {
 			// ReleaseRequestState, so the code below jumps to
 			// `release:' to guarantee it rather than yielding.
 			//
+			StartTimer ("PreRequestHandlerExecute");
 			eventHandler = Events [PreRequestHandlerExecuteEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					if (stop)
 						goto release;
+			StopTimer ();
 			
 				
 #if TARGET_J2EE
@@ -1244,6 +1330,7 @@ namespace System.Web {
 			}
 #endif
 
+			StartTimer ("ProcessRequest");
 			try {
 				context.BeginTimeoutPossible ();
 				if (handler != null){
@@ -1268,6 +1355,7 @@ namespace System.Web {
 				in_begin = false;
 				context.EndTimeoutPossible ();
 			}
+			StopTimer ();
 #if TARGET_J2EE
 			if (doProcessHandler) {
 				yield return false;
@@ -1281,13 +1369,16 @@ namespace System.Web {
 			
 			// These are executed after the application has returned
 			
+			StartTimer ("PostRequestHandlerExecute");
 			eventHandler = Events [PostRequestHandlerExecuteEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					if (stop)
 						goto release;
+			StopTimer ();
 			
 		release:
+			StartTimer ("ReleaseRequestState");
 			eventHandler = Events [ReleaseRequestStateEvent];
 			if (eventHandler != null){
 #pragma warning disable 219
@@ -1299,42 +1390,57 @@ namespace System.Web {
 				}
 #pragma warning restore 219
 			}
+			StopTimer ();
 			
 			if (stop_processing)
 				yield return true;
 
 #if NET_2_0
+			StartTimer ("PostReleaseRequestState");
 			eventHandler = Events [PostReleaseRequestStateEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #endif
 
+			StartTimer ("Filter");
 			if (context.Error == null)
 				context.Response.DoFilter (true);
+			StopTimer ();
 
+			StartTimer ("UpdateRequestCache");
 			eventHandler = Events [UpdateRequestCacheEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 
 #if NET_2_0
+			StartTimer ("PostUpdateRequestCache");
 			eventHandler = Events [PostUpdateRequestCacheEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 
+			StartTimer ("LogRequest");
 			eventHandler = Events [LogRequestEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 
+			StartTimer ("PostLogRequest");
 			eventHandler = Events [PostLogRequestEvent];
 			if (eventHandler != null)
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
+			StopTimer ();
 #endif
+			StartTimer ("PipelineDone");
 			PipelineDone ();
+			StopTimer ();
 		}
 
 
@@ -1425,6 +1531,8 @@ namespace System.Web {
 				context = HttpContext.Current;
 			context.StopTimeoutTimer ();
 #endif
+			context.Request.ReleaseResources ();
+			context.Response.ReleaseResources ();
 			context = null;
 			session = null;
 			HttpContext.Current = null;
@@ -1487,7 +1595,7 @@ namespace System.Web {
 				return ret;
 			
 #if NET_2_0
-			HttpHandlersSection	httpHandlersSection = (HttpHandlersSection) WebConfigurationManager.GetSection ("system.web/httpHandlers");
+			HttpHandlersSection httpHandlersSection = (HttpHandlersSection) WebConfigurationManager.GetWebApplicationSection ("system.web/httpHandlers");
 			ret = httpHandlersSection.LocateHandler (verb, url);
 #else
 			HandlerFactoryConfiguration factory_config = (HandlerFactoryConfiguration) HttpContext.GetAppConfig ("system.web/httpHandlers");
