@@ -33,6 +33,7 @@ using System.Reflection;
 using System.Web.UI;
 using System.Web.SessionState;
 using System.Web.Configuration;
+using System.Threading;
 
 using System.Web.Compilation;
 #if TARGET_J2EE
@@ -69,9 +70,10 @@ namespace System.Web {
 #else
 		static HttpApplicationFactory theFactory = new HttpApplicationFactory();
 #endif
-		MethodInfo session_end;
+		object session_end; // This is a MethodInfo
 		bool needs_init = true;
 		bool app_start_needed = true;
+		bool have_app_events;
 		Type app_type;
 		HttpApplicationState app_state;
 		Hashtable app_event_handlers;
@@ -81,8 +83,11 @@ namespace System.Web {
 #if NET_2_0
 		static bool app_disabled = false;
 		static string[] app_browsers_files = new string[0];
+		static string[] default_machine_browsers_files = new string[0];
+		static string[] app_mono_machine_browsers_files = new string[0];
 #endif
 		Stack available = new Stack ();
+		object next_free;
 		Stack available_for_end = new Stack ();
 		
 		bool IsEventHandler (MethodInfo m)
@@ -128,22 +133,52 @@ namespace System.Web {
 
 			list.Add (method);
 		}
+
+		ArrayList GetMethodsDeep (Type type)
+		{
+			ArrayList al = new ArrayList ();
+			MethodInfo[] methods = type.GetMethods (BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Instance  | BindingFlags.Static | BindingFlags.FlattenHierarchy);
+			al.AddRange (methods);
+
+			Type t = type.BaseType;
+			while (t != null) {
+				methods = t.GetMethods (BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Static);
+				al.AddRange (methods);
+				t = t.BaseType;
+			}
+
+			return al;
+		}
 		
 		Hashtable GetApplicationTypeEvents (Type type)
 		{
+			if (have_app_events)
+				return app_event_handlers;
+
 			lock (this_lock) {
 				if (app_event_handlers != null)
 					return app_event_handlers;
 
 				app_event_handlers = new Hashtable ();
-				BindingFlags flags = BindingFlags.Public    | BindingFlags.NonPublic | 
-					BindingFlags.Instance  | BindingFlags.Static;
-
-				MethodInfo [] methods = type.GetMethods (flags);
-				foreach (MethodInfo m in methods) {
-					if (m.DeclaringType != typeof (HttpApplication) && IsEventHandler (m))
+				ArrayList methods = GetMethodsDeep (type);
+				Hashtable used = null;
+				MethodInfo m;
+				string mname;
+				
+				foreach (object o in methods) {
+					m = o as MethodInfo;
+					if (m.DeclaringType != typeof (HttpApplication) && IsEventHandler (m)) {
+						mname = m.ToString ();
+						if (used == null)
+							used = new Hashtable ();
+						else if (used.ContainsKey (mname))
+							continue;
+						used.Add (mname, m);
 						AddEvent (m, app_event_handlers);
+					}
 				}
+				used = null;
+				have_app_events = true;
 			}
 
 			return app_event_handlers;
@@ -151,12 +186,10 @@ namespace System.Web {
 
 		Hashtable GetApplicationTypeEvents (HttpApplication app)
 		{
-			lock (this_lock) {
-				if (app_event_handlers != null)
-					return app_event_handlers;
+			if (have_app_events)
+				return app_event_handlers;
 
-				return GetApplicationTypeEvents (app.GetType ());
-			}
+			return GetApplicationTypeEvents (app.GetType ());
 		}
 
 		bool FireEvent (string method_name, object target, object [] args)
@@ -249,17 +282,14 @@ namespace System.Web {
 
 				string usualName = moduleName + "_" + eventName;
 				object methodData = possibleEvents [usualName];
-				if (methodData != null && eventName == "End" && moduleName == "Session") {
-					lock (factory) {
-						if (factory.session_end == null)
-							factory.session_end = (MethodInfo) methodData;
-					}
-					continue;
-				}
-
 				if (methodData == null)
 					continue;
 
+				if (eventName == "End" && moduleName == "Session") {
+					Interlocked.CompareExchange (ref factory.session_end, methodData, null);
+					continue;
+				}
+
 				if (methodData is MethodInfo) {
 					factory.AddHandler (evt, target, app, (MethodInfo) methodData);
 					continue;
@@ -279,14 +309,19 @@ namespace System.Web {
 				NoParamsInvoker npi = new NoParamsInvoker (app, method);
 				evt.AddEventHandler (target, npi.FakeDelegate);
 			} else {
-				evt.AddEventHandler (target, Delegate.CreateDelegate (
-							     evt.EventHandlerType, app,
+				if (method.IsStatic) {
+					evt.AddEventHandler (target, Delegate.CreateDelegate (
+						evt.EventHandlerType, method));
+				} else {
+					evt.AddEventHandler (target, Delegate.CreateDelegate (
+								     evt.EventHandlerType, app,
 #if NET_2_0
-							     method
+								     method
 #else
-							     method.Name
+								     method.Name
 #endif
-						     ));
+							     ));
+				}
 			}
 			
 		}
@@ -302,7 +337,7 @@ namespace System.Web {
 			MethodInfo method = null;
 			HttpApplication app = null;
 			lock (factory.available_for_end) {
-				method = factory.session_end;
+				method = (MethodInfo) factory.session_end;
 				if (method == null)
 					return;
 
@@ -395,6 +430,23 @@ namespace System.Web {
 					AppCodeCompiler acc = new AppCodeCompiler ();
 					acc.Compile ();
 
+					// Get the default machine *.browser files.
+					string default_machine_browsers_path = Path.Combine (HttpRuntime.MachineConfigurationDirectory, "Browsers");
+					default_machine_browsers_files = new string[0];
+					if (Directory.Exists (default_machine_browsers_path)) {
+						default_machine_browsers_files 
+							= Directory.GetFiles (default_machine_browsers_path, "*.browser");
+					}
+					
+					// Note whether there are any App_Data/Mono_Machine_Browsers/*.browser files.  If there
+					// are we will be using them instead of the default machine *.browser files.
+					string app_mono_machine_browsers_path = Path.Combine (Path.Combine (physical_app_path, "App_Data"), "Mono_Machine_Browsers");
+					app_mono_machine_browsers_files = new string[0];
+					if (Directory.Exists (app_mono_machine_browsers_path)) {
+						app_mono_machine_browsers_files 
+							= Directory.GetFiles (app_mono_machine_browsers_path, "*.browser");
+					}
+						
 					// Note whether there are any App_Browsers/*.browser files.  If there
 					// are we will be using *.browser files for sniffing in addition to browscap.ini
 					string app_browsers_path = Path.Combine (physical_app_path, "App_Browsers");
@@ -404,7 +456,10 @@ namespace System.Web {
 					}
 #endif
 
-					if (app_file != null) {
+#if NET_2_0
+					app_type = BuildManager.GetPrecompiledApplicationType ();
+#endif
+					if (app_type == null && app_file != null) {
 #if TARGET_J2EE
 						app_file = System.Web.Util.UrlUtils.ResolveVirtualPathFromAppAbsolute("~/" + Path.GetFileName(app_file));
 						app_type = System.Web.J2EE.PageMapper.GetObjectType(context, app_file);
@@ -419,16 +474,16 @@ namespace System.Web {
 							string msg = String.Format ("Error compiling application file ({0}).", app_file);
 							throw new ApplicationException (msg);
 						}
-					} else {
+					} else if (app_type == null) {
 						app_type = typeof (System.Web.HttpApplication);
 						app_state = new HttpApplicationState ();
 					}
 
 					WatchLocationForRestart("Global.asax");
 					WatchLocationForRestart("global.asax");
-					WatchLocationForRestart("Web.config");
-					WatchLocationForRestart("web.config");
-					WatchLocationForRestart("Web.Config");
+					WatchLocationForRestart(String.Empty, "Web.config", true);
+					WatchLocationForRestart(String.Empty, "web.config", true);
+					WatchLocationForRestart(String.Empty, "Web.Config", true);
 					needs_init = false;
 #if NET_2_0
 				} catch (Exception) {
@@ -487,6 +542,12 @@ namespace System.Web {
 				}
 			}
 
+			app = (HttpApplication) Interlocked.Exchange (ref factory.next_free, null);
+			if (app != null) {
+				app.RequestCompleted = false;
+				return app;
+			}
+
 			lock (factory.available) {
 				if (factory.available.Count > 0) {
 					app = (HttpApplication) factory.available.Pop ();
@@ -513,24 +574,33 @@ namespace System.Web {
 
 		internal static void RecycleForSessionEnd (HttpApplication app)
 		{
+			bool dispose = false;
 			HttpApplicationFactory factory = theFactory;
 			lock (factory.available_for_end) {
-				if (factory.available_for_end.Count < 32)
+				if (factory.available_for_end.Count < 64)
 					factory.available_for_end.Push (app);
 				else
-					app.Dispose ();
+					dispose = true;
 			}
+			if (dispose)
+				app.Dispose ();
 		}
 
 		internal static void Recycle (HttpApplication app)
 		{
+			bool dispose = false;
 			HttpApplicationFactory factory = theFactory;
+			if (Interlocked.CompareExchange (ref factory.next_free, app, null) == null)
+				return;
+
 			lock (factory.available) {
-				if (factory.available.Count < 32)
+				if (factory.available.Count < 64)
 					factory.available.Push (app);
 				else
-					app.Dispose ();
+					dispose = true;
 			}
+			if (dispose)
+				app.Dispose ();
 		}
 
 		internal static bool ContextAvailable {
@@ -555,7 +625,7 @@ namespace System.Web {
 			physicalPath = Path.Combine(physicalPath, virtualPath);
 			bool isDir = Directory.Exists(physicalPath);
 			bool isFile = isDir ? false : File.Exists(physicalPath);
-			
+
 			if (isDir || isFile) {
 				// create the watcher
 				FileSystemEventHandler fseh = new FileSystemEventHandler(OnFileChanged);
@@ -590,13 +660,12 @@ namespace System.Web {
 				lock (capabilities_processor_lock) {
 					if (capabilities_processor == null) {
 						capabilities_processor = new System.Web.Configuration.nBrowser.Build();
-						string machine_browsers_path = Path.Combine (HttpRuntime.MachineConfigurationDirectory, "Browsers");
-						if (Directory.Exists (machine_browsers_path)) {
-							string[] machine_browsers_files 
-								= Directory.GetFiles (machine_browsers_path, "*.browser");
-							foreach (string f in machine_browsers_files) {
-								capabilities_processor.AddBrowserFile(f);
-							}
+						string[] machine_browsers_files = app_mono_machine_browsers_files;
+						if (machine_browsers_files.Length == 0)	{
+							machine_browsers_files = default_machine_browsers_files;
+						}
+						foreach (string f in machine_browsers_files) {
+							capabilities_processor.AddBrowserFile(f);
 						}
 						foreach (string f in app_browsers_files) {
 							capabilities_processor.AddBrowserFile(f);
