@@ -36,6 +36,7 @@ using System.IO;
 using System.Runtime.InteropServices;
 using System.Security;
 using System.Security.Permissions;
+using System.Security.Principal;
 using System.Web.Configuration;
 using System.Web.UI;
 using System.Web.Util;
@@ -95,6 +96,8 @@ namespace System.Web {
 		bool validate_cookies, validate_query_string, validate_form;
 		bool checked_cookies, checked_query_string, checked_form;
 
+		readonly static char [] queryTrimChars = {'?'};
+		
 		static HttpRequest ()
 		{
 			host_addresses = GetLocalHostAddresses ();
@@ -110,35 +113,95 @@ namespace System.Web {
 			url_components = new UriBuilder (url);
 			url_components.Query = queryString;
 			
-			query_string_nvc = new WebROCollection ();					
-			HttpUtility.ParseQueryString (queryString, Encoding.Default, query_string_nvc);
+			query_string_nvc = new WebROCollection ();
+			if (queryString != null)
+				HttpUtility.ParseQueryString (queryString, Encoding.Default, query_string_nvc);
 			query_string_nvc.Protect ();
 		}
 
-		UriBuilder UrlComponents {
-			get {
-				if (url_components == null) {
-					url_components = new UriBuilder ();
-					url_components.Scheme = worker_request.GetProtocol ();
-					url_components.Host = worker_request.GetServerName ();
-					url_components.Port = worker_request.GetLocalPort ();
-					url_components.Path = worker_request.GetUriPath ();
-					
-					byte[] queryStringRaw = worker_request.GetQueryStringRawBytes();
-					if(queryStringRaw != null)
-						url_components.Query = ContentEncoding.GetString(queryStringRaw);
-					else
-						url_components.Query = worker_request.GetQueryString();
-				}
-				return url_components;
-			}
-		}
-		
 		internal HttpRequest (HttpWorkerRequest worker_request, HttpContext context)
 		{
 			this.worker_request = worker_request;
 			this.context = context;
 		}
+		
+		internal UriBuilder UrlComponents {
+			get {
+				if (url_components == null) {
+					string query;
+					byte[] queryStringRaw = worker_request.GetQueryStringRawBytes();
+					if(queryStringRaw != null)
+						query = ContentEncoding.GetString(queryStringRaw);
+					else
+						query = worker_request.GetQueryString();
+					
+					BuildUrlComponents (
+#if NET_2_0
+						ApplyUrlMapping (worker_request.GetUriPath ()),
+#else
+						worker_request.GetUriPath (),
+#endif
+						query);
+				}
+				return url_components;
+			}
+		}
+
+		void BuildUrlComponents (string path, string query)
+		{
+			if (url_components != null)
+				return;
+			url_components = new UriBuilder ();
+			url_components.Scheme = worker_request.GetProtocol ();
+			url_components.Host = worker_request.GetServerName ();
+			url_components.Port = worker_request.GetLocalPort ();
+			url_components.Path = path;
+			if (query != null && query.Length > 0)
+				url_components.Query = query.TrimStart (queryTrimChars);
+		}
+
+#if NET_2_0
+		internal string ApplyUrlMapping (string url)
+		{
+			if (WebConfigurationManager.HasConfigErrors)
+				return url;
+			
+			UrlMappingsSection ums = WebConfigurationManager.GetSection ("system.web/urlMappings", ApplicationPath) as UrlMappingsSection;
+			UrlMappingCollection umc;
+
+			if (ums == null || !ums.IsEnabled || (umc = ums.UrlMappings).Count == 0)
+				return url;
+
+			string relUrl = VirtualPathUtility.ToAppRelative (url);
+			UrlMapping um = null;
+			
+			foreach (UrlMapping u in umc) {
+				if (u == null)
+					continue;
+				if (String.Compare (relUrl, u.Url, StringComparison.Ordinal) == 0) {
+					um = u;
+					break;
+				}
+			}
+
+			if (um == null)
+				return url;
+
+			string rawUrl = VirtualPathUtility.ToAbsolute (um.MappedUrl.Trim ());
+			Uri newUrl = new Uri ("http://host.com" + rawUrl);
+
+			if (url_components != null) {
+				url_components.Path = newUrl.AbsolutePath;
+				url_components.Query = newUrl.Query.TrimStart (queryTrimChars);
+				query_string_nvc = new WebROCollection ();
+				HttpUtility.ParseQueryString (newUrl.Query, Encoding.Default, query_string_nvc);
+				query_string_nvc.Protect ();
+			} else
+				BuildUrlComponents (newUrl.AbsolutePath, newUrl.Query);
+
+			return url_components.Path;
+		}
+#endif
 
 		string [] SplitHeader (int header_index)
 		{
@@ -165,6 +228,12 @@ namespace System.Web {
 		}
 
 #if NET_2_0
+#if !TARGET_JVM
+		public WindowsIdentity LogonUserIdentity {
+			get { throw new NotImplementedException (); }
+		}
+#endif
+		
 		string anonymous_id;
 		public string AnonymousID {
 			get {
@@ -198,6 +267,22 @@ namespace System.Web {
 			}
 		}
 
+#if NET_2_0		
+		internal bool BrowserMightHaveSpecialWriter {
+			get {
+				return (browser_capabilities != null 
+					|| HttpApplicationFactory.AppBrowsersFiles.Length > 0);
+			}
+		}
+
+		internal bool BrowserMightHaveAdapters {
+			get {
+				return (browser_capabilities != null 
+					|| HttpApplicationFactory.AppBrowsersFiles.Length > 0);
+			}
+		}
+#endif
+
 		public HttpClientCertificate ClientCertificate {
 			get {
 				if (client_cert == null)
@@ -343,7 +428,13 @@ namespace System.Web {
 					return "/"; // required for 2.0
 
 				if (file_path == null)
-					file_path = UrlUtils.Canonic (worker_request.GetFilePath ());
+					file_path = UrlUtils.Canonic (
+#if NET_2_0
+						ApplyUrlMapping (worker_request.GetFilePath ())
+#else
+						worker_request.GetFilePath ()
+#endif
+					);
 
 				return file_path;
 			}
@@ -353,9 +444,16 @@ namespace System.Web {
 			get {
 				if (base_virtual_dir == null){
 					base_virtual_dir = FilePath;
+					if (UrlUtils.HasSessionId (base_virtual_dir))
+						base_virtual_dir = UrlUtils.RemoveSessionId (VirtualPathUtility.GetDirectory (base_virtual_dir), base_virtual_dir);
+					
 					int p = base_virtual_dir.LastIndexOf ('/');
-					if (p != -1)
+					if (p != -1) {
+						if (p == 0)
+							p = 1;
 						base_virtual_dir = base_virtual_dir.Substring (0, p);
+					} else
+						base_virtual_dir = "/";
 				}
 				return base_virtual_dir;
 			}
@@ -460,7 +558,8 @@ namespace System.Web {
 		//
 		void AddRawKeyValue (StringBuilder key, StringBuilder value)
 		{
-			form.Add (HttpUtility.UrlDecode (key.ToString (), ContentEncoding),
+			string decodedKey = HttpUtility.UrlDecode (key.ToString (), ContentEncoding);
+			form.Add (decodedKey,
 				  HttpUtility.UrlDecode (value.ToString (), ContentEncoding));
 
 			key.Length = 0;
@@ -476,36 +575,37 @@ namespace System.Web {
 		void LoadWwwForm ()
 #endif
 		{
-			Stream input = GetSubStream (InputStream);
-			StreamReader s = new StreamReader (input, ContentEncoding);
+			using (Stream input = GetSubStream (InputStream)) {
+				using (StreamReader s = new StreamReader (input, ContentEncoding)) {
+					StringBuilder key = new StringBuilder ();
+					StringBuilder value = new StringBuilder ();
+					int c;
 
-			StringBuilder key = new StringBuilder ();
-			StringBuilder value = new StringBuilder ();
-			int c;
-
-			while ((c = s.Read ()) != -1){
-				if (c == '='){
-					value.Length = 0;
 					while ((c = s.Read ()) != -1){
-						if (c == '&'){
+						if (c == '='){
+							value.Length = 0;
+							while ((c = s.Read ()) != -1){
+								if (c == '&'){
+									AddRawKeyValue (key, value);
+									break;
+								} else
+									value.Append ((char) c);
+							}
+							if (c == -1){
+								AddRawKeyValue (key, value);
+								return;
+							}
+						} else if (c == '&')
 							AddRawKeyValue (key, value);
-							break;
-						} else
-							value.Append ((char) c);
+						else
+							key.Append ((char) c);
 					}
-					if (c == -1){
+					if (c == -1)
 						AddRawKeyValue (key, value);
-						return;
-					}
-				} else if (c == '&')
-					AddRawKeyValue (key, value);
-				else
-					key.Append ((char) c);
-			}
-			if (c == -1)
-				AddRawKeyValue (key, value);
 
-			EndSubStream (input);
+					EndSubStream (input);
+				}
+			}
 		}
 
 		bool IsContentType (string ct, bool starts_with)
@@ -522,17 +622,18 @@ namespace System.Web {
 					form = new WebROCollection ();
 					files = new HttpFileCollection ();
 
-					if (IsContentType ("application/x-www-form-urlencoded", true))
-						LoadWwwForm ();
-					else if (IsContentType ("multipart/form-data", true))
+					if (IsContentType ("multipart/form-data", true))
 						LoadMultiPart ();
+					else if (
+						IsContentType ("application/x-www-form-urlencoded", true))
+						LoadWwwForm ();
 
 					form.Protect ();
 				}
 
 				if (validate_form && !checked_form){
-					ValidateNameValueCollection ("Form", form);
 					checked_form = true;
+					ValidateNameValueCollection ("Form", form);
 				}
 				
 				return form;
@@ -625,7 +726,7 @@ namespace System.Web {
 			int content_length = ContentLength;
 
 #if NET_2_0
-			HttpRuntimeSection config = (HttpRuntimeSection) WebConfigurationManager.GetSection ("system.web/httpRuntime");
+			HttpRuntimeSection config = (HttpRuntimeSection) WebConfigurationManager.GetWebApplicationSection ("system.web/httpRuntime");
 #else
 			HttpRuntimeConfig config = (HttpRuntimeConfig) HttpContext.GetAppConfig ("system.web/httpRuntime");
 #endif
@@ -705,8 +806,8 @@ namespace System.Web {
 					ms.Write (buffer, 0, total);
 
 				buffer = new byte [INPUT_BUFFER_SIZE];
-				long maxlength = config.MaxRequestLength * 1024;
-				long disk_th = config.RequestLengthDiskThreshold * 1024;
+				long maxlength = config.MaxRequestLength * 1024L;
+				long disk_th = config.RequestLengthDiskThreshold * 1024L;
 				int n;
 				while (true) {
 					n = worker_request.ReadEntityBody (buffer, INPUT_BUFFER_SIZE);
@@ -817,12 +918,27 @@ namespace System.Web {
 
 		public string Path {
 			get {
-				if (unescaped_path == null)
+				if (unescaped_path == null) {
+					string path;
+					if (url_components != null) {
+						// use only if it's already been instantiated, so that we can't go into endless
+						// recursion in some scenarios
+						path = UrlComponents.Path;
+					} else {
 #if NET_2_0
-					unescaped_path = Uri.UnescapeDataString (UrlComponents.Path);
+						path = ApplyUrlMapping (worker_request.GetUriPath ());
 #else
-					unescaped_path = HttpUtility.UrlDecode (UrlComponents.Path);
+						path = worker_request.GetUriPath ();
 #endif
+					}
+						
+#if NET_2_0
+					unescaped_path = Uri.UnescapeDataString (path);
+#else
+					unescaped_path = HttpUtility.UrlDecode (path);
+#endif
+				}
+				
 				return unescaped_path;
 			}
 		}
@@ -886,13 +1002,16 @@ namespace System.Web {
 
 		public NameValueCollection QueryString {
 			get {
-				if (query_string_nvc == null){
+				if (query_string_nvc == null) {
+					query_string_nvc = new WebROCollection ();
 					string q = UrlComponents.Query;
-					if (q.Length != 0)
-						q = q.Remove(0, 1);
-
-					query_string_nvc = new WebROCollection ();					
-					HttpUtility.ParseQueryString (q, ContentEncoding, query_string_nvc);
+					if (q != null) {
+						if (q.Length != 0)
+							q = q.Remove(0, 1);
+					
+						HttpUtility.ParseQueryString (q, ContentEncoding, query_string_nvc);
+					}
+					
 					query_string_nvc.Protect();
 				}
 				
@@ -958,7 +1077,7 @@ namespace System.Web {
 					if (orig_url == null)
 						cached_url = UrlComponents.Uri;
 					else
-						cached_url = new Uri (orig_url); 
+						cached_url = new Uri (orig_url);
 				}
 
 				return cached_url;			
@@ -974,7 +1093,11 @@ namespace System.Web {
 				if (hr == null)
 					return null;
 
-				return new Uri (hr);
+				Uri uri = null;
+				try {
+					uri = new Uri (hr);
+				} catch (UriFormatException) {}
+				return uri;
 			}
 		}
 
@@ -1083,9 +1206,8 @@ namespace System.Web {
 					virtualPath = "~";
 			}
 
-			if (virtualPath.IndexOf (':') != -1)
-				throw new ArgumentNullException (
-					String.Format ("MapPath: Invalid path '{0}', only virtual paths are accepted", virtualPath));
+			if (!VirtualPathUtility.IsValidVirtualPath (virtualPath))
+				throw new HttpException (String.Format ("'{0}' is not a valid virtual path.", virtualPath));
 
 			string appVirtualPath = HttpRuntime.AppDomainAppVirtualPath;
 
@@ -1095,7 +1217,7 @@ namespace System.Web {
 				virtualPath = VirtualPathUtility.Combine (VirtualPathUtility.AppendTrailingSlash (baseVirtualDir), virtualPath);
 			}
 			virtualPath = VirtualPathUtility.ToAbsolute (virtualPath);
-
+			
 			if (!allowCrossAppMapping){
 				if (!StrUtils.StartsWith (virtualPath, appVirtualPath, true))
 					throw new HttpException ("MapPath: Mapping across applications not allowed");
@@ -1213,7 +1335,6 @@ namespace System.Web {
 		{
 			cached_url = null;
 			current_exe_path = path;
-			file_path = path;
 			UrlComponents.Path = path;
 			// recreated on demand
 			root_virtual_dir = null;
@@ -1240,6 +1361,20 @@ namespace System.Web {
 		// Notice: there is nothing raw about this querystring.
 		internal string QueryStringRaw {
 			get {
+				UriBuilder urlComponents = UrlComponents;
+
+				if (urlComponents == null) {
+					string ret = worker_request.GetQueryString ();
+
+					if (ret == null || ret.Length == 0)
+						return String.Empty;
+
+					if (ret [0] == '?')
+						return ret;
+
+					return "?" + ret;
+				}
+				
 				return UrlComponents.Query;
 			}
 
@@ -1275,7 +1410,7 @@ namespace System.Web {
 		
 			foreach (string key in coll.Keys) {
 				string val = coll [key];
-				if (val != null && val != "" && CheckString (val))
+				if (val != null && val.Length > 0 && CheckString (val))
 					ThrowValidationException (name, key, val);
 			}
 		}
@@ -1317,6 +1452,7 @@ namespace System.Web {
 			char current = val [0];
 			for (int idx = 1; idx < len; idx++) {
 				char next = val [idx];
+				// See http://secunia.com/advisories/14325
 				if (current == '<' || current == '\xff1c') {
 					if (next == '!' || next < ' '
 					    || (next >= 'a' && next <= 'z')
@@ -1373,8 +1509,8 @@ namespace System.Web {
 			
 			public override string ToString ()
 			{
-				return string.Format ("ContentType {0}, Name {1}, Filename {2}, Start {3}, Length {4}",
-					ContentType, Name, Filename, Start, Length);
+				return "ContentType " + ContentType + ", Name " + Name + ", Filename " + Filename + ", Start " +
+					Start.ToString () + ", Length " + Length.ToString ();
 			}
 		}
 		
@@ -1585,7 +1721,7 @@ namespace System.Web {
 			while ((header = ReadHeaders ()) != null) {
 				if (StrUtils.StartsWith (header, "Content-Disposition:", true)) {
 					elem.Name = GetContentDispositionAttribute (header, "name");
-					elem.Filename = GetContentDispositionAttributeWithEncoding (header, "filename");      
+					elem.Filename = StripPath (GetContentDispositionAttributeWithEncoding (header, "filename"));
 				} else if (StrUtils.StartsWith (header, "Content-Type:", true)) {
 					elem.ContentType = header.Substring ("Content-Type:".Length).Trim ();
 				}
@@ -1600,7 +1736,16 @@ namespace System.Web {
 			elem.Length = pos - start;
 			return elem;
 		}
-		
+
+		static string StripPath (string path)
+		{
+			if (path == null || path.Length == 0)
+				return path;
+			
+			if (path.IndexOf (":\\") != 1 && !path.StartsWith ("\\\\"))
+				return path;
+			return path.Substring (path.LastIndexOf ('\\') + 1);
+		}
 	}
 #endregion
 }
