@@ -106,16 +106,13 @@ namespace System.Web {
 
 		static HttpResponse ()
 		{
-			string version = null;
 #if NET_2_0
-			HttpRuntimeSection config = WebConfigurationManager.GetSection ("system.web/httpRuntime") as HttpRuntimeSection;
+			HttpRuntimeSection config = WebConfigurationManager.GetWebApplicationSection ("system.web/httpRuntime") as HttpRuntimeSection;
 #else
 			HttpRuntimeConfig config = HttpContext.GetAppConfig ("system.web/httpRuntime") as HttpRuntimeConfig;
-			version = config.VersionHeader;
 #endif
-			if (config.EnableVersionHeader) {
-				if (version == null)
-					version = Environment.Version.ToString (3);
+			if (config != null && config.EnableVersionHeader) {
+				string version = Environment.Version.ToString (3);
 				version_header = new UnknownResponseHeader ("X-AspNet-Version", version);
 			}
 		}
@@ -535,7 +532,7 @@ namespace System.Web {
 
 			bool cookieless = false;
 #if NET_2_0
-			SessionStateSection config = WebConfigurationManager.GetSection ("system.web/sessionState") as SessionStateSection;
+			SessionStateSection config = WebConfigurationManager.GetWebApplicationSection ("system.web/sessionState") as SessionStateSection;
 			cookieless = SessionStateModule.IsCookieLess (context, config);
 #else
 			SessionConfig config = HttpContext.GetAppConfig ("system.web/sessionState") as SessionConfig;
@@ -996,19 +993,28 @@ namespace System.Web {
 		{
 			TransmitFile (vf, false);
 		}
-		
+
+		const int bufLen = 65535;
 		internal void TransmitFile (VirtualFile vf, bool final_flush)
 		{
 			if (vf == null)
 				throw new ArgumentNullException ("vf");
 
+			if (vf is DefaultVirtualFile) {
+				TransmitFile (HostingEnvironment.MapPath (vf.VirtualPath), final_flush);
+				return;
+			}
+			
+			byte[] buf = new byte [bufLen];
 			using (Stream s = vf.Open ()) {
-				long len = s.Length;
-				byte[] buf = new byte [len];
-				int readBytes = s.Read (buf, 0, (int) len);
-				output_stream.Write (buf, 0, readBytes);
-				output_stream.ApplyFilter (final_flush);
-				Flush (final_flush);
+				int readBytes;
+				while ((readBytes = s.Read (buf, 0, bufLen)) > 0) {
+					output_stream.Write (buf, 0, readBytes);
+					output_stream.ApplyFilter (final_flush);
+					Flush (false);
+				}
+				if (final_flush)
+					Flush (true);
 			}
 		}
 #endif
@@ -1027,34 +1033,31 @@ namespace System.Web {
 		}
 
 		internal bool IsCached {
-			get {
-				return cached_response != null;
+			get { return cached_response != null; }
+			set {
+				if (value)
+					cached_response = new CachedRawResponse (cache_policy);
+				else
+					cached_response = null;
 			}
 		}
 
 		public HttpCachePolicy Cache {
 			get {
-				if (cache_policy == null) {
+				if (cache_policy == null)
 					cache_policy = new HttpCachePolicy ();
-					cache_policy.CacheabilityUpdated += new CacheabilityUpdatedCallback (OnCacheabilityUpdated);
-				}
 				
 				return cache_policy;
 			}
-		}
-		
-		private void OnCacheabilityUpdated (object sender, CacheabilityUpdatedEventArgs e)
-		{
-			if (e.Cacheability >= HttpCacheability.Server && !IsCached)
-				cached_response = new CachedRawResponse (cache_policy);
-			else if (e.Cacheability <= HttpCacheability.Private)
-				cached_response = null;
-		}
+		}		
 
 		internal CachedRawResponse GetCachedResponse ()
 		{
-			cached_response.StatusCode = StatusCode;
-			cached_response.StatusDescription = StatusDescription;
+			if (cached_response != null) {
+				cached_response.StatusCode = StatusCode;
+				cached_response.StatusDescription = StatusDescription;
+			}
+			
 			return cached_response;
 		}
 
