@@ -66,6 +66,12 @@ namespace System.Net
 		bool reused;
 		int position;
 		bool busy;
+		HttpWebRequest priority_request;
+		NetworkCredential ntlm_credentials;
+		bool ntlm_authenticated;
+#if NET_1_1
+		bool unsafe_sharing;
+#endif
 
 		bool ssl;
 		bool certsAvailable;
@@ -121,18 +127,32 @@ namespace System.Net
 
 				foreach (IPAddress address in hostEntry.AddressList) {
 					socket = new Socket (address.AddressFamily, SocketType.Stream, ProtocolType.Tcp);
-					try {
-						socket.Connect (new IPEndPoint(address, sPoint.Address.Port));
-						status = WebExceptionStatus.Success;
-						break;
-					} catch (SocketException) {
-						// This might be null if the request is aborted
-						if (socket != null) {
-							socket.Close();
-							socket = null;
-							status = WebExceptionStatus.ConnectFailure;
+
+					IPEndPoint remote = new IPEndPoint (address, sPoint.Address.Port);
+
+#if NET_2_0
+					socket.NoDelay = true;
+					if (!sPoint.CallEndPointDelegate (socket, remote)) {
+						socket.Close ();
+						socket = null;
+						status = WebExceptionStatus.ConnectFailure;
+					} else {
+#endif
+						try {
+							socket.Connect (remote);
+							status = WebExceptionStatus.Success;
+							break;
+						} catch (SocketException) {
+							// This might be null if the request is aborted
+							if (socket != null) {
+								socket.Close ();
+								socket = null;
+								status = WebExceptionStatus.ConnectFailure;
+							}
 						}
+#if NET_2_0
 					}
+#endif
 				}
 			}
 		}
@@ -320,8 +340,10 @@ namespace System.Net
 				req = Data.request;
 
 			Close (true);
-			if (req != null)
+			if (req != null) {
+				req.FinishedReading = true;
 				req.SetResponseError (st, e, where);
+			}
 		}
 		
 		static void ReadDone (IAsyncResult result)
@@ -390,6 +412,7 @@ namespace System.Net
 				stream.ReadBuffer = cnc.buffer;
 				stream.ReadBufferOffset = pos;
 				stream.ReadBufferSize = nread;
+				stream.CheckResponseInBuffer ();
 			} else if (cnc.chunkStream == null) {
 				try {
 					cnc.chunkStream = new ChunkStream (cnc.buffer, pos, nread, data.Headers);
@@ -612,6 +635,7 @@ namespace System.Net
 		internal void NextRead ()
 		{
 			lock (this) {
+				Data.request.FinishedReading = true;
 				string header = (sPoint.UsesProxy) ? "Proxy-Connection" : "Connection";
 				string cncHeader = (Data.Headers != null) ? Data.Headers [header] : null;
 				bool keepAlive = (Data.Version == HttpVersion.Version11 && this.keepAlive);
@@ -626,7 +650,12 @@ namespace System.Net
 				}
 
 				busy = false;
-				SendNext ();
+				if (priority_request != null) {
+					SendRequest (priority_request);
+					priority_request = null;
+				} else {
+					SendNext ();
+				}
 			}
 		}
 		
@@ -907,7 +936,8 @@ namespace System.Net
 		{
 			lock (this) {
 				if (Data.request == sender) {
-					HandleError (WebExceptionStatus.RequestCanceled, null, "Abort");
+					if (!Data.request.FinishedReading)
+						HandleError (WebExceptionStatus.RequestCanceled, null, "Abort");
 					return;
 				}
 
@@ -927,6 +957,13 @@ namespace System.Net
 			}
 		}
 
+		internal void ResetNtlm ()
+		{
+			ntlm_authenticated = false;
+			ntlm_credentials = null;
+			unsafe_sharing = false;
+		}
+
 		internal bool Busy {
 			get { lock (this) return busy; }
 		}
@@ -938,6 +975,29 @@ namespace System.Net
 				}
 			}
 		}
+
+		// -Used for NTLM authentication
+		internal HttpWebRequest PriorityRequest {
+			set { priority_request = value; }
+		}
+
+		internal bool NtlmAuthenticated {
+			get { return ntlm_authenticated; }
+			set { ntlm_authenticated = value; }
+		}
+
+		internal NetworkCredential NtlmCredential {
+			get { return ntlm_credentials; }
+			set { ntlm_credentials = value; }
+		}
+
+#if NET_1_1
+		internal bool UnsafeAuthenticatedConnectionSharing {
+			get { return unsafe_sharing; }
+			set { unsafe_sharing = value; }
+		}
+#endif
+		// -
 	}
 }
 
