@@ -6,15 +6,37 @@
 //
 // (c) 2004 Novell, Inc. (http://www.ximian.com)
 //
-using NUnit.Framework;
+
 using System;
 using System.Collections;
 using System.ComponentModel;
+using DescriptionAttribute = System.ComponentModel.DescriptionAttribute;
 using System.ComponentModel.Design;
 using System.Globalization;
 
+using NUnit.Framework;
+
 namespace MonoTests.System.ComponentModel
 {
+	interface IFoo
+	{
+		event EventHandler Fired;
+		event EventHandler Closed;
+
+		bool HasFired {
+			get;
+		}
+	}
+
+	interface IBar : IFoo
+	{
+		event EventHandler Destroyed;
+
+		bool IsDestroyed {
+			get;
+		}
+	}
+
 	class MyDesigner: IDesigner
 	{
 		public MyDesigner()
@@ -226,6 +248,40 @@ namespace MonoTests.System.ComponentModel
 			get { return prop; }
 			set { prop = value; }
 		}
+
+		[Browsable (false)]
+		public string YetAnotherProperty
+		{
+			get { return null; }
+		}
+
+		public string Name {
+			get { return null; }
+		}
+
+		public string Address {
+			get { return null; }
+		}
+
+		public string Country {
+			get { return null; }
+		}
+
+		private string HairColor {
+			get { return null; }
+		}
+
+		protected int Weight {
+			get { return 5; }
+		}
+
+		internal int Height {
+			get { return 0; }
+		}
+
+		public string WriteOnlyProperty {
+			set { prop = value; }
+		}
 	}
 
 	[DescriptionAttribute ("my test derived component")]
@@ -248,6 +304,19 @@ namespace MonoTests.System.ComponentModel
 			get { return prop; }
 			set { prop = value; }
 		}
+
+
+		[DescriptionAttribute ("test derived")]
+		public new string AnotherProperty
+		{
+			get { return base.AnotherProperty; }
+			set { base.AnotherProperty = value; }
+		}
+
+		public new object YetAnotherProperty
+		{
+			get { return null; }
+		}
 	}
 	
 
@@ -282,11 +351,13 @@ namespace MonoTests.System.ComponentModel
 		}
 	}
 
+	[Browsable (false)]
 	public interface ITestInterface
 	{
 		void TestFunction ();
 	}
-	
+
+	[DesignerAttribute (typeof(MyDesigner), typeof(int))]
 	public class TestClass
 	{
 		public TestClass()
@@ -295,7 +366,14 @@ namespace MonoTests.System.ComponentModel
 		void TestFunction ()
 		{}
 	}
-	
+
+	[DescriptionAttribute ("bla")]
+	public class TestDerivedClass : TestClass, ITestInterface
+	{
+		public void TestFunction ()
+		{}
+	}
+
 	public struct TestStruct
 	{
 		public int TestVal;
@@ -500,6 +578,10 @@ namespace MonoTests.System.ComponentModel
 				}
 			}
 			Assert.IsTrue (found, "#E4");
+
+			// Shows that attributes are retrieved from the current type, the base types 
+			// and the implemented by the type interfaces.
+			Assert.AreEqual (3, TypeDescriptor.GetAttributes (typeof (TestDerivedClass)).Count, "#F1");
 		}
 		
 		[Test]
@@ -532,6 +614,13 @@ namespace MonoTests.System.ComponentModel
 			Assert.AreEqual ("TestName", TypeDescriptor.GetComponentName (sitedcom), "#7");
 			Assert.AreEqual ("TestName", TypeDescriptor.GetComponentName (sitedcom), "#8");
 		}
+
+		[Test]
+		[ExpectedException (typeof (ArgumentNullException))]
+		public void TestGetConverterNullParam ()
+		{
+			TypeDescriptor.GetConverter (null);
+		}
 		
 		[Test]
 		public void TestGetConverter ()
@@ -570,6 +659,12 @@ namespace MonoTests.System.ComponentModel
 #if NET_2_0
 			// Test from bug #76686
 			Assert.AreEqual  (typeof (Int32Converter), TypeDescriptor.GetConverter ((int?) 1).GetType (), "#28");
+#endif
+			Assert.IsTrue (TypeDescriptor.GetConverter (typeof (Component)) is ComponentConverter, "#29");
+			Assert.IsTrue (TypeDescriptor.GetConverter (new Component()) is ComponentConverter, "#30");
+
+#if NET_2_0
+			Assert.AreEqual (typeof (NullableConverter), TypeDescriptor.GetConverter (typeof (int?)).GetType (), "#31");
 #endif
 		}
 		
@@ -690,6 +785,20 @@ namespace MonoTests.System.ComponentModel
 			col = TypeDescriptor.GetEvents (nfscom, filter);
 			Assert.AreEqual (1, col.Count, "#H1");
 			Assert.IsNotNull (col.Find ("AnEvent", true), "#H2");
+
+			col = TypeDescriptor.GetEvents (typeof (IFoo));
+			Assert.AreEqual (2, col.Count, "#I1");
+			Assert.IsNotNull (col.Find ("Fired", true), "#I2");
+			Assert.IsNotNull (col.Find ("Fired", false), "#I3");
+			Assert.IsNotNull (col.Find ("fired", true), "#I4");
+			Assert.IsNull (col.Find ("fired", false), "#I5");
+			Assert.IsNotNull (col.Find ("Closed", true), "#I6");
+
+			col = TypeDescriptor.GetEvents (typeof (IBar));
+			Assert.AreEqual (1, col.Count, "#J1");
+			Assert.IsNull (col.Find ("Fired", true), "#J2");
+			Assert.IsNull (col.Find ("Closed", true), "#J3");
+			Assert.IsNotNull (col.Find ("Destroyed", true), "#J4");
 		}
 		
 		[Test]
@@ -720,6 +829,56 @@ namespace MonoTests.System.ComponentModel
 			col = TypeDescriptor.GetProperties (nfscom, filter);
 			Assert.IsNotNull (col.Find ("TestProperty", true), "#F1");
 			Assert.IsNull (col.Find ("AnotherProperty", true), "#F2");
+
+
+			// GetProperties should return only the last type's implementation of a
+			// property with a matching name in the base types. E.g in the case where 
+			// the "new" keyword is used.
+			//
+			PropertyDescriptorCollection derivedCol = TypeDescriptor.GetProperties (typeof(MyDerivedComponent));
+			Assert.IsNotNull (derivedCol["AnotherProperty"].Attributes[typeof (DescriptionAttribute)], "#G1");
+			int anotherPropsFound = 0;
+			int yetAnotherPropsFound = 0;
+			foreach (PropertyDescriptor props in derivedCol) {
+				if (props.Name == "AnotherProperty")
+					anotherPropsFound++;
+				else if (props.Name == "YetAnotherProperty")
+					yetAnotherPropsFound++;
+			}
+
+			// GetProperties does not return the base type property in the case 
+			// where both the "new" keyword is used and also the Property type is different 
+			// (Type.GetProperties does return both properties)
+			//
+			Assert.AreEqual (1, anotherPropsFound, "#G2");
+			Assert.IsNotNull (derivedCol["AnotherProperty"].Attributes[typeof (DescriptionAttribute)], "#G3");
+			Assert.AreEqual (1, yetAnotherPropsFound, "#G4");
+
+			// Verify that we return the derived "new" property when 
+			// filters are applied that match the parent property
+			// 
+			PropertyDescriptorCollection filteredCol = TypeDescriptor.GetProperties (typeof(MyDerivedComponent), 
+												 new Attribute[] { BrowsableAttribute.Yes });
+			Assert.IsNotNull (filteredCol["YetAnotherProperty"], "#G5");
+
+			// GetProperties does not return write-only properties (ones without a getter)
+			// 
+			Assert.IsNull (col.Find ("WriteOnlyProperty", true), "#H1");
+
+			col = TypeDescriptor.GetProperties (typeof (IFoo));
+			Assert.AreEqual (1, col.Count, "#I1");
+			Assert.IsNotNull (col.Find ("HasFired", true), "#I1");
+			Assert.IsNotNull (col.Find ("HasFired", false), "#I2");
+			Assert.IsNotNull (col.Find ("hasFired", true), "#I3");
+			Assert.IsNull (col.Find ("hasFired", false), "#I4");
+
+			col = TypeDescriptor.GetProperties (typeof (IBar));
+			Assert.AreEqual (1, col.Count, "#J1");
+			Assert.IsNull (col.Find ("HasFired", true), "#J2");
+			Assert.IsNotNull (col.Find ("IsDestroyed", true), "#J3");
+			Assert.IsNotNull (col.Find ("IsDestroyed", false), "#J4");
+			Assert.IsNotNull (col.Find ("isDestroyed", true), "#J5");
+			Assert.IsNull (col.Find ("isDestroyed", false), "#J6");
 		}
 
 		[Test]
@@ -740,6 +899,26 @@ namespace MonoTests.System.ComponentModel
 			Assert.IsNull (col.Find ("AnotherProperty", true), "#B2");
 		}
 
+		[Test]
+#if ONLY_1_1
+		[NUnit.Framework.Category ("NotDotNet")] // .NET 1.x (or csc 1.x) does not retain the original order
+#endif
+		public void GetProperties_Order ()
+		{
+			MyComponent com = new MyComponent (new MyContainer ());
+
+			PropertyDescriptorCollection col = TypeDescriptor.GetProperties (com);
+			Assert.AreEqual (8, col.Count, "#1");
+			Assert.AreEqual ("TestProperty", col [0].Name, "#2");
+			Assert.AreEqual ("AnotherProperty", col [1].Name, "#3");
+			Assert.AreEqual ("YetAnotherProperty", col [2].Name, "#4");
+			Assert.AreEqual ("Name", col [3].Name, "#5");
+			Assert.AreEqual ("Address", col [4].Name, "#6");
+			Assert.AreEqual ("Country", col [5].Name, "#7");
+			Assert.AreEqual ("Site", col [6].Name, "#8");
+			Assert.AreEqual ("Container", col [7].Name, "#9");
+		}
+
 		[TypeConverter (typeof (TestConverter))]
 		class TestConverterClass {
 		}
