@@ -256,6 +256,31 @@ namespace MonoTests.System.Net
 			}
 		}
 
+		[Test] // bug #429200
+		public void GetRequestStream ()
+		{
+			IPEndPoint ep = new IPEndPoint (IPAddress.Loopback, 8000);
+			string url = "http://" + IPAddress.Loopback.ToString () + ":8000/test/";
+
+			using (SocketResponder responder = new SocketResponder (ep, new SocketRequestHandler (EchoRequestHandler))) {
+				responder.Start ();
+
+				HttpWebRequest req = (HttpWebRequest) WebRequest.Create (url);
+				req.Method = "POST";
+				req.Timeout = 2000;
+				req.ReadWriteTimeout = 2000;
+
+				Stream rs1 = req.GetRequestStream ();
+				Stream rs2 = req.GetRequestStream ();
+
+				Assert.IsNotNull (rs1, "#1");
+				Assert.AreSame (rs1, rs2, "#2");
+
+				rs1.Close ();
+				responder.Stop ();
+			}
+		}
+
 		[Test]
 		[Ignore ("This test asserts that our code violates RFC 2616")]
 		public void GetRequestStream_Body_NotAllowed ()
@@ -282,6 +307,7 @@ namespace MonoTests.System.Net
 #if TARGET_JVM
 		[Category("NotWorking")]
 #endif
+		[Ignore ("This does not timeout any more. That's how MS works when reading small responses")]
 		public void ReadTimeout ()
 		{
 			IPEndPoint localEP = new IPEndPoint (IPAddress.Loopback, 8764);
@@ -557,12 +583,11 @@ namespace MonoTests.System.Net
 		}
 #endif
 
-		[Test]
-		[Category ("NotWorking")] // #353495
-		public void LastModifiedKind () // bug #353495
+		[Test] // bug #353495
+		[Category ("NotWorking")]
+		public void LastModifiedKind ()
 		{
 			const string reqURL = "http://coffeefaq.com/site/node/25";
-			const string dtFmt = "ddd, dd MMM yyyy HH:mm:ss";
 			HttpWebRequest req = (HttpWebRequest) WebRequest.Create (reqURL);
 			HttpWebResponse resp = (HttpWebResponse) req.GetResponse ();
 			DateTime lastMod = resp.LastModified;
@@ -635,19 +660,19 @@ namespace MonoTests.System.Net
 
 			StringWriter sw = new StringWriter ();
 			if (statusLine.StartsWith ("POST /original/")) {
-				sw.WriteLine ("HTTP/1.1 302 Found");
+				sw.WriteLine ("HTTP/1.0 302 Found");
 				sw.WriteLine ("Location: " + "http://" + IPAddress.Loopback.ToString () + ":8764/moved/");
 				sw.WriteLine ();
 				sw.Flush ();
 			} else if (statusLine.StartsWith ("GET /moved/")) {
-				sw.WriteLine ("HTTP/1.1 200 OK");
+				sw.WriteLine ("HTTP/1.0 200 OK");
 				sw.WriteLine ("Content-Type: text/plain");
 				sw.WriteLine ("Content-Length: 8");
 				sw.WriteLine ();
 				sw.Write ("LOOKS OK");
 				sw.Flush ();
 			} else {
-				sw.WriteLine ("HTTP/1.1 500 Too Lazy");
+				sw.WriteLine ("HTTP/1.0 500 Too Lazy");
 				sw.WriteLine ();
 				sw.Flush ();
 			}
@@ -676,6 +701,78 @@ namespace MonoTests.System.Net
 			return Encoding.UTF8.GetBytes (sw.ToString ());
 		}
 
+		[Test]
+		public void NtlmAuthentication ()
+		{
+			NtlmServer server = new NtlmServer ();
+			server.Start ();
+
+			string url = String.Format ("http://{0}:{1}/nothing.html", server.IPAddress, server.Port);
+			HttpWebRequest request = (HttpWebRequest) WebRequest.Create (url);
+			request.Timeout = 5000;
+			request.Credentials = new NetworkCredential ("user", "password", "domain");
+			HttpWebResponse resp = (HttpWebResponse) request.GetResponse ();
+			string res = null;
+			using (StreamReader reader = new StreamReader (resp.GetResponseStream ())) {
+				res = reader.ReadToEnd ();	
+			}
+			resp.Close ();
+			server.Stop ();
+			Assert.AreEqual ("OK", res);
+		}
+
+		class NtlmServer : HttpServer {
+			public string Where = "";
+			protected override void Run ()
+			{
+				Where = "before accept";
+				Socket client = sock.Accept ();
+				NetworkStream ns = new NetworkStream (client, false);
+				StreamReader reader = new StreamReader (ns, Encoding.ASCII);
+				string line;
+				Where = "first read";
+				while ((line = reader.ReadLine ()) != null) {
+					if (line.Trim () == String.Empty) {
+						break;
+					}
+				}
+				Where = "first write";
+				StreamWriter writer = new StreamWriter (ns, Encoding.ASCII);
+				writer.Write (  "HTTP/1.1 401 Unauthorized\r\n" +
+						"WWW-Authenticate: NTLM\r\n" +
+						"Content-Length: 5\r\n\r\nWRONG");
+
+				writer.Flush ();
+				Where = "second read";
+				while ((line = reader.ReadLine ()) != null) {
+					if (line.Trim () == String.Empty) {
+						break;
+					}
+				}
+				Where = "second write";
+				writer.Write (  "HTTP/1.1 401 Unauthorized\r\n" +
+						"WWW-Authenticate: NTLM TlRMTVNTUAACAAAAAAAAADgAAAABggAC8GDhqIONH3sAAAAAAAAAAAAAAAA4AAAABQLODgAAAA8=\r\n" +
+						"Content-Length: 5\r\n\r\nWRONG");
+				writer.Flush ();
+
+				Where = "third read";
+				while ((line = reader.ReadLine ()) != null) {
+					if (line.Trim () == String.Empty) {
+						break;
+					}
+				}
+				Where = "third write";
+				writer.Write (  "HTTP/1.1 200 OK\r\n" +
+						"Keep-Alive: true\r\n" +
+						"Content-Length: 2\r\n\r\nOK");
+				writer.Flush ();
+				Thread.Sleep (1000);
+				writer.Close ();
+				reader.Close ();
+				client.Close ();
+			}
+		}
+
 		class BadChunkedServer : HttpServer {
 			protected override void Run ()
 			{
