@@ -263,7 +263,12 @@ namespace System.IO {
 
 		public override string ToString ()
 		{
+#if NET_2_1
+			// for Moonlight we *never* return paths, since ToString is not [SecurityCritical] we simply return the Name
+			return Name;
+#else
 			return OriginalPath;
+#endif
 		}
 
 #if NET_2_0
