@@ -69,7 +69,7 @@ namespace System.IO {
 		internal static readonly string DirectorySeparatorStr;
 		public static readonly char VolumeSeparatorChar;
 
-		private static readonly char[] PathSeparatorChars;
+		internal static readonly char[] PathSeparatorChars;
 		private static readonly bool dirEqualsVolume;
 
 		// class methods
@@ -79,13 +79,13 @@ namespace System.IO {
 				return null;
 
 			if (path.IndexOfAny (InvalidPathChars) != -1)
-				throw new ArgumentException ("Illegal characters in path", "path");
+				throw new ArgumentException ("Illegal characters in path.");
 
 			int iExt = findExtension (path);
 
 			if (extension == null)
 				return iExt < 0 ? path : path.Substring (0, iExt);
-			else if (extension == String.Empty)
+			else if (extension.Length == 0)
 				return iExt < 0 ? path + '.' : path.Substring (0, iExt + 1);
 
 			else if (path.Length != 0) {
@@ -112,43 +112,114 @@ namespace System.IO {
 			if (path2 == null)
 				throw new ArgumentNullException ("path2");
 
-			if (path1 == String.Empty)
+			if (path1.Length == 0)
 				return path2;
 
-			if (path2 == String.Empty)
+			if (path2.Length == 0)
 				return path1;
 
 			if (path1.IndexOfAny (InvalidPathChars) != -1)
-				throw new ArgumentException ("Illegal characters in path", "path1");
+				throw new ArgumentException ("Illegal characters in path.");
 
 			if (path2.IndexOfAny (InvalidPathChars) != -1)
-				throw new ArgumentException ("Illegal characters in path", "path2");
+				throw new ArgumentException ("Illegal characters in path.");
 
 			//TODO???: UNC names
-			// LAMESPEC: MS says that if path1 is not empty and path2 is a full path
-			// it should throw ArgumentException
 			if (IsPathRooted (path2))
 				return path2;
 			
 			char p1end = path1 [path1.Length - 1];
 			if (p1end != DirectorySeparatorChar && p1end != AltDirectorySeparatorChar && p1end != VolumeSeparatorChar)
-				return path1 + DirectorySeparatorChar + path2;
+				return path1 + DirectorySeparatorStr + path2;
 
 			return path1 + path2;
 		}
+	
+		//
+		// This routine:
+		//   * Removes duplicat path separators from a string
+		//   * If the string starts with \\, preserves the first two (hostname on Windows)
+		//   * Removes the trailing path separator.
+		//   * Returns the DirectorySeparatorChar for the single input DirectorySeparatorChar or AltDirectorySeparatorChar
+		//
+		// Unlike CanonicalizePath, this does not do any path resolution
+		// (which GetDirectoryName is not supposed to do).
+		//
+		internal static string CleanPath (string s)
+		{
+			int l = s.Length;
+			int sub = 0;
+			int start = 0;
+
+			// Host prefix?
+			char s0 = s [0];
+			if (l > 2 && s0 == '\\' && s [1] == '\\'){
+				start = 2;
+			}
+
+			// We are only left with root
+			if (l == 1 && (s0 == DirectorySeparatorChar || s0 == AltDirectorySeparatorChar))
+				return s;
+
+			// Cleanup
+			for (int i = start; i < l; i++){
+				char c = s [i];
+				
+				if (c != DirectorySeparatorChar && c != AltDirectorySeparatorChar)
+					continue;
+				if (i+1 == l)
+					sub++;
+				else {
+					c = s [i + 1];
+					if (c == DirectorySeparatorChar || c == AltDirectorySeparatorChar)
+						sub++;
+				}
+			}
+
+			if (sub == 0)
+				return s;
+
+			char [] copy = new char [l-sub];
+			if (start != 0){
+				copy [0] = '\\';
+				copy [1] = '\\';
+			}
+			for (int i = start, j = start; i < l && j < copy.Length; i++){
+				char c = s [i];
+
+				if (c != DirectorySeparatorChar && c != AltDirectorySeparatorChar){
+					copy [j++] = c;
+					continue;
+				}
+
+				// For non-trailing cases.
+				if (j+1 != copy.Length){
+					copy [j++] = DirectorySeparatorChar;
+					for (;i < l-1; i++){
+						c = s [i+1];
+						if (c != DirectorySeparatorChar && c != AltDirectorySeparatorChar)
+							break;
+					}
+				}
+			}
+			return new String (copy);
+		}
 
 		public static string GetDirectoryName (string path)
 		{
 			// LAMESPEC: For empty string MS docs say both
 			// return null AND throw exception.  Seems .NET throws.
 			if (path == String.Empty)
-				throw new ArgumentException();
+				throw new ArgumentException("Invalid path");
 
 			if (path == null || GetPathRoot (path) == path)
 				return null;
 
-			CheckArgument.WhitespaceOnly (path);
-			CheckArgument.PathChars (path);
+			if (path.Trim ().Length == 0)
+				throw new ArgumentException ("Argument string consists of whitespace characters only.");
+
+			if (path.IndexOfAny (System.IO.Path.InvalidPathChars) > -1)
+				throw new ArgumentException ("Path contains invalid characters");
 
 			int nLast = path.LastIndexOfAny (PathSeparatorChars);
 			if (nLast == 0)
@@ -157,10 +228,17 @@ namespace System.IO {
 			if (nLast > 0) {
 				string ret = path.Substring (0, nLast);
 				int l = ret.Length;
-				if (l >= 2 && ret [l - 1] == VolumeSeparatorChar)
+
+				if (l >= 2 && DirectorySeparatorChar == '\\' && ret [l - 1] == VolumeSeparatorChar)
 					return ret + DirectorySeparatorChar;
-				else
-					return ret;
+				else {
+					//
+					// Important: do not use CanonicalizePath here, use
+					// the custom CleanPath here, as this should not
+					// return absolute paths
+					//
+					return CleanPath (ret);
+				}
 			}
 
 			return String.Empty;
@@ -172,7 +250,7 @@ namespace System.IO {
 				return null;
 
 			if (path.IndexOfAny (InvalidPathChars) != -1)
-				throw new ArgumentException ("Illegal characters in path", "path");
+				throw new ArgumentException ("Illegal characters in path.");
 
 			int iExt = findExtension (path);
 
@@ -186,11 +264,11 @@ namespace System.IO {
 
 		public static string GetFileName (string path)
 		{
-			if (path == null || path == String.Empty)
+			if (path == null || path.Length == 0)
 				return path;
 
 			if (path.IndexOfAny (InvalidPathChars) != -1)
-				throw new ArgumentException ("Illegal characters in path", "path");
+				throw new ArgumentException ("Illegal characters in path.");
 
 			int nLast = path.LastIndexOfAny (PathSeparatorChars);
 			if (nLast >= 0)
@@ -251,7 +329,7 @@ namespace System.IO {
 
 			if (path.Trim ().Length == 0) {
 				string msg = Locale.GetText ("The specified path is not of a legal form (empty).");
-				throw new ArgumentException (msg, "path");
+				throw new ArgumentException (msg);
 			}
 
 			// adjust for drives, i.e. a special case for windows
@@ -269,6 +347,8 @@ namespace System.IO {
 
 				if (path [0] != DirectorySeparatorChar)
 					path = path.Replace (AltDirectorySeparatorChar, DirectorySeparatorChar);
+
+				path = CanonicalizePath (path);
 			} else {
 				if (!IsPathRooted (path))
 					path = Directory.GetCurrentDirectory () + DirectorySeparatorStr + path;
@@ -301,8 +381,8 @@ namespace System.IO {
 			if (path == null)
 				return null;
 
-			if (path == String.Empty)
-				throw new ArgumentException ("This specified path is invalid.");
+			if (path.Trim ().Length == 0)
+				throw new ArgumentException ("The specified path is not of a legal form.");
 
 			if (!IsPathRooted (path))
 				return String.Empty;
@@ -390,10 +470,13 @@ namespace System.IO {
 		private static extern string get_temp_path ();
 
 		public static bool HasExtension (string path)
-		{  
-			if (path == null || path.Trim () == String.Empty)
+		{
+			if (path == null || path.Trim ().Length == 0)
 				return false;
 
+			if (path.IndexOfAny (InvalidPathChars) != -1)
+				throw new ArgumentException ("Illegal characters in path.");
+
 			int pos = findExtension (path);
 			return 0 <= pos && pos < path.Length - 1;
 		}
@@ -404,7 +487,7 @@ namespace System.IO {
 				return false;
 
 			if (path.IndexOfAny (InvalidPathChars) != -1)
-				throw new ArgumentException ("Illegal characters in path", "path");
+				throw new ArgumentException ("Illegal characters in path.");
 
 			char c = path [0];
 			return (c == DirectorySeparatorChar 	||
@@ -441,24 +524,23 @@ namespace System.IO {
 
 		public static string GetRandomFileName ()
 		{
-			char[] invalid = GetInvalidFileNameChars ();
 			// returns a 8.3 filename (total size 12)
 			StringBuilder sb = new StringBuilder (12);
 			// using strong crypto but without creating the file
 			RandomNumberGenerator rng = RandomNumberGenerator.Create ();
-			byte[] buffer = new byte [16];
-			while (sb.Length < 12) {
-				int i = 0;
-				rng.GetNonZeroBytes (buffer);
-				while ((i < buffer.Length) && (sb.Length < 12)) {
-					char c = (char) buffer [i];
-					if (Array.IndexOf (invalid, c) == -1)
-						sb.Append (c);
-					i++;
-					if (sb.Length == 8)
-						sb.Append ('.');
-				}
+			byte [] buffer = new byte [11];
+			rng.GetBytes (buffer);
+
+			for (int i = 0; i < buffer.Length; i++) {
+				if (sb.Length == 8)
+					sb.Append ('.');
+
+				// restrict to length of range [a..z0..9]
+				int b = (buffer [i] % 36);
+				char c = (char) (b < 26 ? (b + 'a') : (b - 26 + '0'));
+				sb.Append (c);
 			}
+
 			return sb.ToString ();
 		}
 #endif
@@ -507,16 +589,43 @@ namespace System.IO {
 
 			dirEqualsVolume = (DirectorySeparatorChar == VolumeSeparatorChar);
 		}
-		
+
+		// returns the server and share part of a UNC. Assumes "path" is a UNC.
+		static string GetServerAndShare (string path)
+		{
+			int len = 2;
+			while (len < path.Length && !IsDsc (path [len])) len++;
+
+			if (len < path.Length) {
+				len++;
+				while (len < path.Length && !IsDsc (path [len])) len++;
+			}
+
+			return path.Substring (2, len - 2).Replace (AltDirectorySeparatorChar, DirectorySeparatorChar);
+		}
+
+		// assumes Environment.IsRunningOnWindows == true
 		static bool SameRoot (string root, string path)
 		{
 			// compare root - if enough details are available
 			if ((root.Length < 2) || (path.Length < 2))
 				return false;
+
+			// UNC handling
+			if (IsDsc (root[0]) && IsDsc (root[1])) {
+				if (!(IsDsc (path[0]) && IsDsc (path[1])))
+					return false;
+
+				string rootShare = GetServerAndShare (root);
+				string pathShare = GetServerAndShare (path);
+
+				return String.Compare (rootShare, pathShare, true, CultureInfo.InvariantCulture) == 0;
+			}
+			
 			// same volume/drive
 			if (!root [0].Equals (path [0]))
 				return false;
-			// presence if the separator
+			// presence of the separator
 			if (path[1] != Path.VolumeSeparatorChar)
 				return false;
 			if ((root.Length > 2) && (path.Length > 2)) {
@@ -540,7 +649,7 @@ namespace System.IO {
 			// STEP 2: Check to see if this is only a root
 			string root = Path.GetPathRoot (path);
 			// it will return '\' for path '\', while it should return 'c:\' or so.
-			// Note: commenting this out makes the ened for the (target == 1...) check in step 5
+			// Note: commenting this out makes the need for the (target == 1...) check in step 5
 			//if (root == path) return path;
 
 			// STEP 3: split the directories, this gets rid of consecutative "/"'s
@@ -548,11 +657,23 @@ namespace System.IO {
 			// STEP 4: Get rid of directories containing . and ..
 			int target = 0;
 
+			bool isUnc = Environment.IsRunningOnWindows &&
+				root.Length > 2 && IsDsc (root[0]) && IsDsc (root[1]);
+
+			// Set an overwrite limit for UNC paths since '\' + server + share
+			// must not be eliminated by the '..' elimination algorithm.
+			int limit = isUnc ? 3 : 0;
+
 			for (int i = 0; i < dirs.Length; i++) {
+				// WIN32 path components must be trimmed
+				if (Environment.IsRunningOnWindows)
+					dirs[i] = dirs[i].Trim ();
+				
 				if (dirs[i] == "." || (i != 0 && dirs[i].Length == 0))
 					continue;
 				else if (dirs[i] == "..") {
-					if (target != 0)
+					// don't overwrite path segments below the limit
+					if (target > limit)
 						target--;
 				} else
 					dirs[target++] = dirs[i];
@@ -564,10 +685,16 @@ namespace System.IO {
 			else {
 				string ret = String.Join (DirectorySeparatorStr, dirs, 0, target);
 				if (Environment.IsRunningOnWindows) {
+					// append leading '\' of the UNC path that was lost in STEP 3.
+					if (isUnc)
+						ret = Path.DirectorySeparatorStr + ret;
+
 					if (!SameRoot (root, ret))
 						ret = root + ret;
-					// In GetFullPath(), it is assured that here never comes UNC. So this must only applied to such path that starts with '\', without drive specification.
-					if (!IsDsc (path[0]) && SameRoot (root, path)) {
+
+					if (isUnc) {
+						return ret;
+					} else if (!IsDsc (path[0]) && SameRoot (root, path)) {
 						if (ret.Length <= 2 && !ret.EndsWith (DirectorySeparatorStr)) // '\' after "c:"
 							ret += Path.DirectorySeparatorChar;
 						return ret;
