@@ -52,6 +52,7 @@ namespace System.Reflection.Emit
 	[ClassInterface (ClassInterfaceType.None)]
 	public sealed class TypeBuilder : Type, _TypeBuilder
 	{
+#pragma warning disable 169		
 		#region Sync with reflection.h
 		private string tname;
 		private string nspace;
@@ -81,12 +82,14 @@ namespace System.Reflection.Emit
 		private RefEmitPermissionSet[] permissions;
 		private Type created;
 		#endregion
+#pragma warning restore 169		
+		
 		string fullname;
 		bool createTypeCalled;
 		private Type underlying_type;
 
-	public const int UnspecifiedTypeSize = 0;
-
+		public const int UnspecifiedTypeSize = 0;
+		
 		protected override TypeAttributes GetAttributeFlagsImpl ()
 		{
 			return attrs;
@@ -98,8 +101,10 @@ namespace System.Reflection.Emit
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern void create_internal_class (TypeBuilder tb);
 		
+#if NET_2_0 || BOOTSTRAP_NET_2_0
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern void setup_generic_class ();
+#endif
 
 		[MethodImplAttribute(MethodImplOptions.InternalCall)]
 		private extern void create_generic_class ();
@@ -1414,6 +1419,14 @@ namespace System.Reflection.Emit
 				return created.TypeHandle;
 			}
 		}
+		
+		//
+		// Used internally by mcs only
+		//
+		internal void SetCharSet (TypeAttributes ta)
+		{
+			this.attrs = ta;
+		}
 
 		public void SetCustomAttribute (CustomAttributeBuilder customBuilder)
 		{
@@ -1564,8 +1577,6 @@ namespace System.Reflection.Emit
 			return res;
 		}
 
-		static int UnmanagedDataCount = 0;
-		
 		public FieldBuilder DefineUninitializedData (string name, int size, FieldAttributes attributes)
 		{
 			if (name == null)
@@ -1763,6 +1774,7 @@ namespace System.Reflection.Emit
 			get { return IsGenericTypeDefinition; }
 		}
 
+		[MonoTODO]
 		public override int GenericParameterPosition {
 			get {
 				throw new NotImplementedException ();
@@ -1796,13 +1808,44 @@ namespace System.Reflection.Emit
 				return res;
 		}
 
+		static bool IsValidGetMethodType (Type type)
+		{
+			if (type is TypeBuilder || type is MonoGenericClass)
+				return true;
+			/*GetMethod() must work with TypeBuilders after CreateType() was called.*/
+			if (type.Module is ModuleBuilder)
+				return true;
+			if (type.IsGenericParameter)
+				return false;
+
+			Type[] inst = type.GetGenericArguments ();
+			if (inst == null)
+				return false;
+			for (int i = 0; i < inst.Length; ++i) {
+				if (IsValidGetMethodType (inst [i]))
+					return true;
+			}
+			return false;
+		}
+
 		public static MethodInfo GetMethod (Type type, MethodInfo method)
 		{
+			if (!IsValidGetMethodType (type))
+				throw new ArgumentException ("type is not TypeBuilder but " + type.GetType (), "type");
+
+			if (!type.IsGenericType)
+				throw new ArgumentException ("type is not a generic type", "type");
+
+			if (!method.DeclaringType.IsGenericTypeDefinition)
+				throw new ArgumentException ("method declaring type is not a generic type definition", "method");
+			if (method.DeclaringType != type.GetGenericTypeDefinition ())
+				throw new ArgumentException ("method declaring type is not the generic type definition of type", "method");
+
 			MethodInfo res = type.GetMethod (method);
 			if (res == null)
-				throw new System.Exception ("method not found");
-			else
-				return res;
+				throw new ArgumentException (String.Format ("method {0} not found in type {1}", method.Name, type));
+				
+			return res;
 		}
 
 		public static FieldInfo GetField (Type type, FieldInfo field)
