@@ -51,6 +51,7 @@ using System.Security.Principal;
 using System.Configuration.Assemblies;
 
 #if NET_2_0
+using System.Collections.Generic;
 using System.Runtime.ConstrainedExecution;
 #endif
 
@@ -58,6 +59,7 @@ namespace System {
 
 #if NET_2_0
 	[ComVisible (true)]
+	[ComDefaultInterface (typeof (_AppDomain))]
 #endif
 	[ClassInterface(ClassInterfaceType.None)]
 	public sealed class AppDomain : MarshalByRefObject , _AppDomain , IEvidenceFactory
@@ -80,6 +82,8 @@ namespace System {
 
 		[ThreadStatic]
 		private static IPrincipal _principal;
+		
+		static AppDomain default_domain;
 
 		private AppDomain ()
 		{
@@ -99,6 +103,13 @@ namespace System {
 			}
 		}
 
+#if NET_2_0
+		[MonoTODO]
+		public ApplicationTrust ApplicationTrust {
+			get { throw new NotImplementedException (); }
+		}
+#endif
+
 		public string BaseDirectory {
 			get {
 				string path = SetupInformationNoCopy.ApplicationBase;
@@ -158,12 +169,18 @@ namespace System {
 					// ... we will provide our own
 					lock (this) {
 						// the executed assembly from the "default" appdomain
-						// or null if we're not in the default appdomain
+						// or null if we're not in the default appdomain or
+						// if there is no entry assembly (embedded mono)
 						Assembly a = Assembly.GetEntryAssembly ();
-						if (a == null)
-							_evidence = AppDomain.DefaultDomain.Evidence;
-						else
+						if (a == null) {
+							if (this == DefaultDomain)
+								// mono is embedded
+								return new Evidence ();
+							else
+								_evidence = AppDomain.DefaultDomain.Evidence;
+						} else {
 							_evidence = Evidence.GetDefaultHostEvidence (a);
+						}
 					}
 				}
 				return new Evidence (_evidence);	// return a copy
@@ -207,12 +224,19 @@ namespace System {
 
 		internal static AppDomain DefaultDomain {
 			get {
-				return getRootDomain ();
+				if (default_domain == null) {
+					AppDomain rd = getRootDomain ();
+					if (rd == CurrentDomain)
+						default_domain = rd;
+					else
+						default_domain = (AppDomain) RemotingServices.GetDomainProxy (rd);
+				}
+				return default_domain;
 			}
 		}
 
 #if NET_2_0
-		[Obsolete ("")]
+		[Obsolete ("AppDomain.AppendPrivatePath has been deprecated. Please investigate the use of AppDomainSetup.PrivateBinPath instead.")]
 #endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void AppendPrivatePath (string path)
@@ -236,7 +260,7 @@ namespace System {
 		}
 
 #if NET_2_0
-		[Obsolete ("")]
+		[Obsolete ("AppDomain.ClearPrivatePath has been deprecated. Please investigate the use of AppDomainSetup.PrivateBinPath instead.")]
 #endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void ClearPrivatePath ()
@@ -244,6 +268,9 @@ namespace System {
 			SetupInformationNoCopy.PrivateBinPath = String.Empty;
 		}
 
+#if NET_2_0
+		[Obsolete ("Use AppDomainSetup.ShadowCopyDirectories")]
+#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void ClearShadowCopyPath ()
 		{
@@ -256,10 +283,10 @@ namespace System {
 		}
 
 #if NET_1_1
-		public ObjectHandle CreateComInstanceFrom (string assemblyName, string typeName,
+		public ObjectHandle CreateComInstanceFrom (string assemblyFile, string typeName,
 			byte [] hashValue, AssemblyHashAlgorithm hashAlgorithm)
 		{
-			return Activator.CreateComInstanceFrom (assemblyName, typeName, hashValue ,hashAlgorithm);
+			return Activator.CreateComInstanceFrom (assemblyFile, typeName, hashValue ,hashAlgorithm);
 		}
 #endif
 
@@ -311,30 +338,30 @@ namespace System {
 			return (oh != null) ? oh.Unwrap () : null;
 		}
 
-		public ObjectHandle CreateInstanceFrom (string assemblyName, string typeName)
+		public ObjectHandle CreateInstanceFrom (string assemblyFile, string typeName)
 		{
-			if (assemblyName == null)
-				throw new ArgumentNullException ("assemblyName");
+			if (assemblyFile == null)
+				throw new ArgumentNullException ("assemblyFile");
 
-			return Activator.CreateInstanceFrom (assemblyName, typeName);
+			return Activator.CreateInstanceFrom (assemblyFile, typeName);
 		}
 
-		public ObjectHandle CreateInstanceFrom (string assemblyName, string typeName, object[] activationAttributes)
+		public ObjectHandle CreateInstanceFrom (string assemblyFile, string typeName, object[] activationAttributes)
 		{
-			if (assemblyName == null)
-				throw new ArgumentNullException ("assemblyName");
+			if (assemblyFile == null)
+				throw new ArgumentNullException ("assemblyFile");
 
-			return Activator.CreateInstanceFrom (assemblyName, typeName, activationAttributes);
+			return Activator.CreateInstanceFrom (assemblyFile, typeName, activationAttributes);
 		}
 
-		public ObjectHandle CreateInstanceFrom (string assemblyName, string typeName, bool ignoreCase,
+		public ObjectHandle CreateInstanceFrom (string assemblyFile, string typeName, bool ignoreCase,
 		                                        BindingFlags bindingAttr, Binder binder, object[] args, CultureInfo culture,
 		                                        object[] activationAttributes, Evidence securityAttributes)
 		{
-			if (assemblyName == null)
-				throw new ArgumentNullException ("assemblyName");
+			if (assemblyFile == null)
+				throw new ArgumentNullException ("assemblyFile");
 
-			return Activator.CreateInstanceFrom (assemblyName, typeName, ignoreCase, bindingAttr, binder, args,
+			return Activator.CreateInstanceFrom (assemblyFile, typeName, ignoreCase, bindingAttr, binder, args,
 			                                     culture, activationAttributes, securityAttributes);
 		}
 
@@ -425,6 +452,10 @@ namespace System {
 		                                              PermissionSet optionalPermissions,
 		                                              PermissionSet refusedPermissions, bool isSynchronized)
 		{
+			if (name == null)
+				throw new ArgumentNullException ("name");
+			ValidateAssemblyName (name.Name);
+
 			// FIXME: examine all other parameters
 			
 			AssemblyBuilder ab = new AssemblyBuilder (name, dir, access, false);
@@ -432,15 +463,49 @@ namespace System {
 			return ab;
 		}
 
+#if NET_2_0
+		// NET 3.5 method
+		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, string dir,
+		                                              Evidence evidence,
+		                                              PermissionSet requiredPermissions,
+		                                              PermissionSet optionalPermissions,
+		                                              PermissionSet refusedPermissions, bool isSynchronized, IEnumerable<CustomAttributeBuilder> assemblyAttributes)
+		{
+			AssemblyBuilder ab = DefineDynamicAssembly (name, access, dir, evidence, requiredPermissions, optionalPermissions, refusedPermissions, isSynchronized);
+			if (assemblyAttributes != null)
+				foreach (CustomAttributeBuilder cb in assemblyAttributes) {
+					ab.SetCustomAttribute (cb);
+				}
+			return ab;
+		}
+
+		// NET 3.5 method
+		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, IEnumerable<CustomAttributeBuilder> assemblyAttributes) {
+			return DefineDynamicAssembly (name, access, null, null, null, null, null, false, assemblyAttributes);
+		}
+#endif
+
+#if NET_2_1
+		// TODO: the last parameter is ignored for now
+		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, bool emitSymbolInfo)
+		{
+			return DefineDynamicAssembly (name, access, null, null, null, null, null, false);
+		}
+#endif
+
 		internal AssemblyBuilder DefineInternalDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access)
 		{
 			return new AssemblyBuilder (name, null, access, true);
 		}
 
-		public void DoCallBack (CrossAppDomainDelegate theDelegate)
+		//
+		// AppDomain.DoCallBack works because AppDomain is a MarshalByRefObject
+		// so, when you call AppDomain.DoCallBack, that's a remote call
+		//
+		public void DoCallBack (CrossAppDomainDelegate callBackDelegate)
 		{
-			if (theDelegate != null)
-				theDelegate ();
+			if (callBackDelegate != null)
+				callBackDelegate ();
 		}
 
 		public int ExecuteAssembly (string assemblyFile)
@@ -453,14 +518,31 @@ namespace System {
 			return ExecuteAssembly (assemblyFile, assemblySecurity, null);
 		}
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		public extern int ExecuteAssembly (string assemblyFile, Evidence assemblySecurity, string[] args);
+		public int ExecuteAssembly (string assemblyFile, Evidence assemblySecurity, string[] args)
+		{
+			Assembly a = Assembly.LoadFrom (assemblyFile, assemblySecurity);
+			return ExecuteAssemblyInternal (a, args);
+		}
 
-		[MonoTODO ("No support for ExecuteAssembly")]
 		public int ExecuteAssembly (string assemblyFile, Evidence assemblySecurity, string[] args, byte[] hashValue, AssemblyHashAlgorithm hashAlgorithm)
 		{
-			throw new NotImplementedException ();
+			Assembly a = Assembly.LoadFrom (assemblyFile, assemblySecurity, hashValue, hashAlgorithm);
+			return ExecuteAssemblyInternal (a, args);
 		}
+
+		int ExecuteAssemblyInternal (Assembly a, string[] args)
+		{
+			if (a.EntryPoint == null)
+#if NET_2_0
+				throw new MissingMethodException ("Entry point not found in assembly '" + a.FullName + "'.");
+#else
+				throw new COMException ("Unspecified error.", -2147467259);
+#endif
+			return ExecuteAssembly (a, args);
+		}
+
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private extern int ExecuteAssembly (Assembly a, string[] args);
 		
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern Assembly [] GetAssemblies (bool refOnly);
@@ -491,6 +573,17 @@ namespace System {
 			return Load (assemblyRef, null);
 		}
 
+		internal Assembly LoadSatellite (AssemblyName assemblyRef)
+		{
+			if (assemblyRef == null)
+				throw new ArgumentNullException ("assemblyRef");
+
+			Assembly result = LoadAssembly (assemblyRef.FullName, null, false);
+			if (result == null)
+				throw new FileNotFoundException (null, assemblyRef.Name);
+			return result;
+		}
+
 		public Assembly Load (AssemblyName assemblyRef, Evidence assemblySecurity)
 		{
 			if (assemblyRef == null)
@@ -503,15 +596,48 @@ namespace System {
 					throw new ArgumentException (Locale.GetText ("assemblyRef.Name cannot be empty."), "assemblyRef");
 			}
 
-			return LoadAssembly (assemblyRef.FullName, assemblySecurity, false);
+			Assembly assembly = LoadAssembly (assemblyRef.FullName, assemblySecurity, false);
+			if (assembly != null)
+				return assembly;
+
+			if (assemblyRef.CodeBase == null)
+				throw new FileNotFoundException (null, assemblyRef.Name);
+
+			string cb = assemblyRef.CodeBase;
+			if (cb.ToLower (CultureInfo.InvariantCulture).StartsWith ("file://"))
+				cb = new Mono.Security.Uri (cb).LocalPath;
+
+			try {
+				assembly = Assembly.LoadFrom (cb, assemblySecurity);
+			} catch {
+				throw new FileNotFoundException (null, assemblyRef.Name);
+			}
+			AssemblyName aname = assembly.GetName ();
+			// Name, version, culture, publickeytoken. Anything else?
+			if (assemblyRef.Name != aname.Name)
+				throw new FileNotFoundException (null, assemblyRef.Name);
+
+			if (assemblyRef.Version != new Version () && assemblyRef.Version != aname.Version)
+				throw new FileNotFoundException (null, assemblyRef.Name);
+
+			if (assemblyRef.CultureInfo != null && assemblyRef.CultureInfo.Equals (aname))
+				throw new FileNotFoundException (null, assemblyRef.Name);
+
+			byte [] pt = assemblyRef.GetPublicKeyToken ();
+			if (pt != null) {
+				byte [] loaded_pt = aname.GetPublicKeyToken ();
+				if (loaded_pt == null || (pt.Length != loaded_pt.Length))
+					throw new FileNotFoundException (null, assemblyRef.Name);
+				for (int i = pt.Length - 1; i >= 0; i--)
+					if (loaded_pt [i] != pt [i])
+						throw new FileNotFoundException (null, assemblyRef.Name);
+			}
+			return assembly;
 		}
 
 		public Assembly Load (string assemblyString)
 		{
-			if (assemblyString == null)
-				throw new ArgumentNullException ("assemblyString");
-
-			return LoadAssembly (assemblyString, null, false);
+			return Load (assemblyString, null, false);
 		}
 
 		public Assembly Load (string assemblyString, Evidence assemblySecurity)
@@ -523,8 +649,14 @@ namespace System {
 		{
 			if (assemblyString == null)
 				throw new ArgumentNullException ("assemblyString");
+				
+			if (assemblyString.Length == 0)
+				throw new ArgumentException ("assemblyString cannot have zero length");
 
-			return LoadAssembly (assemblyString, assemblySecurity, refonly);
+			Assembly assembly = LoadAssembly (assemblyString, assemblySecurity, refonly);
+			if (assembly == null)
+				throw new FileNotFoundException (null, assemblyString);
+			return assembly;
 		}
 
 		public Assembly Load (byte[] rawAssembly)
@@ -549,7 +681,7 @@ namespace System {
 		{
 			if (rawAssembly == null)
 				throw new ArgumentNullException ("rawAssembly");
-				
+
 			Assembly assembly = LoadAssemblyRaw (rawAssembly, rawSymbolStore, securityEvidence, refonly);
 			assembly.FromByteArray = true;
 			return assembly;
@@ -571,6 +703,9 @@ namespace System {
 			_granted = ps.PermissionSet;
 		}
 
+#if NET_2_0
+		[Obsolete ("Use AppDomainSetup.SetCachePath")]
+#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetCachePath (string path)
 		{
@@ -587,12 +722,18 @@ namespace System {
 			_principal = null;
 		}
 
+#if NET_2_0
+		[Obsolete ("Use AppDomainSetup.ShadowCopyFiles")]
+#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetShadowCopyFiles()
 		{
 			SetupInformationNoCopy.ShadowCopyFiles = "true";
 		}
 
+#if NET_2_0
+		[Obsolete ("Use AppDomainSetup.ShadowCopyDirectories")]
+#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetShadowCopyPath (string path)
 		{
@@ -654,10 +795,14 @@ namespace System {
 			bool pushed = false;
 
 			try {
+				Exception exc;
 				InternalPushDomainRef (domain);
 				pushed = true;
 				InternalSetDomain (domain);
-				return ((MonoMethod) method).InternalInvoke (obj, args);
+				object o = ((MonoMethod) method).InternalInvoke (obj, args, out exc);
+				if (exc != null)
+					throw exc;
+				return o;
 			}
 			finally {
 				InternalSetDomain (current);
@@ -672,10 +817,14 @@ namespace System {
 			bool pushed = false;
 
 			try {
+				Exception exc;
 				InternalPushDomainRefByID (domain_id);
 				pushed = true;
 				InternalSetDomainByID (domain_id);
-				return ((MonoMethod) method).InternalInvoke (obj, args);
+				object o = ((MonoMethod) method).InternalInvoke (obj, args, out exc);
+				if (exc != null)
+					throw exc;
+				return o;
 			}
 			finally {
 				InternalSetDomain (current);
@@ -746,11 +895,68 @@ namespace System {
 			else
 				ad._evidence = new Evidence (securityInfo);	// copy
 
+#if NET_2_0
+			if (info.AppDomainInitializer != null) {
+				if (!info.AppDomainInitializer.Method.IsStatic)
+					throw new ArgumentException ("Non-static methods cannot be invoked as an appdomain initializer");
+
+				Loader loader = new Loader (
+					info.AppDomainInitializer.Method.DeclaringType.Assembly.Location);
+				ad.DoCallBack (loader.Load);
+
+				Initializer initializer = new Initializer (
+					info.AppDomainInitializer,
+					info.AppDomainInitializerArguments);
+				ad.DoCallBack (initializer.Initialize);
+			}
+#endif
+
 			return ad;
 		}
 
+#if NET_2_0
+		[Serializable]
+		class Loader {
+
+			string assembly;
+
+			public Loader (string assembly)
+			{
+				this.assembly = assembly;
+			}
+
+			public void Load ()
+			{
+				Assembly.LoadFrom (assembly);
+			}
+		}
+
+		[Serializable]
+		class Initializer {
+
+			AppDomainInitializer initializer;
+			string [] arguments;
+
+			public Initializer (AppDomainInitializer initializer, string [] arguments)
+			{
+				this.initializer = initializer;
+				this.arguments = arguments;
+			}
+
+			public void Initialize ()
+			{
+				initializer (arguments);
+			}
+		}
+#endif
+
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo,string appBasePath,
 		                                      string appRelativeSearchPath, bool shadowCopyFiles)
+		{
+			return CreateDomain (friendlyName, securityInfo, CreateDomainSetup (appBasePath, appRelativeSearchPath, shadowCopyFiles));
+		}
+
+		static AppDomainSetup CreateDomainSetup (string appBasePath, string appRelativeSearchPath, bool shadowCopyFiles)
 		{
 			AppDomainSetup info = new AppDomainSetup ();
 
@@ -760,11 +966,24 @@ namespace System {
 			if (shadowCopyFiles)
 				info.ShadowCopyFiles = "true";
 			else
+#if NET_2_0
 				info.ShadowCopyFiles = "false";
+#else
+				info.ShadowCopyFiles = null;
+#endif
 
-			return CreateDomain (friendlyName, securityInfo, info);
+			return info;
 		}
 
+#if NET_2_0
+		[MonoTODO]
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info,
+		                                      PermissionSet grantSet, params StrongName [] fullTrustAssemblies)
+		{
+			throw new NotImplementedException ();
+		}
+#endif
+
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern bool InternalIsFinalizingForUnload (int domain_id);
 
@@ -799,6 +1018,17 @@ namespace System {
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public extern void SetData (string name, object data);
 
+#if NET_2_0
+		[MonoTODO]
+		public void SetData (string name, object data, IPermission permission)
+		{
+			throw new NotImplementedException ();
+		}
+#endif
+
+#if NET_2_0
+		[Obsolete ("Use AppDomainSetup.DynamicBase")]
+#endif
 		[SecurityPermission (SecurityAction.LinkDemand, ControlAppDomain = true)]
 		public void SetDynamicBase (string path)
 		{
@@ -806,7 +1036,11 @@ namespace System {
 		}
 
 #if NET_2_0
-		[Obsolete ("")]
+		[Obsolete ("AppDomain.GetCurrentThreadId has been deprecated"
+			+ " because it does not provide a stable Id when managed"
+			+ " threads are running on fibers (aka lightweight"
+			+ " threads). To get a stable identifier for a managed"
+			+ " thread, use the ManagedThreadId property on Thread.'")]
 #endif
 		public static int GetCurrentThreadId ()
 		{
@@ -818,7 +1052,40 @@ namespace System {
 			return getFriendlyName ();
 		}
 
+		private static void ValidateAssemblyName (string name)
+		{
+			if (name == null || name.Length == 0)
+				throw new ArgumentException ("The Name of " +
+					"AssemblyName cannot be null or a " +
+					"zero-length string.");
+
+			bool isValid = true;
+
+			for (int i = 0; i < name.Length; i++) {
+				char c = name [i];
+
+				// do not allow leading whitespace
+				if (i == 0 && char.IsWhiteSpace (c)) {
+					isValid = false;
+					break;
+				}
+
+				// do not allow /,\ or : in name
+				if (c == '/' || c == '\\' || c == ':') {
+					isValid = false;
+					break;
+				}
+			}
+
+			if (!isValid)
+				throw new ArgumentException ("The Name of " +
+					"AssemblyName cannot start with " +
+					"whitespace, or contain '/', '\\' " +
+					" or ':'.");
+		}
+
 		// The following methods are called from the runtime. Don't change signatures.
+#pragma warning disable 169		
 		private void DoAssemblyLoad (Assembly assembly)
 		{
 			if (AssemblyLoad == null)
@@ -829,11 +1096,16 @@ namespace System {
 
 		private Assembly DoAssemblyResolve (string name, bool refonly)
 		{
+			ResolveEventHandler del;
 #if NET_2_0
-			if (refonly && ReflectionOnlyAssemblyResolve == null)
-				return null;
+			if (refonly)
+				del = ReflectionOnlyAssemblyResolve;
+			else
+				del = AssemblyResolve;
+#else
+			del = AssemblyResolve;
 #endif
-			if (AssemblyResolve == null)
+			if (del == null)
 				return null;
 			
 			/* Prevent infinite recursion */
@@ -853,13 +1125,7 @@ namespace System {
 #endif
 			ht [name] = name;
 			try {
-				
-#if NET_2_0
-				Delegate [] invocation_list = refonly ? ReflectionOnlyAssemblyResolve.GetInvocationList () : 
-					AssemblyResolve.GetInvocationList ();
-#else
-				Delegate [] invocation_list = AssemblyResolve.GetInvocationList ();
-#endif
+				Delegate[] invocation_list = del.GetInvocationList ();
 
 				foreach (Delegate eh in invocation_list) {
 					ResolveEventHandler handler = (ResolveEventHandler) eh;
@@ -943,6 +1209,7 @@ namespace System {
 			else
 				arrResponse = null;
 		}
+#pragma warning restore 169
 
 		// End of methods called from the runtime
 		
@@ -986,13 +1253,6 @@ namespace System {
 		public event UnhandledExceptionEventHandler UnhandledException;
 #endif
 
-		/* Avoid warnings for events used only by the runtime */
-		private void DummyUse () {
-			ProcessExit += (EventHandler)null;
-			ResourceResolve += (ResolveEventHandler)null;
-			UnhandledException += (UnhandledExceptionEventHandler)null;
-		}
-
 #if NET_2_0
 
 		public event ResolveEventHandler ReflectionOnlyAssemblyResolve;
@@ -1036,42 +1296,39 @@ namespace System {
 
 		// static methods
 
-		[MonoTODO ("add support for new delegate")]
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, string appBasePath,
-			string appRelativeSearchPath, bool shadowCopy, AppDomainInitializer adInit, string[] adInitArgs)
+			string appRelativeSearchPath, bool shadowCopyFiles, AppDomainInitializer adInit, string[] adInitArgs)
 		{
-			return CreateDomain (friendlyName, securityInfo, appBasePath, appRelativeSearchPath, shadowCopy);
+			AppDomainSetup info = CreateDomainSetup (appBasePath, appRelativeSearchPath, shadowCopyFiles);
+
+			info.AppDomainInitializerArguments = adInitArgs;
+			info.AppDomainInitializer = adInit;
+
+			return CreateDomain (friendlyName, securityInfo, info);
 		}
 
-		[MonoTODO ("resolve assemblyName to location")]
 		public int ExecuteAssemblyByName (string assemblyName)
 		{
 			return ExecuteAssemblyByName (assemblyName, null, null);
 		}
 
-		[MonoTODO ("resolve assemblyName to location")]
 		public int ExecuteAssemblyByName (string assemblyName, Evidence assemblySecurity)
 		{
 			return ExecuteAssemblyByName (assemblyName, assemblySecurity, null);
 		}
 
-		[MonoTODO ("resolve assemblyName to location")]
 		public int ExecuteAssemblyByName (string assemblyName, Evidence assemblySecurity, params string[] args)
 		{
-			if (assemblyName == null)
-				throw new ArgumentNullException ("assemblyName");
+			Assembly a = Assembly.Load (assemblyName, assemblySecurity);
 
-			AssemblyName an = new AssemblyName (assemblyName);
-			return ExecuteAssemblyByName (an, assemblySecurity, args);
+			return ExecuteAssemblyInternal (a, args);
 		}
 
-		[MonoTODO ("assemblyName may not have a codebase")]
 		public int ExecuteAssemblyByName (AssemblyName assemblyName, Evidence assemblySecurity, params string[] args)
 		{
-			if (assemblyName == null)
-				throw new ArgumentNullException ("assemblyName");
+			Assembly a = Assembly.Load (assemblyName, assemblySecurity);
 
-			return ExecuteAssembly (assemblyName.CodeBase, assemblySecurity, args);
+			return ExecuteAssemblyInternal (a, args);
 		}
 
 		public bool IsDefaultAppDomain ()
