@@ -45,14 +45,19 @@ namespace System
 		internal Type utype;
 		internal Array values;
 		internal string[] names;
+		internal Hashtable name_hash;
 		[ThreadStatic]
 		static Hashtable cache;
+		static Hashtable global_cache;
+		static object global_cache_monitor;
 		
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern void get_enum_info (Type enumType, out MonoEnumInfo info);
 
 		static MonoEnumInfo ()
 		{
+			global_cache_monitor = new object ();
+			global_cache = new Hashtable ();
 		}
 
 		static Hashtable Cache {
@@ -68,17 +73,36 @@ namespace System
 			utype = other.utype;
 			values = other.values;
 			names = other.names;
+			name_hash = other.name_hash;
 		}
 
 		internal static void GetInfo (Type enumType, out MonoEnumInfo info)
 		{
+			/* First check the thread-local cache without locking */
 			if (Cache.ContainsKey (enumType)) {
 				info = (MonoEnumInfo) cache [enumType];
 				return;
 			}
+			/* Threads could die, so keep a global cache too */
+			lock (global_cache_monitor) {
+				if (global_cache.ContainsKey (enumType)) {
+					info = (MonoEnumInfo) global_cache [enumType];
+					cache [enumType] = info;
+					return;
+				}
+			}
+
 			get_enum_info (enumType, out info);
 			Array.Sort (info.values, info.names);
-			cache.Add (enumType, new MonoEnumInfo (info));
+			if (info.names.Length > 50) {
+				info.name_hash = new Hashtable (info.names.Length);
+				for (int i = 0; i <  info.names.Length; ++i)
+					info.name_hash [info.names [i]] = i;
+			}
+			MonoEnumInfo cached = new MonoEnumInfo (info);
+			lock (global_cache_monitor) {
+				global_cache [enumType] = cached;
+			}
 		}
 	};
 
@@ -143,10 +167,17 @@ namespace System
 			return Convert.ToInt64 (get_value (), provider);
 		}
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		sbyte IConvertible.ToSByte (IFormatProvider provider)
 		{
 			return Convert.ToSByte (get_value (), provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
 		float IConvertible.ToSingle (IFormatProvider provider)
 		{
@@ -158,20 +189,42 @@ namespace System
 			return Convert.ToType (get_value (), type, provider);
 		}
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		ushort IConvertible.ToUInt16 (IFormatProvider provider)
 		{
 			return Convert.ToUInt16 (get_value (), provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		uint IConvertible.ToUInt32 (IFormatProvider provider)
 		{
 			return Convert.ToUInt32 (get_value (), provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		ulong IConvertible.ToUInt64 (IFormatProvider provider)
 		{
 			return Convert.ToUInt64 (get_value (), provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
+
 		// <-- End IConvertible methods
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
@@ -285,12 +338,19 @@ namespace System
 			return Parse (enumType, value, false);
 		}
 
-		private static int FindName (string [] names, string name,  bool ignoreCase)
+		private static int FindName (Hashtable name_hash, string [] names, string name,  bool ignoreCase)
 		{
 			if (!ignoreCase) {
-				for (int i = 0; i < names.Length; ++i) {
-					if (name == names [i])
-						return i;
+				/* For enums with many values, use a hash table */
+				if (name_hash != null) {
+					object val = name_hash [name];
+					if (val != null)
+						return (int)val;
+				} else {
+					for (int i = 0; i < names.Length; ++i) {
+						if (name == names [i])
+							return i;
+					}
 				}
 			} else {
 				for (int i = 0; i < names.Length; ++i) {
@@ -349,7 +409,7 @@ namespace System
 			MonoEnumInfo.GetInfo (enumType, out info);
 
 			// is 'value' a named constant?
-			int loc = FindName (info.names, value, ignoreCase);
+			int loc = FindName (info.name_hash, info.names, value, ignoreCase);
 			if (loc >= 0)
 				return info.values.GetValue (loc);
 
@@ -360,7 +420,7 @@ namespace System
 				string [] names = value.Split (split_char);
 				ulong retVal = 0;
 				for (int i = 0; i < names.Length; ++i) {
-					loc = FindName (info.names, names [i].Trim (), ignoreCase);
+					loc = FindName (info.name_hash, info.names, names [i].Trim (), ignoreCase);
 					if (loc < 0)
 						throw new ArgumentException ("The requested value was not found.");
 					retVal |= GetValue (info.values.GetValue (loc), typeCode);
@@ -516,10 +576,12 @@ namespace System
 			return DefaultEquals (this, obj);
 		}
 
+		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+		private extern int get_hashcode ();
+
 		public override int GetHashCode ()
 		{
-			object v = this.get_value ();
-			return v.GetHashCode ();
+			return get_hashcode ();
 		}
 
 		private static string FormatSpecifier_X (Type enumType, object value, bool upper)
