@@ -634,7 +634,33 @@ namespace System
 		{
 			return String.Compare (value, this, comparisonType) == 0;
 		}
+
+		public static int Compare (string strA, string strB, CultureInfo culture, CompareOptions options)
+		{
+			if (culture == null)
+				throw new ArgumentNullException ("culture");
+
+			return culture.CompareInfo.Compare (strA, strB, options);
+		}
+
+		public static int Compare (string strA, int indexA, string strB, int indexB, int length, CultureInfo culture, CompareOptions options)
+		{
+			if (culture == null)
+				throw new ArgumentNullException ("culture");
+
+			int len1 = length;
+			int len2 = length;
+			
+			if (length > (strA.Length - indexA))
+				len1 = strA.Length - indexA;
+
+			if (length > (strB.Length - indexB))
+				len2 = strB.Length - indexB;
+
+			return culture.CompareInfo.Compare (strA, indexA, len1, strB, indexB, len2, options);
+		}
 #endif
+
 		public int CompareTo (Object value)
 		{
 			if (value == null)
@@ -1406,9 +1432,9 @@ namespace System
 			return Normalization.Normalize (this, 0);
 		}
 
-		public string Normalize (NormalizationForm form)
+		public string Normalize (NormalizationForm normalizationForm)
 		{
-			switch (form) {
+			switch (normalizationForm) {
 			default:
 				return Normalization.Normalize (this, 0);
 			case NormalizationForm.FormD:
@@ -1425,9 +1451,9 @@ namespace System
 			return Normalization.IsNormalized (this, 0);
 		}
 
-		public bool IsNormalized (NormalizationForm form)
+		public bool IsNormalized (NormalizationForm normalizationForm)
 		{
-			switch (form) {
+			switch (normalizationForm) {
 			default:
 				return Normalization.IsNormalized (this, 0);
 			case NormalizationForm.FormD:
@@ -1661,6 +1687,8 @@ namespace System
 					}
 					i = found + oldValue.length;
 				}
+				if (count == 0)
+					return this;
 				int nlen = this.length + ((newValue.length - oldValue.length) * count);
 				String tmp = InternalAllocateStr (nlen);
 
@@ -1825,18 +1853,35 @@ namespace System
 	
 		public static string Format (IFormatProvider provider, string format, params object[] args)
 		{
-			StringBuilder b = new StringBuilder ();
-			FormatHelper (b, provider, format, args);
+			StringBuilder b = FormatHelper (null, provider, format, args);
 			return b.ToString ();
 		}
 		
-		internal static void FormatHelper (StringBuilder result, IFormatProvider provider, string format, params object[] args)
+		internal static StringBuilder FormatHelper (StringBuilder result, IFormatProvider provider, string format, params object[] args)
 		{
 			if (format == null)
 				throw new ArgumentNullException ("format");
 			if (args == null)
 				throw new ArgumentNullException ("args");
 
+			if (result == null) {
+				/* Try to approximate the size of result to avoid reallocations */
+				int i, len;
+
+				len = 0;
+				for (i = 0; i < args.Length; ++i) {
+					string s = args [i] as string;
+					if (s != null)
+						len += s.length;
+					else
+						break;
+				}
+				if (i == args.Length)
+					result = new StringBuilder (len + format.length);
+				else
+					result = new StringBuilder ();
+			}
+
 			int ptr = 0;
 			int start = ptr;
 			while (ptr < format.length) {
@@ -1911,6 +1956,8 @@ namespace System
 
 			if (start < format.length)
 				result.Append (format, start, format.Length - start);
+
+			return result;
 		}
 
 		public unsafe static String Copy (String str)
@@ -1929,20 +1976,20 @@ namespace System
 			return tmp;
 		}
 
-		public static String Concat (Object obj)
+		public static String Concat (Object arg0)
 		{
-			if (obj == null)
+			if (arg0 == null)
 				return String.Empty;
 
-			return obj.ToString ();
+			return arg0.ToString ();
 		}
 
-		public unsafe static String Concat (Object obj1, Object obj2)
+		public unsafe static String Concat (Object arg0, Object arg1)
 		{
 			string s1, s2;
 
-			s1 = (obj1 != null) ? obj1.ToString () : null;
-			s2 = (obj2 != null) ? obj2.ToString () : null;
+			s1 = (arg0 != null) ? arg0.ToString () : null;
+			s2 = (arg1 != null) ? arg1.ToString () : null;
 			
 			if (s1 == null) {
 				if (s2 == null)
@@ -1967,55 +2014,55 @@ namespace System
 			return tmp;
 		}
 
-		public static String Concat (Object obj1, Object obj2, Object obj3)
+		public static String Concat (Object arg0, Object arg1, Object arg2)
 		{
 			string s1, s2, s3;
-			if (obj1 == null)
+			if (arg0 == null)
 				s1 = String.Empty;
 			else
-				s1 = obj1.ToString ();
+				s1 = arg0.ToString ();
 
-			if (obj2 == null)
+			if (arg1 == null)
 				s2 = String.Empty;
 			else
-				s2 = obj2.ToString ();
+				s2 = arg1.ToString ();
 
-			if (obj3 == null)
+			if (arg2 == null)
 				s3 = String.Empty;
 			else
-				s3 = obj3.ToString ();
+				s3 = arg2.ToString ();
 
 			return Concat (s1, s2, s3);
 		}
 
 #if ! BOOTSTRAP_WITH_OLDLIB
 		[CLSCompliant(false)]
-		public static String Concat (Object obj1, Object obj2, Object obj3,
-					     Object obj4, __arglist)
+		public static String Concat (Object arg0, Object arg1, Object arg2,
+					     Object arg3, __arglist)
 		{
 			string s1, s2, s3, s4;
 
-			if (obj1 == null)
+			if (arg0 == null)
 				s1 = String.Empty;
 			else
-				s1 = obj1.ToString ();
+				s1 = arg0.ToString ();
 
-			if (obj2 == null)
+			if (arg1 == null)
 				s2 = String.Empty;
 			else
-				s2 = obj2.ToString ();
+				s2 = arg1.ToString ();
 
-			if (obj3 == null)
+			if (arg2 == null)
 				s3 = String.Empty;
 			else
-				s3 = obj3.ToString ();
+				s3 = arg2.ToString ();
 
 			ArgIterator iter = new ArgIterator (__arglist);
 			int argCount = iter.GetRemainingCount();
 
 			StringBuilder sb = new StringBuilder ();
-			if (obj4 != null)
-				sb.Append (obj4.ToString ());
+			if (arg3 != null)
+				sb.Append (arg3.ToString ());
 
 			for (int i = 0; i < argCount; i++) {
 				TypedReference typedRef = iter.GetNextArg ();
@@ -2028,106 +2075,106 @@ namespace System
 		}
 #endif
 
-		public unsafe static String Concat (String s1, String s2)
+		public unsafe static String Concat (String str0, String str1)
 		{
-			if (s1 == null || s1.Length == 0) {
-				if (s2 == null || s2.Length == 0)
+			if (str0 == null || str0.Length == 0) {
+				if (str1 == null || str1.Length == 0)
 					return String.Empty;
-				return s2;
+				return str1;
 			}
 
-			if (s2 == null || s2.Length == 0)
-				return s1; 
+			if (str1 == null || str1.Length == 0)
+				return str0; 
 
-			String tmp = InternalAllocateStr (s1.length + s2.length);
+			String tmp = InternalAllocateStr (str0.length + str1.length);
 
-			fixed (char *dest = tmp, src = s1)
-				CharCopy (dest, src, s1.length);
-			fixed (char *dest = tmp, src = s2)
-				CharCopy (dest + s1.Length, src, s2.length);
+			fixed (char *dest = tmp, src = str0)
+				CharCopy (dest, src, str0.length);
+			fixed (char *dest = tmp, src = str1)
+				CharCopy (dest + str0.Length, src, str1.length);
 
 			return tmp;
 		}
 
-		public unsafe static String Concat (String s1, String s2, String s3)
+		public unsafe static String Concat (String str0, String str1, String str2)
 		{
-			if (s1 == null || s1.Length == 0){
-				if (s2 == null || s2.Length == 0){
-					if (s3 == null || s3.Length == 0)
+			if (str0 == null || str0.Length == 0){
+				if (str1 == null || str1.Length == 0){
+					if (str2 == null || str2.Length == 0)
 						return String.Empty;
-					return s3;
+					return str2;
 				} else {
-					if (s3 == null || s3.Length == 0)
-						return s2;
+					if (str2 == null || str2.Length == 0)
+						return str1;
 				}
-				s1 = String.Empty;
+				str0 = String.Empty;
 			} else {
-				if (s2 == null || s2.Length == 0){
-					if (s3 == null || s3.Length == 0)
-						return s1;
+				if (str1 == null || str1.Length == 0){
+					if (str2 == null || str2.Length == 0)
+						return str0;
 					else
-						s2 = String.Empty;
+						str1 = String.Empty;
 				} else {
-					if (s3 == null || s3.Length == 0)
-						s3 = String.Empty;
+					if (str2 == null || str2.Length == 0)
+						str2 = String.Empty;
 				}
 			}
 
-			String tmp = InternalAllocateStr (s1.length + s2.length + s3.length);
+			String tmp = InternalAllocateStr (str0.length + str1.length + str2.length);
 
-			if (s1.Length != 0) {
-				fixed (char *dest = tmp, src = s1) {
-					CharCopy (dest, src, s1.length);
+			if (str0.Length != 0) {
+				fixed (char *dest = tmp, src = str0) {
+					CharCopy (dest, src, str0.length);
 				}
 			}
-			if (s2.Length != 0) {
-				fixed (char *dest = tmp, src = s2) {
-					CharCopy (dest + s1.Length, src, s2.length);
+			if (str1.Length != 0) {
+				fixed (char *dest = tmp, src = str1) {
+					CharCopy (dest + str0.Length, src, str1.length);
 				}
 			}
-			if (s3.Length != 0) {
-				fixed (char *dest = tmp, src = s3) {
-					CharCopy (dest + s1.Length + s2.Length, src, s3.length);
+			if (str2.Length != 0) {
+				fixed (char *dest = tmp, src = str2) {
+					CharCopy (dest + str0.Length + str1.Length, src, str2.length);
 				}
 			}
 
 			return tmp;
 		}
 
-		public unsafe static String Concat (String s1, String s2, String s3, String s4)
+		public unsafe static String Concat (String str0, String str1, String str2, String str3)
 		{
-			if (s1 == null && s2 == null && s3 == null && s4 == null)
+			if (str0 == null && str1 == null && str2 == null && str3 == null)
 				return String.Empty;
 
-			if (s1 == null)
-				s1 = String.Empty;
-			if (s2 == null)
-				s2 = String.Empty;
-			if (s3 == null)
-				s3 = String.Empty;
-			if (s4 == null)
-				s4 = String.Empty;
+			if (str0 == null)
+				str0 = String.Empty;
+			if (str1 == null)
+				str1 = String.Empty;
+			if (str2 == null)
+				str2 = String.Empty;
+			if (str3 == null)
+				str3 = String.Empty;
 
-			String tmp = InternalAllocateStr (s1.length + s2.length + s3.length + s4.length);
+			String tmp = InternalAllocateStr (str0.length + str1.length + str2.length + str3.length);
 
-			if (s1.Length != 0) {
-				fixed (char *dest = tmp, src = s1) {
-					CharCopy (dest, src, s1.length);
+			if (str0.Length != 0) {
+				fixed (char *dest = tmp, src = str0) {
+					CharCopy (dest, src, str0.length);
 				}
 			}
-			if (s2.Length != 0) {
-				fixed (char *dest = tmp, src = s2) {
-					CharCopy (dest + s1.Length, src, s2.length);
+			if (str1.Length != 0) {
+				fixed (char *dest = tmp, src = str1) {
+					CharCopy (dest + str0.Length, src, str1.length);
 				}
 			}
-			if (s3.Length != 0) {
-				fixed (char *dest = tmp, src = s3) {
-					CharCopy (dest + s1.Length + s2.Length, src, s3.length);
+			if (str2.Length != 0) {
+				fixed (char *dest = tmp, src = str2) {
+					CharCopy (dest + str0.Length + str1.Length, src, str2.length);
 				}
 			}
-			if (s4.Length != 0) {
-				fixed (char *dest = tmp, src = s4) {
-					CharCopy (dest + s1.Length + s2.Length + s3.Length, src, s4.length);
+			if (str3.Length != 0) {
+				fixed (char *dest = tmp, src = str3) {
+					CharCopy (dest + str0.Length + str1.Length + str2.Length, src, str3.length);
 				}
 			}
 
@@ -2357,36 +2404,64 @@ namespace System
 		{
 			return Convert.ToInt64 (this, provider);
 		}
-	
+
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		sbyte IConvertible.ToSByte (IFormatProvider provider)
 		{
 			return Convert.ToSByte (this, provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
 		float IConvertible.ToSingle (IFormatProvider provider)
 		{
 			return Convert.ToSingle (this, provider);
 		}
 
-		object IConvertible.ToType (Type conversionType, IFormatProvider provider)
+		object IConvertible.ToType (Type type, IFormatProvider provider)
 		{
-			return Convert.ToType (this, conversionType,  provider);
+			return Convert.ToType (this, type, provider);
 		}
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		ushort IConvertible.ToUInt16 (IFormatProvider provider)
 		{
 			return Convert.ToUInt16 (this, provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		uint IConvertible.ToUInt32 (IFormatProvider provider)
 		{
 			return Convert.ToUInt32 (this, provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		ulong IConvertible.ToUInt64 (IFormatProvider provider)
 		{
 			return Convert.ToUInt64 (this, provider);
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
 		public int Length {
 			get {
@@ -2561,7 +2636,7 @@ namespace System
 
 		// Certain constructors are redirected to CreateString methods with
 		// matching argument list. The this pointer should not be used.
-
+#pragma warning disable 169
 		private unsafe String CreateString (sbyte* value)
 		{
 			if (value == null)
@@ -2730,6 +2805,7 @@ namespace System
 			}
 			return result;
 		}
+#pragma warning restore 169
 
 		/* helpers used by the runtime as well as above or eslewhere in corlib */
 		internal static unsafe void memset (byte *dest, int val, int len)
@@ -2952,34 +3028,34 @@ namespace System
 		unsafe public extern String (sbyte *value, int startIndex, int length, Encoding enc);
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		public extern String (char [] val, int startIndex, int length);
+		public extern String (char [] value, int startIndex, int length);
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		public extern String (char [] val);
+		public extern String (char [] value);
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		public extern String (char c, int count);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern static string InternalJoin (string separator, string[] value, int sIndex, int count);
+//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+//		private extern static string InternalJoin (string separator, string[] value, int sIndex, int count);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern String InternalReplace (String oldValue, string newValue, CompareInfo comp);
+//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+//		private extern String InternalReplace (String oldValue, string newValue, CompareInfo comp);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern void InternalCopyTo (int sIndex, char[] dest, int destIndex, int count);
+//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+//		private extern void InternalCopyTo (int sIndex, char[] dest, int destIndex, int count);
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern String[] InternalSplit (char[] separator, int count, int options);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern String InternalTrim (char[] chars, int typ);
+//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+//		private extern String InternalTrim (char[] chars, int typ);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern int InternalLastIndexOfAny (char [] anyOf, int sIndex, int count);
+//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+//		private extern int InternalLastIndexOfAny (char [] anyOf, int sIndex, int count);
 
-		[MethodImplAttribute (MethodImplOptions.InternalCall)]
-		private extern String InternalPad (int width, char chr, bool right);
+//		[MethodImplAttribute (MethodImplOptions.InternalCall)]
+//		private extern String InternalPad (int width, char chr, bool right);
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		internal extern static String InternalAllocateStr (int length);
