@@ -7,6 +7,7 @@
 
 using System;
 using System.Collections;
+using System.Reflection;
 
 using System.IO;
 using System.Runtime.Serialization;
@@ -292,6 +293,19 @@ public class HashtableTest : Assertion {
 			     0, h.Count);
 	}
 
+#if NET_2_0
+	public class MyEqualityComparer : IEqualityComparer {
+		public bool Equals (object x, object y) { return x == y; }
+		public int GetHashCode (object obj) { return 1; }
+	}
+
+	static IEqualityComparer GetEqualityComparer (Hashtable h)
+	{
+		return (IEqualityComparer) typeof (Hashtable).GetField ("equalityComparer",
+			BindingFlags.NonPublic | BindingFlags.Instance).GetValue (h);
+	}
+#endif
+	
         [Test]
 	public void TestClone() {
 		{
@@ -329,6 +343,18 @@ public class HashtableTest : Assertion {
 
 			((char[])h1[c1[0]])[0] = 'z';
 			AssertEquals("shallow copy", h1[c1[0]], h2[c1[0]]);
+
+#if NET_2_0
+			// NET 2.0 stuff
+			MyEqualityComparer a = new MyEqualityComparer ();
+			Hashtable mh1 = new Hashtable (a);
+			Hashtable mh1clone = (Hashtable) mh1.Clone ();
+			
+			// warning, depends on the field name.
+			AssertEquals ("EqualityComparer",
+				GetEqualityComparer (mh1),
+				GetEqualityComparer (mh1clone));
+#endif
 		}
 	}
 
@@ -824,6 +850,46 @@ public class HashtableTest : Assertion {
 		Assert ("GetEnumerator.IsSerializable", ht.GetEnumerator ().GetType ().IsSerializable);
 		Assert ("Synchronized.IsSerializable", Hashtable.Synchronized (ht).GetType ().IsSerializable);
 	}
+
+	[Test]
+	public void TestHashtableWithCustomComparer ()
+	{
+		// see bug #324761
+		IDHashtable dd = new IDHashtable ();
+		Random r = new Random (1000);
+		for (int n = 0; n < 10000; n++) {
+			int v = r.Next (0, 1000);
+			dd [v] = v;
+			v = r.Next (0, 1000);
+			dd.Remove (v);
+		}
+	}
+}
+
+class IDHashtable : Hashtable {
+
+	class IDComparer : IComparer {
+		public int Compare (object x, object y)
+		{
+			if ((int) x == (int) y)
+				return 0;
+			else
+				return 1;
+		}
+	}
+
+	class IDHashCodeProvider : IHashCodeProvider {
+		public int GetHashCode (object o)
+		{
+			return (int) o;
+		}
+	}
+
+	public IDHashtable ()
+		: base (new IDHashCodeProvider (),
+				new IDComparer ())
+	{
+	}
 }
 
 [Serializable]
