@@ -4,8 +4,8 @@
 // Author:
 //   Miguel de Icaza (miguel@ximian.com)
 //
-// (C) 2001, 2002, 2003 Ximian, Inc.
-// (C) 2004 Novell, Inc.
+// Copyright 2001, 2002, 2003 Ximian, Inc.
+// Copyright 2004 Novell, Inc.
 //
 
 //
@@ -15,12 +15,13 @@
 //
 // Only remove it if you need to debug locally on your tree.
 //
-//#define PRODUCTION
+#define PRODUCTION
 
 using System;
 using System.IO;
 using System.Collections;
 using System.Collections.Specialized;
+using System.Globalization;
 using System.Reflection;
 using System.Reflection.Emit;
 using System.Runtime.InteropServices;
@@ -37,7 +38,6 @@ namespace Mono.CSharp {
 	/// </summary>
 	public class CodeGen {
 		static AppDomain current_domain;
-		static public SymbolWriter SymbolWriter;
 
 		public static AssemblyClass Assembly;
 		public static ModuleClass Module;
@@ -83,23 +83,28 @@ namespace Mono.CSharp {
 
 		static public string FileName;
 
+#if MS_COMPATIBLE
+		const AssemblyBuilderAccess COMPILER_ACCESS = 0;
+#else
+		/* Keep this in sync with System.Reflection.Emit.AssemblyBuilder */
+		const AssemblyBuilderAccess COMPILER_ACCESS = (AssemblyBuilderAccess) 0x800;
+#endif
+				
 		//
-		// Initializes the symbol writer
+		// Initializes the code generator variables for interactive use (repl)
 		//
-		static void InitializeSymbolWriter (string filename)
+		static public void InitDynamic (string name)
 		{
-			SymbolWriter = SymbolWriter.GetSymbolWriter (Module.Builder, filename);
-
-			//
-			// If we got an ISymbolWriter instance, initialize it.
-			//
-			if (SymbolWriter == null) {
-				Report.Warning (
-					-18, 1, "Could not find the symbol writer assembly (Mono.CompilerServices.SymbolWriter.dll). This is normally an installation problem. Please make sure to compile and install the mcs/class/Mono.CompilerServices.SymbolWriter directory.");
-				return;
-			}
+			current_domain = AppDomain.CurrentDomain;
+			AssemblyName an = Assembly.GetAssemblyName (name, name);
+			
+			Assembly.Builder = current_domain.DefineDynamicAssembly (an, AssemblyBuilderAccess.Run | COMPILER_ACCESS);
+			Module.Builder = Assembly.Builder.DefineDynamicModule (Basename (name), false);
+#if GMCS_SOURCE
+			Assembly.Name = Assembly.Builder.GetName ();
+#endif
 		}
-
+		
 		//
 		// Initializes the code generator variables
 		//
@@ -130,7 +135,7 @@ namespace Mono.CSharp {
 
 			try {
 				Assembly.Builder = current_domain.DefineDynamicAssembly (an,
-					AssemblyBuilderAccess.Save, Dirname (name));
+					AssemblyBuilderAccess.RunAndSave | COMPILER_ACCESS, Dirname (name));
 			}
 			catch (ArgumentException) {
 				// specified key may not be exportable outside it's container
@@ -139,7 +144,7 @@ namespace Mono.CSharp {
 						RootContext.StrongNameKeyContainer + "'.");
 					Environment.Exit (1);
 				}
-				return false;
+				throw;
 			}
 			catch (CryptographicException) {
 				if ((RootContext.StrongNameKeyContainer != null) || (RootContext.StrongNameKeyFile != null)) {
@@ -163,22 +168,31 @@ namespace Mono.CSharp {
 			// If the third argument is true, the ModuleBuilder will dynamically
 			// load the default symbol writer.
 			//
-			Module.Builder = Assembly.Builder.DefineDynamicModule (
-				Basename (name), Basename (output), false);
+			try {
+				Module.Builder = Assembly.Builder.DefineDynamicModule (
+					Basename (name), Basename (output), want_debugging_support);
 
-			if (want_debugging_support)
-				InitializeSymbolWriter (output);
+#if !MS_COMPATIBLE
+				// TODO: We should use SymbolWriter from DefineDynamicModule
+				if (want_debugging_support && !SymbolWriter.Initialize (Module.Builder, output)) {
+					Report.Error (40, "Unexpected debug information initialization error `{0}'",
+						"Could not find the symbol writer assembly (Mono.CompilerServices.SymbolWriter.dll)");
+					return false;
+				}
+#endif
+			} catch (ExecutionEngineException e) {
+				Report.Error (40, "Unexpected debug information initialization error `{0}'",
+					e.Message);
+				return false;
+			}
 
 			return true;
 		}
 
-		static public void Save (string name)
+		static public void Save (string name, bool saveDebugInfo)
 		{
 			try {
 				Assembly.Builder.Save (Basename (name));
-
-				if (SymbolWriter != null)
-					SymbolWriter.WriteSymbolFile ();
 			}
 			catch (COMException) {
 				if ((RootContext.StrongNameKeyFile == null) || (!RootContext.StrongNameDelaySign))
@@ -191,14 +205,25 @@ namespace Mono.CSharp {
 			}
 			catch (System.IO.IOException io) {
 				Report.Error (16, "Could not write to file `"+name+"', cause: " + io.Message);
+				return;
 			}
 			catch (System.UnauthorizedAccessException ua) {
 				Report.Error (16, "Could not write to file `"+name+"', cause: " + ua.Message);
+				return;
+			}
+
+			//
+			// Write debuger symbol file
+			//
+			if (saveDebugInfo)
+				SymbolWriter.WriteSymbolFile ();
 			}
-		}
 	}
 
 
+	/// <summary>
+	/// An interface to hold all the information needed in the resolving phase.
+	/// </summary>
 	public interface IResolveContext
 	{
 		DeclSpace DeclContainer { get; }
@@ -220,12 +245,18 @@ namespace Mono.CSharp {
 	/// </summary>
 	public class EmitContext : IResolveContext {
 
-		DeclSpace declSpace;
+		//
+		// Holds a varible used during collection or object initialization.
+		//
+		public Expression CurrentInitializerVariable;
+
+		DeclSpace decl_space;
+		
 		public DeclSpace TypeContainer;
-		public ILGenerator   ig;
+		public ILGenerator ig;
 
 		[Flags]
-		public enum Flags : byte {
+		public enum Flags : int {
 			/// <summary>
 			///   This flag tracks the `checked' state of the compilation,
 			///   it controls whether we should generate code that does overflow
@@ -264,7 +295,23 @@ namespace Mono.CSharp {
 			///   Whether control flow analysis is disabled on structs
 			///   (only meaningful when DoFlowAnalysis is set)
 			/// </summary>
-			OmitStructFlowAnalysis = 1 << 6
+			OmitStructFlowAnalysis = 1 << 6,
+
+			///
+			/// Indicates the current context is in probing mode, no errors are reported. 
+			///
+			ProbingMode = 1	<<	7,
+
+			//
+			// Inside field intializer expression
+			//
+			InFieldInitializer = 1 << 8,
+			
+			InferReturnType = 1 << 9,
+			
+			InCompoundAssignment = 1 << 10,
+
+			OmitDebuggingInfo = 1 << 11
 		}
 
 		Flags flags;
@@ -282,26 +329,19 @@ namespace Mono.CSharp {
 		///   So this reflects the low-level staticness of the method, while
 		///   IsStatic represents the semantic, high-level staticness.
 		/// </summary>
-		public bool MethodIsStatic;
-
-		/// <summary>
-		///   Whether we are emitting a field initializer
-		/// </summary>
-		public bool IsFieldInitializer;
-
-		public bool InferReturnType;
+		//public bool MethodIsStatic;
 
 		/// <summary>
 		///   The value that is allowed to be returned or NULL if there is no
 		///   return type.
 		/// </summary>
-		Type returnType;
+		Type return_type;
 
 		/// <summary>
 		///   Points to the Type (extracted from the TypeContainer) that
 		///   declares this body of code
 		/// </summary>
-		public Type ContainerType;
+		public readonly Type ContainerType;
 		
 		/// <summary>
 		///   Whether this is generating code for a constructor
@@ -335,13 +375,6 @@ namespace Mono.CSharp {
 		/// </summary>
 		public bool HasReturnLabel;
 
-		/// <summary>
-		///   Whether we are inside an iterator block.
-		/// </summary>
-		public bool InIterator;
-
-		public bool IsLastStatement;
-
 		/// <summary>
 		///  Whether we are in a `fixed' initialization
 		/// </summary>
@@ -350,7 +383,7 @@ namespace Mono.CSharp {
 		/// <summary>
 		///  Whether we are inside an anonymous method.
 		/// </summary>
-		public AnonymousContainer CurrentAnonymousMethod;
+		public AnonymousExpression CurrentAnonymousMethod;
 		
 		/// <summary>
 		///   Location for this EmitContext
@@ -385,8 +418,6 @@ namespace Mono.CSharp {
 			Emitting
 		}
 
-		public static EmitContext TempEc;
-		
 		bool isAnonymousMethodAllowed = true;
 
 		Phase current_phase;
@@ -408,19 +439,19 @@ namespace Mono.CSharp {
 			this.ig = ig;
 
 			TypeContainer = parent;
-			this.declSpace = ds;
+			this.decl_space = ds;
 			if (RootContext.Checked)
 				flags |= Flags.CheckState;
 			flags |= Flags.ConstantCheckState;
 
+			if (return_type == null)
+				throw new ArgumentNullException ("return_type");
 #if GMCS_SOURCE
 			if ((return_type is TypeBuilder) && return_type.IsGenericTypeDefinition)
 				throw new InternalErrorException ();
 #endif
 
 			IsStatic = (code_flags & Modifiers.STATIC) != 0;
-			MethodIsStatic = IsStatic;
-			InIterator = (code_flags & Modifiers.METHOD_YIELDS) != 0;
 			ReturnType = return_type;
 			IsConstructor = is_constructor;
 			CurrentBlock = null;
@@ -448,11 +479,13 @@ namespace Mono.CSharp {
 		{
 		}
 
+		// IResolveContext.DeclContainer
 		public DeclSpace DeclContainer { 
-			get { return declSpace; }
-			set { declSpace = value; }
+			get { return decl_space; }
+			set { decl_space = value; }
 		}
 
+		// IResolveContext.GenericDeclContainer
 		public DeclSpace GenericDeclContainer {
 			get { return DeclContainer; }
 		}
@@ -490,21 +523,39 @@ namespace Mono.CSharp {
 		public struct FlagsHandle : IDisposable
 		{
 			EmitContext ec;
-			Flags invmask, oldval;
+			readonly Flags invmask, oldval;
+
+			public FlagsHandle (EmitContext ec, Flags flagsToSet)
+				: this (ec, flagsToSet, flagsToSet)
+			{
+			}
+
 			internal FlagsHandle (EmitContext ec, Flags mask, Flags val)
 			{
 				this.ec = ec;
 				invmask = ~mask;
 				oldval = ec.flags & mask;
 				ec.flags = (ec.flags & invmask) | (val & mask);
+
+				if ((mask & Flags.ProbingMode) != 0)
+					Report.DisableReporting ();
 			}
+
 			public void Dispose ()
 			{
+				if ((invmask & Flags.ProbingMode) == 0)
+					Report.EnableReporting ();
+
 				ec.flags = (ec.flags & invmask) | oldval;
 			}
 		}
 
 		// Temporarily set all the given flags to the given value.  Should be used in an 'using' statement
+		public FlagsHandle Set (Flags flagsToSet)
+		{
+			return new FlagsHandle (this, flagsToSet);
+		}
+
 		public FlagsHandle With (Flags bits, bool enable)
 		{
 			return new FlagsHandle (this, bits, enable ? bits : 0);
@@ -517,11 +568,30 @@ namespace Mono.CSharp {
 				(omit_struct_analysis ? Flags.OmitStructFlowAnalysis : 0);
 			return new FlagsHandle (this, Flags.DoFlowAnalysis | Flags.OmitStructFlowAnalysis, newflags);
 		}
-
-		public bool IsInObsoleteScope {
-			get { return ResolveContext.IsInObsoleteScope; }
+		
+		/// <summary>
+		///   If this is true, then Return and ContextualReturn statements
+		///   will set the ReturnType value based on the expression types
+		///   of each return statement instead of the method return type
+		///   (which is initially null).
+		/// </summary>
+		public bool InferReturnType {
+			get { return (flags & Flags.InferReturnType) != 0; }
 		}
 
+		// IResolveContext.IsInObsoleteScope
+		public bool IsInObsoleteScope {
+			get {
+				// Disables obsolete checks when probing is on
+				return IsInProbingMode || ResolveContext.IsInObsoleteScope;
+			}
+		}
+
+		public bool IsInProbingMode {
+			get { return (flags & Flags.ProbingMode) != 0; }
+		}
+
+		// IResolveContext.IsInUnsafeScope
 		public bool IsInUnsafeScope {
 			get { return InUnsafe || ResolveContext.IsInUnsafeScope; }
 		}
@@ -531,10 +601,34 @@ namespace Mono.CSharp {
 			set { isAnonymousMethodAllowed = value; }
 		}
 
+		public bool IsInFieldInitializer {
+			get { return (flags & Flags.InFieldInitializer) != 0; }
+		}
+		
+		public bool IsInCompoundAssignment {
+			get { return (flags & Flags.InCompoundAssignment) != 0; }
+		}
+
+		public bool IsVariableCapturingRequired {
+			get {
+				return !IsInProbingMode && (CurrentBranching == null || !CurrentBranching.CurrentUsageVector.IsUnreachable);
+			}
+		}
+
 		public FlowBranching CurrentBranching {
 			get { return current_flow_branching; }
 		}
 
+		public bool OmitDebuggingInfo {
+			get { return (flags & Flags.OmitDebuggingInfo) != 0; }
+			set {
+				if (value)
+					flags |= Flags.OmitDebuggingInfo;
+				else
+					flags &= ~Flags.OmitDebuggingInfo;
+			}
+		}
+
 		// <summary>
 		//   Starts a new code branching.  This inherits the state of all local
 		//   variables and parameters from the current branching.
@@ -550,21 +644,20 @@ namespace Mono.CSharp {
 		// </summary>
 		public FlowBranching StartFlowBranching (Block block)
 		{
-			FlowBranching.BranchingType type;
-
-			if ((CurrentBranching != null) &&
-			    (CurrentBranching.Type == FlowBranching.BranchingType.Switch))
-				type = FlowBranching.BranchingType.SwitchSection;
-			else
-				type = FlowBranching.BranchingType.Block;
-
 			flags |= Flags.DoFlowAnalysis;
 
 			current_flow_branching = FlowBranching.CreateBranching (
-				CurrentBranching, type, block, block.StartLocation);
+				CurrentBranching, FlowBranching.BranchingType.Block, block, block.StartLocation);
 			return current_flow_branching;
 		}
 
+		public FlowBranchingTryCatch StartFlowBranching (TryCatch stmt)
+		{
+			FlowBranchingTryCatch branching = new FlowBranchingTryCatch (CurrentBranching, stmt);
+			current_flow_branching = branching;
+			return branching;
+		}
+
 		public FlowBranchingException StartFlowBranching (ExceptionStatement stmt)
 		{
 			FlowBranchingException branching = new FlowBranchingException (CurrentBranching, stmt);
@@ -579,6 +672,13 @@ namespace Mono.CSharp {
 			return branching;
 		}
 
+		public FlowBranchingIterator StartFlowBranching (Iterator iterator)
+		{
+			FlowBranchingIterator branching = new FlowBranchingIterator (CurrentBranching, iterator);
+			current_flow_branching = branching;
+			return branching;
+		}
+
 		public FlowBranchingToplevel StartFlowBranching (ToplevelBlock stmt)
 		{
 			FlowBranchingToplevel branching = new FlowBranchingToplevel (CurrentBranching, stmt);
@@ -590,22 +690,12 @@ namespace Mono.CSharp {
 		//   Ends a code branching.  Merges the state of locals and parameters
 		//   from all the children of the ending branching.
 		// </summary>
-		public FlowBranching.UsageVector DoEndFlowBranching ()
+		public bool EndFlowBranching ()
 		{
 			FlowBranching old = current_flow_branching;
 			current_flow_branching = current_flow_branching.Parent;
 
-			return current_flow_branching.MergeChild (old);
-		}
-
-		// <summary>
-		//   Ends a code branching.  Merges the state of locals and parameters
-		//   from all the children of the ending branching.
-		// </summary>
-		public bool EndFlowBranching ()
-		{
-			FlowBranching.UsageVector vector = DoEndFlowBranching ();
-
+			FlowBranching.UsageVector vector = current_flow_branching.MergeChild (old);
 			return vector.IsUnreachable;
 		}
 
@@ -613,6 +703,7 @@ namespace Mono.CSharp {
 		//   Kills the current code branching.  This throws away any changed state
 		//   information and should only be used in case of an error.
 		// </summary>
+		// FIXME: this is evil
 		public void KillFlowBranching ()
 		{
 			current_flow_branching = current_flow_branching.Parent;
@@ -622,8 +713,12 @@ namespace Mono.CSharp {
 		{
 			if (CurrentAnonymousMethod == null)
 				return false;
+
+			// FIXME: IsIterator is too aggressive, we should capture only if child
+			// block contains yield
 			if (CurrentAnonymousMethod.IsIterator)
 				return true;
+
 			return local.Block.Toplevel != CurrentBlock.Toplevel;
 		}
 		
@@ -648,24 +743,41 @@ namespace Mono.CSharp {
 			bool unreachable;
 			
 			if (ResolveTopBlock (null, block, md.ParameterInfo, md, out unreachable)){
+				if (Report.Errors > 0)
+					return;
+
 				EmitMeta (block);
 
 				current_phase = Phase.Emitting;
+#if PRODUCTION
+				try {
+#endif
 				EmitResolvedTopBlock (block, unreachable);
+#if PRODUCTION
+				} catch (Exception e){
+					Console.WriteLine ("Exception caught by the compiler while emitting:");
+					Console.WriteLine ("   Block that caused the problem begin at: " + block.loc);
+					
+					Console.WriteLine (e.GetType ().FullName + ": " + e.Message);
+					throw;
+				}
+#endif
 			}
 		}
 
 		bool resolved;
+		bool unreachable;
 
 		public bool ResolveTopBlock (EmitContext anonymous_method_host, ToplevelBlock block,
 					     Parameters ip, IMethodData md, out bool unreachable)
 		{
-			current_phase = Phase.Resolving;
-			
-			unreachable = false;
-
-			if (resolved)
+			if (resolved) {
+				unreachable = this.unreachable;
 				return true;
+			}
+
+			current_phase = Phase.Resolving;
+			unreachable = false;
 
 			if (!loc.IsNull)
 				CurrentFile = loc.File;
@@ -676,11 +788,6 @@ namespace Mono.CSharp {
 				if (!block.ResolveMeta (this, ip))
 					return false;
 
-				if ((md != null) && (md.Iterator != null)) {
-					if (!md.Iterator.Resolve (this))
-						return false;
-				}
-
 				using (this.With (EmitContext.Flags.DoFlowAnalysis, true)) {
 					FlowBranchingToplevel top_level;
 					if (anonymous_method_host != null)
@@ -697,7 +804,7 @@ namespace Mono.CSharp {
 
 					bool flow_unreachable = top_level.End ();
 					if (flow_unreachable)
-						unreachable = true;
+						this.unreachable = unreachable = true;
 				}
 #if PRODUCTION
 			} catch (Exception e) {
@@ -713,7 +820,7 @@ namespace Mono.CSharp {
 			}
 #endif
 
-			if (returnType != null && !unreachable) {
+			if (return_type != TypeManager.void_type && !unreachable) {
 				if (CurrentAnonymousMethod == null) {
 					Report.Error (161, md.Location, "`{0}': not all code paths return a value", md.GetSignatureForError ());
 					return false;
@@ -724,20 +831,16 @@ namespace Mono.CSharp {
 				}
 			}
 
-			if (!block.CompleteContexts (this))
-				return false;
-
 			resolved = true;
 			return true;
 		}
 
 		public Type ReturnType {
 			set {
-				returnType = value == TypeManager.void_type ?
-					null : value;
+				return_type = value;
 			}
 			get {
-				return returnType;
+				return return_type;
 			}
 		}
 
@@ -766,13 +869,8 @@ namespace Mono.CSharp {
 				// this case.
 				//
 
-				bool in_iterator = (CurrentAnonymousMethod != null) &&
-					CurrentAnonymousMethod.IsIterator && InIterator;
-
-				if ((block != null) && block.IsDestructor) {
-					// Nothing to do; S.R.E automatically emits a leave.
-				} else if (HasReturnLabel || (!unreachable && !in_iterator)) {
-					if (returnType != null)
+				if (HasReturnLabel || !unreachable) {
+					if (return_type != TypeManager.void_type)
 						ig.Emit (OpCodes.Ldloc, TemporaryReturn ());
 					ig.Emit (OpCodes.Ret);
 				}
@@ -783,39 +881,29 @@ namespace Mono.CSharp {
 		///   This is called immediately before emitting an IL opcode to tell the symbol
 		///   writer to which source line this opcode belongs.
 		/// </summary>
-		public void Mark (Location loc, bool check_file)
+		public void Mark (Location loc)
 		{
-			if ((CodeGen.SymbolWriter == null) || loc.IsNull)
+			if (!SymbolWriter.HasSymbolWriter || OmitDebuggingInfo || loc.IsNull)
 				return;
 
-			if (check_file && (CurrentFile != loc.File))
-				return;
-
-			CodeGen.SymbolWriter.MarkSequencePoint (ig, loc.Row, loc.Column);
+			SymbolWriter.MarkSequencePoint (ig, loc);
 		}
 
 		public void DefineLocalVariable (string name, LocalBuilder builder)
 		{
-			if (CodeGen.SymbolWriter == null)
-				return;
-
-			CodeGen.SymbolWriter.DefineLocalVariable (name, builder);
+			SymbolWriter.DefineLocalVariable (name, builder);
 		}
 
 		public void BeginScope ()
 		{
 			ig.BeginScope();
-
-			if (CodeGen.SymbolWriter != null)
-				CodeGen.SymbolWriter.OpenScope(ig);
+			SymbolWriter.OpenScope(ig);
 		}
 
 		public void EndScope ()
 		{
 			ig.EndScope();
-
-			if (CodeGen.SymbolWriter != null)
-				CodeGen.SymbolWriter.CloseScope(ig);
+			SymbolWriter.CloseScope(ig);
 		}
 
 		/// <summary>
@@ -842,8 +930,6 @@ namespace Mono.CSharp {
 
 		public void FreeTemporaryLocal (LocalBuilder b, Type t)
 		{
-			Stack s;
-
 			if (temporary_storage == null) {
 				temporary_storage = new Hashtable ();
 				temporary_storage [t] = b;
@@ -854,9 +940,8 @@ namespace Mono.CSharp {
 				temporary_storage [t] = b;
 				return;
 			}
-			if (o is Stack) {
-				s = (Stack) o;
-			} else {
+			Stack s = o as Stack;
+			if (s == null) {
 				s = new Stack ();
 				s.Push (o);
 				temporary_storage [t] = s;
@@ -895,7 +980,7 @@ namespace Mono.CSharp {
 		public LocalBuilder TemporaryReturn ()
 		{
 			if (return_value == null){
-				return_value = ig.DeclareLocal (returnType);
+				return_value = ig.DeclareLocal (return_type);
 				if (!HasReturnLabel){
 					ReturnLabel = ig.DefineLabel ();
 					HasReturnLabel = true;
@@ -919,7 +1004,7 @@ namespace Mono.CSharp {
 				throw new Exception ("NeedReturnLabel called from Emit phase, should only be called during Resolve");
 			}
 			
-			if (!InIterator && !HasReturnLabel) 
+			if (!HasReturnLabel)
 				HasReturnLabel = true;
 		}
 
@@ -1006,12 +1091,13 @@ namespace Mono.CSharp {
 		public AssemblyBuilder Builder;
 		bool is_cls_compliant;
 		bool wrap_non_exception_throws;
-		bool has_extension_method;
+		Type runtime_compatibility_attr_type;
 
 		public Attribute ClsCompliantAttribute;
 
 		ListDictionary declarative_security;
 #if GMCS_SOURCE
+		bool has_extension_method;		
 		public AssemblyName Name;
 		MethodInfo add_type_forwarder;
 		ListDictionary emitted_forwarders;
@@ -1028,7 +1114,11 @@ namespace Mono.CSharp {
 		}
 
 		public bool HasExtensionMethods {
-			set { has_extension_method = value; }
+			set {
+#if GMCS_SOURCE				
+				has_extension_method = value;
+#endif
+			}
 		}
 
 		public bool IsClsCompliant {
@@ -1056,6 +1146,36 @@ namespace Mono.CSharp {
 
 		public void Resolve ()
 		{
+			runtime_compatibility_attr_type = TypeManager.CoreLookupType (
+				"System.Runtime.CompilerServices", "RuntimeCompatibilityAttribute", Kind.Class, false);
+
+			if (RootContext.Unsafe) {
+				//
+				// Emits [assembly: SecurityPermissionAttribute (SecurityAction.RequestMinimum, SkipVerification = true)]
+				// when -unsafe option was specified
+				//
+				
+				Location loc = Location.Null;
+
+				MemberAccess system_security_permissions = new MemberAccess (new MemberAccess (
+					new QualifiedAliasMember (QualifiedAliasMember.GlobalAlias, "System", loc), "Security", loc), "Permissions", loc);
+
+				ArrayList pos = new ArrayList (1);
+				pos.Add (new Argument (new MemberAccess (new MemberAccess (system_security_permissions, "SecurityAction", loc), "RequestMinimum")));
+
+				ArrayList named = new ArrayList (1);
+				named.Add (new DictionaryEntry ("SkipVerification", new Argument (new BoolLiteral (true, loc))));
+
+				GlobalAttribute g = new GlobalAttribute (new NamespaceEntry (null, null, null), "assembly", system_security_permissions,
+					"SecurityPermissionAttribute", new object[] { pos, named }, loc, false);
+				g.AttachTo (this);
+
+				if (g.Resolve () != null) {
+					declarative_security = new ListDictionary ();
+					g.ExtractSecurityPermissionSet (declarative_security);
+				}
+			}
+
 			if (OptAttributes == null)
 				return;
 
@@ -1063,19 +1183,21 @@ namespace Mono.CSharp {
 			if (!OptAttributes.CheckTargets())
 				return;
 
-			ClsCompliantAttribute = ResolveAttribute (TypeManager.cls_compliant_attribute_type);
+			if (TypeManager.cls_compliant_attribute_type != null)
+				ClsCompliantAttribute = ResolveAttribute (TypeManager.cls_compliant_attribute_type);
+
 			if (ClsCompliantAttribute != null) {
 				is_cls_compliant = ClsCompliantAttribute.GetClsCompliantAttributeValue ();
 			}
 
-#if GMCS_SOURCE
-			Attribute a = ResolveAttribute (TypeManager.runtime_compatibility_attr_type);
-			if (a != null) {
-				object val = a.GetPropertyValue ("WrapNonExceptionThrows");
-				if (val != null)
-					wrap_non_exception_throws = (bool)val;
+			if (runtime_compatibility_attr_type != null) {
+				Attribute a = ResolveAttribute (runtime_compatibility_attr_type);
+				if (a != null) {
+					object val = a.GetPropertyValue ("WrapNonExceptionThrows");
+					if (val != null)
+						wrap_non_exception_throws = (bool) val;
+				}
 			}
-#endif
 		}
 
 		// fix bug #56621
@@ -1122,43 +1244,41 @@ namespace Mono.CSharp {
 					//       are loaded yet.
 					// TODO: Does not handle quoted attributes properly
 					switch (a.Name) {
-						case "AssemblyKeyFile":
-						case "AssemblyKeyFileAttribute":
-						case "System.Reflection.AssemblyKeyFileAttribute":
-							if (RootContext.StrongNameKeyFile != null) {
-								Report.SymbolRelatedToPreviousError (a.Location, a.Name);
-								Report.Warning (1616, 1, "Option `{0}' overrides attribute `{1}' given in a source file or added module",
-                                    "keyfile", "System.Reflection.AssemblyKeyFileAttribute");
-							}
-							else {
-								string value = a.GetString ();
-								if (value.Length != 0)
-									RootContext.StrongNameKeyFile = value;
-							}
-							break;
-						case "AssemblyKeyName":
-						case "AssemblyKeyNameAttribute":
-						case "System.Reflection.AssemblyKeyNameAttribute":
-							if (RootContext.StrongNameKeyContainer != null) {
-								Report.SymbolRelatedToPreviousError (a.Location, a.Name);
-								Report.Warning (1616, 1, "Option `{0}' overrides attribute `{1}' given in a source file or added module",
+					case "AssemblyKeyFile":
+					case "AssemblyKeyFileAttribute":
+					case "System.Reflection.AssemblyKeyFileAttribute":
+						if (RootContext.StrongNameKeyFile != null) {
+							Report.SymbolRelatedToPreviousError (a.Location, a.Name);
+							Report.Warning (1616, 1, "Option `{0}' overrides attribute `{1}' given in a source file or added module",
+									"keyfile", "System.Reflection.AssemblyKeyFileAttribute");
+						} else {
+							string value = a.GetString ();
+							if (value.Length != 0)
+								RootContext.StrongNameKeyFile = value;
+						}
+						break;
+					case "AssemblyKeyName":
+					case "AssemblyKeyNameAttribute":
+					case "System.Reflection.AssemblyKeyNameAttribute":
+						if (RootContext.StrongNameKeyContainer != null) {
+							Report.SymbolRelatedToPreviousError (a.Location, a.Name);
+							Report.Warning (1616, 1, "Option `{0}' overrides attribute `{1}' given in a source file or added module",
 									"keycontainer", "System.Reflection.AssemblyKeyNameAttribute");
-							}
-							else {
-								string value = a.GetString ();
-								if (value.Length != 0)
-									RootContext.StrongNameKeyContainer = value;
-							}
-							break;
-						case "AssemblyDelaySign":
-						case "AssemblyDelaySignAttribute":
-						case "System.Reflection.AssemblyDelaySignAttribute":
-							RootContext.StrongNameDelaySign = a.GetBoolean ();
-							break;
+						} else {
+							string value = a.GetString ();
+							if (value.Length != 0)
+								RootContext.StrongNameKeyContainer = value;
+						}
+						break;
+					case "AssemblyDelaySign":
+					case "AssemblyDelaySignAttribute":
+					case "System.Reflection.AssemblyDelaySignAttribute":
+						RootContext.StrongNameDelaySign = a.GetBoolean ();
+						break;
 					}
 				}
 			}
-
+			
 			AssemblyName an = new AssemblyName ();
 			an.Name = Path.GetFileNameWithoutExtension (name);
 
@@ -1249,7 +1369,7 @@ namespace Mono.CSharp {
 			else if (aname.Version != null || aname.CultureInfo != null)
 				throw new Exception ("Friend assembly `" + a.GetString () + 
 						"' is invalid. InternalsVisibleTo cannot have version or culture specified.");
-			else if (aname.GetPublicKey () == null && Name.GetPublicKey () != null) {
+			else if (aname.GetPublicKey () == null && Name.GetPublicKey () != null && Name.GetPublicKey ().Length != 0) {
 				Report.Error (1726, a.Location, "Friend assembly reference `" + aname.FullName + "' is invalid." +
 						" Strong named assemblies must specify a public key in their InternalsVisibleTo declarations");
 				return false;
@@ -1259,9 +1379,31 @@ namespace Mono.CSharp {
 		}
 #endif
 
+		static bool IsValidAssemblyVersion (string version)
+		{
+			Version v;
+			try {
+				v = new Version (version);
+			} catch {
+				try {
+					int major = int.Parse (version, CultureInfo.InvariantCulture);
+					v = new Version (major, 0);
+				} catch {
+					return false;
+				}
+			}
+
+			foreach (int candidate in new int [] { v.Major, v.Minor, v.Build, v.Revision }) {
+				if (candidate > ushort.MaxValue)
+					return false;
+			}
+
+			return true;
+		}
+
 		public override void ApplyAttributeBuilder (Attribute a, CustomAttributeBuilder customBuilder)
 		{
-			if (a.Type.IsSubclassOf (TypeManager.security_attr_type) && a.CheckSecurityActionValidity (true)) {
+			if (a.IsValidSecurityAttribute ()) {
 				if (declarative_security == null)
 					declarative_security = new ListDictionary ();
 
@@ -1280,6 +1422,19 @@ namespace Mono.CSharp {
 				}
 			}
 
+			if (a.Type == TypeManager.assembly_version_attribute_type) {
+				string value = a.GetString ();
+				if (value == null || value.Length == 0)
+					return;
+
+				value = value.Replace ('*', '0');
+
+				if (!IsValidAssemblyVersion (value)) {
+					a.Error_AttributeEmitError (string.Format ("Specified version `{0}' is not valid", value));
+					return;
+				}
+			}
+
 #if GMCS_SOURCE
 			if (a.Type == TypeManager.internals_visible_attr_type && !CheckInternalsVisibleAttribute (a))
 				return;
@@ -1349,20 +1504,22 @@ namespace Mono.CSharp {
 #if GMCS_SOURCE
 			if (has_extension_method)
 				Builder.SetCustomAttribute (TypeManager.extension_attribute_attr);
-
-			// FIXME: Does this belong inside SRE.AssemblyBuilder instead?
-			if (OptAttributes == null || !OptAttributes.Contains (TypeManager.runtime_compatibility_attr_type)) {
-				ConstructorInfo ci = TypeManager.GetConstructor (
-					TypeManager.runtime_compatibility_attr_type, Type.EmptyTypes);
-				PropertyInfo [] pis = new PropertyInfo [1];
-				pis [0] = TypeManager.GetProperty (
-					TypeManager.runtime_compatibility_attr_type, "WrapNonExceptionThrows");
-				object [] pargs = new object [1];
-				pargs [0] = true;
-				Builder.SetCustomAttribute (new CustomAttributeBuilder (ci, new object [0], pis, pargs));
-			}
 #endif
 
+			if (runtime_compatibility_attr_type != null) {
+				// FIXME: Does this belong inside SRE.AssemblyBuilder instead?
+				if (OptAttributes == null || !OptAttributes.Contains (runtime_compatibility_attr_type)) {
+					ConstructorInfo ci = TypeManager.GetPredefinedConstructor (
+						runtime_compatibility_attr_type, Location.Null, Type.EmptyTypes);
+					PropertyInfo [] pis = new PropertyInfo [1];
+					pis [0] = TypeManager.GetPredefinedProperty (runtime_compatibility_attr_type,
+						"WrapNonExceptionThrows", Location.Null, TypeManager.bool_type);
+					object [] pargs = new object [1];
+					pargs [0] = true;
+					Builder.SetCustomAttribute (new CustomAttributeBuilder (ci, new object [0], pis, pargs));
+				}
+			}
+
 			if (declarative_security != null) {
 
 				MethodInfo add_permission = typeof (AssemblyBuilder).GetMethod ("AddPermissionRequests", BindingFlags.Instance | BindingFlags.NonPublic);
@@ -1424,7 +1581,9 @@ namespace Mono.CSharp {
 		// TODO: make it private and move all builder based methods here
 		public ModuleBuilder Builder;
 		bool m_module_is_unsafe;
+#if GMCS_SOURCE		
 		bool has_default_charset;
+#endif		
 
 		public CharSet DefaultCharSet = CharSet.Ansi;
 		public TypeAttributes DefaultCharSetType = TypeAttributes.AnsiClass;
@@ -1451,15 +1610,14 @@ namespace Mono.CSharp {
 		{
 			base.Emit (tc);
 
-			if (!m_module_is_unsafe)
-				return;
-
-			if (TypeManager.unverifiable_code_ctor == null) {
-				Console.WriteLine ("Internal error ! Cannot set unverifiable code attribute.");
-				return;
+			if (m_module_is_unsafe) {
+				Type t = TypeManager.CoreLookupType ("System.Security", "UnverifiableCodeAttribute", Kind.Class, true);
+				if (t != null) {
+					ConstructorInfo unverifiable_code_ctor = TypeManager.GetPredefinedConstructor (t, Location.Null, Type.EmptyTypes);
+					if (unverifiable_code_ctor != null)
+						Builder.SetCustomAttribute (new CustomAttributeBuilder (unverifiable_code_ctor, new object [0]));
+				}
 			}
-				
-			Builder.SetCustomAttribute (new CustomAttributeBuilder (TypeManager.unverifiable_code_ctor, new object [0]));
 		}
                 
 		public override void ApplyAttributeBuilder (Attribute a, CustomAttributeBuilder customBuilder)
@@ -1470,7 +1628,7 @@ namespace Mono.CSharp {
 				}
 				else if (CodeGen.Assembly.IsClsCompliant != a.GetBoolean ()) {
 					Report.SymbolRelatedToPreviousError (CodeGen.Assembly.ClsCompliantAttribute.Location, CodeGen.Assembly.ClsCompliantAttribute.GetSignatureForError ());
-					Report.Error (3017, a.Location, "You cannot specify the CLSCompliant attribute on a module that differs from the CLSCompliant attribute on the assembly");
+					Report.Warning (3017, 1, a.Location, "You cannot specify the CLSCompliant attribute on a module that differs from the CLSCompliant attribute on the assembly");
 					return;
 				}
 			}
@@ -1480,7 +1638,11 @@ namespace Mono.CSharp {
 
 		public bool HasDefaultCharSet {
 			get {
+#if GMCS_SOURCE		
 				return has_default_charset;
+#else
+				return false;
+#endif								
 			}
 		}
 
@@ -1496,23 +1658,26 @@ namespace Mono.CSharp {
 			if (!OptAttributes.CheckTargets())
 				return;
 
+			if (TypeManager.default_charset_type == null)
+				return;
+
 			Attribute a = ResolveAttribute (TypeManager.default_charset_type);
 			if (a != null) {
 				has_default_charset = true;
 				DefaultCharSet = a.GetCharSetValue ();
 				switch (DefaultCharSet) {
-					case CharSet.Ansi:
-					case CharSet.None:
-						break;
-					case CharSet.Auto:
-						DefaultCharSetType = TypeAttributes.AutoClass;
-						break;
-					case CharSet.Unicode:
-						DefaultCharSetType = TypeAttributes.UnicodeClass;
-						break;
-					default:
-						Report.Error (1724, a.Location, "Value specified for the argument to 'System.Runtime.InteropServices.DefaultCharSetAttribute' is not valid");
-						break;
+				case CharSet.Ansi:
+				case CharSet.None:
+					break;
+				case CharSet.Auto:
+					DefaultCharSetType = TypeAttributes.AutoClass;
+					break;
+				case CharSet.Unicode:
+					DefaultCharSetType = TypeAttributes.UnicodeClass;
+					break;
+				default:
+					Report.Error (1724, a.Location, "Value specified for the argument to 'System.Runtime.InteropServices.DefaultCharSetAttribute' is not valid");
+					break;
 				}
 			}
 #endif
