@@ -538,6 +538,7 @@ static dis_map_t method_flags_map [] = {
 /*	{ METHOD_ATTRIBUTE_HAS_SECURITY,        "hassecurity" }, */
 	{ METHOD_ATTRIBUTE_REQUIRE_SEC_OBJECT,  "requiresecobj" },
 	{ METHOD_ATTRIBUTE_PINVOKE_IMPL,        "pinvokeimpl " }, 
+	{ METHOD_ATTRIBUTE_STRICT,	            "strict " }, 
 	{ 0, NULL }
 };
 
@@ -717,24 +718,25 @@ dis_code (MonoImage *m, guint32 token, guint32 rva, MonoGenericContainer *contai
 	}
 
 	mh = mono_metadata_parse_mh_full (m, container, ptr);
-	if ((entry_point = mono_image_get_entry_point (m))){
+	if ((entry_point = mono_image_get_entry_point (m)) && mono_metadata_token_index (entry_point)){
 		loc = mono_metadata_locate_token (m, entry_point);
 		if (rva == read32 (loc))
 			fprintf (output, "\t.entrypoint\n");
 	}
 	
-	fprintf (output, "\t// Code size %d (0x%x)\n", mh->code_size, mh->code_size);
-	fprintf (output, "\t.maxstack %d\n", mh->max_stack);
-	if (mh->num_locals)
-		dis_locals (m, mh, ptr);
-	disassemble_cil (m, mh, container);
-	
+	if (mh) {
+		fprintf (output, "\t// Code size %d (0x%x)\n", mh->code_size, mh->code_size);
+		fprintf (output, "\t.maxstack %d\n", mh->max_stack);
+		if (mh->num_locals)
+			dis_locals (m, mh, ptr);
+		disassemble_cil (m, mh, container);
 /*
   hex_dump (mh->code, 0, mh->code_size);
   printf ("\nAfter the code\n");
   hex_dump (mh->code + mh->code_size, 0, 64);
 */
-	mono_metadata_free_mh (mh);
+		mono_metadata_free_mh (mh);
+	}
 }
 
 static char *
@@ -1016,14 +1018,12 @@ dis_property_signature (MonoImage *m, guint32 prop_idx, MonoGenericContainer *co
 	g_string_sprintfa (res, "%s %s (", blurb, qk);
 	g_free (qk);
 	g_free (blurb);
-	mono_metadata_free_type (type);
 	for (i = 0; i < pcount; i++) {
 		if (i)
 			g_string_append (res, ", ");
 		param = mono_metadata_parse_type_full (m, container, MONO_PARSE_PARAM, 0, ptr, &ptr);
 		blurb = dis_stringify_param (m, param);
 		g_string_append (res, blurb);
-		mono_metadata_free_type (param);
 		g_free (blurb);
 	}
 	g_string_append_c (res, ')');
@@ -1388,6 +1388,24 @@ dis_types (MonoImage *m, int forward)
 	}
 }
 
+static const char *
+get_uninitialized_data_type (guint32 size)
+{
+	switch (size) {
+	case 1:
+		return "int8";
+	case 2:
+		return "int16";
+	case 4:
+		return "int32";
+	case 8:
+		return "int64";
+	default:
+		g_error ("get_uninitialized_data_type for size: %d\n", size);
+	}
+	return NULL;
+}
+
 /**
  * dis_data:
  * @m: metadata context
@@ -1416,13 +1434,18 @@ dis_data (MonoImage *m)
 		type = mono_metadata_parse_field_type (m, 0, sig + 1, &sig);
 		mono_class_init (mono_class_from_mono_type (type));
 		size = mono_type_size (type, &align);
-		fprintf (output, ".data D_%08x = bytearray (", cols [MONO_FIELD_RVA_RVA]);
-		for (b = 0; b < size; ++b) {
-			if (!(b % 16))
-				fprintf (output, "\n\t");
-			fprintf (output, " %02X", rva [b] & 0xff);
-		}
-		fprintf (output, ") // size: %d\n", size);
+
+		if (rva) {
+			fprintf (output, ".data D_%08x = bytearray (", cols [MONO_FIELD_RVA_RVA]);
+			for (b = 0; b < size; ++b) {
+				if (!(b % 16))
+					fprintf (output, "\n\t");
+				fprintf (output, " %02X", rva [b] & 0xff);
+			}
+			fprintf (output, ") // size: %d\n", size);
+		} else
+			fprintf (output, ".data D_%08x = %s [%d]\n",
+				cols [MONO_FIELD_RVA_RVA], get_uninitialized_data_type (size), size);
 	}
 }
 
@@ -1483,21 +1506,17 @@ struct {
 static void
 disassemble_file (const char *file)
 {
-	MonoAssembly *ass;
 	MonoImageOpenStatus status;
 	MonoImage *img;
+	MonoAssembly *assembly;
 
-	ass = mono_assembly_open (file, &status);
-	if (ass == NULL){
-		img = mono_image_open (file, &status);
-		if (!img) {
-			fprintf (stderr, "Error while trying to process %s\n", file);
-			return;
-		} else {
-			mono_assembly_load_references (img, &status);
-		}
+
+	img = mono_image_open (file, &status);
+	if (!img) {
+		fprintf (stderr, "Error while trying to process %s\n", file);
+		return;
 	} else {
-		img = ass->image;
+		assembly = mono_assembly_load_from_full (img, file, &status, FALSE);
 	}
 
 	setup_filter (img);
