@@ -12,13 +12,21 @@
 #include <config.h>
 #include <glib.h>
 #include <string.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#ifdef HAVE_UTIME_H
+#include <utime.h>
+#else
+#ifdef HAVE_SYS_UTIME_H
+#include <sys/utime.h>
+#endif
+#endif
 
-#include <mono/os/gc_wrapper.h>
-
+#include <mono/metadata/gc-internal.h>
 #include <mono/metadata/object.h>
 #include <mono/metadata/domain-internals.h>
 #include "mono/metadata/metadata-internals.h"
@@ -33,23 +41,53 @@
 #include <mono/metadata/monitor.h>
 #include <mono/metadata/threadpool.h>
 #include <mono/metadata/mono-debug.h>
+#include <mono/metadata/attach.h>
+#include <mono/metadata/file-io.h>
 #include <mono/utils/mono-uri.h>
 #include <mono/utils/mono-logger.h>
 #include <mono/utils/mono-path.h>
 #include <mono/utils/mono-stdlib.h>
+#include <mono/utils/mono-io-portability.h>
 #ifdef PLATFORM_WIN32
 #include <direct.h>
 #endif
 
-#define MONO_CORLIB_VERSION 56
+/*
+ * This is the version number of the corlib-runtime interface. When
+ * making changes to this interface (by changing the layout
+ * of classes the runtime knows about, changing icall signature or
+ * semantics etc), increment this variable. Also increment the
+ * pair of this variable in mscorlib in:
+ *       mcs/class/mscorlib/System/Environment.cs
+ *
+ * Changes which are already detected at runtime, like the addition
+ * of icalls, do not require an increment.
+ */
+#define MONO_CORLIB_VERSION 69
+
+typedef struct
+{
+	int runtime_count;
+	int assemblybinding_count;
+	MonoDomain *domain;
+} RuntimeConfig;
 
 CRITICAL_SECTION mono_delegate_section;
 
+#ifdef _EGLIB_MAJOR
+/* Need to lock here because EGLIB has locking defined as no-ops, we can not depend on mono_strtod do the right locking */
+/* Ideally this will be fixed in eglib */
+CRITICAL_SECTION mono_strtod_mutex;
+#endif
+
+
 static gunichar2 process_guid [36];
 static gboolean process_guid_set = FALSE;
 
 static gboolean shutting_down = FALSE;
 
+static gboolean no_exec = FALSE;
+
 static MonoAssembly *
 mono_domain_assembly_preload (MonoAssemblyName *aname,
 			      gchar **assemblies_path,
@@ -69,8 +107,43 @@ mono_domain_fire_assembly_load (MonoAssembly *assembly, gpointer user_data);
 static void
 add_assemblies_to_domain (MonoDomain *domain, MonoAssembly *ass, GHashTable *hash);
 
-static void
-mono_domain_unload (MonoDomain *domain);
+static MonoAppDomain *
+mono_domain_create_appdomain_internal (char *friendly_name, MonoAppDomainSetup *setup);
+
+static MonoLoadFunc load_function = NULL;
+
+void
+mono_install_runtime_load (MonoLoadFunc func)
+{
+	load_function = func;
+}
+
+MonoDomain*
+mono_runtime_load (const char *filename, const char *runtime_version)
+{
+	g_assert (load_function);
+	return load_function (filename, runtime_version);
+}
+
+/*
+ * mono_runtime_set_no_exec:
+ *
+ *   Instructs the runtime to operate in static mode, i.e. avoid/do not allow managed
+ * code execution. This is useful for running the AOT compiler on platforms which
+ * allow full-aot execution only.
+ * This should be called before mono_runtime_init ().
+ */
+void
+mono_runtime_set_no_exec (gboolean val)
+{
+	no_exec = val;
+}
+
+gboolean
+mono_runtime_get_no_exec (void)
+{
+	return no_exec;
+}
 
 /**
  * mono_runtime_init:
@@ -92,8 +165,10 @@ mono_runtime_init (MonoDomain *domain, MonoThreadStartCB start_cb,
 	MonoAppDomain *ad;
 	MonoClass *class;
 	MonoString *arg;
+
+	mono_portability_helpers_init ();
 	
-	MONO_GC_PRE_INIT ();
+	mono_gc_base_init ();
 	mono_monitor_init ();
 	mono_thread_pool_init ();
 	mono_marshal_init ();
@@ -120,36 +195,44 @@ mono_runtime_init (MonoDomain *domain, MonoThreadStartCB start_cb,
 
 	InitializeCriticalSection (&mono_delegate_section);
 
+#ifdef _EGLIB_MAJOR
+	/* Needed until EGLIB is fixed #464316 */
+	InitializeCriticalSection (&mono_strtod_mutex);
+#endif
+	
 	mono_thread_attach (domain);
 	mono_context_init (domain);
 	mono_context_set (domain->default_context);
 
 	mono_type_initialization_init ();
 
+	if (!mono_runtime_get_no_exec ()) {
+		/*
+		 * Create an instance early since we can't do it when there is no memory.
+		 */
+		arg = mono_string_new (domain, "Out of memory");
+		domain->out_of_memory_ex = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "OutOfMemoryException", arg, NULL);
 	
-	/*
-	 * Create an instance early since we can't do it when there is no memory.
-	 */
-	arg = mono_string_new (domain, "Out of memory");
-	domain->out_of_memory_ex = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "OutOfMemoryException", arg, NULL);
-	
-	/* 
-	 * These two are needed because the signal handlers might be executing on
-	 * an alternate stack, and Boehm GC can't handle that.
-	 */
-	arg = mono_string_new (domain, "A null value was found where an object instance was required");
-	domain->null_reference_ex = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "NullReferenceException", arg, NULL);
-	arg = mono_string_new (domain, "The requested operation caused a stack overflow.");
-	domain->stack_overflow_ex = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "StackOverflowException", arg, NULL);
-	
+		/* 
+		 * These two are needed because the signal handlers might be executing on
+		 * an alternate stack, and Boehm GC can't handle that.
+		 */
+		arg = mono_string_new (domain, "A null value was found where an object instance was required");
+		domain->null_reference_ex = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "NullReferenceException", arg, NULL);
+		arg = mono_string_new (domain, "The requested operation caused a stack overflow.");
+		domain->stack_overflow_ex = mono_exception_from_name_two_strings (mono_defaults.corlib, "System", "StackOverflowException", arg, NULL);
+	}
+
 	/* GC init has to happen after thread init */
 	mono_gc_init ();
 
 	mono_network_init ();
 
+	mono_attach_init ();
+
 	/* mscorlib is loaded before we install the load hook */
 	mono_domain_fire_assembly_load (mono_defaults.corlib->assembly, NULL);
-
+	
 	return;
 }
 
@@ -208,6 +291,8 @@ mono_runtime_cleanup (MonoDomain *domain)
 {
 	shutting_down = TRUE;
 
+	mono_attach_cleanup ();
+
 	/* This ends up calling any pending pending (for at most 2 seconds) */
 	mono_gc_cleanup ();
 
@@ -220,6 +305,10 @@ mono_runtime_cleanup (MonoDomain *domain)
 	mono_type_initialization_cleanup ();
 
 	mono_monitor_cleanup ();
+
+#ifndef PLATFORM_WIN32
+	_wapi_cleanup ();
+#endif
 }
 
 static MonoDomainFunc quit_function = NULL;
@@ -264,6 +353,68 @@ mono_runtime_is_shutting_down (void)
 	return shutting_down;
 }
 
+/**
+ * mono_domain_create_appdomain:
+ * @friendly_name: The friendly name of the appdomain to create
+ * @configuration_file: The configuration file to initialize the appdomain with
+ * 
+ * Returns a MonoDomain initialized with the appdomain
+ */
+MonoDomain *
+mono_domain_create_appdomain (char *friendly_name, char *configuration_file)
+{
+	MonoAppDomain *ad;
+	MonoAppDomainSetup *setup;
+	MonoClass *class;
+
+	class = mono_class_from_name (mono_defaults.corlib, "System", "AppDomainSetup");
+	setup = (MonoAppDomainSetup *) mono_object_new (mono_domain_get (), class);
+	setup->configuration_file = configuration_file != NULL ? mono_string_new (mono_domain_get (), configuration_file) : NULL;
+
+	ad = mono_domain_create_appdomain_internal (friendly_name, setup);
+
+	return mono_domain_from_appdomain (ad);
+}
+
+static MonoAppDomain *
+mono_domain_create_appdomain_internal (char *friendly_name, MonoAppDomainSetup *setup)
+{
+	MonoClass *adclass;
+	MonoAppDomain *ad;
+	MonoDomain *data;
+	
+	MONO_ARCH_SAVE_REGS;
+
+	adclass = mono_class_from_name (mono_defaults.corlib, "System", "AppDomain");
+
+	/* FIXME: pin all those objects */
+	data = mono_domain_create();
+
+	ad = (MonoAppDomain *) mono_object_new (data, adclass);
+	ad->data = data;
+	data->domain = ad;
+	data->setup = setup;
+	data->friendly_name = g_strdup (friendly_name);
+	// FIXME: The ctor runs in the current domain
+	// FIXME: Initialize null_reference_ex and stack_overflow_ex
+	data->out_of_memory_ex = mono_exception_from_name_domain (data, mono_defaults.corlib, "System", "OutOfMemoryException");
+
+	if (!setup->application_base) {
+		/* Inherit from the root domain since MS.NET does this */
+		MonoDomain *root = mono_get_root_domain ();
+		if (root->setup->application_base)
+			MONO_OBJECT_SETREF (setup, application_base, mono_string_new_utf16 (data, mono_string_chars (root->setup->application_base), mono_string_length (root->setup->application_base)));
+	}
+
+	mono_context_init (data);
+
+	mono_set_private_bin_path_from_config (data);
+	
+	add_assemblies_to_domain (data, mono_defaults.corlib->assembly, NULL);
+
+	return ad;
+}
+
 /**
  * mono_domain_has_type_resolve:
  * @domain: application domains being looked up
@@ -478,37 +629,118 @@ ves_icall_System_AppDomain_getRootDomain ()
 	return root->domain;
 }
 
+static char*
+get_attribute_value (const gchar **attribute_names, 
+		     const gchar **attribute_values, 
+		     const char *att_name)
+{
+	int n;
+	for (n = 0; attribute_names [n] != NULL; n++) {
+		if (strcmp (attribute_names [n], att_name) == 0)
+			return g_strdup (attribute_values [n]);
+	}
+	return NULL;
+}
+
+static void
+start_element (GMarkupParseContext *context, 
+	       const gchar         *element_name,
+	       const gchar        **attribute_names,
+	       const gchar        **attribute_values,
+	       gpointer             user_data,
+	       GError             **error)
+{
+	RuntimeConfig *runtime_config = user_data;
+	
+	if (strcmp (element_name, "runtime") == 0) {
+		runtime_config->runtime_count++;
+		return;
+	}
+
+	if (strcmp (element_name, "assemblyBinding") == 0) {
+		runtime_config->assemblybinding_count++;
+		return;
+	}
+	
+	if (runtime_config->runtime_count != 1 || runtime_config->assemblybinding_count != 1)
+		return;
+
+	if (strcmp (element_name, "probing") != 0)
+		return;
+
+	g_free (runtime_config->domain->private_bin_path);
+	runtime_config->domain->private_bin_path = get_attribute_value (attribute_names, attribute_values, "privatePath");
+	if (runtime_config->domain->private_bin_path && !runtime_config->domain->private_bin_path [0]) {
+		g_free (runtime_config->domain->private_bin_path);
+		runtime_config->domain->private_bin_path = NULL;
+		return;
+	}
+}
+
+static void
+end_element (GMarkupParseContext *context,
+	     const gchar         *element_name,
+	     gpointer             user_data,
+	     GError             **error)
+{
+	RuntimeConfig *runtime_config = user_data;
+	if (strcmp (element_name, "runtime") == 0)
+		runtime_config->runtime_count--;
+	else if (strcmp (element_name, "assemblyBinding") == 0)
+		runtime_config->assemblybinding_count--;
+}
+
+static const GMarkupParser
+mono_parser = {
+	start_element,
+	end_element,
+	NULL,
+	NULL,
+	NULL
+};
+
+void
+mono_set_private_bin_path_from_config (MonoDomain *domain)
+{
+	gchar *config_file, *text;
+	gsize len;
+	struct stat sbuf;
+	GMarkupParseContext *context;
+	RuntimeConfig runtime_config;
+	
+	if (!domain || !domain->setup || !domain->setup->configuration_file)
+		return;
+
+	config_file = mono_string_to_utf8 (domain->setup->configuration_file);
+	if (stat (config_file, &sbuf) != 0) {
+		g_free (config_file);
+		return;
+	}
+
+	if (!g_file_get_contents (config_file, &text, &len, NULL)) {
+		g_free (config_file);
+		return;
+	}
+	g_free (config_file);
+
+	runtime_config.runtime_count = 0;
+	runtime_config.assemblybinding_count = 0;
+	runtime_config.domain = domain;
+	
+	context = g_markup_parse_context_new (&mono_parser, 0, &runtime_config, NULL);
+	if (g_markup_parse_context_parse (context, text, len, NULL))
+		g_markup_parse_context_end_parse (context, NULL);
+	g_markup_parse_context_free (context);
+	g_free (text);
+}
+
 MonoAppDomain *
 ves_icall_System_AppDomain_createDomain (MonoString *friendly_name, MonoAppDomainSetup *setup)
 {
-	MonoClass *adclass;
-	MonoAppDomain *ad;
-	MonoDomain *data;
+	char *fname = mono_string_to_utf8 (friendly_name);
+	MonoAppDomain *ad = mono_domain_create_appdomain_internal (fname, setup);
 	
-	MONO_ARCH_SAVE_REGS;
-
-	adclass = mono_class_from_name (mono_defaults.corlib, "System", "AppDomain");
-
-	/* FIXME: pin all those objects */
-	data = mono_domain_create();
-
-	ad = (MonoAppDomain *) mono_object_new (data, adclass);
-	ad->data = data;
-	data->domain = ad;
-	data->setup = setup;
-	data->friendly_name = mono_string_to_utf8 (friendly_name);
-	data->out_of_memory_ex = mono_exception_from_name_domain (data, mono_defaults.corlib, "System", "OutOfMemoryException");
-
-	if (!setup->application_base) {
-		/* Inherit from the root domain since MS.NET does this */
-		MonoDomain *root = mono_get_root_domain ();
-		if (root->setup->application_base)
-			MONO_OBJECT_SETREF (setup, application_base, mono_string_new_utf16 (data, mono_string_chars (root->setup->application_base), mono_string_length (root->setup->application_base)));
-	}
-
-	mono_context_init (data);
-
-	add_assemblies_to_domain (data, mono_defaults.corlib->assembly, NULL);
+	g_free (fname);
 
 	return ad;
 }
@@ -521,42 +753,45 @@ ves_icall_System_AppDomain_GetAssemblies (MonoAppDomain *ad, MonoBoolean refonly
 	static MonoClass *System_Reflection_Assembly;
 	MonoArray *res;
 	GSList *tmp;
-	int i, count;
-	
+	int i;
+	GPtrArray *assemblies;
+
 	MONO_ARCH_SAVE_REGS;
 
 	if (!System_Reflection_Assembly)
 		System_Reflection_Assembly = mono_class_from_name (
 			mono_defaults.corlib, "System.Reflection", "Assembly");
 
-	count = 0;
+	/* 
+	 * Make a copy of the list of assemblies because we can't hold the assemblies
+	 * lock while creating objects etc.
+	 */
+	assemblies = g_ptr_array_new ();
 	/* Need to skip internal assembly builders created by remoting */
 	mono_domain_assemblies_lock (domain);
-	for (tmp = domain->domain_assemblies; tmp; tmp = tmp->next) {
-		ass = tmp->data;
-		if (refonly && !ass->ref_only)
-			continue;
-		if (!ass->corlib_internal)
-			count++;
-	}
-	res = mono_array_new (domain, System_Reflection_Assembly, count);
-	i = 0;
 	for (tmp = domain->domain_assemblies; tmp; tmp = tmp->next) {
 		ass = tmp->data;
 		if (refonly && !ass->ref_only)
 			continue;
 		if (ass->corlib_internal)
 			continue;
-		mono_array_setref (res, i, mono_assembly_get_object (domain, ass));
-		++i;
+		g_ptr_array_add (assemblies, ass);
 	}
 	mono_domain_assemblies_unlock (domain);
 
+	res = mono_array_new (domain, System_Reflection_Assembly, assemblies->len);
+	for (i = 0; i < assemblies->len; ++i) {
+		ass = g_ptr_array_index (assemblies, i);
+		mono_array_setref (res, i, mono_assembly_get_object (domain, ass));
+	}
+
+	g_ptr_array_free (assemblies, TRUE);
+
 	return res;
 }
 
-static MonoReflectionAssembly *
-try_assembly_resolve (MonoDomain *domain, MonoString *fname, gboolean refonly)
+MonoReflectionAssembly *
+mono_try_assembly_resolve (MonoDomain *domain, MonoString *fname, gboolean refonly)
 {
 	MonoClass *klass;
 	MonoMethod *method;
@@ -592,7 +827,7 @@ mono_domain_assembly_postload_search (MonoAssemblyName *aname,
 	aname_str = mono_stringify_assembly_name (aname);
 
 	/* FIXME: We invoke managed code here, so there is a potential for deadlocks */
-	assembly = try_assembly_resolve (domain, mono_string_new (domain, aname_str), refonly);
+	assembly = mono_try_assembly_resolve (domain, mono_string_new (domain, aname_str), refonly);
 
 	g_free (aname_str);
 
@@ -633,9 +868,10 @@ add_assemblies_to_domain (MonoDomain *domain, MonoAssembly *ass, GHashTable *ht)
 
 	if (ass->image->references) {
 		for (i = 0; ass->image->references [i] != NULL; i++) {
-			if (!g_hash_table_lookup (ht, ass->image->references [i])) {
-				add_assemblies_to_domain (domain, ass->image->references [i], ht);
-			}
+			if (ass->image->references [i] != REFERENCE_MISSING)
+				if (!g_hash_table_lookup (ht, ass->image->references [i])) {
+					add_assemblies_to_domain (domain, ass->image->references [i], ht);
+				}
 		}
 	}
 	if (destroy_ht)
@@ -688,32 +924,80 @@ mono_domain_fire_assembly_load (MonoAssembly *assembly, gpointer user_data)
 	mono_runtime_invoke (assembly_load_method, domain->domain, params, NULL);
 }
 
+/*
+ * LOCKING: Acquires the domain assemblies lock.
+ */
 static void
 set_domain_search_path (MonoDomain *domain)
 {
 	MonoAppDomainSetup *setup;
 	gchar **tmp;
-	gchar *utf8;
+	gchar *search_path = NULL;
 	gint i;
 	gint npaths = 0;
 	gchar **pvt_split = NULL;
 	GError *error = NULL;
 	gint appbaselen = -1;
 
-	if ((domain->search_path != NULL) && !domain->setup->path_changed)
-		return;
-	if (!domain->setup)
-		return;
+	/* 
+	 * We use the low-level domain assemblies lock, since this is called from
+	 * assembly loads hooks, which means this thread might hold the loader lock.
+	 */
+	mono_domain_assemblies_lock (domain);
 
+	if (!domain->setup) {
+		mono_domain_assemblies_unlock (domain);
+		return;
+	}
+
+	if ((domain->search_path != NULL) && !domain->setup->path_changed) {
+		mono_domain_assemblies_unlock (domain);
+		return;
+	}
 	setup = domain->setup;
-	if (!setup->application_base)
+	if (!setup->application_base) {
+		mono_domain_assemblies_unlock (domain);
 		return; /* Must set application base to get private path working */
+	}
 
 	npaths++;
-	if (setup->private_bin_path) {
-		utf8 = mono_string_to_utf8 (setup->private_bin_path);
-		pvt_split = g_strsplit (utf8, G_SEARCHPATH_SEPARATOR_S, 1000);
-		g_free (utf8);
+	
+	if (setup->private_bin_path)
+		search_path = mono_string_to_utf8 (setup->private_bin_path);
+	
+	if (domain->private_bin_path) {
+		if (search_path == NULL)
+			search_path = domain->private_bin_path;
+		else {
+			gchar *tmp2 = search_path;
+			search_path = g_strjoin (";", search_path, domain->private_bin_path, NULL);
+			g_free (tmp2);
+		}
+	}
+	
+	if (search_path) {
+		/*
+		 * As per MSDN documentation, AppDomainSetup.PrivateBinPath contains a list of
+		 * directories relative to ApplicationBase separated by semicolons (see
+		 * http://msdn2.microsoft.com/en-us/library/system.appdomainsetup.privatebinpath.aspx)
+		 * The loop below copes with the fact that some Unix applications may use ':' (or
+		 * System.IO.Path.PathSeparator) as the path search separator. We replace it with
+		 * ';' for the subsequent split.
+		 *
+		 * The issue was reported in bug #81446
+		 */
+
+#ifndef PLATFORM_WIN32
+		gint slen;
+
+		slen = strlen (search_path);
+		for (i = 0; i < slen; i++)
+			if (search_path [i] == ':')
+				search_path [i] = ';';
+#endif
+		
+		pvt_split = g_strsplit (search_path, ";", 1000);
+		g_free (search_path);
 		for (tmp = pvt_split; *tmp; tmp++, npaths++);
 	}
 
@@ -727,6 +1011,7 @@ set_domain_search_path (MonoDomain *domain)
 		 * domain->search_path = g_malloc (sizeof (char *));
 		 * domain->search_path [0] = NULL;
 		*/
+		mono_domain_assemblies_unlock (domain);
 		return;
 	}
 
@@ -800,6 +1085,8 @@ set_domain_search_path (MonoDomain *domain)
 	domain->setup->path_changed = FALSE;
 
 	g_strfreev (pvt_split);
+
+	mono_domain_assemblies_unlock (domain);
 }
 
 static gboolean
@@ -816,6 +1103,7 @@ shadow_copy_sibling (gchar *src, gint srclen, const char *extension, gchar *targ
 	strcpy (target + targetlen - tail_len, extension);
 	dest = g_utf8_to_utf16 (target, strlen (target), NULL, NULL, NULL);
 	
+	DeleteFile (dest);
 	copy_result = CopyFile (orig, dest, FALSE);
 	g_free (orig);
 	g_free (dest);
@@ -843,26 +1131,59 @@ get_cstring_hash (const char *str)
 	return h;
 }
 
+static char *
+get_shadow_assembly_location_base (MonoDomain *domain)
+{
+	MonoAppDomainSetup *setup;
+	char *cache_path, *appname;
+	char *userdir;
+	char *location;
+	
+	setup = domain->setup;
+	if (setup->cache_path != NULL && setup->application_name != NULL) {
+		cache_path = mono_string_to_utf8 (setup->cache_path);
+#ifndef PLATFORM_WIN32
+		{
+			gint i;
+			for (i = strlen (cache_path) - 1; i >= 0; i--)
+				if (cache_path [i] == '\\')
+					cache_path [i] = '/';
+		}
+#endif
+
+		appname = mono_string_to_utf8 (setup->application_name);
+		location = g_build_filename (cache_path, appname, "assembly", "shadow", NULL);
+		g_free (appname);
+		g_free (cache_path);
+	} else {
+		userdir = g_strdup_printf ("%s-mono-cachepath", g_get_user_name ());
+		location = g_build_filename (g_get_tmp_dir (), userdir, "assembly", "shadow", NULL);
+		g_free (userdir);
+	}
+	return location;
+}
+
 static char *
 get_shadow_assembly_location (const char *filename)
 {
 	gint32 hash = 0, hash2 = 0;
 	char name_hash [9];
-	char path_hash [19];
+	char path_hash [30];
 	char *bname = g_path_get_basename (filename);
+	char *dirname = g_path_get_dirname (filename);
+	char *location, *tmploc;
 	MonoDomain *domain = mono_domain_get ();
 	
 	hash = get_cstring_hash (bname);
-	hash2 = get_cstring_hash (g_path_get_dirname (filename));
+	hash2 = get_cstring_hash (dirname);
 	g_snprintf (name_hash, sizeof (name_hash), "%08x", hash);
-	g_snprintf (path_hash, sizeof (path_hash), "%08x_%08x", hash ^ hash2, hash2);
-	return g_build_filename (mono_string_to_utf8 (domain->setup->dynamic_base), 
-				 "assembly", 
-				 "shadow", 
-				 name_hash,
-				 path_hash,
-				 bname, 
-				 NULL);
+	g_snprintf (path_hash, sizeof (path_hash), "%08x_%08x_%08x", hash ^ hash2, hash2, domain->shadow_serial);
+	tmploc = get_shadow_assembly_location_base (domain);
+	location = g_build_filename (tmploc, name_hash, path_hash, bname, NULL);
+	g_free (tmploc);
+	g_free (bname);
+	g_free (dirname);
+	return location;
 }
 
 static gboolean
@@ -921,10 +1242,18 @@ ensure_directory_exists (const char *filename)
 	char *p;
 	gchar *dir = g_path_get_dirname (filename);
 	int retval;
+	struct stat sbuf;
 	
-	if (!dir || !dir [0])
+	if (!dir || !dir [0]) {
+		g_free (dir);
 		return FALSE;
-
+	}
+	
+	if (stat (dir, &sbuf) == 0 && S_ISDIR (sbuf.st_mode)) {
+		g_free (dir);
+		return TRUE;
+	}
+	
 	p = dir;
 	while (*p == '/')
 		p++;
@@ -948,8 +1277,116 @@ ensure_directory_exists (const char *filename)
 #endif
 }
 
-static char *
-make_shadow_copy (const char *filename)
+static gboolean
+private_file_needs_copying (const char *src, struct stat *sbuf_src, char *dest)
+{
+	struct stat sbuf_dest;
+	
+	if (stat (src, sbuf_src) == -1 || stat (dest, &sbuf_dest) == -1)
+		return TRUE;
+
+	if (sbuf_src->st_mode == sbuf_dest.st_mode &&
+	    sbuf_src->st_size == sbuf_dest.st_size &&
+	    sbuf_src->st_mtime == sbuf_dest.st_mtime)
+		return FALSE;
+
+	return TRUE;
+}
+
+static gboolean
+shadow_copy_create_ini (const char *shadow, const char *filename)
+{
+	char *dir_name;
+	char *ini_file;
+	guint16 *u16_ini;
+	gboolean result;
+	guint32 n;
+	HANDLE *handle;
+	gchar *full_path;
+
+	dir_name = g_path_get_dirname (shadow);
+	ini_file = g_build_filename (dir_name, "__AssemblyInfo__.ini", NULL);
+	g_free (dir_name);
+	if (g_file_test (ini_file, G_FILE_TEST_IS_REGULAR)) {
+		g_free (ini_file);
+		return TRUE;
+	}
+
+	u16_ini = g_utf8_to_utf16 (ini_file, strlen (ini_file), NULL, NULL, NULL);
+	g_free (ini_file);
+	if (!u16_ini) {
+		return FALSE;
+	}
+	handle = CreateFile (u16_ini, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
+				NULL, CREATE_NEW, FileAttributes_Normal, NULL);
+	g_free (u16_ini);
+	if (handle == INVALID_HANDLE_VALUE) {
+		return FALSE;
+	}
+
+	full_path = mono_path_resolve_symlinks (filename);
+	result = WriteFile (handle, full_path, strlen (full_path), &n, NULL);
+	g_free (full_path);
+	CloseHandle (handle);
+	return result;
+}
+
+gboolean
+mono_is_shadow_copy_enabled (MonoDomain *domain, const gchar *dir_name)
+{
+	const char *version;
+	MonoAppDomainSetup *setup;
+	gchar *all_dirs;
+	gchar **dir_ptr;
+	gchar **directories;
+	gchar *shadow_status_string;
+	gchar *base_dir;
+	gboolean shadow_enabled;
+	gboolean found = FALSE;
+
+	if (domain == NULL)
+		return FALSE;
+
+	setup = domain->setup;
+	if (setup == NULL || setup->shadow_copy_files == NULL)
+		return FALSE;
+
+	version = mono_get_runtime_info ()->framework_version;
+	shadow_status_string = mono_string_to_utf8 (setup->shadow_copy_files);
+	/* For 1.x, not NULL is enough. In 2.0 it has to be "true" */
+	shadow_enabled = (*version <= '1' || !g_ascii_strncasecmp (shadow_status_string, "true", 4));
+	g_free (shadow_status_string);
+	if (!shadow_enabled)
+		return FALSE;
+
+	if (setup->shadow_copy_directories == NULL)
+		return TRUE;
+
+	/* Is dir_name a shadow_copy destination already? */
+	base_dir = get_shadow_assembly_location_base (domain);
+	if (strstr (dir_name, base_dir) == dir_name) {
+		g_free (base_dir);
+		return FALSE;
+	}
+	g_free (base_dir);
+
+	all_dirs = mono_string_to_utf8 (setup->shadow_copy_directories);
+	directories = g_strsplit (all_dirs, G_SEARCHPATH_SEPARATOR_S, 1000);
+	dir_ptr = directories;
+	while (*dir_ptr) {
+		if (**dir_ptr != '\0' && !strcmp (*dir_ptr, dir_name)) {
+			found = TRUE;
+			break;
+		}
+		dir_ptr++;
+	}
+	g_strfreev (directories);
+	g_free (all_dirs);
+	return found;
+}
+
+char *
+mono_make_shadow_copy (const char *filename)
 {
 	gchar *sibling_source, *sibling_target;
 	gint sibling_source_len, sibling_target_len;
@@ -957,15 +1394,31 @@ make_shadow_copy (const char *filename)
 	char *shadow;
 	gboolean copy_result;
 	MonoException *exc;
+	struct stat src_sbuf;
+	struct utimbuf utbuf;
+	char *dir_name = g_path_get_dirname (filename);
+	MonoDomain *domain = mono_domain_get ();
+	set_domain_search_path (domain);
+
+	if (!mono_is_shadow_copy_enabled (domain, dir_name)) {
+		g_free (dir_name);
+		return (char *) filename;
+	}
+	g_free (dir_name);
 	
 	shadow = get_shadow_assembly_location (filename);
 	if (ensure_directory_exists (shadow) == FALSE) {
+		g_free (shadow);
 		exc = mono_get_exception_execution_engine ("Failed to create shadow copy (ensure directory exists).");
 		mono_raise_exception (exc);
 	}	
 
+	if (!private_file_needs_copying (filename, &src_sbuf, shadow))
+		return (char*) shadow;
+
 	orig = g_utf8_to_utf16 (filename, strlen (filename), NULL, NULL, NULL);
 	dest = g_utf8_to_utf16 (shadow, strlen (shadow), NULL, NULL, NULL);
+	DeleteFile (dest);
 	copy_result = CopyFile (orig, dest, FALSE);
 	g_free (dest);
 	g_free (orig);
@@ -994,29 +1447,54 @@ make_shadow_copy (const char *filename)
 		exc = mono_get_exception_execution_engine ("Failed to create shadow copy of sibling data (CopyFile).");
 		mono_raise_exception (exc);
 	}
+
+	/* Create a .ini file containing the original assembly location */
+	if (!shadow_copy_create_ini (shadow, filename)) {
+		g_free (shadow);
+		exc = mono_get_exception_execution_engine ("Failed to create shadow copy .ini file.");
+		mono_raise_exception (exc);
+	}
+
+	utbuf.actime = src_sbuf.st_atime;
+	utbuf.modtime = src_sbuf.st_mtime;
+	utime (shadow, &utbuf);
 	
 	return shadow;
 }
 
+
+MonoDomain *
+mono_domain_from_appdomain (MonoAppDomain *appdomain)
+{
+	if (appdomain == NULL)
+		return NULL;
+	
+	return appdomain->data;
+}
+
 static gboolean
 try_load_from (MonoAssembly **assembly, const gchar *path1, const gchar *path2,
 					const gchar *path3, const gchar *path4,
 					gboolean refonly, gboolean is_private)
 {
-	
 	gchar *fullpath;
-
+	gboolean found = FALSE;
+	
 	*assembly = NULL;
 	fullpath = g_build_filename (path1, path2, path3, path4, NULL);
-	if (g_file_test (fullpath, G_FILE_TEST_IS_REGULAR)) {
-		if (is_private) {
-			char *new_path = make_shadow_copy (fullpath);
-			g_free (fullpath);
-			fullpath = new_path;
-		}
 
+	if (IS_PORTABILITY_SET) {
+		gchar *new_fullpath = mono_portability_find_file (fullpath, TRUE);
+		if (new_fullpath) {
+			g_free (fullpath);
+			fullpath = new_fullpath;
+			found = TRUE;
+		}
+	} else
+		found = g_file_test (fullpath, G_FILE_TEST_IS_REGULAR);
+	
+	if (found)
 		*assembly = mono_assembly_open_full (fullpath, NULL, refonly);
-	}
 
 	g_free (fullpath);
 	return (*assembly != NULL);
@@ -1076,6 +1554,8 @@ real_load (gchar **search_path, const gchar *culture, const gchar *name, gboolea
 /*
  * Try loading the assembly from ApplicationBase and PrivateBinPath 
  * and then from assemblies_path if any.
+ * LOCKING: This is called from the assembly loading code, which means the caller
+ * might hold the loader lock. Thus, this function must not acquire the domain lock.
  */
 static MonoAssembly *
 mono_domain_assembly_preload (MonoAssemblyName *aname,
@@ -1142,7 +1622,7 @@ ves_icall_System_Reflection_Assembly_LoadFrom (MonoString *fname, MonoBoolean re
 	}
 		
 	name = filename = mono_string_to_utf8 (fname);
-
+	
 	ass = mono_assembly_open_full (filename, &status, refOnly);
 	
 	if (!ass){
@@ -1163,9 +1643,9 @@ ves_icall_System_Reflection_Assembly_LoadFrom (MonoString *fname, MonoBoolean re
 
 MonoReflectionAssembly *
 ves_icall_System_AppDomain_LoadAssemblyRaw (MonoAppDomain *ad, 
-											MonoArray *raw_assembly,
-											MonoArray *raw_symbol_store, MonoObject *evidence,
-											MonoBoolean refonly)
+					    MonoArray *raw_assembly,
+					    MonoArray *raw_symbol_store, MonoObject *evidence,
+					    MonoBoolean refonly)
 {
 	MonoAssembly *ass;
 	MonoReflectionAssembly *refass = NULL;
@@ -1180,7 +1660,7 @@ ves_icall_System_AppDomain_LoadAssemblyRaw (MonoAppDomain *ad,
 	}
 
 	if (raw_symbol_store != NULL)
-		mono_debug_init_2_memory (image, mono_array_addr (raw_symbol_store, guint8, 0), mono_array_length (raw_symbol_store));
+		mono_debug_open_image_from_memory (image, mono_array_addr (raw_symbol_store, guint8, 0), mono_array_length (raw_symbol_store));
 
 	ass = mono_assembly_load_from_full (image, "", &status, refonly);
 
@@ -1216,20 +1696,22 @@ ves_icall_System_AppDomain_LoadAssembly (MonoAppDomain *ad,  MonoString *assRef,
 	g_free (name);
 
 	if (!parsed) {
-		MonoException *exc;
-
 		/* This is a parse error... */
-		exc = mono_get_exception_file_not_found2 (NULL, assRef);
-		mono_raise_exception (exc);
+		return NULL;
 	}
 
-	ass = mono_assembly_load_full (&aname, NULL, &status, refOnly);
+	ass = mono_assembly_load_full_nosearch (&aname, NULL, &status, refOnly);
 	mono_assembly_name_free (&aname);
 
-	if (!ass && (refass = try_assembly_resolve (domain, assRef, refOnly)) == NULL){
-		/* FIXME: it doesn't make much sense since we really don't have a filename ... */
-		MonoException *exc = mono_get_exception_file_not_found2 (NULL, assRef);
-		mono_raise_exception (exc);
+	if (!ass) {
+		/* MS.NET doesn't seem to call the assembly resolve handler for refonly assemblies */
+		if (!refOnly)
+			refass = mono_try_assembly_resolve (domain, assRef, refOnly);
+		else
+			refass = NULL;
+		if (!refass) {
+			return NULL;
+		}
 	}
 
 	if (refass == NULL)
@@ -1278,26 +1760,17 @@ ves_icall_System_AppDomain_InternalIsFinalizingForUnload (gint32 domain_id)
 }
 
 gint32
-ves_icall_System_AppDomain_ExecuteAssembly (MonoAppDomain *ad, MonoString *file, 
-					    MonoObject *evidence, MonoArray *args)
+ves_icall_System_AppDomain_ExecuteAssembly (MonoAppDomain *ad, 
+											MonoReflectionAssembly *refass, MonoArray *args)
 {
-	MonoAssembly *assembly;
 	MonoImage *image;
 	MonoMethod *method;
-	char *filename;
-	gint32 res;
-	MonoReflectionAssembly *refass;
 
 	MONO_ARCH_SAVE_REGS;
 
-	filename = mono_string_to_utf8 (file);
-	assembly = mono_assembly_open (filename, NULL);
-	g_free (filename);
-
-	if (!assembly)
-		mono_raise_exception (mono_get_exception_file_not_found2 (NULL, file));
-
-	image = assembly->image;
+	g_assert (refass);
+	image = refass->assembly->image;
+	g_assert (image);
 
 	method = mono_get_method (image, mono_image_get_entry_point (image), NULL);
 
@@ -1307,12 +1780,7 @@ ves_icall_System_AppDomain_ExecuteAssembly (MonoAppDomain *ad, MonoString *file,
 	if (!args)
 		args = (MonoArray *) mono_array_new (ad->data, mono_defaults.string_class, 0);
 
-	refass = mono_assembly_get_object (ad->data, assembly);
-	MONO_OBJECT_SETREF (refass, evidence, evidence);
-
-	res = mono_runtime_exec_main (method, (MonoArray *)args, NULL);
-
-	return res;
+	return mono_runtime_exec_main (method, (MonoArray *)args, NULL);
 }
 
 gint32 
@@ -1452,23 +1920,32 @@ typedef struct unload_data {
 	char *failure_reason;
 } unload_data;
 
+
 static guint32 WINAPI
 unload_thread_main (void *arg)
 {
 	unload_data *data = (unload_data*)arg;
 	MonoDomain *domain = data->domain;
 
+	/* Have to attach to the runtime so shutdown can wait for this thread */
+	mono_thread_attach (mono_get_root_domain ());
+
 	/* 
 	 * FIXME: Abort our parent thread last, so we can return a failure 
 	 * indication if aborting times out.
 	 */
-	if (!mono_threads_abort_appdomain_threads (domain, 10000)) {
+	if (!mono_threads_abort_appdomain_threads (domain, -1)) {
 		data->failure_reason = g_strdup_printf ("Aborting of threads in domain %s timed out.", domain->friendly_name);
 		return 1;
 	}
 
+	if (!mono_thread_pool_remove_domain_jobs (domain, -1)) {
+		data->failure_reason = g_strdup_printf ("Cleanup of threadpool jobs of domain %s timed out.", domain->friendly_name);
+		return 1;
+	}
+
 	/* Finalize all finalizable objects in the doomed appdomain */
-	if (!mono_domain_finalize (domain, 10000)) {
+	if (!mono_domain_finalize (domain, -1)) {
 		data->failure_reason = g_strdup_printf ("Finalization of domain %s timed out.", domain->friendly_name);
 		return 1;
 	}
@@ -1510,7 +1987,7 @@ unload_thread_main (void *arg)
  *  'simple' way, which means do everything needed to avoid crashes and
  *  memory leaks, but not much else.
  */
-static void
+void
 mono_domain_unload (MonoDomain *domain)
 {
 	HANDLE thread_handle;
@@ -1526,16 +2003,21 @@ mono_domain_unload (MonoDomain *domain)
 
 	/* Atomically change our state to UNLOADING */
 	prev_state = InterlockedCompareExchange ((gint32*)&domain->state,
-											 MONO_APPDOMAIN_UNLOADING,
+											 MONO_APPDOMAIN_UNLOADING_START,
 											 MONO_APPDOMAIN_CREATED);
 	if (prev_state != MONO_APPDOMAIN_CREATED) {
-		if (prev_state == MONO_APPDOMAIN_UNLOADING)
+		switch (prev_state) {
+		case MONO_APPDOMAIN_UNLOADING_START:
+		case MONO_APPDOMAIN_UNLOADING:
 			mono_raise_exception (mono_get_exception_cannot_unload_appdomain ("Appdomain is already being unloaded."));
-		else
-			if (prev_state == MONO_APPDOMAIN_UNLOADED)
-				mono_raise_exception (mono_get_exception_cannot_unload_appdomain ("Appdomain is already unloaded."));
-		else
+			break;
+		case MONO_APPDOMAIN_UNLOADED:
+			mono_raise_exception (mono_get_exception_cannot_unload_appdomain ("Appdomain is already unloaded."));
+			break;
+		default:
+			g_warning ("Incalid appdomain state %d", prev_state);
 			g_assert_not_reached ();
+		}
 	}
 
 	mono_domain_set (domain, FALSE);
@@ -1555,6 +2037,8 @@ mono_domain_unload (MonoDomain *domain)
 	thread_data.domain = domain;
 	thread_data.failure_reason = NULL;
 
+	/*The managed callback finished successfully, now we start tearing down the appdomain*/
+	domain->state = MONO_APPDOMAIN_UNLOADING;
 	/* 
 	 * First we create a separate thread for unloading, since
 	 * we might have to abort some threads, including the current one.
