@@ -2,11 +2,32 @@
 #define __MONO_OBJECT_INTERNALS_H__
 
 #include <mono/metadata/object.h>
+#include <mono/metadata/threads.h>
 #include <mono/metadata/reflection.h>
 #include <mono/metadata/mempool.h>
 #include <mono/io-layer/io-layer.h>
 #include "mono/utils/mono-compiler.h"
 
+/* 
+ * We should find a better place for this stuff. We can't put it in mono-compiler.h,
+ * since that is included by libgc.
+ */
+#ifndef G_LIKELY
+#define G_LIKELY(a) (a)
+#define G_UNLIKELY(a) (a)
+#endif
+
+/*
+ * glib defines this macro and uses it in the definition of G_LIKELY, and thus,
+ * g_assert (). The macro expands to a complex piece of code, preventing some
+ * gcc versions like 4.3.0 from handling the __builtin_expect construct properly,
+ * causing the generation of the unlikely branch into the middle of the code.
+ */
+#ifdef _G_BOOLEAN_EXPR
+#undef _G_BOOLEAN_EXPR
+#define _G_BOOLEAN_EXPR(expr) (gsize)(expr)
+#endif
+
 #if 1
 #ifdef __GNUC__
 #define mono_assert(expr)		   G_STMT_START{		  \
@@ -119,6 +140,8 @@ typedef struct {
 
 struct _MonoException {
 	MonoObject object;
+	/* Stores the IPs and the generic sharing infos
+	   (vtable/MRGCTX) of the frames. */
 	MonoArray  *trace_ips;
 	MonoObject *inner_ex;
 	MonoString *message;
@@ -129,6 +152,7 @@ struct _MonoException {
 	gint32	    remote_stack_index;
 	gint32	    hresult;
 	MonoString *source;
+	MonoObject *_data;
 };
 
 typedef struct {
@@ -257,27 +281,34 @@ struct _MonoThread {
 	int stack_size;
 	MonoObject *start_obj;
 	GSList *appdomain_refs;
-	MonoBoolean interruption_requested;
+	/* This is modified using atomic ops, so keep it a gint32 */
+	gint32 interruption_requested;
 	gpointer suspend_event;
 	gpointer suspended_event;
 	gpointer resume_event;
-	MonoObject *synch_lock;
+	CRITICAL_SECTION *synch_cs;
 	guint8* serialized_culture_info;
 	guint32 serialized_culture_info_len;
 	guint8* serialized_ui_culture_info;
 	guint32 serialized_ui_culture_info_len;
 	MonoObject *execution_context;
+	MonoBoolean thread_dump_requested;
+	gpointer end_stack; /* This is only used when running in the debugger. */
+	MonoBoolean thread_interrupt_requested;
+	guint8	apartment_state;
+	gint32 critical_region_level;
+	guint32 small_id; /* A small, unique id, used for the hazard pointer table. */
+	MonoThreadManageCallback manage_callback;
+	MonoException *pending_exception;
 	/* 
 	 * These fields are used to avoid having to increment corlib versions
 	 * when a new field is added to the unmanaged MonoThread structure.
 	 */
-	MonoBoolean thread_dump_requested;
-	gpointer end_stack; /* This is only used when running in the debugger. */
-	MonoBoolean thread_interrupt_requested;
+	gpointer unused2;
+	gpointer unused3;
 	gpointer unused4;
 	gpointer unused5;
 	gpointer unused6;
-	gpointer unused7;
 };
 
 typedef struct {
@@ -387,6 +418,7 @@ typedef struct {
 	MonoString *iso2lang;
 	MonoString *icu_name;
 	MonoString *win3lang;
+	MonoString *territory;
 	MonoCompareInfo *compareinfo;
 	const gint32 *calendar_data;
 	const void* text_info_data;
@@ -414,7 +446,7 @@ typedef struct {
 
 typedef struct {
 	MonoObject object;
-	guint16 intType;
+	guint32 intType;
 } MonoInterfaceTypeAttribute;
 
 /* used to free a dynamic method */
@@ -423,6 +455,8 @@ typedef void        (*MonoFreeMethodFunc)	 (MonoDomain *domain, MonoMethod *meth
 /* Used to initialize the method pointers inside vtables */
 typedef gboolean    (*MonoInitVTableFunc)    (MonoVTable *vtable);
 
+void mono_set_pending_exception (MonoException *exc) MONO_INTERNAL;
+
 /* remoting and async support */
 
 MonoAsyncResult *
@@ -454,6 +488,9 @@ mono_method_call_message_new (MonoMethod *method, gpointer *params, MonoMethod *
 void
 mono_method_return_message_restore (MonoMethod *method, gpointer *params, MonoArray *out_args) MONO_INTERNAL;
 
+void
+mono_delegate_ctor_with_method (MonoObject *this, MonoObject *target, gpointer addr, MonoMethod *method) MONO_INTERNAL;
+
 void
 mono_delegate_ctor	    (MonoObject *this_obj, MonoObject *target, gpointer addr) MONO_INTERNAL;
 
@@ -478,9 +515,6 @@ mono_install_compile_method (MonoCompileFunc func) MONO_INTERNAL;
 void
 mono_install_free_method    (MonoFreeMethodFunc func) MONO_INTERNAL;
 
-void
-mono_install_init_vtable    (MonoInitVTableFunc func) MONO_INTERNAL;
-
 void
 mono_type_initialization_init (void) MONO_INTERNAL;
 
@@ -519,12 +553,21 @@ struct _MonoReflectionGenericMethod {
 
 struct _MonoDelegate {
 	MonoObject object;
-	MonoObject *target_type;
-	MonoObject *target;
-	MonoString *method_name;
+	/* The compiled code of the target method */
 	gpointer method_ptr;
+	/* The invoke code */
+	gpointer invoke_impl;
+	MonoObject *target;
+	MonoMethod *method;
 	gpointer delegate_trampoline;
+	/* 
+	 * If non-NULL, this points to a memory location which stores the address of 
+	 * the compiled code of the method, or NULL if it is not yet compiled.
+	 */
+	guint8 **method_code;
 	MonoReflectionMethod *method_info;
+	MonoReflectionMethod *original_method_info;
+	MonoObject *data;
 };
 
 typedef struct _MonoMulticastDelegate MonoMulticastDelegate;
@@ -929,7 +972,7 @@ typedef struct {
 typedef struct _MonoReflectionGenericClass MonoReflectionGenericClass;
 struct _MonoReflectionGenericClass {
 	MonoReflectionType type;
-	MonoReflectionType *generic_type;
+	MonoReflectionTypeBuilder *generic_type;
 	guint32 initialized;
 };
 
@@ -946,6 +989,7 @@ typedef struct {
 	MonoArray   *keyToken;
 	guint32     versioncompat;
 	MonoObject *version;
+	guint32     processor_architecture;
 } MonoReflectionAssemblyName;
 
 typedef struct {
@@ -1010,6 +1054,7 @@ typedef struct {
 	gint32 nrefs;
 	MonoArray *refs;
 	GSList *referenced_by;
+	MonoReflectionType *owner;
 } MonoReflectionDynamicMethod;	
 
 typedef struct {
@@ -1020,8 +1065,33 @@ typedef struct {
 	MonoReflectionType *return_type;
 	guint32 call_conv;
 	guint32 unmanaged_call_conv;
+	MonoArray *modreqs;
+	MonoArray *modopts;
 } MonoReflectionSigHelper;
 
+typedef struct {
+	MonoObject object;
+	MonoReflectionGenericClass *inst;
+	MonoReflectionFieldBuilder *fb;
+} MonoReflectionFieldOnTypeBuilderInst;
+
+typedef struct {
+	MonoObject object;
+	MonoReflectionGenericClass *inst;
+	MonoReflectionCtorBuilder *cb;
+} MonoReflectionCtorOnTypeBuilderInst;
+
+typedef struct {
+	MonoObject object;
+	MonoReflectionGenericClass *inst;
+	MonoReflectionMethodBuilder *mb;
+} MonoReflectionMethodOnTypeBuilderInst;
+
+typedef struct {
+	MonoObject object;
+	MonoBoolean visible;
+} MonoReflectionComVisibleAttribute;
+
 enum {
 	RESOURCE_LOCATION_EMBEDDED = 1,
 	RESOURCE_LOCATION_ANOTHER_ASSEMBLY = 2,
@@ -1051,11 +1121,13 @@ typedef enum {
 
 void          mono_image_create_pefile (MonoReflectionModuleBuilder *module, HANDLE file) MONO_INTERNAL;
 void          mono_image_basic_init (MonoReflectionAssemblyBuilder *assembly) MONO_INTERNAL;
-MonoReflectionModule * mono_image_load_module (MonoReflectionAssemblyBuilder *assembly, MonoString *file_name) MONO_INTERNAL;
+MonoReflectionModule * mono_image_load_module_dynamic (MonoReflectionAssemblyBuilder *assembly, MonoString *file_name) MONO_INTERNAL;
 guint32       mono_image_insert_string (MonoReflectionModuleBuilder *module, MonoString *str) MONO_INTERNAL;
-guint32       mono_image_create_token  (MonoDynamicImage *assembly, MonoObject *obj, gboolean create_methodspec) MONO_INTERNAL;
+guint32       mono_image_create_token  (MonoDynamicImage *assembly, MonoObject *obj, gboolean create_methodspec, gboolean register_token) MONO_INTERNAL;
 guint32       mono_image_create_method_token (MonoDynamicImage *assembly, MonoObject *obj, MonoArray *opt_param_types) MONO_INTERNAL;
 void          mono_image_module_basic_init (MonoReflectionModuleBuilder *module) MONO_INTERNAL;
+void          mono_image_register_token (MonoDynamicImage *assembly, guint32 token, MonoObject *obj) MONO_INTERNAL;
+void          mono_dynamic_image_free (MonoDynamicImage *image) MONO_INTERNAL;
 
 void        mono_reflection_setup_internal_class  (MonoReflectionTypeBuilder *tb) MONO_INTERNAL;
 
@@ -1070,11 +1142,12 @@ MonoReflectionType* mono_reflection_create_runtime_class  (MonoReflectionTypeBui
 void        mono_reflection_get_dynamic_overrides (MonoClass *klass, MonoMethod ***overrides, int *num_overrides) MONO_INTERNAL;
 
 void mono_reflection_create_dynamic_method (MonoReflectionDynamicMethod *m) MONO_INTERNAL;
+void mono_reflection_destroy_dynamic_method (MonoReflectionDynamicMethod *mb) MONO_INTERNAL;
 
 void        mono_reflection_initialize_generic_parameter (MonoReflectionGenericParam *gparam) MONO_INTERNAL;
 
-MonoType*
-mono_class_bind_generic_parameters (MonoType *type, int type_argc, MonoType **types) MONO_INTERNAL;
+MonoClass*
+mono_class_bind_generic_parameters (MonoClass *klass, int type_argc, MonoType **types, gboolean is_dynamic) MONO_INTERNAL;
 MonoType*
 mono_reflection_bind_generic_parameters (MonoReflectionType *type, int type_argc, MonoType **types) MONO_INTERNAL;
 MonoReflectionMethod*
@@ -1091,11 +1164,14 @@ MonoArray  *mono_reflection_sighelper_get_signature_field (MonoReflectionSigHelp
 MonoReflectionMarshal* mono_reflection_marshal_from_marshal_spec (MonoDomain *domain, MonoClass *klass, MonoMarshalSpec *spec) MONO_INTERNAL;
 
 gpointer
-mono_reflection_lookup_dynamic_token (MonoImage *image, guint32 token, MonoClass **handle_class) MONO_INTERNAL;
+mono_reflection_lookup_dynamic_token (MonoImage *image, guint32 token, gboolean valid_token, MonoClass **handle_class, MonoGenericContext *context) MONO_INTERNAL;
 
 gboolean
 mono_reflection_call_is_assignable_to (MonoClass *klass, MonoClass *oklass) MONO_INTERNAL;
 
+gboolean
+mono_reflection_is_valid_dynamic_token (MonoDynamicImage *image, guint32 token) MONO_INTERNAL;
+
 void
 mono_image_build_metadata (MonoReflectionModuleBuilder *module) MONO_INTERNAL;
 
@@ -1120,6 +1196,9 @@ mono_remote_class_vtable (MonoDomain *domain, MonoRemoteClass *remote_class, Mon
 void
 mono_upgrade_remote_class (MonoDomain *domain, MonoObject *tproxy, MonoClass *klass) MONO_INTERNAL;
 
+gpointer
+mono_create_ftnptr (MonoDomain *domain, gpointer addr) MONO_INTERNAL;
+
 gpointer
 mono_get_addr_from_ftnptr (gpointer descr) MONO_INTERNAL;
 
@@ -1129,5 +1208,79 @@ mono_nullable_init (guint8 *buf, MonoObject *value, MonoClass *klass) MONO_INTER
 MonoObject*
 mono_nullable_box (guint8 *buf, MonoClass *klass) MONO_INTERNAL;
 
+#define MONO_IMT_SIZE 19
+
+typedef union {
+	int vtable_slot;
+	gpointer target_code;
+} MonoImtItemValue;
+
+typedef struct _MonoImtBuilderEntry {
+	gpointer key;
+	struct _MonoImtBuilderEntry *next;
+	MonoImtItemValue value;
+	int children;
+} MonoImtBuilderEntry;
+
+typedef struct _MonoIMTCheckItem MonoIMTCheckItem;
+
+struct _MonoIMTCheckItem {
+	gpointer          key;
+	int               check_target_idx;
+	MonoImtItemValue  value;
+	guint8           *jmp_code;
+	guint8           *code_target;
+	guint8            is_equals;
+	guint8            compare_done;
+	guint8            chunk_size;
+	guint8            short_branch;
+};
+
+typedef gpointer (*MonoImtThunkBuilder) (MonoVTable *vtable, MonoDomain *domain,
+		MonoIMTCheckItem **imt_entries, int count, gpointer fail_trunk);
+
+void
+mono_install_imt_thunk_builder (MonoImtThunkBuilder func) MONO_INTERNAL;
+
+void
+mono_install_imt_trampoline (gpointer tramp) MONO_INTERNAL;
+
+void
+mono_install_vtable_trampoline (gpointer tramp) MONO_INTERNAL;
+
+void
+mono_vtable_build_imt_slot (MonoVTable* vtable, int imt_slot) MONO_INTERNAL;
+
+guint32
+mono_method_get_imt_slot (MonoMethod *method) MONO_INTERNAL;
+
+void
+mono_method_add_generic_virtual_invocation (MonoDomain *domain, gpointer *vtable_slot,
+	MonoGenericInst *method_inst, gpointer code) MONO_INTERNAL;
+
+gpointer
+mono_method_alloc_generic_virtual_thunk (MonoDomain *domain, int size) MONO_INTERNAL;
+
+typedef enum {
+	MONO_UNHANDLED_POLICY_LEGACY,
+	MONO_UNHANDLED_POLICY_CURRENT
+} MonoRuntimeUnhandledExceptionPolicy;
+
+MonoRuntimeUnhandledExceptionPolicy
+mono_runtime_unhandled_exception_policy_get (void) MONO_INTERNAL;
+void
+mono_runtime_unhandled_exception_policy_set (MonoRuntimeUnhandledExceptionPolicy policy) MONO_INTERNAL;
+
+MonoVTable *
+mono_class_try_get_vtable (MonoDomain *domain, MonoClass *class) MONO_INTERNAL;
+
+MonoException *
+mono_runtime_class_init_full (MonoVTable *vtable, gboolean raise_exception) MONO_INTERNAL;
+
+void
+mono_method_clear_object (MonoDomain *domain, MonoMethod *method) MONO_INTERNAL;
+
 #endif /* __MONO_OBJECT_INTERNALS_H__ */
 
+
+
