@@ -5,8 +5,8 @@
  *	Dick Porter (dick@ximian.com)
  *	Gonzalo Paniagua Javier (gonzalo@ximian.com)
  *
- * (C) 2001 Ximian, Inc.
- * Copyright (c) 2005-2006 Novell, Inc. (http://www.novell.com)
+ * Copyright 2001-2003 Ximian, Inc (http://www.ximian.com)
+ * Copyright 2004-2009 Novell, Inc (http://www.novell.com)
  */
 
 #include <config.h>
@@ -14,9 +14,21 @@
 #include <glib.h>
 #include <string.h>
 #include <stdlib.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 #include <errno.h>
 
+#include <sys/types.h>
+#ifndef PLATFORM_WIN32 
+#include <sys/socket.h>
+#include <sys/ioctl.h>
+#include <netinet/in.h>
+#include <netinet/tcp.h>
+#include <netdb.h>
+#include <arpa/inet.h>
+#endif
+
 #include <mono/metadata/object.h>
 #include <mono/io-layer/io-layer.h>
 #include <mono/metadata/socket-io.h>
@@ -31,7 +43,9 @@
 #include <mono/metadata/threadpool-internals.h>
 #include <mono/metadata/domain-internals.h>
 
-#include <sys/time.h> 
+#ifdef HAVE_SYS_TIME_H
+#include <sys/time.h>
+#endif
 #ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
 #endif
@@ -64,6 +78,11 @@
 #undef AF_INET6
 #endif
 
+#ifdef PLATFORM_ANDROID
+// not yet actually implemented...
+#undef AF_INET6
+#endif
+
 #undef DEBUG
 
 static gint32 convert_family(MonoAddressFamily mono_family)
@@ -284,11 +303,18 @@ static gint32 convert_socketflags (gint32 sflags)
 		flags |= MSG_PEEK;
 	if (sflags & SocketFlags_DontRoute)
 		flags |= MSG_DONTROUTE;
+#if 0
+	/* Ignore Partial - see bug 349688.  Don't return -1, because
+	 * according to the comment in that bug ms runtime doesn't for
+	 * UDP sockets (this means we will silently ignore it for TCP
+	 * too)
+	 */
 	if (sflags & SocketFlags_Partial)
 #ifdef MSG_MORE
 		flags |= MSG_MORE;
 #else
 		return -1;	
+#endif
 #endif
 	if (sflags & SocketFlags_MaxIOVectorLength)
 		/* FIXME: Don't know what to do for MaxIOVectorLength query */
@@ -376,11 +402,23 @@ static gint32 convert_sockopt_level_and_name(MonoSocketOptionLevel mono_level,
 			break;
 #endif
 		case SocketOptionName_ExclusiveAddressUse:
+#ifdef SO_EXCLUSIVEADDRUSE
+			*system_name = SO_EXCLUSIVEADDRUSE;
+			break;
+#endif
 		case SocketOptionName_UseLoopback:
+#ifdef SO_USELOOPBACK
+			*system_name = SO_USELOOPBACK;
+			break;
+#endif
 		case SocketOptionName_MaxConnections:
-			/* Can't figure out how to map these, so fall
-			 * through
-			 */
+#ifdef SO_MAXCONN
+			*system_name = SO_MAXCONN;
+			break;
+#elif defined(SOMAXCONN)
+			*system_name = SOMAXCONN;
+			break;
+#endif
 		default:
 			g_warning("System.Net.Sockets.SocketOptionName 0x%x is not supported at Socket level", mono_name);
 			return(-1);
@@ -597,44 +635,82 @@ static gint32 convert_sockopt_level_and_name(MonoSocketOptionLevel mono_level,
 	return(0);
 }
 
-#define STASH_SYS_ASS(this) \
-	if(system_assembly == NULL) { \
-		system_assembly=mono_image_loaded ("System"); \
-		if (!system_assembly) {	\
-			MonoAssembly *sa = mono_assembly_open ("System.dll", NULL);	\
-			if (!sa) g_assert_not_reached ();	\
-			else {system_assembly = mono_assembly_get_image (sa);}	\
-		}	\
+static MonoImage *get_socket_assembly (void)
+{
+	static const char *version = NULL;
+	static gboolean moonlight;
+	static MonoImage *socket_assembly = NULL;
+	
+	if (version == NULL) {
+		version = mono_get_runtime_info ()->framework_version;
+		moonlight = !strcmp (version, "2.1");
 	}
-
-static MonoImage *system_assembly=NULL;
-
+	
+	if (socket_assembly == NULL) {
+		if (moonlight) {
+			socket_assembly = mono_image_loaded ("System.Net");
+			if (!socket_assembly) {
+				MonoAssembly *sa = mono_assembly_open ("System.Net.dll", NULL);
+			
+				if (!sa) {
+					g_assert_not_reached ();
+				} else {
+					socket_assembly = mono_assembly_get_image (sa);
+				}
+			}
+		} else {
+			socket_assembly = mono_image_loaded ("System");
+			if (!socket_assembly) {
+				MonoAssembly *sa = mono_assembly_open ("System.dll", NULL);
+			
+				if (!sa) {
+					g_assert_not_reached ();
+				} else {
+					socket_assembly = mono_assembly_get_image (sa);
+				}
+			}
+		}
+	}
+	
+	return(socket_assembly);
+}
 
 #ifdef AF_INET6
 static gint32 get_family_hint(void)
 {
-	MonoClass *socket_class;
-	MonoClassField *ipv6_field, *ipv4_field;
-	gint32 ipv6_enabled = -1, ipv4_enabled = -1;
-	MonoVTable *vtable;
+	MonoDomain *domain = mono_domain_get ();
 
-	socket_class = mono_class_from_name (system_assembly,
-					     "System.Net.Sockets", "Socket");
-	ipv4_field = mono_class_get_field_from_name (socket_class,
-						     "ipv4Supported");
-	ipv6_field = mono_class_get_field_from_name (socket_class,
-						     "ipv6Supported");
-	vtable = mono_class_vtable (mono_domain_get (), socket_class);
+	if (!domain->inet_family_hint) {
+		MonoClass *socket_class;
+		MonoClassField *ipv6_field, *ipv4_field;
+		gint32 ipv6_enabled = -1, ipv4_enabled = -1;
+		MonoVTable *vtable;
 
-	mono_field_static_get_value(vtable, ipv4_field, &ipv4_enabled);
-	mono_field_static_get_value(vtable, ipv6_field, &ipv6_enabled);
+		socket_class = mono_class_from_name (get_socket_assembly (), "System.Net.Sockets", "Socket");
+		ipv4_field = mono_class_get_field_from_name (socket_class, "ipv4Supported");
+		ipv6_field = mono_class_get_field_from_name (socket_class, "ipv6Supported");
+		vtable = mono_class_vtable (mono_domain_get (), socket_class);
+		mono_runtime_class_init (vtable);
 
-	if(ipv4_enabled == 1 && ipv6_enabled == 1) {
-		return(PF_UNSPEC);
-	} else if(ipv4_enabled == 1) {
-		return(PF_INET);
-	} else {
-		return(PF_INET6);
+		mono_field_static_get_value (vtable, ipv4_field, &ipv4_enabled);
+		mono_field_static_get_value (vtable, ipv6_field, &ipv6_enabled);
+
+		mono_domain_lock (domain);
+		if (ipv4_enabled == 1 && ipv6_enabled == 1) {
+			domain->inet_family_hint = 1;
+		} else if (ipv4_enabled == 1) {
+			domain->inet_family_hint = 2;
+		} else {
+			domain->inet_family_hint = 3;
+		}
+		mono_domain_unlock (domain);
+	}
+	switch (domain->inet_family_hint) {
+	case 1: return PF_UNSPEC;
+	case 2: return PF_INET;
+	case 3: return PF_INET6;
+	default:
+		return PF_UNSPEC;
 	}
 }
 #endif
@@ -648,8 +724,6 @@ gpointer ves_icall_System_Net_Sockets_Socket_Socket_internal(MonoObject *this, g
 	
 	MONO_ARCH_SAVE_REGS;
 
-	STASH_SYS_ASS(this);
-	
 	*error = 0;
 	
 	sock_family=convert_family(family);
@@ -688,33 +762,6 @@ gpointer ves_icall_System_Net_Sockets_Socket_Socket_internal(MonoObject *this, g
 	}
 #endif
 
-#ifndef PLATFORM_WIN32
-	/* .net seems to set this by default for SOCK_STREAM,
-	 * not for SOCK_DGRAM (see bug #36322)
-	 *
-	 * It seems winsock has a rather different idea of what
-	 * SO_REUSEADDR means.  If it's set, then a new socket can be
-	 * bound over an existing listening socket.  There's a new
-	 * windows-specific option called SO_EXCLUSIVEADDRUSE but
-	 * using that means the socket MUST be closed properly, or a
-	 * denial of service can occur.  Luckily for us, winsock
-	 * behaves as though any other system would when SO_REUSEADDR
-	 * is true, so we don't need to do anything else here.  See
-	 * bug 53992.
-	 */
-	{
-	int ret, true = 1;
-	
-	ret = _wapi_setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &true, sizeof (true));
-	if(ret==SOCKET_ERROR) {
-		*error = WSAGetLastError ();
-		
-		closesocket(sock);
-		return(NULL);
-	}
-	}
-#endif
-	
 	return(GUINT_TO_POINTER (sock));
 }
 
@@ -753,13 +800,20 @@ gint32 ves_icall_System_Net_Sockets_Socket_Available_internal(SOCKET sock,
 							      gint32 *error)
 {
 	int ret;
-	gulong amount;
+	int amount;
 	
 	MONO_ARCH_SAVE_REGS;
 
 	*error = 0;
 	
+#if defined(PLATFORM_MACOSX)
+	{
+		socklen_t optlen = sizeof (amount);
+		ret=getsockopt (sock, SOL_SOCKET, SO_NREAD, &amount, &optlen);
+	}
+#else
 	ret=ioctlsocket(sock, FIONREAD, &amount);
+#endif
 	if(ret==SOCKET_ERROR) {
 		*error = WSAGetLastError ();
 		return(0);
@@ -791,13 +845,51 @@ void ves_icall_System_Net_Sockets_Socket_Blocking_internal(SOCKET sock,
 }
 
 gpointer ves_icall_System_Net_Sockets_Socket_Accept_internal(SOCKET sock,
-							     gint32 *error)
+							     gint32 *error,
+							     gboolean blocking)
 {
 	SOCKET newsock;
 	
 	MONO_ARCH_SAVE_REGS;
 
 	*error = 0;
+
+#ifdef PLATFORM_WIN32
+	/* Several applications are getting stuck during shutdown on Windows 
+	 * when an accept call is on a background thread.
+	 * 
+	 */
+	if (blocking) {
+		MonoThread* curthread = mono_thread_current ();
+
+		if (curthread) {
+			for (;;) {
+				int selectret;
+				int optlen = sizeof (gint);
+				TIMEVAL timeout; 
+				fd_set readfds;
+				FD_ZERO (&readfds);
+				FD_SET(sock, &readfds);
+				timeout.tv_sec = 0;
+				timeout.tv_usec = 1000;
+				selectret = select (0, &readfds, NULL, NULL, &timeout);
+				if (selectret > 0)
+					break;
+				if (curthread->state & ThreadState_StopRequested)
+					return NULL;
+				/* A negative return from select means that an error has occurred.
+				 * Let _wapi_accept handle that.*/
+				if (selectret != 0)
+					break;
+				/* The socket's state may have changed.  If it is no longer listening, stop.*/
+				if (!getsockopt (sock, SOL_SOCKET, SO_ACCEPTCONN, (char*)&selectret, &optlen)) {
+					if (!selectret)
+						break;
+				}
+			}
+		}
+	}
+#endif
 	
 	newsock = _wapi_accept (sock, NULL, 0);
 	if(newsock==INVALID_SOCKET) {
@@ -835,7 +927,7 @@ static MonoObject *create_object_from_sockaddr(struct sockaddr *saddr,
 	MonoAddressFamily family;
 
 	/* Build a System.Net.SocketAddress object instance */
-	sockaddr_class=mono_class_from_name(system_assembly, "System.Net", "SocketAddress");
+	sockaddr_class=mono_class_from_name(get_socket_assembly (), "System.Net", "SocketAddress");
 	sockaddr_obj=mono_object_new(domain, sockaddr_class);
 	
 	/* Locate the SocketAddress data buffer in the object */
@@ -1179,10 +1271,7 @@ ves_icall_System_Net_Sockets_Socket_Poll_internal (SOCKET sock, gint mode,
 				thread = mono_thread_current ();
 			}
 			
-			mono_monitor_enter (thread->synch_lock);
-			leave = ((thread->state & ThreadState_AbortRequested) != 0 ||
-				 (thread->state & ThreadState_StopRequested) != 0);
-			mono_monitor_exit (thread->synch_lock);
+			leave = mono_thread_test_state (thread, ThreadState_AbortRequested | ThreadState_StopRequested);
 			
 			if (leave != 0) {
 				g_free (pfds);
@@ -1360,6 +1449,35 @@ gint32 ves_icall_System_Net_Sockets_Socket_Receive_internal(SOCKET sock, MonoArr
 	return(ret);
 }
 
+gint32 ves_icall_System_Net_Sockets_Socket_Receive_array_internal(SOCKET sock, MonoArray *buffers, gint32 flags, gint32 *error)
+{
+	int ret, count;
+	DWORD recv;
+	WSABUF *wsabufs;
+	DWORD recvflags = 0;
+	
+	MONO_ARCH_SAVE_REGS;
+
+	*error = 0;
+	
+	wsabufs = mono_array_addr (buffers, WSABUF, 0);
+	count = mono_array_length (buffers);
+	
+	recvflags = convert_socketflags (flags);
+	if (recvflags == -1) {
+		*error = WSAEOPNOTSUPP;
+		return(0);
+	}
+	
+	ret = WSARecv (sock, wsabufs, count, &recv, &recvflags, NULL, NULL);
+	if (ret == SOCKET_ERROR) {
+		*error = WSAGetLastError ();
+		return(0);
+	}
+	
+	return(recv);
+}
+
 gint32 ves_icall_System_Net_Sockets_Socket_RecvFrom_internal(SOCKET sock, MonoArray *buffer, gint32 offset, gint32 count, gint32 flags, MonoObject **sockaddr, gint32 *error)
 {
 	int ret;
@@ -1453,6 +1571,35 @@ gint32 ves_icall_System_Net_Sockets_Socket_Send_internal(SOCKET sock, MonoArray
 	return(ret);
 }
 
+gint32 ves_icall_System_Net_Sockets_Socket_Send_array_internal(SOCKET sock, MonoArray *buffers, gint32 flags, gint32 *error)
+{
+	int ret, count;
+	DWORD sent;
+	WSABUF *wsabufs;
+	DWORD sendflags = 0;
+	
+	MONO_ARCH_SAVE_REGS;
+
+	*error = 0;
+	
+	wsabufs = mono_array_addr (buffers, WSABUF, 0);
+	count = mono_array_length (buffers);
+	
+	sendflags = convert_socketflags (flags);
+	if (sendflags == -1) {
+		*error = WSAEOPNOTSUPP;
+		return(0);
+	}
+	
+	ret = WSASend (sock, wsabufs, count, &sent, sendflags, NULL, NULL);
+	if (ret == SOCKET_ERROR) {
+		*error = WSAGetLastError ();
+		return(0);
+	}
+	
+	return(sent);
+}
+
 gint32 ves_icall_System_Net_Sockets_Socket_SendTo_internal(SOCKET sock, MonoArray *buffer, gint32 offset, gint32 count, gint32 flags, MonoObject *sockaddr, gint32 *error)
 {
 	int ret;
@@ -1508,7 +1655,7 @@ static SOCKET Socket_to_SOCKET(MonoObject *sockobj)
 	MonoClassField *field;
 	
 	field=mono_class_get_field_from_name(sockobj->vtable->klass, "socket");
-	sock=*(SOCKET *)(((char *)sockobj)+field->offset);
+	sock=GPOINTER_TO_INT (*(gpointer *)(((char *)sockobj)+field->offset));
 
 	return(sock);
 }
@@ -1526,6 +1673,7 @@ void ves_icall_System_Net_Sockets_Socket_Select_internal(MonoArray **sockets, gi
 	MonoClass *sock_arr_class;
 	MonoArray *socks;
 	time_t start;
+	mono_array_size_t socks_size;
 	
 	MONO_ARCH_SAVE_REGS;
 
@@ -1548,7 +1696,7 @@ void ves_icall_System_Net_Sockets_Socket_Select_internal(MonoArray **sockets, gi
 			return;
 		}
 
-		pfds [idx].fd = GPOINTER_TO_INT (Socket_to_SOCKET (obj));
+		pfds [idx].fd = Socket_to_SOCKET (obj);
 		pfds [idx].events = (mode == 0) ? MONO_POLLIN : (mode == 1) ? MONO_POLLOUT : POLL_ERRORS;
 		idx++;
 	}
@@ -1573,10 +1721,8 @@ void ves_icall_System_Net_Sockets_Socket_Select_internal(MonoArray **sockets, gi
 			if (thread == NULL)
 				thread = mono_thread_current ();
 
-			mono_monitor_enter (thread->synch_lock);
-			leave = ((thread->state & ThreadState_AbortRequested) != 0 || 
-				 (thread->state & ThreadState_StopRequested) != 0);
-			mono_monitor_exit (thread->synch_lock);
+			leave = mono_thread_test_state (thread, ThreadState_AbortRequested | ThreadState_StopRequested);
+			
 			if (leave != 0) {
 				g_free (pfds);
 				*sockets = NULL;
@@ -1606,9 +1752,9 @@ void ves_icall_System_Net_Sockets_Socket_Select_internal(MonoArray **sockets, gi
 	}
 
 	sock_arr_class= ((MonoObject *)*sockets)->vtable->klass;
-	ret += 3; /* space for the NULL delimiters */
-	socks = mono_array_new_full (mono_domain_get (), sock_arr_class, (guint32*)&ret, NULL);
-	ret -= 3;
+	socks_size = ((mono_array_size_t)ret) + 3; /* space for the NULL delimiters */
+	socks = mono_array_new_full (mono_domain_get (), sock_arr_class, &socks_size, NULL);
+
 	mode = idx = 0;
 	for (i = 0; i < count && ret > 0; i++) {
 		mono_pollfd *pfd;
@@ -1715,7 +1861,7 @@ void ves_icall_System_Net_Sockets_Socket_GetSocketOption_obj_internal(SOCKET soc
 	switch(name) {
 	case SocketOptionName_Linger:
 		/* build a System.Net.Sockets.LingerOption */
-		obj_class=mono_class_from_name(system_assembly,
+		obj_class=mono_class_from_name(get_socket_assembly (),
 					       "System.Net.Sockets",
 					       "LingerOption");
 		obj=mono_object_new(domain, obj_class);
@@ -2210,29 +2356,32 @@ static gboolean hostent_to_IPHostEntry(struct hostent *he, MonoString **h_name,
 				       gboolean add_local_ips)
 {
 	MonoDomain *domain = mono_domain_get ();
-	int i;
+	int i = 0;
 	struct in_addr *local_in = NULL;
 	int nlocal_in = 0;
 
-	if(he->h_length!=4 || he->h_addrtype!=AF_INET) {
-		return(FALSE);
-	}
+	if (he != NULL) {
+		if(he->h_length!=4 || he->h_addrtype!=AF_INET) {
+			return(FALSE);
+		}
 
-	*h_name=mono_string_new(domain, he->h_name);
+		*h_name=mono_string_new(domain, he->h_name);
 
-	i=0;
-	while(he->h_aliases[i]!=NULL) {
-		i++;
-	}
-	
-	*h_aliases=mono_array_new(domain, mono_get_string_class (), i);
-	i=0;
-	while(he->h_aliases[i]!=NULL) {
-		MonoString *alias;
+		while(he->h_aliases[i]!=NULL) {
+			i++;
+		}
 		
-		alias=mono_string_new(domain, he->h_aliases[i]);
-		mono_array_setref (*h_aliases, i, alias);
-		i++;
+		*h_aliases=mono_array_new(domain, mono_get_string_class (), i);
+		i=0;
+		while(he->h_aliases[i]!=NULL) {
+			MonoString *alias;
+			
+			alias=mono_string_new(domain, he->h_aliases[i]);
+			mono_array_setref (*h_aliases, i, alias);
+			i++;
+		}
+	} else if (!add_local_ips) {
+		return FALSE;
 	}
 
 	if (add_local_ips) {
@@ -2256,10 +2405,15 @@ static gboolean hostent_to_IPHostEntry(struct hostent *he, MonoString **h_name,
 			}
 
 			g_free (local_in);
+		} else if (he == NULL) {
+			/* If requesting "" and there are no other interfaces up, MS returns 127.0.0.1 */
+			*h_addr_list = mono_array_new(domain, mono_get_string_class (), 1);
+			mono_array_setref (*h_addr_list, 0, mono_string_new (domain, "127.0.0.1"));
+			return TRUE;
 		}
 	}
 	
-	if (nlocal_in == 0) {
+	if (nlocal_in == 0 && he != NULL) {
 		i = 0;
 		while (he->h_addr_list[i]!=NULL) {
 			i++;
@@ -2315,10 +2469,6 @@ static gboolean hostent_to_IPHostEntry2(struct hostent *he1,struct hostent *he2,
 
 	family_hint = get_family_hint ();
 
-	if(he1 == NULL && he2 == NULL) {
-		return(FALSE);
-	}
-
 	/*
 	 * Check if address length and family are correct
 	 */
@@ -2372,7 +2522,7 @@ static gboolean hostent_to_IPHostEntry2(struct hostent *he1,struct hostent *he2,
 			mono_array_setref (*h_aliases, i, alias);
 			i++;
 		}
-	} else {
+	} else if (!add_local_ips) {
 		return(FALSE);
 	}
 
@@ -2450,6 +2600,13 @@ static gboolean hostent_to_IPHostEntry2(struct hostent *he1,struct hostent *he2,
 			g_free (local_in);
 			g_free (local_in6);
 			return TRUE;
+		} else if (he1 == NULL && he2 == NULL) {
+			/* If requesting "" and there are no other interfaces up, MS returns 127.0.0.1 */
+			*h_addr_list = mono_array_new(domain, mono_get_string_class (), 1);
+			mono_array_setref (*h_addr_list, 0, mono_string_new (domain, "127.0.0.1"));
+			g_free (local_in);
+			g_free (local_in6);
+			return TRUE;
 		}
 
 		g_free (local_in);
@@ -2599,8 +2756,12 @@ addrinfo_to_IPHostEntry(struct addrinfo *info, MonoString **h_name,
 
 		mono_array_setref (*h_addr_list, addr_index, addr_string);
 
-		if(!i && ai->ai_canonname != NULL) {
-			*h_name=mono_string_new(domain, ai->ai_canonname);
+		if(!i) {
+			if (ai->ai_canonname != NULL) {
+				*h_name=mono_string_new(domain, ai->ai_canonname);
+			} else {
+				*h_name=mono_string_new(domain, buffer);
+			}
 		}
 
 		addr_index++;
@@ -2628,8 +2789,10 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 	MONO_ARCH_SAVE_REGS;
 	
 	hostname=mono_string_to_utf8 (host);
+	if (*hostname == '\0')
+		add_local_ips = TRUE;
 #ifdef HAVE_SIOCGIFCONF
-	if (gethostname (this_hostname, sizeof (this_hostname)) != -1) {
+	if (!add_local_ips && gethostname (this_hostname, sizeof (this_hostname)) != -1) {
 		if (!strcmp (hostname, this_hostname))
 			add_local_ips = TRUE;
 	}
@@ -2640,7 +2803,7 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_flags = AI_CANONNAME;
 
-	if (getaddrinfo(hostname, NULL, &hints, &info) == -1) {
+	if (*hostname && getaddrinfo(hostname, NULL, &hints, &info) == -1) {
 		return(FALSE);
 	}
 	
@@ -2658,9 +2821,11 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 	MONO_ARCH_SAVE_REGS;
 	
 	hostname=mono_string_to_utf8 (host);
+	if (*hostname == '\0')
+		add_local_ips = TRUE;
 
 #ifdef HAVE_SIOCGIFCONF
-	if (gethostname (this_hostname, sizeof (this_hostname)) != -1) {
+	if (!add_local_ips && gethostname (this_hostname, sizeof (this_hostname)) != -1) {
 		if (!strcmp (hostname, this_hostname))
 			add_local_ips = TRUE;
 	}
@@ -2671,13 +2836,15 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 	buffer1 = g_malloc0(buffer_size1);
 	buffer2 = g_malloc0(buffer_size2);
 
-	while (gethostbyname2_r(hostname, AF_INET, &he1, buffer1, buffer_size1,
+	hp1 = NULL;
+	hp2 = NULL;
+	while (*hostname && gethostbyname2_r(hostname, AF_INET, &he1, buffer1, buffer_size1,
 				&hp1, &herr) == ERANGE) {
 		buffer_size1 *= 2;
 		buffer1 = g_realloc(buffer1, buffer_size1);
 	}
 
-	if (hp1 == NULL)
+	if (*hostname && hp1 == NULL)
 	{
 		while (gethostbyname2_r(hostname, AF_INET6, &he2, buffer2,
 					buffer_size2, &hp2, &herr) == ERANGE) {
@@ -2685,8 +2852,6 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 			buffer2 = g_realloc(buffer2, buffer_size2);
 		}
 	}
-	else
-		hp2 = NULL;
 
 	return_value = hostent_to_IPHostEntry2(hp1, hp2, h_name, h_aliases,
 					       h_addr_list, add_local_ips);
@@ -2711,19 +2876,22 @@ MonoBoolean ves_icall_System_Net_Dns_GetHostByName_internal(MonoString *host, Mo
 	MONO_ARCH_SAVE_REGS;
 
 	hostname=mono_string_to_utf8(host);
+	if (*hostname == '\0')
+		add_local_ips = TRUE;
 #ifdef HAVE_SIOCGIFCONF
-	if (gethostname (this_hostname, sizeof (this_hostname)) != -1) {
+	if (!add_local_ips && gethostname (this_hostname, sizeof (this_hostname)) != -1) {
 		if (!strcmp (hostname, this_hostname))
 			add_local_ips = TRUE;
 	}
 #endif
 
-	he = _wapi_gethostbyname (hostname);
+	he = NULL;
+	if (*hostname)
+		he = _wapi_gethostbyname (hostname);
 	g_free(hostname);
 
-	if(he==NULL) {
+	if (*hostname && he==NULL)
 		return(FALSE);
-	}
 
 	return(hostent_to_IPHostEntry(he, h_name, h_aliases, h_addr_list, add_local_ips));
 }
@@ -2771,7 +2939,6 @@ inet_pton (int family, const char *address, void *inaddrp)
 extern MonoBoolean ves_icall_System_Net_Dns_GetHostByAddr_internal(MonoString *addr, MonoString **h_name, MonoArray **h_aliases, MonoArray **h_addr_list)
 {
 	char *address;
-	const char *version;
 	gboolean v1;
 	
 #ifdef AF_INET6
@@ -2787,10 +2954,7 @@ extern MonoBoolean ves_icall_System_Net_Dns_GetHostByAddr_internal(MonoString *a
 	gboolean ret;
 #endif
 
-	MONO_ARCH_SAVE_REGS;
-
-	version = mono_get_runtime_info ()->framework_version;
-	v1 = (version[0] == '1');
+	v1 = mono_framework_version () == 1;
 
 	address = mono_string_to_utf8 (addr);
 
@@ -2817,12 +2981,18 @@ extern MonoBoolean ves_icall_System_Net_Dns_GetHostByAddr_internal(MonoString *a
 	}
 	
 	if(family == AF_INET) {
+#if HAVE_SOCKADDR_IN_SIN_LEN
+		saddr.sin_len = sizeof (saddr);
+#endif
 		if(getnameinfo ((struct sockaddr*)&saddr, sizeof(saddr),
 				hostname, sizeof(hostname), NULL, 0,
 				flags) != 0) {
 			return(FALSE);
 		}
 	} else if(family == AF_INET6) {
+#if HAVE_SOCKADDR_IN6_SIN_LEN
+		saddr6.sin6_len = sizeof (saddr6);
+#endif
 		if(getnameinfo ((struct sockaddr*)&saddr6, sizeof(saddr6),
 				hostname, sizeof(hostname), NULL, 0,
 				flags) != 0) {
