@@ -1,5 +1,6 @@
 using System;
 using System.Reflection;
+using System.Runtime.CompilerServices;
 
 /*
  * Regression tests for the mono JIT.
@@ -2207,6 +2208,30 @@ class Tests {
 		return i;
 	}
 
+	public static void rethrow () {
+		try {
+			throw new ApplicationException();
+		} catch (ApplicationException) {
+			try {
+				throw new OverflowException();
+			} catch (Exception) {
+				throw;
+			}
+		}
+	}
+
+	// Test that a rethrow rethrows the correct exception
+	public static int test_0_rethrow_nested () {
+		try {
+			rethrow ();
+		} catch (OverflowException) {
+			return 0;
+		} catch (Exception) {
+			return 1;
+		}
+		return 2;
+	}
+
 	/* MarshalByRefObject prevents the methods from being inlined */
 	class ThrowClass : MarshalByRefObject {
 		public static void rethrow1 () {
@@ -2352,6 +2377,39 @@ class Tests {
 		return 1;
 	}
 
+	public int amethod () {
+		return 1;
+	}
+
+	public static int test_0_nonvirt_nullref_at_clause_start () {
+		Tests t = null;
+		try {
+			t.amethod ();
+		} catch (NullReferenceException) {
+			return 0;
+		}
+
+		return 1;
+	}
+
+	public static int throw_only () {
+		throw new Exception ();
+	}
+
+	[MethodImpl(MethodImplOptions.NoInlining)] 
+	public static int throw_only2 () {
+		return throw_only ();
+	}
+
+	public static int test_0_inline_throw_only () {
+		try {
+			return throw_only2 ();
+		}
+		catch (Exception ex) {
+			return 0;
+		}
+	}
+
 	// bug #78633
 	public static int test_0_throw_to_branch_opt_outer_clause () {
 		int i = 0;
