@@ -117,6 +117,33 @@ class Tests {
 		return ldelem_any (arr);
 	}
 
+	public static T return_ref<T> (ref T t) {
+		return t;
+	}
+
+	public static T ldelema_any<T> (T[] arr) {
+		return return_ref<T> (ref arr [0]);
+	}
+
+	public static int test_0_ldelema () {
+		string[] arr = new string [1];
+
+		arr [0] = "Hello";
+
+		if (ldelema_any <string> (arr) == "Hello")
+			return 0;
+		else
+			return 1;
+	}
+
+	public static T[,] newarr_multi<T> () {
+		return new T [1, 1];
+	}
+
+	public static int test_0_newarr_multi_dim () {
+		return newarr_multi<string> ().GetType () == typeof (string[,]) ? 0 : 1;
+	}
+
 	interface ITest
 	{
 		void Foo<T> ();
@@ -213,6 +240,10 @@ class Tests {
 		return 0;
 	}
 
+	public static int test_0_nullable_ldflda () {
+		return GenericClass<string>.BIsAClazz == false ? 0 : 1;
+	}
+
 	public struct GenericStruct<T> {
 		public T t;
 
@@ -258,6 +289,15 @@ class Tests {
 		{
 			return x [index];
 		}
+
+        protected static T NullB = default(T);       
+        private static Nullable<bool>  _BIsA = null;
+        public static bool BIsAClazz {
+            get {
+                _BIsA = false;
+                return _BIsA.Value;
+            }
+        }
 	}
 
 	public class MRO : MarshalByRefObject {
@@ -356,6 +396,11 @@ class Tests {
 		return the_type == typeof (string) ? 0 : 1;
 	}
 
+	public static int test_0_throw_dead_this () {
+        new Foo<string> ("").throw_dead_this ();
+		return 0;
+	}
+
 	public static Type the_type;
 
 	public void ldvirtftn<T> () {
@@ -392,6 +437,14 @@ class Tests {
 			}
 		}
 
+		public void throw_dead_this () {
+			try {
+				new SomeClass().ThrowAnException();
+			}
+			catch {
+			}
+		}
+
 		public T1 get_default () {
 			return default (T1);
 		}
@@ -408,6 +461,12 @@ class Tests {
 
 	}
 
+	public class SomeClass {
+		public void ThrowAnException() {
+			throw new Exception ("Something went wrong");
+		}
+	}		
+
 	public interface IMyHandler {
 		object Bar<T>();
 	}
