@@ -5,7 +5,7 @@
 #include "mini-x86.h"
 #elif defined(__x86_64__)
 #include "mini-amd64.h"
-#elif defined(__ppc__) || defined(__powerpc__)
+#elif defined(__mono_ppc__)
 #include "mini-ppc.h"
 #elif defined(__sparc__) || defined(sparc)
 #include "mini-sparc.h"
@@ -23,6 +23,8 @@
 #include "mini-alpha.h"
 #elif defined(__mips__)
 #include "mini-mips.h"
+#elif defined(__hppa__)
+#include "mini-hppa.h"
 #else
 #error add arch specific include file in mini-arch.h
 #endif
