@@ -106,7 +106,7 @@ class Tests {
 		return TestDriver.RunTests (typeof (Tests));
 	}
 	
-	static int test_0_return () {
+	public static int test_0_return () {
 		Simple s;
 		s.a = 1;
 		s.b = 2;
@@ -115,14 +115,14 @@ class Tests {
 		return s.a - 1;
 	}
 
-	static int test_0_string_access () {
+	public static int test_0_string_access () {
 		string s = "Hello";
 		if (s [1] != 'e')
 			return 1;
 		return 0;
 	}
 
-	static int test_0_string_virtual_call () {
+	public static int test_0_string_virtual_call () {
 		string s = "Hello";
 		string s2 = s.ToString ();
 		if (s2 [1] != 'e')
@@ -130,23 +130,23 @@ class Tests {
 		return 0;
 	}
 
-	static int test_0_iface_call () {
+	public static int test_0_iface_call () {
 		string s = "Hello";
 		object o = ((ICloneable)s).Clone ();
 		return 0;
 	}
 
-	static int test_5_newobj () {
+	public static int test_5_newobj () {
 		Sample s = new Sample (5);
 		return s.a;
 	}
 
-	static int test_4_box () {
+	public static int test_4_box () {
 		object obj = 4;
 		return (int)obj;
 	}
 
-	static int test_0_enum_unbox () {
+	public static int test_0_enum_unbox () {
 		SampleEnum x = SampleEnum.A;
 		object o = x;
 		
@@ -167,7 +167,7 @@ class Tests {
 		return r;
 	}
 
-	static int test_3_return_struct () {
+	public static int test_3_return_struct () {
 		Simple v = get_simple (1);
 
 		if (v.a != 1)
@@ -186,7 +186,7 @@ class Tests {
 		return get_simple (v);
 	}
 	
-	static int test_2_return_struct_virtual () {
+	public static int test_2_return_struct_virtual () {
 		Tests t = new Tests ();
 		Simple v = t.v_get_simple (2);
 
@@ -217,7 +217,7 @@ class Tests {
 		return 0;
 	}
 	
-	static int test_5_pass_struct () {
+	public static int test_5_pass_struct () {
 		Simple v = get_simple (1);
 		if (receive_simple (7, v, 9) != 0)
 			return 0;
@@ -227,7 +227,7 @@ class Tests {
 	}
 
 	static Simple s_v;
-	static int test_5_pass_static_struct () {
+	public static int test_5_pass_static_struct () {
 		s_v = get_simple (1);
 		if (receive_simple (7, s_v, 9) != 0)
 			return 0;
@@ -265,7 +265,7 @@ class Tests {
 		return 0;
 	}
 
-	static int test_5_pass_small_struct () {
+	public static int test_5_pass_small_struct () {
 		Small v = get_small (1);
 		if (receive_small (7, v, 9) != 0)
 			return 0;
@@ -313,7 +313,7 @@ class Tests {
 		return check_struct1(x);
 	}
 
-	static int test_0_struct1_args () {
+	public static int test_0_struct1_args () {
 		int r;
 		struct1 x;
 
@@ -369,7 +369,7 @@ class Tests {
 		return check_struct2(x);
 	}
 
-	static int test_0_struct2_args () {
+	public static int test_0_struct2_args () {
 		int r;
 		struct2 x;
 
@@ -387,6 +387,54 @@ class Tests {
 		return 0;
 	}
 
+	// 128 bits
+	struct Struct3 {
+		public long i, j, k, l;
+	}
+
+	static int pass_struct3 (int i, int j, int k, int l, int m, int n, int o, int p, Struct3 s, int q) {
+		if (s.i + s.j + s.k + s.l != 10)
+			return 1;
+		else
+			return 0;
+	}
+
+	public static int test_0_struct3_args () {
+		Struct3 s = new Struct3 ();
+		s.i = 1;
+		s.j = 2;
+		s.k = 3;
+		s.l = 4;
+
+		return pass_struct3 (1, 2, 3, 4, 5, 6, 7, 8, s, 9);
+	}
+
+	// Struct with unaligned size on 64 bit machines
+	struct Struct4 {
+        public int i, j, k, l, m;
+		public int i1, i2, i3, i4, i5, i6;
+	}
+
+	static int pass_struct4 (Struct4 s) {
+		if (s.i + s.j + s.k + s.l + s.m != 15)
+			return 1;
+		else
+			return 0;
+	}
+
+	public static int test_0_struct4_args () {
+		Struct4 s = new Struct4 ();
+		s.i = 1;
+		s.j = 2;
+		s.k = 3;
+		s.l = 4;
+		s.m = 5;
+
+		return pass_struct4 (s);
+	}
+
+
+
 	struct AStruct {
 		public int i;
 
@@ -400,13 +448,13 @@ class Tests {
 	}
 
 	// Test that vtypes are unboxed during a virtual call
-	static int test_44_unbox_trampoline () {
+	public static int test_44_unbox_trampoline () {
 		AStruct s = new AStruct (44);
 		object o = s;
 		return o.GetHashCode ();
 	}
 
-	static int test_0_unbox_trampoline2 () {
+	public static int test_0_unbox_trampoline2 () {
 		int i = 12;
 		object o = i;
 			
@@ -420,7 +468,7 @@ class Tests {
 	}
 
 	// Test fields with big offsets
-	static int test_0_fields_with_big_offsets () {
+	public static int test_0_fields_with_big_offsets () {
 		StructWithBigOffsets s = new StructWithBigOffsets ();
 		StructWithBigOffsets s2 = new StructWithBigOffsets ();
 
@@ -490,7 +538,7 @@ class Tests {
 
 	}
 
-	static int test_0_seektest () {
+	public static int test_0_seektest () {
 		TestRegA t = new TestRegA ();
 		return (int)t.Seek (0);
 	}
@@ -503,7 +551,7 @@ class Tests {
 	class Duper: Super {
 	}
 
-	static int test_0_null_cast () {
+	public static int test_0_null_cast () {
 		object o = null;
 
 		Super s = (Super)o;
@@ -511,7 +559,7 @@ class Tests {
 		return 0;
 	}
 	
-	static int test_0_super_cast () {
+	public static int test_0_super_cast () {
 		Duper d = new Duper ();
 		Super sup = d;
 		Object o = d;
@@ -533,7 +581,7 @@ class Tests {
 		return 0;
 	}
 
-	static int test_0_super_cast_array () {
+	public static int test_0_super_cast_array () {
 		Duper[] d = new Duper [0];
 		Super[] sup = d;
 		Object[] o = d;
@@ -555,7 +603,7 @@ class Tests {
 		return 0;
 	}
 
-	static int test_0_multi_array_cast () {
+	public static int test_0_multi_array_cast () {
 		Duper[,] d = new Duper [1, 1];
 		object[,] o = d;
 
@@ -569,7 +617,7 @@ class Tests {
 		return 0;
 	}
 
-	static int test_0_vector_array_cast () {
+	public static int test_0_vector_array_cast () {
 		Array arr1 = Array.CreateInstance (typeof (int), new int[] {1}, new int[] {0});
 		Array arr2 = Array.CreateInstance (typeof (int), new int[] {1}, new int[] {10});
 
@@ -593,10 +641,14 @@ class Tests {
 		if (arr2 is int[])
 			return 4;
 
+		int [,] [] arr3 = new int [1, 1] [];
+		object o = arr3;
+		int [,] [] arr4 = (int [,] [])o;
+
 		return 0;
 	}
 
-	static int test_0_enum_array_cast () {
+	public static int test_0_enum_array_cast () {
 		TypeCode[] tc = new TypeCode [0];
 		object[] oa;
 		ValueType[] vta;
@@ -637,7 +689,7 @@ class Tests {
 		return 0;
 	}
 
-	static int test_0_more_cast_corner_cases () {
+	public static int test_0_more_cast_corner_cases () {
 		ValueType[] vta = new ValueType [0];
 		Enum[] ea = new Enum [0];
 		Array a = vta;
@@ -715,7 +767,7 @@ class Tests {
 		return 0;
 	}
 
-	static int test_0_cast_iface_array () {
+	public static int test_0_cast_iface_array () {
 		object o = new ICloneable [0];
 		object o2 = new Duper [0];
 		object t;
@@ -771,7 +823,7 @@ class Tests {
 		return ((day-1) + temp + (365* (year-1)) + ((year-1)/4) - ((year-1)/100) + ((year-1)/400));
 	}
 
-	static int test_719162_complex_div () {
+	public static int test_719162_complex_div () {
 		int adays = AbsoluteDays (1970, 1, 1);
 		return adays;
 	}
@@ -786,7 +838,7 @@ class Tests {
 		return 5;
 	}
 
-	static int test_2_static_delegate () {
+	public static int test_2_static_delegate () {
 		GetIntDel del = new GetIntDel (return4);
 		int v = del ();
 		if (v != 4)
@@ -794,7 +846,7 @@ class Tests {
 		return 2;
 	}
 
-	static int test_2_instance_delegate () {
+	public static int test_2_instance_delegate () {
 		Tests t = new Tests ();
 		GetIntDel del = new GetIntDel (t.return5);
 		int v = del ();
@@ -803,7 +855,7 @@ class Tests {
 		return 2;
 	}
 
-	static int test_1_store_decimal () {
+	public static int test_1_store_decimal () {
 		decimal[,] a = {{1}};
 
 		if (a[0,0] != 1m)
@@ -811,7 +863,7 @@ class Tests {
 		return 1;
 	}
 
-	static int test_2_intptr_stobj () {
+	public static int test_2_intptr_stobj () {
 		System.IntPtr [] arr = { new System.IntPtr () };
 
 		if (arr [0] != (System.IntPtr)0)
@@ -827,7 +879,7 @@ class Tests {
 	 * Test that evaluation of complex arguments does not overwrite the
 	 * arguments already in outgoing registers.
 	 */
-	static int test_155_regalloc () {
+	public static int test_155_regalloc () {
 		int a = 10;
 		int b = 10;
 
@@ -847,7 +899,7 @@ class Tests {
 		return true;
 	}
 
-	static int test_2_large_struct_pass ()
+	public static int test_2_large_struct_pass ()
 	{
 		Large a, b, c, d;
 		a = new Large ();
@@ -968,7 +1020,7 @@ class Tests {
 
 
 	static uint dum_de_dum = 1;
-	static int test_0_long_arg_opt ()
+	public static int test_0_long_arg_opt ()
 	{
 		return Foo (0x1234567887654321, dum_de_dum);
 	}
@@ -984,7 +1036,7 @@ class Tests {
 		return 0;
 	}
 	
-	static int test_0_long_ret_opt ()
+	public static int test_0_long_ret_opt ()
 	{
 		ulong x = X ();
 		if (x != 0x1234567887654321)
@@ -1012,7 +1064,7 @@ class Tests {
 		counter = 10;
 		return true;
 	}
-	static int test_0_cond_branch_side_effects () {
+	public static int test_0_cond_branch_side_effects () {
 		counter = 5;
 		if (WriteStuff());
 		if (counter == 10)
@@ -1064,7 +1116,7 @@ ncells ) {
 		}
 	}
 
-	static int test_0_calls_opcode_emulation () {
+	public static int test_0_calls_opcode_emulation () {
 		// Test that emulated opcodes do not clobber arguments already in
 		// out registers
 		checked {
@@ -1074,25 +1126,25 @@ ncells ) {
 		return 0;
 	}
 
-	static int test_0_intrins_string_length () {
+	public static int test_0_intrins_string_length () {
 		string s = "ABC";
 
 		return (s.Length == 3) ? 0 : 1;
 	}
 
-	static int test_0_intrins_string_chars () {
+	public static int test_0_intrins_string_chars () {
 		string s = "ABC";
 
 		return (s [0] == 'A' && s [1] == 'B' && s [2] == 'C') ? 0 : 1;
 	}
 
-	static int test_0_intrins_object_gettype () {
+	public static int test_0_intrins_object_gettype () {
 		object o = 1;
 
 		return (o.GetType () == typeof (int)) ? 0 : 1;
 	}
 
-	static int test_0_intrins_object_gethashcode () {
+	public static int test_0_intrins_object_gethashcode () {
 		object o = new Object ();
 
 		return (o.GetHashCode () == o.GetHashCode ()) ? 0 : 1;
@@ -1101,32 +1153,32 @@ ncells ) {
 	class FooClass {
 	}
 
-	static int test_0_intrins_object_ctor () {
+	public static int test_0_intrins_object_ctor () {
 		object o = new FooClass ();
 
 		return (o != null) ? 0 : 1;
 	}
 
-	static int test_0_intrins_array_rank () {
+	public static int test_0_intrins_array_rank () {
 		int[,] a = new int [10, 10];
 
 		return (a.Rank == 2) ? 0 : 1;
 	}
 
-	static int test_0_intrins_array_length () {
+	public static int test_0_intrins_array_length () {
 		int[,] a = new int [10, 10];
 		Array a2 = a;
 
 		return (a2.Length == 100) ? 0 : 1;
 	}
 
-	static int test_0_intrins_runtimehelpers_offset_to_string_data () {
+	public static int test_0_intrins_runtimehelpers_offset_to_string_data () {
 		int i = RuntimeHelpers.OffsetToStringData;
 		
 		return i - i;
 	}
 
-	static int test_0_intrins_string_setchar () {
+	public static int test_0_intrins_string_setchar () {
 		StringBuilder sb = new StringBuilder ("ABC");
 
 		sb [1] = 'D';
@@ -1139,13 +1191,13 @@ ncells ) {
         Foo f = new Foo ();	
 	}
 
-	static int test_0_regress_78990_unaligned_structs () {
+	public static int test_0_regress_78990_unaligned_structs () {
 		new Bar ();
 
 		return 0;
 	}
 
-	static unsafe int test_97_negative_index () {
+	public static unsafe int test_97_negative_index () {
 		char[] arr = new char[] {'a', 'b'};
 		fixed (char *p = arr) {
 			char *i = p + 2;
@@ -1155,7 +1207,7 @@ ncells ) {
 	}
 
 	/* bug #82281 */
-	static int test_0_unsigned_right_shift_imm0 () {
+	public static int test_0_unsigned_right_shift_imm0 () {
 		uint temp = 0;
 		byte[] data = new byte[256];
 		for (int i = 0; i < 1; i ++)
@@ -1175,7 +1227,7 @@ ncells ) {
 		}
 	}
 
-	static int test_0_abcrem_check_this_removal () {
+	public static int test_0_abcrem_check_this_removal () {
 		Bar2 b = new Bar2 ();
 
 		// The check_this generated here by the JIT should be removed
@@ -1192,7 +1244,7 @@ ncells ) {
 		return 0;
 	}
 
-	static int test_0_abcrem_check_this_removal2 () {
+	public static int test_0_abcrem_check_this_removal2 () {
 		Bar2 b = new Bar2 ();
 
 		invoke_twice (b);
@@ -1201,7 +1253,7 @@ ncells ) {
 	}
 
 	/* #346563 */
-	static int test_0_array_access_64_bit () {
+	public static int test_0_array_access_64_bit () {
 		int[] arr2 = new int [10];
 		for (int i = 0; i < 10; ++i)
 			arr2 [i] = i;
@@ -1226,20 +1278,32 @@ ncells ) {
 		return 0;
 	}
 
+	public static float return_float () {
+		return 1.4e-45f;
+	}
+
+	public static int test_0_float_return_spill () {
+		// The return value of return_float () is spilled because of the
+		// boxing call
+		object o = return_float ();
+		float f = return_float ();
+		return (float)o == f ? 0 : 1;
+	}
+
 	class R4Holder {
 		public static float pi = 3.14f;
 
 		public float float_field;
 	}
 
-	static int test_0_ldsfld_soft_float () {
+	public static int test_0_ldsfld_soft_float () {
 		if (R4Holder.pi == 3.14f)
 			return 0;
 		else
 			return 1;
 	}
 
-	static int test_0_ldfld_stfld_soft_float () {
+	public static int test_0_ldfld_stfld_soft_float () {
 		R4Holder h = new R4Holder ();
 		h.float_field = 3.14f;
 
@@ -1255,7 +1319,7 @@ ncells ) {
 		public float float_field;
 	}
 
-	static int test_0_ldfld_stfld_soft_float_remote () {
+	public static int test_0_ldfld_stfld_soft_float_remote () {
 		R4HolderRemote h = new R4HolderRemote ();
 		h.float_field = 3.14f;
 
@@ -1265,7 +1329,7 @@ ncells ) {
 			return 1;
 	}
 
-	static int test_0_locals_soft_float () {
+	public static int test_0_locals_soft_float () {
 		float f = 0.0f;
 		
 		f = 3.14f;
@@ -1275,5 +1339,30 @@ ncells ) {
 		else
 			return 1;
 	}
+
+	static int range_check_strlen (int i, string s) {
+		if (i < 0 || i > s.Length)
+			return 1;
+		else
+			return 0;
+	}
+		
+	public static int test_0_range_check_opt () {
+		if (range_check_strlen (0, "A") != 0)
+			return 1;
+		if (range_check_strlen (1, "A") != 0)
+			return 2;
+		if (range_check_strlen (2, "A") != 1)
+			return 3;
+		if (range_check_strlen (-100, "A") != 1)
+			return 4;
+		return 0;
+	}
+
+	static int test_0_array_get_set_soft_float () {
+		float[,] arr = new float [2, 2];
+		arr [0, 0] = 256f;
+		return arr [0, 0] == 256f ? 0 : 1;
+	}
 }
 
