@@ -15,6 +15,7 @@
 #include <mono/metadata/marshal.h>
 #include <mono/metadata/tabledefs.h>
 #include <mono/metadata/mono-debug-debugger.h>
+#include <mono/metadata/monitor.h>
 #include <mono/arch/amd64/amd64-codegen.h>
 
 #ifdef HAVE_VALGRIND_MEMCHECK_H
@@ -222,90 +223,77 @@ guchar*
 mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *code_size, MonoJumpInfo **ji, gboolean aot)
 {
 	guint8 *buf, *code, *tramp, *br [2], *r11_save_code, *after_r11_save_code;
-	int i, lmf_offset, offset, res_offset, arg_offset, tramp_offset, saved_regs_offset;
-	int saved_fpregs_offset, rbp_offset, framesize, orig_rsp_to_rbp_offset;
+	int i, lmf_offset, offset, res_offset, arg_offset, rax_offset, tramp_offset, saved_regs_offset;
+	int saved_fpregs_offset, rbp_offset, framesize, orig_rsp_to_rbp_offset, cfa_offset;
 	gboolean has_caller;
+	GSList *unwind_ops = NULL;
+	GSList *l;
 
 	if (tramp_type == MONO_TRAMPOLINE_JUMP)
 		has_caller = FALSE;
 	else
 		has_caller = TRUE;
 
-	code = buf = mono_global_codeman_reserve (524);
+	code = buf = mono_global_codeman_reserve (538);
 
 	*ji = NULL;
 
-	framesize = 524 + sizeof (MonoLMF);
+	framesize = 538 + sizeof (MonoLMF);
 	framesize = (framesize + (MONO_ARCH_FRAME_ALIGNMENT - 1)) & ~ (MONO_ARCH_FRAME_ALIGNMENT - 1);
 
-	if (tramp_type == MONO_TRAMPOLINE_GENERIC_CLASS_INIT) {
-		static int byte_offset = -1;
-		static guint8 bitmask;
-
-		guint8 *jump;
-
-		if (byte_offset < 0)
-			mono_marshal_find_bitfield_offset (MonoVTable, initialized, &byte_offset, &bitmask);
-
-		amd64_test_membase_imm_size (code, MONO_ARCH_VTABLE_REG, byte_offset, bitmask, 1);
-		jump = code;
-		amd64_branch8 (code, X86_CC_Z, -1, 1);
-
-		amd64_ret (code);
-
-		x86_patch (jump, code);
-	}
-
 	orig_rsp_to_rbp_offset = 0;
 	r11_save_code = code;
 	/* Reserve 5 bytes for the mov_membase_reg to save R11 */
 	code += 5;
 	after_r11_save_code = code;
 
-	/*
-	 * The generic class init trampoline is called directly by
-	 * JITted code, there is no specific trampoline.  The lazy
-	 * fetch trampolines behave like generic class init
-	 * trampolines.
-	 */
-	if (tramp_type != MONO_TRAMPOLINE_GENERIC_CLASS_INIT &&
-			tramp_type != MONO_TRAMPOLINE_RGCTX_LAZY_FETCH) {
-		/* Pop the return address off the stack */
-		amd64_pop_reg (code, AMD64_R11);
-		orig_rsp_to_rbp_offset += 8;
-	}
+	// CFA = sp + 16 (the trampoline address is on the stack)
+	cfa_offset = 16;
+	mono_add_unwind_op_def_cfa (unwind_ops, code, buf, AMD64_RSP, 16);
+	// IP saved at CFA - 8
+	mono_add_unwind_op_offset (unwind_ops, code, buf, AMD64_RIP, -8);
+
+	/* Pop the return address off the stack */
+	amd64_pop_reg (code, AMD64_R11);
+	orig_rsp_to_rbp_offset += 8;
+
+	cfa_offset -= 8;
+	mono_add_unwind_op_def_cfa_offset (unwind_ops, code, buf, cfa_offset);
 
 	/* 
 	 * Allocate a new stack frame
 	 */
 	amd64_push_reg (code, AMD64_RBP);
+	cfa_offset += 8;
+	mono_add_unwind_op_def_cfa_offset (unwind_ops, code, buf, cfa_offset);
+	mono_add_unwind_op_offset (unwind_ops, code, buf, AMD64_RBP, - cfa_offset);
+
 	orig_rsp_to_rbp_offset -= 8;
 	amd64_mov_reg_reg (code, AMD64_RBP, AMD64_RSP, 8);
+	mono_add_unwind_op_def_cfa_reg (unwind_ops, code, buf, AMD64_RBP);
 	amd64_alu_reg_imm (code, X86_SUB, AMD64_RSP, framesize);
 
 	offset = 0;
 	rbp_offset = - offset;
 
+	offset += 8;
+	rax_offset = - offset;
+
 	offset += 8;
 	tramp_offset = - offset;
 
 	offset += 8;
 	arg_offset = - offset;
 
-	if (tramp_type != MONO_TRAMPOLINE_GENERIC_CLASS_INIT &&
-			tramp_type != MONO_TRAMPOLINE_RGCTX_LAZY_FETCH) {
-		/* Compute the trampoline address from the return address */
-		if (aot) {
-			/* 7 = length of call *<offset>(rip) */
-			amd64_alu_reg_imm (code, X86_SUB, AMD64_R11, 7);
-		} else {
-			/* 5 = length of amd64_call_membase () */
-			amd64_alu_reg_imm (code, X86_SUB, AMD64_R11, 5);
-		}
-		amd64_mov_membase_reg (code, AMD64_RBP, tramp_offset, AMD64_R11, 8);
+	/* Compute the trampoline address from the return address */
+	if (aot) {
+		/* 7 = length of call *<offset>(rip) */
+		amd64_alu_reg_imm (code, X86_SUB, AMD64_R11, 7);
 	} else {
-		amd64_mov_membase_imm (code, AMD64_RBP, tramp_offset, 0, 8);
+		/* 5 = length of amd64_call_membase () */
+		amd64_alu_reg_imm (code, X86_SUB, AMD64_R11, 5);
 	}
+	amd64_mov_membase_reg (code, AMD64_RBP, tramp_offset, AMD64_R11, 8);
 
 	offset += 8;
 	res_offset = - offset;
@@ -322,7 +310,6 @@ mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *c
 		} else if (i != AMD64_R11) {
 			amd64_mov_membase_reg (code, AMD64_RBP, saved_regs_offset + (i * 8), i, 8);
 		} else {
-			
 			/* We have to save R11 right at the start of
 			   the trampoline code because it's used as a
 			   scratch register */
@@ -336,7 +323,8 @@ mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *c
 		amd64_movsd_membase_reg (code, AMD64_RBP, saved_fpregs_offset + (i * 8), i);
 
 	if (tramp_type != MONO_TRAMPOLINE_GENERIC_CLASS_INIT &&
-			tramp_type != MONO_TRAMPOLINE_RGCTX_LAZY_FETCH) {
+			tramp_type != MONO_TRAMPOLINE_MONITOR_ENTER &&
+			tramp_type != MONO_TRAMPOLINE_MONITOR_EXIT) {
 		/* Obtain the trampoline argument which is encoded in the instruction stream */
 		if (aot) {
 			/* Load the GOT offset */
@@ -364,7 +352,8 @@ mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *c
 		}
 		amd64_mov_membase_reg (code, AMD64_RBP, arg_offset, AMD64_R11, 8);
 	} else {
-		amd64_mov_membase_reg (code, AMD64_RBP, arg_offset, MONO_ARCH_VTABLE_REG, 8);
+		amd64_mov_reg_membase (code, AMD64_R11, AMD64_RBP, saved_regs_offset + (MONO_AMD64_ARG_REG1 * 8), 8);
+		amd64_mov_membase_reg (code, AMD64_RBP, arg_offset, AMD64_R11, 8);
 	}
 
 	/* Save LMF begin */
@@ -471,20 +460,18 @@ mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *c
 	amd64_mov_reg_membase (code, AMD64_R11, AMD64_RBP, lmf_offset + G_STRUCT_OFFSET (MonoLMF, lmf_addr), 8);
 	amd64_mov_membase_reg (code, AMD64_R11, 0, AMD64_RCX, 8);
 
-	/* Restore argument registers, r10 (needed to pass rgctx to
-	   static shared generic methods) and r11 (imt register for
-	   interface calls). */
-	for (i = 0; i < AMD64_NREG; ++i)
-		if (AMD64_IS_ARGUMENT_REG (i) || i == AMD64_R10 || i == AMD64_R11)
-			amd64_mov_reg_membase (code, i, AMD64_RBP, saved_regs_offset + (i * 8), 8);
-
 	/* 
-	 * FIXME: When using aot-only, the called code might be a C vararg function 
-	 * which uses %rax as well.
-	 * We could restore it, but we would have to use another register to store the
-	 * target address, and we don't have any left.
-	 * Also, the default AOT plt trampolines overwrite 'rax'.
+	 * Save rax to the stack, after the leave instruction, this will become part of
+	 * the red zone.
 	 */
+	amd64_mov_membase_reg (code, AMD64_RBP, rax_offset, AMD64_RAX, 8);
+
+	/* Restore argument registers, r10 (needed to pass rgctx to
+	   static shared generic methods), r11 (imt register for
+	   interface calls), and rax (needed for direct calls to C vararg functions). */
+	for (i = 0; i < AMD64_NREG; ++i)
+		if (AMD64_IS_ARGUMENT_REG (i) || i == AMD64_R10 || i == AMD64_R11 || i == AMD64_RAX)
+			amd64_mov_reg_membase (code, i, AMD64_RBP, saved_regs_offset + (i * 8), 8);
 
 	for (i = 0; i < 8; ++i)
 		amd64_movsd_reg_membase (code, i, AMD64_RBP, saved_fpregs_offset + (i * 8));
@@ -492,16 +479,16 @@ mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *c
 	/* Restore stack */
 	amd64_leave (code);
 
-	if (tramp_type == MONO_TRAMPOLINE_CLASS_INIT ||
-			tramp_type == MONO_TRAMPOLINE_GENERIC_CLASS_INIT ||
-			tramp_type == MONO_TRAMPOLINE_RGCTX_LAZY_FETCH)
+	if (MONO_TRAMPOLINE_TYPE_MUST_RETURN (tramp_type)) {
+		/* Load result */
+		amd64_mov_reg_membase (code, AMD64_RAX, AMD64_RSP, rax_offset - 0x8, 8);
 		amd64_ret (code);
-	else {
-		/* call the compiled method */
-		amd64_jump_reg (code, AMD64_RAX);
+	} else {
+		/* call the compiled method using the saved rax */
+		amd64_jump_membase (code, AMD64_RSP, rax_offset - 0x8);
 	}
 
-	g_assert ((code - buf) <= 524);
+	g_assert ((code - buf) <= 538);
 
 	mono_arch_flush_icache (buf, code - buf);
 
@@ -514,6 +501,12 @@ mono_arch_create_trampoline_code_full (MonoTrampolineType tramp_type, guint32 *c
 		nullified_class_init_trampoline = mono_arch_get_nullified_class_init_trampoline (&code_len);
 	}
 
+	mono_save_trampoline_xdebug_info ("<generic_trampoline>", buf, *code_size, unwind_ops);
+
+	for (l = unwind_ops; l; l = l->next)
+		g_free (l->data);
+	g_slist_free (unwind_ops);
+	
 	return buf;
 }
 
@@ -574,7 +567,16 @@ mono_arch_create_specific_trampoline (gpointer arg1, MonoTrampolineType tramp_ty
 gpointer
 mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot)
 {
-	guint8 *tramp = mono_get_trampoline_code (MONO_TRAMPOLINE_RGCTX_LAZY_FETCH);
+	guint32 code_size;
+	MonoJumpInfo *ji;
+
+	return mono_arch_create_rgctx_lazy_fetch_trampoline_full (slot, &code_size, &ji, FALSE);
+}
+
+gpointer
+mono_arch_create_rgctx_lazy_fetch_trampoline_full (guint32 slot, guint32 *code_size, MonoJumpInfo **ji, gboolean aot)
+{
+	guint8 *tramp;
 	guint8 *code, *buf;
 	guint8 **rgctx_null_jumps;
 	int tramp_size;
@@ -582,7 +584,7 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot)
 	int i;
 	gboolean mrgctx;
 
-	g_assert (tramp);
+	*ji = NULL;
 
 	mrgctx = MONO_RGCTX_SLOT_IS_MRGCTX (slot);
 	index = MONO_RGCTX_SLOT_INDEX (slot);
@@ -596,7 +598,7 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot)
 		index -= size - 1;
 	}
 
-	tramp_size = 36 + 8 * depth;
+	tramp_size = 64 + 8 * depth;
 
 	code = buf = mono_global_codeman_reserve (tramp_size);
 
@@ -604,65 +606,332 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot)
 
 	if (mrgctx) {
 		/* get mrgctx ptr */
-		amd64_mov_reg_reg (buf, AMD64_RAX, AMD64_ARG_REG1, 8);
+		amd64_mov_reg_reg (code, AMD64_RAX, AMD64_ARG_REG1, 8);
 	} else {
 		/* load rgctx ptr from vtable */
-		amd64_mov_reg_membase (buf, AMD64_RAX, AMD64_ARG_REG1, G_STRUCT_OFFSET (MonoVTable, runtime_generic_context), 8);
+		amd64_mov_reg_membase (code, AMD64_RAX, AMD64_ARG_REG1, G_STRUCT_OFFSET (MonoVTable, runtime_generic_context), 8);
 		/* is the rgctx ptr null? */
-		amd64_test_reg_reg (buf, AMD64_RAX, AMD64_RAX);
+		amd64_test_reg_reg (code, AMD64_RAX, AMD64_RAX);
 		/* if yes, jump to actual trampoline */
-		rgctx_null_jumps [0] = buf;
-		amd64_branch8 (buf, X86_CC_Z, -1, 1);
+		rgctx_null_jumps [0] = code;
+		amd64_branch8 (code, X86_CC_Z, -1, 1);
 	}
 
 	for (i = 0; i < depth; ++i) {
 		/* load ptr to next array */
 		if (mrgctx && i == 0)
-			amd64_mov_reg_membase (buf, AMD64_RAX, AMD64_RAX, sizeof (MonoMethodRuntimeGenericContext), 8);
+			amd64_mov_reg_membase (code, AMD64_RAX, AMD64_RAX, sizeof (MonoMethodRuntimeGenericContext), 8);
 		else
-			amd64_mov_reg_membase (buf, AMD64_RAX, AMD64_RAX, 0, 8);
+			amd64_mov_reg_membase (code, AMD64_RAX, AMD64_RAX, 0, 8);
 		/* is the ptr null? */
-		amd64_test_reg_reg (buf, AMD64_RAX, AMD64_RAX);
+		amd64_test_reg_reg (code, AMD64_RAX, AMD64_RAX);
 		/* if yes, jump to actual trampoline */
-		rgctx_null_jumps [i + 1] = buf;
-		amd64_branch8 (buf, X86_CC_Z, -1, 1);
+		rgctx_null_jumps [i + 1] = code;
+		amd64_branch8 (code, X86_CC_Z, -1, 1);
 	}
 
 	/* fetch slot */
-	amd64_mov_reg_membase (buf, AMD64_RAX, AMD64_RAX, sizeof (gpointer) * (index + 1), 8);
+	amd64_mov_reg_membase (code, AMD64_RAX, AMD64_RAX, sizeof (gpointer) * (index + 1), 8);
 	/* is the slot null? */
-	amd64_test_reg_reg (buf, AMD64_RAX, AMD64_RAX);
+	amd64_test_reg_reg (code, AMD64_RAX, AMD64_RAX);
 	/* if yes, jump to actual trampoline */
-	rgctx_null_jumps [depth + 1] = buf;
-	amd64_branch8 (buf, X86_CC_Z, -1, 1);
+	rgctx_null_jumps [depth + 1] = code;
+	amd64_branch8 (code, X86_CC_Z, -1, 1);
 	/* otherwise return */
-	amd64_ret (buf);
+	amd64_ret (code);
 
 	for (i = mrgctx ? 1 : 0; i <= depth + 1; ++i)
-		x86_patch (rgctx_null_jumps [i], buf);
+		x86_patch (rgctx_null_jumps [i], code);
 
 	g_free (rgctx_null_jumps);
 
 	/* move the rgctx pointer to the VTABLE register */
-	amd64_mov_reg_reg (buf, MONO_ARCH_VTABLE_REG, AMD64_ARG_REG1, 8);
-	/* store the slot in RAX */
-	amd64_mov_reg_imm (buf, AMD64_RAX, slot);
+	amd64_mov_reg_reg (code, MONO_ARCH_VTABLE_REG, AMD64_ARG_REG1, 8);
+
+	if (aot) {
+		*ji = mono_patch_info_list_prepend (*ji, code - buf, MONO_PATCH_INFO_JIT_ICALL_ADDR, g_strdup_printf ("specific_trampoline_lazy_fetch_%u", slot));
+		amd64_mov_reg_membase (code, AMD64_R11, AMD64_RIP, 0, 8);
+		amd64_jump_reg (code, AMD64_R11);
+	} else {
+		tramp = mono_arch_create_specific_trampoline (GUINT_TO_POINTER (slot), MONO_TRAMPOLINE_RGCTX_LAZY_FETCH, mono_get_root_domain (), NULL);
+
+		/* jump to the actual trampoline */
+		amd64_jump_code (code, tramp);
+	}
+
+	mono_arch_flush_icache (buf, code - buf);
+
+	g_assert (code - buf <= tramp_size);
+
+	*code_size = code - buf;
+
+	return buf;
+}
+
+gpointer
+mono_arch_create_generic_class_init_trampoline (void)
+{
+	guint8 *tramp;
+	guint8 *code, *buf;
+	static int byte_offset = -1;
+	static guint8 bitmask;
+	guint8 *jump;
+	int tramp_size;
+
+	tramp_size = 64;
+
+	code = buf = mono_global_codeman_reserve (tramp_size);
+
+	if (byte_offset < 0)
+		mono_marshal_find_bitfield_offset (MonoVTable, initialized, &byte_offset, &bitmask);
+
+	amd64_test_membase_imm_size (code, MONO_AMD64_ARG_REG1, byte_offset, bitmask, 1);
+	jump = code;
+	amd64_branch8 (code, X86_CC_Z, -1, 1);
+
+	amd64_ret (code);
+
+	x86_patch (jump, code);
+
+	tramp = mono_arch_create_specific_trampoline (NULL, MONO_TRAMPOLINE_GENERIC_CLASS_INIT, mono_get_root_domain (), NULL);
+
 	/* jump to the actual trampoline */
-	amd64_jump_code (buf, tramp);
+	amd64_jump_code (code, tramp);
 
-	mono_arch_flush_icache (code, buf - code);
+	mono_arch_flush_icache (buf, code - buf);
 
+	g_assert (code - buf <= tramp_size);
+
+	return buf;
+}
+
+#ifdef MONO_ARCH_MONITOR_OBJECT_REG
+
+gpointer
+mono_arch_create_monitor_enter_trampoline (void)
+{
+	guint32 code_size;
+	MonoJumpInfo *ji;
+
+	return mono_arch_create_monitor_enter_trampoline_full (&code_size, &ji, FALSE);
+}
+
+gpointer
+mono_arch_create_monitor_enter_trampoline_full (guint32 *code_size, MonoJumpInfo **ji, gboolean aot)
+{
+	guint8 *tramp;
+	guint8 *code, *buf;
+	guint8 *jump_obj_null, *jump_sync_null, *jump_cmpxchg_failed, *jump_other_owner, *jump_tid;
+	int tramp_size;
+	int owner_offset, nest_offset, dummy;
+
+	g_assert (MONO_ARCH_MONITOR_OBJECT_REG == AMD64_RDI);
+
+	mono_monitor_threads_sync_members_offset (&owner_offset, &nest_offset, &dummy);
+	g_assert (MONO_THREADS_SYNC_MEMBER_SIZE (owner_offset) == sizeof (gpointer));
+	g_assert (MONO_THREADS_SYNC_MEMBER_SIZE (nest_offset) == sizeof (guint32));
+	owner_offset = MONO_THREADS_SYNC_MEMBER_OFFSET (owner_offset);
+	nest_offset = MONO_THREADS_SYNC_MEMBER_OFFSET (nest_offset);
+
+	tramp_size = 96;
+
+	code = buf = mono_global_codeman_reserve (tramp_size);
+
+	if (mono_thread_get_tls_offset () != -1) {
+		/* MonoObject* obj is in RDI */
+		/* is obj null? */
+		amd64_test_reg_reg (buf, AMD64_RDI, AMD64_RDI);
+		/* if yes, jump to actual trampoline */
+		jump_obj_null = buf;
+		amd64_branch8 (buf, X86_CC_Z, -1, 1);
+
+		/* load obj->synchronization to RCX */
+		amd64_mov_reg_membase (buf, AMD64_RCX, AMD64_RDI, G_STRUCT_OFFSET (MonoObject, synchronisation), 8);
+		/* is synchronization null? */
+		amd64_test_reg_reg (buf, AMD64_RCX, AMD64_RCX);
+		/* if yes, jump to actual trampoline */
+		jump_sync_null = buf;
+		amd64_branch8 (buf, X86_CC_Z, -1, 1);
+
+		/* load MonoThread* into RDX */
+		buf = mono_amd64_emit_tls_get (buf, AMD64_RDX, mono_thread_get_tls_offset ());
+		/* load TID into RDX */
+		amd64_mov_reg_membase (buf, AMD64_RDX, AMD64_RDX, G_STRUCT_OFFSET (MonoThread, tid), 8);
+
+		/* is synchronization->owner null? */
+		amd64_alu_membase_imm_size (buf, X86_CMP, AMD64_RCX, owner_offset, 0, 8);
+		/* if not, jump to next case */
+		jump_tid = buf;
+		amd64_branch8 (buf, X86_CC_NZ, -1, 1);
+
+		/* if yes, try a compare-exchange with the TID */
+		/* zero RAX */
+		amd64_alu_reg_reg (buf, X86_XOR, AMD64_RAX, AMD64_RAX);
+		/* compare and exchange */
+		amd64_prefix (buf, X86_LOCK_PREFIX);
+		amd64_cmpxchg_membase_reg_size (buf, AMD64_RCX, owner_offset, AMD64_RDX, 8);
+		/* if not successful, jump to actual trampoline */
+		jump_cmpxchg_failed = buf;
+		amd64_branch8 (buf, X86_CC_NZ, -1, 1);
+		/* if successful, return */
+		amd64_ret (buf);
+
+		/* next case: synchronization->owner is not null */
+		x86_patch (jump_tid, buf);
+		/* is synchronization->owner == TID? */
+		amd64_alu_membase_reg_size (buf, X86_CMP, AMD64_RCX, owner_offset, AMD64_RDX, 8);
+		/* if not, jump to actual trampoline */
+		jump_other_owner = buf;
+		amd64_branch8 (buf, X86_CC_NZ, -1, 1);
+		/* if yes, increment nest */
+		amd64_inc_membase_size (buf, AMD64_RCX, nest_offset, 4);
+		/* return */
+		amd64_ret (buf);
+
+		x86_patch (jump_obj_null, buf);
+		x86_patch (jump_sync_null, buf);
+		x86_patch (jump_cmpxchg_failed, buf);
+		x86_patch (jump_other_owner, buf);
+	}
+
+	/* jump to the actual trampoline */
+#if MONO_AMD64_ARG_REG1 != AMD64_RDI
+	amd64_mov_reg_reg (buf, MONO_AMD64_ARG_REG1, AMD64_RDI);
+#endif
+
+	if (aot) {
+		*ji = mono_patch_info_list_prepend (*ji, buf - code, MONO_PATCH_INFO_JIT_ICALL_ADDR, "specific_trampoline_monitor_enter");
+		amd64_mov_reg_membase (buf, AMD64_R11, AMD64_RIP, 0, 8);
+		amd64_jump_reg (buf, AMD64_R11);
+	} else {
+		tramp = mono_arch_create_specific_trampoline (NULL, MONO_TRAMPOLINE_MONITOR_ENTER, mono_get_root_domain (), NULL);
+
+		/* jump to the actual trampoline */
+		amd64_jump_code (buf, tramp);
+	}
+
+	mono_arch_flush_icache (buf, buf - code);
 	g_assert (buf - code <= tramp_size);
 
+	*code_size = buf - code;
+
 	return code;
 }
 
-guint32
-mono_arch_get_rgctx_lazy_fetch_offset (gpointer *regs)
+gpointer
+mono_arch_create_monitor_exit_trampoline (void)
 {
-	return (guint32)(gulong)(regs [AMD64_RAX]);
+	guint32 code_size;
+	MonoJumpInfo *ji;
+
+	return mono_arch_create_monitor_exit_trampoline_full (&code_size, &ji, FALSE);
 }
 
+gpointer
+mono_arch_create_monitor_exit_trampoline_full (guint32 *code_size, MonoJumpInfo **ji, gboolean aot)
+{
+	guint8 *tramp;
+	guint8 *code, *buf;
+	guint8 *jump_obj_null, *jump_have_waiters;
+	guint8 *jump_next;
+	int tramp_size;
+	int owner_offset, nest_offset, entry_count_offset;
+
+	g_assert (MONO_ARCH_MONITOR_OBJECT_REG == AMD64_RDI);
+
+	mono_monitor_threads_sync_members_offset (&owner_offset, &nest_offset, &entry_count_offset);
+	g_assert (MONO_THREADS_SYNC_MEMBER_SIZE (owner_offset) == sizeof (gpointer));
+	g_assert (MONO_THREADS_SYNC_MEMBER_SIZE (nest_offset) == sizeof (guint32));
+	g_assert (MONO_THREADS_SYNC_MEMBER_SIZE (entry_count_offset) == sizeof (gint32));
+	owner_offset = MONO_THREADS_SYNC_MEMBER_OFFSET (owner_offset);
+	nest_offset = MONO_THREADS_SYNC_MEMBER_OFFSET (nest_offset);
+	entry_count_offset = MONO_THREADS_SYNC_MEMBER_OFFSET (entry_count_offset);
+
+	tramp_size = 94;
+
+	code = buf = mono_global_codeman_reserve (tramp_size);
+
+	if (mono_thread_get_tls_offset () != -1) {
+		/* MonoObject* obj is in RDI */
+		/* is obj null? */
+		amd64_test_reg_reg (buf, AMD64_RDI, AMD64_RDI);
+		/* if yes, jump to actual trampoline */
+		jump_obj_null = buf;
+		amd64_branch8 (buf, X86_CC_Z, -1, 1);
+
+		/* load obj->synchronization to RCX */
+		amd64_mov_reg_membase (buf, AMD64_RCX, AMD64_RDI, G_STRUCT_OFFSET (MonoObject, synchronisation), 8);
+		/* is synchronization null? */
+		amd64_test_reg_reg (buf, AMD64_RCX, AMD64_RCX);
+		/* if not, jump to next case */
+		jump_next = buf;
+		amd64_branch8 (buf, X86_CC_NZ, -1, 1);
+		/* if yes, just return */
+		amd64_ret (buf);
+
+		/* next case: synchronization is not null */
+		x86_patch (jump_next, buf);
+		/* load MonoThread* into RDX */
+		buf = mono_amd64_emit_tls_get (buf, AMD64_RDX, mono_thread_get_tls_offset ());
+		/* load TID into RDX */
+		amd64_mov_reg_membase (buf, AMD64_RDX, AMD64_RDX, G_STRUCT_OFFSET (MonoThread, tid), 8);
+		/* is synchronization->owner == TID */
+		amd64_alu_membase_reg_size (buf, X86_CMP, AMD64_RCX, owner_offset, AMD64_RDX, 8);
+		/* if yes, jump to next case */
+		jump_next = buf;
+		amd64_branch8 (buf, X86_CC_Z, -1, 1);
+		/* if not, just return */
+		amd64_ret (buf);
+
+		/* next case: synchronization->owner == TID */
+		x86_patch (jump_next, buf);
+		/* is synchronization->nest == 1 */
+		amd64_alu_membase_imm_size (buf, X86_CMP, AMD64_RCX, nest_offset, 1, 4);
+		/* if not, jump to next case */
+		jump_next = buf;
+		amd64_branch8 (buf, X86_CC_NZ, -1, 1);
+		/* if yes, is synchronization->entry_count zero? */
+		amd64_alu_membase_imm_size (buf, X86_CMP, AMD64_RCX, entry_count_offset, 0, 4);
+		/* if not, jump to actual trampoline */
+		jump_have_waiters = buf;
+		amd64_branch8 (buf, X86_CC_NZ, -1 , 1);
+		/* if yes, set synchronization->owner to null and return */
+		amd64_mov_membase_imm (buf, AMD64_RCX, owner_offset, 0, 8);
+		amd64_ret (buf);
+
+		/* next case: synchronization->nest is not 1 */
+		x86_patch (jump_next, buf);
+		/* decrease synchronization->nest and return */
+		amd64_dec_membase_size (buf, AMD64_RCX, nest_offset, 4);
+		amd64_ret (buf);
+
+		x86_patch (jump_obj_null, buf);
+		x86_patch (jump_have_waiters, buf);
+	}
+
+	/* jump to the actual trampoline */
+#if MONO_AMD64_ARG_REG1 != AMD64_RDI
+	amd64_mov_reg_reg (buf, MONO_AMD64_ARG_REG1, AMD64_RDI);
+#endif
+
+	if (aot) {
+		*ji = mono_patch_info_list_prepend (*ji, buf - code, MONO_PATCH_INFO_JIT_ICALL_ADDR, "specific_trampoline_monitor_exit");
+		amd64_mov_reg_membase (buf, AMD64_R11, AMD64_RIP, 0, 8);
+		amd64_jump_reg (buf, AMD64_R11);
+	} else {
+		tramp = mono_arch_create_specific_trampoline (NULL, MONO_TRAMPOLINE_MONITOR_EXIT, mono_get_root_domain (), NULL);
+		amd64_jump_code (buf, tramp);
+	}
+
+	mono_arch_flush_icache (buf, buf - code);
+	g_assert (buf - code <= tramp_size);
+
+	*code_size = buf - code;
+
+	return code;
+}
+#endif
+
 void
 mono_arch_invalidate_method (MonoJitInfo *ji, void *func, gpointer func_arg)
 {
